/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Checksum representation
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: Checksum.cpp,v 1.1 2009/10/26 16:32:20 molnarzs Exp $
 *  Release: $Name:  $
 */
#include "Checksum.h"
#include "logger-utility.h"
#include "StatCommon.h"

#include <assert.h>
#include <boost/algorithm/string.hpp>

namespace glite {
namespace data {
namespace transfer {
namespace urlcopy {

Checksum::Checksum(const char* alg, const char* val)
{
    const std::string a = (alg == NULL ? "" : alg);
    const std::string v = (val == NULL ? "" : val);
    _init(a, v);
}

Checksum::Checksum(const std::string& alg, const std::string& val)
{
    _init(alg, val);
}

bool Checksum::operator == (const Checksum& op) const
{
    return m_alg == op.m_alg && m_val == op.m_val;
}

bool Checksum::operator != (const Checksum& op) const
{
    return ! (*this == op);
}

const std::string& Checksum::value() const
{
    return m_val;
}

const std::string& Checksum::algorithm() const
{
    return m_alg;
}

void Checksum::clear()
{
    m_alg.clear();
    m_val.clear();
}

std::string Checksum::_convertToStandardAlgorithmName(const std::string& alg)
{
    std::string newAlg(alg);
    boost::to_upper(newAlg);

    if (newAlg == "AD")
        newAlg = "ADLER32";
    else if (newAlg == "CS")
        newAlg = "CRC32";

    if (alg != newAlg)
        INFO_LOG(Checksum) << "Algorithm '"  << alg << "' is not standard, "
            << "converted to '" << newAlg << "'";

    return newAlg;
}

Checksum& Checksum::operator = (const Checksum& csum)
{
    if (&csum != this)
    {
        m_alg = csum.m_alg;
        m_val = csum.m_val;
    }

    return *this;
}

void Checksum::copyChecksumAlgorithm(char* d)
{
    assert(d);
    strncpy(d, m_alg.c_str(), _ChecksumAlgorithmMaxLen() - 1);
}

void Checksum::copyChecksumValue(char* d)
{
    assert(d);
    strncpy(d, m_val.c_str(), _ChecksumValueMaxLen() - 1);
}

void Checksum::_init(const std::string& alg, const std::string& val)
{
    m_alg = _convertToStandardAlgorithmName(alg);
    m_val = _doChecksumDeviations(m_alg, val);
}

std::string Checksum::_doChecksumDeviations(const std::string& alg, const std::string& chksum)
{
    std::string chksum_copy(chksum);
    // Checksum comparison must be case-independent
    boost::to_upper(chksum_copy);

    if (alg == "ADLER32") {
        // Handle the trailing spaces and zeros in ADLER32 case.
        std::size_t valueStarts = chksum_copy.find_first_not_of(" 0");

        if (valueStarts == std::string::npos) {
            chksum_copy.clear();
        } else {
            chksum_copy = chksum_copy.substr(valueStarts);
        }
    }

    return chksum_copy;
}

const std::string& Checksum::DefaultChecksumAlgorithm()
{
    static const std::string s("ADLER32");
    return s;
}

size_t Checksum::_ChecksumAlgorithmMaxLen()
{
    return CHECKSUM_ALGORITHM_LEN;
}

size_t Checksum::_ChecksumValueMaxLen()
{
    return CHECKSUM_VALUE_LEN;
}

std::ostream& operator << (std::ostream& out, const Checksum& c)
{
    out << c.m_alg << ":" << c.m_val;
    return out;
}

} // namespace urlcopy
} // namespace transfer
} // namespace data
} // namespace glite
