/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Checksum representation
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: Checksum.h,v 1.1 2009/10/26 16:32:20 molnarzs Exp $
 *  Release: $Name:  $
 */
#ifndef GLITE_DATA_TRANSFER_URL_COPY_URL_CHECKSUM_H_
#define GLITE_DATA_TRANSFER_URL_COPY_URL_CHECKSUM_H_

#include <string>
#include <iostream>

namespace glite {
namespace data {
namespace transfer {
namespace urlcopy {

class Checksum {
public:
    Checksum(const std::string& alg, const std::string& val);
    Checksum(const char* alg, const char* val);

    bool operator == (const Checksum& op) const;
    bool operator != (const Checksum& op) const;
    Checksum& operator = (const Checksum& csum);

    /**
     * Copy a checksum algorithm name. Truncate it if the source is longer than the maximum length.
     *
     * @param d destination
     */
    void copyChecksumAlgorithm(char* d);

    /**
     * Copy a checksum value.
     *
     * @param d destination
     * @param s source
     */
    void copyChecksumValue(char* d);

    /**
     * If neither srmLs nor the user gives checksum algorithm name, then ADLER32 is used.
     */
    static const std::string& DefaultChecksumAlgorithm();

    /** Return the checksum value */
    const std::string& value() const;

    /** Return the checksum algorithm */
    const std::string& algorithm() const;

    /** Maket the checksum an empty one ("not set") */
    void clear();

private:
    /** Return maximum allowed length of the checksum algorithm name */
    static size_t _ChecksumAlgorithmMaxLen();

    /** Return maximum allowed length of the checksum value */
    static size_t _ChecksumValueMaxLen();

    void _init(const std::string& alg, const std::string& val);

    std::string _convertToStandardAlgorithmName(const std::string& alg);

    /**
     * See: https://savannah.cern.ch/bugs/?57046
     * There are deviations from the checksum format specifications. Do them here.
     */
    std::string _doChecksumDeviations(const std::string& alg, const std::string& chksum);

    std::string m_alg;
    std::string m_val;

    friend std::ostream& operator << (std::ostream& out, const Checksum& c);
};

std::ostream& operator << (std::ostream& out, const Checksum& c);

} // namespace urlcopy
} // namespace transfer
} // namespace data
} // namespace glite

#endif //GLITE_DATA_TRANSFER_URL_COPY_URL_CHECKSUM_H_
