/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Checksum checking logic for Srm/Url copy
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: ChecksumChecker.cpp,v 1.4 2009/10/26 16:32:20 molnarzs Exp $
 *  Release: $Name:  $
 */
#include "ChecksumChecker.h"
#include "TransferUtils.h"
#include "logger-utility.h"
#include "FileType.h"
#include "ChecksumGetter_gridFTP.h"
#include "FileRemover_gridFTP.h"
#include "Checksum.h"
#include "Instrument.h"

#include <glite/data/agents/AgentExceptions.h>

#include <boost/lexical_cast.hpp>
#include <assert.h>
#include <cstring>

namespace glite {
namespace data {
namespace transfer {
namespace urlcopy {

using glite::data::agents::RuntimeError;

boost::scoped_ptr<IChecksumGetter> ChecksumChecker::_ChecksumGetter(
    dynamic_cast<IChecksumGetter*> (new ChecksumGetter_gridFTP));

ChecksumChecker::ChecksumChecker(
    UrlCopyError::Phase p,
    RefreshTimeUpdater::IUpdateRefreshTime* urt,
    Context::Ptr& ctx,
    bool isSourceScope) :
    _phase(p), _updater(urt), _isSource(isSourceScope), _ctx(ctx)
{
    assert(urt);
}

UrlCopyError::Scope ChecksumChecker::_getScope() const
{
    return _isSource ? UrlCopyError::SCOPE_SOURCE : UrlCopyError::SCOPE_DESTINATION;
}

void ChecksumChecker::getMissingChecksumsWithGridFTP(const unsigned int timeout)
{
    for (_FilePairPtrVector_type::iterator itr = _filesWithMissingChecksum.begin();
         itr != _filesWithMissingChecksum.end();
         ++itr)
    {
        assert(*itr);
        FilePair& filePair = **itr;
        File& file = _isSource ? filePair.src : filePair.dst;
        INFO_LOG(ChecksumChecker) << "Obtaining checksum by using gridftp for file " << file.name;
        UrlCopyError::Category c;
        std::string m;
        _ChecksumGetter->getChecksumFromGridFTP(filePair, file, timeout, _updater, c, m);

        if (c != UrlCopyError::SUCCESS) {
            ERROR_LOG(ChecksumChecker) << m;
            filePair.setFileError(_getScope(), _phase, c, m);
            filePair.setFinalError(_getScope());
            _filesToRemove.push_back(*itr);
        } else {
            INFO_LOG(ChecksumChecker) << "gridftp returned checksum	" << file.algorithm << ":"
                    << file.value;

            _filesWithChecksum.push_back(*itr);
        }
    }

    _filesWithMissingChecksum.clear();
}

void ChecksumChecker::compareSourceChecksums()
{
    for (_FilePairPtrVector_type::iterator itr = _filesWithChecksum.begin();
         itr != _filesWithChecksum.end();
         ++itr)
    {
        FilePair& filePair = **itr;
        
        bool mustBeAvailable = (filePair.useCase == FilePair::e_compareOnly || 
            filePair.useCase == FilePair::e_checksumAndCompare);

        Checksum src(filePair.src.algorithm, filePair.src.value);
       
        // INSTRUMENTING START
        if (Instrument::Instance().srcChecksumNotAvailable.get()) {
            src.clear();
        }
        
        if (Instrument::Instance().srcForceChecksum.get()) {
            src = *Instrument::Instance().srcForceChecksum; 
        }
        // INSTRUMENTING END
        
        if (mustBeAvailable && src == Checksum("", "")) {
            ERROR_LOG(ChecksumChecker) << "Source checksum is not available!";
            filePair.setFileError(_getScope(), _phase, UrlCopyError::ERROR_CHECKSUM_UNKNOWN,
                "Source checksum is not available");
            filePair.setFinalError(_getScope());
            continue;
        }

        // Source checksum comparison with the user value is required when:
        bool toCompare = (filePair.useCase == FilePair::e_checksumOnly ||
            filePair.useCase == FilePair::e_checksumAndCompare);

        // Compare checksum with the user value if necessary (Use Case 3)
        if (toCompare) 
        {
            INFO_LOG(ChecksumChecker) << "Comparing source checksum with user value for file "
                    << filePair.src.name;
		    
	    INFO_LOG(ChecksumChecker) << "User checksum algorithm: "
                    << filePair.userAlg;	    

	    INFO_LOG(ChecksumChecker) << "User provided checksum:  "
                    << filePair.userVal;	    
		    
            Checksum user(filePair.userAlg, filePair.userVal);

            if (src != user) {
                ERROR_LOG(ChecksumChecker) << "User checksum '" << filePair.userAlg << ":"
                        << filePair.userVal << "' differs from the source checksum!";
                filePair.setFileError(_getScope(), _phase, UrlCopyError::ERROR_CHECKSUM_MISMATCH,
                    "Source file/user checksum mismatch");
            } else {
                INFO_LOG(ChecksumChecker) << "Source/user checksum OK";
            }
        } else if (filePair.useCase == FilePair::e_checksumOnly) {
            INFO_LOG(ChecksumChecker) << "Source checksum comparison with user value not required";
        }
    }
}

void ChecksumChecker::compareDestinationChecksums()
{
    for (_FilePairPtrVector_type::iterator itr = _filesWithChecksum.begin();
         itr != _filesWithChecksum.end();
         ++itr)
    {
        assert(*itr);
        FilePair& filePair = **itr;
        bool match = false;
        Checksum dst(filePair.dst.algorithm, filePair.dst.value);

        // INSTRUMENTING START
        if (Instrument::Instance().dstChecksumNotAvailable.get()) {
            dst.clear();
        }
    
        if (Instrument::Instance().dstForceChecksum.get()) {
            dst = *Instrument::Instance().dstForceChecksum; 
        }
        // INSTRUMENTING END

        if (dst == Checksum("", "")) {
            ERROR_LOG(ChecksumChecker) << "Destination checksum is not available!";
            filePair.setFileError(_getScope(), _phase, UrlCopyError::ERROR_CHECKSUM_UNKNOWN,
                "Destination checksum is not available");
            filePair.setFinalError(_getScope());
            _filesToRemove.push_back(*itr);
            continue;
        }

        if (filePair.useCase == FilePair::e_compareOnly) {
            INFO_LOG(ChecksumChecker) << "Comparing source / destination checksum for files '"
                    << filePair.src.name << "' and '" << filePair.dst.name << "'";

            Checksum src(filePair.src.algorithm, filePair.src.value);
            match = (src == dst);
        } else {
            INFO_LOG(ChecksumChecker) << "Comparing destination checksum with user value for file '"
                    << filePair.dst.name << "'";

            Checksum user(filePair.userAlg, filePair.userVal);
            match = (user == dst);
        }

        if (!match) {
            filePair.setFileError(_getScope(), _phase, UrlCopyError::ERROR_CHECKSUM_MISMATCH,
                "Checksum mismatch");
            filePair.setFinalError(_getScope());
            _filesToRemove.push_back(*itr);
        } else {
            INFO_LOG(ChecksumChecker) << "Checksum matches";
        }
    }
}

void ChecksumChecker::removeFailedDestinationFiles(const int timeout)
{
    for (_FilePairPtrVector_type::iterator it = _filesToRemove.begin();
         it != _filesToRemove.end();
         ++it)
    {
        assert(*it);
        FilePair& filePair = **it;
        INFO_LOG(ChecksumChecker) << "Removing destination file " << filePair.dst.name;
        UrlCopyError::Category c = UrlCopyError::SUCCESS;
        std::string m;

        if (filePair.dst.isSurl) {
            TransferUtils::RmFile(_updater, _ctx, filePair.dst.name, c, m);
        } else {
            FileRemover_gridFTP::RemoveFile(timeout, _updater, filePair.dst.name);    
        }
    }

    _filesToRemove.clear();
}

void ChecksumChecker::_getChecksumFromSrmLs(
    TransferUtils::PathInfo& pinfo,
    boost::shared_ptr<FilePair>& fPair)
{
    assert(fPair.get());

    fPair->isUserProvidedChecksum = (fPair->userAlg && strlen(fPair->userAlg) > 0)
            && (fPair->userVal && strlen(fPair->userVal) > 0);

    // decide the use case
    if (!fPair->compare && !fPair->isUserProvidedChecksum)
    {
    	    INFO_LOG(ChecksumChecker) << "User checksum algorithm: "
                    << fPair->userAlg;	    

	    INFO_LOG(ChecksumChecker) << "User provided checksum:  "
                    << fPair->userVal;	    
    
        fPair->useCase = FilePair::e_noChecksumComparison;
    }
    else if (!fPair->compare && fPair->isUserProvidedChecksum)
        fPair->useCase = FilePair::e_checksumOnly;
    else if (fPair->compare && !fPair->isUserProvidedChecksum)
        fPair->useCase = FilePair::e_compareOnly;
    else
        fPair->useCase = FilePair::e_checksumAndCompare;

    File& file = _isSource ? fPair->src : fPair->dst;
    INFO_LOG(ChecksumChecker) << "Checksum operations on file '" << file.name << "'";
    // Handle use case 0
    if (fPair->useCase == FilePair::e_noChecksumComparison) {
        INFO_LOG(ChecksumChecker) << "Skipping checksum verification";
        return;
    }
    // Record the file pair information, then use the recorded one for further  processing
    _filePairs.push_back(fPair);
    ChecksumCheckerUtils::FilePair& filePair = *(_filePairs.back());
    // Get SrmLs checksum
    std::string& csumSrmLsVal = pinfo.cksm_value;
    std::string& csumSrmLsAlg = pinfo.cksm_algorithm;

    // INSTRUMENTING START
    if ( ( Instrument::Instance().srcSrmlsNoChecksum.get() && _isSource ) ||
         ( Instrument::Instance().dstSrmlsNoChecksum.get() && !_isSource ) )
    {
        csumSrmLsVal.clear();
        csumSrmLsAlg.clear();
    }

    if ( Instrument::Instance().srcSrmlsYesChecksum.get() && _isSource )
    {
        csumSrmLsVal = Instrument::Instance().srcSrmlsYesChecksum->value();
        csumSrmLsAlg = Instrument::Instance().srcSrmlsYesChecksum->algorithm();
    }

    if ( Instrument::Instance().dstSrmlsYesChecksum.get() && !_isSource )
    {
        csumSrmLsVal = Instrument::Instance().dstSrmlsYesChecksum->value();
        csumSrmLsAlg = Instrument::Instance().dstSrmlsYesChecksum->algorithm();
    }
    // INSTRUMENTING END
    file.isSrmLsProvidedChecksum = !(csumSrmLsVal.empty() || csumSrmLsAlg.empty());

    if (file.isSrmLsProvidedChecksum) {
        INFO_LOG(ChecksumChecker) << "SrmLs returned checksum " << csumSrmLsAlg << ":"
                << csumSrmLsVal;
        Checksum c(csumSrmLsAlg, csumSrmLsVal);
        c.copyChecksumAlgorithm(file.algorithm);
        c.copyChecksumValue(file.value);
        _filesWithChecksum.push_back(&filePair);
    } else {
        // Use Case 1: client provides checksum, requires no comparison, and
        // SrmLs does not provide checksum. No fallback to gridftp in source case.
        if (_isSource &&
            filePair.useCase == FilePair::e_checksumOnly &&
            !file.isSrmLsProvidedChecksum)
        {
            if (file.isSurl) {
                INFO_LOG(ChecksumChecker)
                    << "SrmLs returned no checksum and no fallback to gridFTP in source file case...";
            } else {
                INFO_LOG(ChecksumChecker) 
                    << "Checksum is not available, skipping gridFTP checksum retrieval."; 
            }
        } else {
            // In other cases, get the checksum from gridFTP later.
            _filesWithMissingChecksum.push_back(&filePair);
        }
    }
}

void ChecksumChecker::_setErrorForAllGridFTPFiles(UrlCopyError::Category c, const std::string& m)
{
    for (_FilePairPtrVector_type::iterator it = _filesWithMissingChecksum.begin();
         it != _filesWithMissingChecksum.end();
         ++it)
    {
        assert(*it);
        (*it)->setFileError(_getScope(), _phase, c, m);
    }

    _filesToRemove = _filesWithMissingChecksum;
}

} // namespace urlcopy
} // namespace transfer
} // namespace data
} // namespace glite

