/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Checksum checking logic for Srm/Url copy
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: ChecksumChecker.h,v 1.2 2009/08/04 17:49:05 molnarzs Exp $
 *  Release: $Name:  $
 */
#ifndef GLITE_DATA_TRANSFER_URL_COPY_URL_CHECKSUM_CHECKER_H_
#define GLITE_DATA_TRANSFER_URL_COPY_URL_CHECKSUM_CHECKER_H_

#include "RefreshTimeUpdater.h"
#include "ChecksumCheckerUtils.h"
#include <glite/data/srm/util/Context.h>
#include <glite/data/transfer/urlcopy/UrlCopyError.h>

#include <boost/lexical_cast.hpp>
#include <boost/scoped_ptr.hpp>
#include <vector>
#include <string>

namespace glite {
namespace data {
namespace transfer {
namespace urlcopy {

namespace TransferUtils {
class PathInfo;
} // namespace TransferUtils

class IChecksumGetter;

using glite::data::srm::util::Context;
using namespace ChecksumCheckerUtils;

/**
 * Implements the checksum checking logic, specified in
 *
 * https://twiki.cern.ch/twiki/bin/view/EGEE/FtsChecksums
 *
 * The class cannot be used as is, children should complete its functionality. It must implement
 * functionality that fills in the following containers:
 *
 * _filePairs: all source/destination file pairs in a job
 * _filesWithMissingSrmLsChecksum: all source/destination file pairs in a job where the SrmLs call
 * did not provide checksum information, depending on the scope (source or destination)
 *
 * The methods in this base class will process those containers.
 *
 * Generally, the class methods should be used in the following way (no arguments, etc. now):
 *
 * In PREPARATION phase, working with source files:
 *
 * ChecksumChecker_Child checker;
 * checker.getSourceChecksumFromSrmLs(); // for all source/destination pairs in a job...
 * checker.sendPrepareToGetForMissing(); // Prepare the TURL for files with missing checksums
 * checker.getMissingChecksumsWithGridFTP();
 * checker.compareSourceChecksums();
 *
 * In FINALIZATION phase, working with destination files:
 *
 * ChecksumChecker_Child checker;
 * checker.getDestinationChecksumFromSrmLs(); // for all source/destination pairs in a job...
 * checker.sendPrepareToGetForMissing(); // Prepare the TURL for files with missing checksums
 * checker.getMissingChecksumsWithGridFTP();
 * checker.removeFailedDestinationFiles(); // Failed files removed in this phase
 * checker.compareDestinationChecksums();
 * checker.removeFailedDestinationFiles(); // Again, remove the files failed here
 *
 * getSourceChecksumFromSrmLs, getDestinationChecksumFromSrmLs and sendPrepareToGetForMissing should
 * be methods in the children. The base class does not introduce unnecessary constraints about the
 * signature, etc. of those methods. This is not a polymorphic class!
 *
 *
 * Testing
 *
 * The class depends on gridftp operations. To be able to unit-test the class, those operations
 * may be replaced with mock functionalities (dependency injection).
 */
class ChecksumChecker
{
public:
    /**
     * The method retrieves checksum information using gridFTP. Those files must be loaded into
     * _filesWithMissingChecksum before calling this method, the method consumes that
     * container. It retrieves the info one-by-one.
     *
     * If checksum retrieval is successful, the file is placed into _filesWithChecksum container,
     * else to the  _filesToRemove. In the latter case, the file error information is also set
     * (will be reflected in the referenced stat structure).
     *
     * The _filesWithMissingChecksum is consumed up totally (empty at return).
     */
    void getMissingChecksumsWithGridFTP(const unsigned int timeout);

    /**
     * Compare the source file checksums. According to the use cases, it will either skip comparison
     * or compare with the user value. Comparison is decided file-by-file in a job.
     *
     * The method iterates through _filesWithChecksum, so files with valid source checksums must
     * be stored there before calling this method.
     */
    void compareSourceChecksums();

    /**
     * Compare the destination file checksums. According to the use cases, it will either skip
     * comparison, compare with the user value or compare with the source checksum. Comparison is
     * decided file-by-file in a job.
     *
     * The method iterates through _filesWithChecksum, so files with valid destination checksums
     * must be stored there there before calling this method.
     *
     * If the comparison fails, the file pair is added to _filesToRemove.
     */
    void compareDestinationChecksums();

    /**
     * The method iterates through _filesToRemove and removes all the destination files referenced
     * stored in the container. The files in that container stores references to files with required
     * but unavailable or mismatching checksums.
     */
    void removeFailedDestinationFiles(const int timeout);

    /**
     * Replace the class implementing the gridFTP operations with something else. The method should
     * be called by the unit tests with classes implementing mock functionalities, in production
     * code it does not make sense.
     *
     * @param getter The new object. The ChecksumChecker object will delete the pointer, be careful
     *               with the ownership!
     */
    template<class T> static void Test_ChangeChecksumGetter(T* getter)
    {
        _ChecksumGetter.reset(dynamic_cast<IChecksumGetter*> (getter));
    }

protected:
    /**
     * Constructor.
     *
     * @param p The phase in which the object is used. The phase cannot be changed during the
     *          lifetime of the object.
     * @param urt RefreshTimeUpdater class for the SRM method calls and gridFTP operations.
     * @param ctx The actual application context.
     * @param isSourceScope true: The object is used in source scope, else in destination scope.
     */
    ChecksumChecker(
        UrlCopyError::Phase p,
        RefreshTimeUpdater::IUpdateRefreshTime* urt,
        Context::Ptr& ctx,
        bool isSourceScope);

    /**
     * Helper method. Fill in error information in all the files that failed to provide checksum
     * either with SrmLS or gridFTP, and checksum is required. Those files are also placed into
     * _filesToRemove. The method goes through the _filesWithMissingChecksum.
     *
     * @param c The error category to be set.
     * @param m The error message to be set.
     */
    void _setErrorForAllGridFTPFiles(UrlCopyError::Category c, const std::string& m);

    /**
     * Fetch the checksum information from a PathInfo structure. The structure is obtained from a
     * previous SrmLs call. The method also determines which use case of the specification is the
     * actual case, and fills in the FilePair and checksum information accordingly. If the PathInfo
     * does not contain checksum information, and the actual use case and scope requires such
     * information, then the FilePair reference is stored in the _filesWithMissingChecksum as well,
     * for later gridFTP operations.
     *
     * @param pinfo The result from a previous SrmLs call
     * @param filePair The checksum info gets here (and to the under-laying stat structure)
     */
    void _getChecksumFromSrmLs(
        TransferUtils::PathInfo& pinfo,
        boost::shared_ptr<FilePair>& filePair);

    /**
     * Return the scope in which the ChecksumChecker is used.
     */
    UrlCopyError::Scope _getScope() const;


    /**
     * The source/destination file pairs of a job.
     */
    typedef std::vector<boost::shared_ptr<FilePair> > _FilePairs_type;
    _FilePairs_type _filePairs;

    typedef std::vector<FilePair*> _FilePairPtrVector_type;

    /**
     * Files having no checksum information. Files get here after failed SrmLs or gridFTP checksum
     * retrieval attempts.
     */
    _FilePairPtrVector_type _filesWithMissingChecksum;

    /**
     * File pair references get here if destination checksum should be required (but unavailable),
     * or mismatches. The system should remove those destination files.
     */
    _FilePairPtrVector_type _filesToRemove;

    /**
     * Files with successfully obtained checksum information get here. They will participate in the
     * comparison proceduer.
     */
    _FilePairPtrVector_type _filesWithChecksum;

    /**
     * The phase in which the ChecksumChecker is used.
     */
    const UrlCopyError::Phase _phase;

    /**
     * Reference to the RefreshTimeUpdater object that should be used during SRM and gridFTP
     * operations.
     */
    RefreshTimeUpdater::IUpdateRefreshTime* _updater;

    /**
     * TRUE: The scope in which the ChecksumChecker is used is SOURCE
     * FALSE: The scope in which the ChecksumChecker is used is DESTINATION
     */
    const bool _isSource;

    /**
     * The actual application context
     */
    Context::Ptr& _ctx;

    /**
     * An instance of the object responsible for obtaining checksum information by using gridFTP.
     * May be replaced for unit tests.
     */
    static boost::scoped_ptr<IChecksumGetter> _ChecksumGetter;
};

} // namespace urlcopy
} // namespace transfer
} // namespace data
} // namespace glite

#endif //GLITE_DATA_TRANSFER_URL_COPY_URL_CHECKSUM_CHECKER_H_
