/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Checksum checking logic for Srm/Url copy - utilities
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: ChecksumCheckerUtils.h,v 1.4 2009/10/26 16:32:20 molnarzs Exp $
 *  Release: $Name:  $
 */
#ifndef GLITE_DATA_TRANSFER_URL_COPY_URL_CHECKSUM_CHECKER_UTILS_H_
#define GLITE_DATA_TRANSFER_URL_COPY_URL_CHECKSUM_CHECKER_UTILS_H_

#include <glite/data/transfer/urlcopy/UrlCopyError.h>
#include <string>

namespace glite {
namespace data {
namespace transfer {
namespace urlcopy {

namespace ChecksumCheckerUtils {

/**
 * A file (source or destination). It is a controlled and isolated
 * subset of the underlaying stat structure, with operations for the ChecksumChecker. When changing
 * the information in the File, the changes are also reflected in the stat!
 */
struct File
{
    File();

    /** The algorithm name returned by SrmLs or gridFTP */
    char* algorithm;
    /** The value returned by SrmLs or gridFTP */
    char* value;
    /** The file name (SURL) */
    char* name;
    /** The TURL */
    char* turl;
    /** true if the checksum algorithm/value is obtained by an SrmLS call */
    bool isSrmLsProvidedChecksum;
    bool isSurl;
};

/**
 * The special representation of a source/destination file pair. It is a controlled and isolated
 * subset of the underlaying stat structure, with operations for the ChecksumChecker. When changing
 * the information in the FilePair, the changes are also reflected in the stat!
 *
 * The class hides the differences between the SrmCopy_FileStat and the UrlCopy_FileStat.
 */
struct FilePair
{
    typedef enum {
        e_noChecksumComparison, // Use case 0: no checksum comparison required at all
        e_checksumOnly, // User provides checksum, compare flag is false (use case 1)
        e_compareOnly, // User does not provide checksum, compare flag is true (use case 2)
        e_checksumAndCompare // User provides checksum, compare flag is true (use case 3)
    } UseCase_type;

    FilePair();
    virtual ~FilePair();

    /**
     * Set error condition in the under-laying stat file.
     *
     * @param s The scope of the actual checksum checking operations (source or destination) where
     *          the error occurred
     * @param p The phase of the actual checksum checking operations (source or destination) where
     *          the error occurred
     * @param c The error category to be set
     * @param m The error message to be set
     */
    virtual void setFileError(
        UrlCopyError::Scope s,
        UrlCopyError::Phase p,
        UrlCopyError::Category c,
        const std::string& m) = 0;

    /**
     * If the file error is also final transfer error, set it with this method (if such a
     * functionality is required).
     *
     * @param s The scope of the actual checksum checking operations (source or destination) where
     *          the error occurred
     * @param p The phase of the actual checksum checking operations (source or destination) where
     *          the error occurred
     * @param c The error category to be set
     * @param m The error message to be set
     */

    virtual void setFinalError(
        UrlCopyError::Scope s,
        UrlCopyError::Phase p = UrlCopyError::PHASE_UNDEF,
        UrlCopyError::Category c = UrlCopyError::SUCCESS,
        const std::string& m = "")
    {
        /* EMPTY */
    }

    /** The value of the compare-checksums flag */
    bool compare;
    /** The checksum algorithm name given by the user - direct reference to the corresponding
     * field of the stat */
    char* userAlg;
    /** The checksum value given by the user - direct reference to the corresponding
     * field of the stat */
    char* userVal;
    /** true if the user provided checksum else false */
    bool isUserProvidedChecksum;
    /** The checksum comparison use case (see the specs) */
    UseCase_type useCase;
    /** The source file */
    File src;
    /** The destination file */
    File dst;
};

} // ChecksumCheckerUtils

} // namespace urlcopy
} // namespace transfer
} // namespace data
} // namespace glite

#endif // GLITE_DATA_TRANSFER_URL_COPY_URL_CHECKSUM_CHECKER_UTILS_H_
