/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Checksum checking logic for SRM copy
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: ChecksumChecker_SrmCopy.cpp,v 1.2 2009/08/04 17:49:05 molnarzs Exp $
 *  Release: $Name:  $
 */
#include "ChecksumChecker_SrmCopy.h"
#include "SrmCopyUtils.h"
#include "SrmCopyStat.h"
#include "TransferUtils.h"
#include "logger-utility.h"
#include "SrmMethodCaller.h"
#include "stat-utility.h"

#include <glite/data/srm/util/PrepareToGet.h>
#include <glite/data/srm/util/types.h>
#include <glite/data/agents/AgentExceptions.h>

namespace glite {
namespace data {
namespace transfer {
namespace urlcopy {

using namespace glite::data::srm::util;
using namespace ChecksumCheckerUtils;
using glite::data::agents::LogicError;

ChecksumChecker_SrmCopy::_SrmCopyFilePair::~_SrmCopyFilePair()
{
    // EMPTY
}

ChecksumChecker_SrmCopy::_SrmCopyFilePair::_SrmCopyFilePair(SrmCopy_FileStat& filePair) :
    FilePair(), _filePair(filePair)
{
    compare = static_cast<bool> (filePair.mm_compare_checksum);
    userAlg = filePair.mm_checksum_user_algorithm;
    userVal = filePair.mm_checksum_user_value;
    src.algorithm = filePair.mm_src_checksum_algorithm;
    src.value = filePair.mm_src_checksum_value;
    dst.algorithm = filePair.mm_dst_checksum_algorithm;
    dst.value = filePair.mm_dst_checksum_value;
    src.name = filePair.mm_src_name;
    dst.name = filePair.mm_dest_name;
    src.turl = filePair.mm_src_tx_name;
    dst.turl = filePair.mm_dst_tx_name;
}

void ChecksumChecker_SrmCopy::_SrmCopyFilePair::setFileError(
    UrlCopyError::Scope s,
    UrlCopyError::Phase p,
    UrlCopyError::Category c,
    const std::string& m)
{
    SrmCopyUtils::SetFileError(_filePair, s, p, c, m);
}

ChecksumChecker_SrmCopy::ChecksumChecker_SrmCopy(
    UrlCopyError::Phase p,
    RefreshTimeUpdater::IUpdateRefreshTime* urt,
    Context::Ptr& ctx,
    bool isSourceScope) :
    ChecksumChecker(p, urt, ctx, isSourceScope)
{
    // EMPTY
}

void ChecksumChecker_SrmCopy::getSourceChecksumFromSrmLs(
    TransferUtils::PathInfo& pinfo,
    SrmCopy_FileStat& filePair)
{
    boost::shared_ptr<FilePair> fpair(dynamic_cast<FilePair*> (new _SrmCopyFilePair(filePair)));
    _getChecksumFromSrmLs(pinfo, fpair);
}

void ChecksumChecker_SrmCopy::getDestinationChecksumFromSrmLs(SrmCopy_FileStat& filePair)
{
    UrlCopyError::Category c = UrlCopyError::SUCCESS;
    std::string m;
    TransferUtils::PathInfo pinfo;

    if (!TransferUtils::PathExists(_updater, _ctx, filePair.mm_dest_name, true, c, m, pinfo)) {
        ERROR_LOG(ChecksumChecker) << "srmLs failed for file " << filePair.mm_dest_name << " [" << c
                << "]: " << m;
    }

    boost::shared_ptr<FilePair> fpair(dynamic_cast<FilePair*> (new _SrmCopyFilePair(filePair)));
    _getChecksumFromSrmLs(pinfo, fpair);
}

bool ChecksumChecker_SrmCopy::sendPrepareToGetForMissing(time_t startTime, int timeout)
{
    if (_filesWithMissingChecksum.empty())
        return true;

    UrlCopyError::Category c = UrlCopyError::SUCCESS;
    std::string m;
    boost::scoped_ptr<PrepareToGet> request;

    try {
        request.reset(PrepareToGet::create(_ctx));
    } catch (const LogicError& exc) {
        _setErrorForAllGridFTPFiles(UrlCopyError::ERROR_INTERNAL_ERROR, exc.what());
        return false;
    }
    // Set request parameters
    request->files.resize(_filesWithMissingChecksum.size());
    unsigned int i = 0;

    for (_FilePairPtrVector_type::iterator it = _filesWithMissingChecksum.begin();
         it != _filesWithMissingChecksum.end();
         ++it)
    {
        assert(*it);
        request->files[i].surl = _isSource ? (*it)->src.name : (*it)->dst.name;
        ++i;
    }

    request->transferParameters.accessPattern = AccessPattern::PROCESSING_MODE;
    request->transferParameters.connectionType = ConnectionType::LAN;
    request->transferParameters.protocol = PROTOCOL_GSIFTP;

    // Send the request
    if (!TransferUtils::SendPrepareToGet(_ctx, _updater, request, c, m)) {
        _setErrorForAllGridFTPFiles(UrlCopyError::ERROR_CHECKSUM_UNKNOWN, m);
        return false;
    }

    // poll the request status
    if (!TransferUtils::PollPrepareToGet(_ctx, _updater, request, timeout, startTime,
        MAX_GET_SLEEP_TIME, c, m)) {
        _setErrorForAllGridFTPFiles(UrlCopyError::ERROR_CHECKSUM_UNKNOWN, m);
        return false;
    }

    SrmMethodCaller srmMethodCaller(_updater);

    if (UrlCopyError::ERROR_ABORTED == c) {
        INFO_LOG(ChecksumChecker) << "Aborting PrepareToGet request " << request->token();

        if (srmMethodCaller.abortRequest(_ctx, *request, c, m, -1)) {
            INFO_LOG(ChecksumChecker) << "PrepareToGet request " << request->token()
                    << " successfully aborted.";
        } else {
            ERROR_LOG(ChecksumChecker) << "Error aborting PrepareToGet request " << request->token()
                    << " [" << c << "]: " << m;
        }

        _setErrorForAllGridFTPFiles(UrlCopyError::ERROR_CHECKSUM_UNKNOWN,
            "PrepareToGet was aborted. Cannot get checksum.");
        return false;
    }

    // the request completed; check if it was successful (at least for some files).
    if ((SrmStatus::SRM_SUCCESS != request->status.code) && (SrmStatus::SRM_PARTIAL_SUCCESS
            != request->status.code)) {
        const std::string msg =
                "PrepareToGet request failed for all files. Cannot determine checksum.";
        ERROR_LOG(ChecksumChecker) << msg;
        _setErrorForAllGridFTPFiles(UrlCopyError::ERROR_CHECKSUM_UNKNOWN, msg);
        return false;
    }

    i = 0;
    // check the individual file statuses
    for (_FilePairPtrVector_type::iterator it = _filesWithMissingChecksum.begin();
         it != _filesWithMissingChecksum.end();
         ++it)
    {
        assert(*it);
        FilePair& filePair = **it;

        if (request->files[i].status.code != SrmStatus::SRM_FILE_PINNED) {
            filePair.setFileError(_getScope(), _phase, UrlCopyError::ERROR_CHECKSUM_UNKNOWN,
                "PrepareToGet request failed for the file. Cannot get checksum.");
        } else {
            File& file = (_isSource ? filePair.src : filePair.dst);
            copy_url(file.turl, request->files[i].turl.c_str());
        }

        ++i;
    }
    // Release the PTG request
    srmMethodCaller.completeRequest(_ctx, *request, c, m, -1);
    return true;
}

} // namespace urlcopy
} // namespace transfer
} // namespace data
} // namespace glite
