/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Checksum checking logic for Url copy
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: ChecksumChecker_UrlCopy.cpp,v 1.2 2009/08/04 17:49:05 molnarzs Exp $
 *  Release: $Name:  $
 */
#include "ChecksumChecker_UrlCopy.h"
#include "UrlCopyUtils.h"
#include "UrlCopyStat.h"
#include "TransferUtils.h"
#include "logger-utility.h"
#include "FileType.h"

#include <glite/data/srm/util/PrepareToGet.h>
#include <glite/data/srm/util/types.h>
#include <glite/data/agents/AgentExceptions.h>

#include <assert.h>

namespace glite {
namespace data {
namespace transfer {
namespace urlcopy {

using namespace glite::data::srm::util;
using namespace ChecksumCheckerUtils;

ChecksumChecker_UrlCopy::ChecksumChecker_UrlCopy(
    UrlCopyError::Phase p,
    RefreshTimeUpdater::IUpdateRefreshTime* urt,
    Context::Ptr& ctx,
    bool isSourceScope) :
    ChecksumChecker(p, urt, ctx, isSourceScope)
{
    // EMPTY
}

void ChecksumChecker_UrlCopy::getSourceChecksumFromSrmLs(
    TransferUtils::PathInfo& pinfo,
    UrlCopy_FileStat& filePair)
{
    boost::shared_ptr<FilePair> fpair(dynamic_cast<FilePair*> (new _UrlCopyFilePair(filePair)));
    _getChecksumFromSrmLs(pinfo, fpair);
}

bool ChecksumChecker_UrlCopy::sendPrepareToGetForMissing(
    int finalTimeout,
    int globalTimeout,
    int startTime,
    const char* destSpaceToken)
{
    if (_filesWithMissingChecksum.empty()) {
        return true;
    }

    // Count the SURL-s
    int surlNo = 0;

    for (_FilePairPtrVector_type::iterator it = _filesWithMissingChecksum.begin();
         it != _filesWithMissingChecksum.end();
         ++it)
    {
        assert(*it);

        if (_isSource ? (*it)->src.isSurl : (*it)->dst.isSurl) 
        {
            ++surlNo;
        }
    }

    if (surlNo == 0) 
    {
        INFO_LOG(ChecksumChecker_UrlCopy) << "No SURL with missing checksum found, PrepareToGet for checksum"
            " is not sent.";

        return true;
    }

    // -------------------------------------------------------------------------
    // Create PrepareToGet Request
    // -------------------------------------------------------------------------
    // Create SRM Context
    boost::scoped_ptr<PrepareToGet> request;
    try {
        // Create PrepareToGet Request
        request.reset(PrepareToGet::create(_ctx));
    } catch (const LogicError& exc) {
        // PrepareToGet method is not supported
        _setFinalErrorForAllGridFTPFiles(UrlCopyError::ERROR_INTERNAL_ERROR,
            "Cannot send PrepareToGet");
        return false;
    }

    // Set Request Parameters
    request->files.resize(surlNo);
    unsigned int i = 0;

    for (_FilePairPtrVector_type::iterator it = _filesWithMissingChecksum.begin();
         it != _filesWithMissingChecksum.end();
         ++it)
    {
        assert(*it);
        File& file = (_isSource ? (*it)->src : (*it)->dst);

        if (file.isSurl) {
            request->files[i].surl = file.name;
            ++i;
        }
    }

#ifdef USE_SAFE_CLEANUP
    if(finalTimeout> 0) {
        // Set desired Request Time
        request->desiredTotalRequestTime = SAFE_CLEANUP_MULIPLY_FACTOR * finalTimeout;
    }
    if(globalTimeout> 0) {
        // Set desired Pin Lifetime
        request->desiredPinLifeTime = SAFE_CLEANUP_MULIPLY_FACTOR * globalTimeout;
    }
#endif //USE_SAFE_CLEANUP
    request->transferParameters.accessPattern = AccessPattern::PROCESSING_MODE;
    request->transferParameters.connectionType = ConnectionType::LAN;
    request->transferParameters.protocol = PROTOCOL_GSIFTP;

    if (destSpaceToken && 0 != strlen(destSpaceToken))
        request->targetSpaceToken = destSpaceToken;

    UrlCopyError::Category c;
    std::string m;

    // Send the request
    if (!TransferUtils::SendPrepareToGet(_ctx, _updater, request, c, m)) {
        _setFinalErrorForAllGridFTPFiles(c, m);
        return false;
    }

    c = UrlCopyError::ERROR_GENERAL_FAILURE;
    int timeout = (finalTimeout > 0 ? finalTimeout : DEFAULT_GET_TIMEOUT);

    if (!TransferUtils::PollPrepareToGet(_ctx, _updater, request, timeout, startTime,
        MAX_GET_SLEEP_TIME, c, m)) {
        _setFinalErrorForAllGridFTPFiles(c, m);
        return false;
    }

    return true;
}

void ChecksumChecker_UrlCopy::getDestinationChecksumFromSrmLs(UrlCopy_FileStat& filePair)
{
    UrlCopyError::Category c = UrlCopyError::SUCCESS;
    std::string m;
    TransferUtils::PathInfo pinfo;

    if (filePair.mm_destination.mm_type == FileType::SURL) {
        if (!TransferUtils::PathExists(
            _updater, _ctx, filePair.mm_destination.mm_name, true, c, m, pinfo)) 
        {
            ERROR_LOG(ChecksumChecker) << "srmLs failed for file " 
                << filePair.mm_destination.mm_name << " [" << c << "]: " << m;
        }
    }

    boost::shared_ptr<FilePair> fpair(dynamic_cast<FilePair*> (new _UrlCopyFilePair(filePair)));
    _getChecksumFromSrmLs(pinfo, fpair);
}

void ChecksumChecker_UrlCopy::_setFinalErrorForAllGridFTPFiles(
    UrlCopyError::Category c,
    const std::string& m)
{
    _setErrorForAllGridFTPFiles(c, m);

    for (_FilePairPtrVector_type::iterator it = _filesWithMissingChecksum.begin();
         it != _filesWithMissingChecksum.end();
         ++it)
    {
        assert(*it);
        (*it)->setFinalError(_getScope(), _phase, c, m);
    }
}

ChecksumChecker_UrlCopy::_UrlCopyFilePair::~_UrlCopyFilePair()
{
    // EMPTY
}

ChecksumChecker_UrlCopy::_UrlCopyFilePair::_UrlCopyFilePair(UrlCopy_FileStat& filePair) :
    FilePair(),
    _filePair(filePair)
{
    compare = static_cast<bool> (filePair.mm_compare_checksum);
    userAlg = filePair.mm_checksum_user_algorithm;
    userVal = filePair.mm_checksum_user_value;
    src.algorithm = filePair.mm_src_checksum_algorithm;
    src.value = filePair.mm_src_checksum_value;
    dst.algorithm = filePair.mm_dst_checksum_algorithm;
    dst.value = filePair.mm_dst_checksum_value;
    src.name = filePair.mm_source.mm_name;
    dst.name = filePair.mm_destination.mm_name;
    src.turl = filePair.mm_source.mm_tx_name;
    dst.turl = filePair.mm_destination.mm_tx_name;

    src.isSurl = ((static_cast<FileType::Value>(filePair.mm_source.mm_type) == FileType::SURL));
    dst.isSurl = ((static_cast<FileType::Value>(filePair.mm_destination.mm_type) == FileType::SURL));
}

void ChecksumChecker_UrlCopy::_UrlCopyFilePair::setFileError(
    UrlCopyError::Scope s,
    UrlCopyError::Phase p,
    UrlCopyError::Category c,
    const std::string& m)
{
    if (s == UrlCopyError::SCOPE_SOURCE)
        UrlCopyUtils::SetScopeError(_filePair.mm_source, s, p, c, m);
    else if (s == UrlCopyError::SCOPE_DESTINATION)
        UrlCopyUtils::SetScopeError(_filePair.mm_destination, s, p, c, m);
    else
        assert(0);
}

void ChecksumChecker_UrlCopy::_UrlCopyFilePair::setFinalError(
    UrlCopyError::Scope s,
    UrlCopyError::Phase p,
    UrlCopyError::Category c,
    const std::string& m)
{
    UrlCopyUtils::SetFinalError(_filePair, s, p, c, m);
}

} // namespace urlcopy
} // namespace transfer
} // namespace data
} // namespace glite
