/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Checksum checking logic for Url copy
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: ChecksumChecker_UrlCopy.h,v 1.2 2009/08/04 17:49:05 molnarzs Exp $
 *  Release: $Name:  $
 */
#ifndef GLITE_DATA_TRANSFER_URL_COPY_URL_CHECKSUM_CHECKER_URLCOPY_H_
#define GLITE_DATA_TRANSFER_URL_COPY_URL_CHECKSUM_CHECKER_URLCOPY_H_

#include "ChecksumChecker.h"

namespace glite {
namespace data {
namespace transfer {
namespace urlcopy {

class UrlCopy_FileStat;

namespace TransferUtils {
class PathInfo;
} // TransferUtils

/**
 * Checksum checking logic specialized for URL copy.
 */
class ChecksumChecker_UrlCopy: public ChecksumChecker
{
public:
    /**
     * Constructor.
     *
     * @param p The phase in which the object is used. The phase cannot be changed during the
     *          lifetime of the object.
     * @param urt RefreshTimeUpdater class for the SRM method calls and gridFTP operations.
     * @param ctx The actual application context.
     * @param isSourceScope true: The object is used in source scope, else in destination scope.
     */
    ChecksumChecker_UrlCopy(
        UrlCopyError::Phase p,
        RefreshTimeUpdater::IUpdateRefreshTime* urt,
        Context::Ptr& ctx,
        bool isSourceScope);

    /**
     * The method fetches the checksum information from the data returned by a previous srmLs call.
     * It inserts a source/destination file pair into the ChecksumChecker system.
     *
     * Basically, it creates a proper _UrlCopyFilePair object and calls the general
     * _getChecksumFromSrmLs to do the real work.
     *
     * @param pinfo The information returned by the srmLs call, about the source file.
     * @param filePair The source/destination file pair (and the under-laying stat structure)
     */
    void getSourceChecksumFromSrmLs(TransferUtils::PathInfo& pinfo, UrlCopy_FileStat& filePair);

    /**
     * If SrmLs does not provide checksum information, we need to use gridFTP. GridFTP requires a
     * TURL. This method sends an SrmPrepareToGet call for the files with missing checksum and
     * obtaines the TURL in that way.
     *
     * @param finalTimeout Timeout of the SrmPrepareToGet call
     * @param globalTimeout Timeout for the whole transfer
     * @param startTime Start time of the operation (to be able to calculate global timeout, etc.)
     * @param destSpaceToken The destination space token (NULL if not required)
     *
     * @return true of the operation is successful else false
     */
    bool sendPrepareToGetForMissing(
        int finalTimeout,
        int globalTimeout,
        int startTime,
        const char* destSpaceToken);

    /**
     * The method obtains the checksum information for the destination file. The difference between
     * this and getSourceChecksumFromSrmLs is that getSourceChecksumFromSrmLs reuses the result
     * of a previous SrmLs call on the source file, while this method will executes the SrmLs
     * on the destination, then calls the _getChecksumFromSrmLs.
     *
     * @param pinfo The information returned by the srmLs call, about the source file.
     * @param filePair The source/destination file pair (the stat itself)
     */
    void getDestinationChecksumFromSrmLs(UrlCopy_FileStat& filePair);

private:
    /**
     * In UrlCopy case, the checksum error is also final error for the job. We need to set it in
     * the stat-s. This method sets this condition for each file pairs: sets both file and final
     * error information.
     *
     * @param c The error category
     * @param m The error message.
     */
    void _setFinalErrorForAllGridFTPFiles(UrlCopyError::Category c, const std::string& m);

    /**
     * Specialization of ChecksumCheckerUtils::FilePair in the UrlCopy case.
     */
    class _UrlCopyFilePair: public ChecksumCheckerUtils::FilePair
    {
    public:
        /**
         * Constructor.
         *
         * @param filePair The underlaying stat of the transfer. The reference to it is recorded.
         */
        _UrlCopyFilePair(UrlCopy_FileStat& filePair);

        /** Destructor */
        virtual ~_UrlCopyFilePair();

        /** See parent */
        virtual void setFileError(
            UrlCopyError::Scope s,
            UrlCopyError::Phase p,
            UrlCopyError::Category c,
            const std::string& m);

        /** See parent */
        virtual void setFinalError(
            UrlCopyError::Scope s,
            UrlCopyError::Phase p = UrlCopyError::PHASE_UNDEF,
            UrlCopyError::Category c = UrlCopyError::SUCCESS,
            const std::string& m = "");

    private:
        /** The under-laying stat */
        UrlCopy_FileStat& _filePair;
    };
};

} // namespace urlcopy
} // namespace transfer
} // namespace data
} // namespace glite

#endif //GLITE_DATA_TRANSFER_URL_COPY_URL_CHECKSUM_CHECKER_URLCOPY_H_
