/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Checksum retrieval with gridFTP
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: ChecksumGetter_gridFTP.cpp,v 1.3 2009/10/26 16:32:20 molnarzs Exp $
 *  Release: $Name:  $
 */
#include "ChecksumGetter_gridFTP.h"
#include "gridftp-utility.h"
#include "logger-utility.h"
#include "FileType.h"
#include "gridftp-utility.h"
#include "Checksum.h"
#include <cstring>
#include <boost/lexical_cast.hpp>

namespace glite {
namespace data {
namespace transfer {
namespace urlcopy {

ChecksumGetter_gridFTP::~ChecksumGetter_gridFTP()
{
    // EMPTY
}

void ChecksumGetter_gridFTP::getChecksumFromGridFTP(
    ChecksumCheckerUtils::FilePair& filePair,
    ChecksumCheckerUtils::File& file,
    const unsigned int timeout,
    RefreshTimeUpdater::IUpdateRefreshTime* updater,
    UrlCopyError::Category& c,
    std::string& m)
{
    INFO_LOG(ChecksumGetter_gridFTP) << "Calling gridftp for file " << file.name;
    // Get the source File Type
    // NOTE: explicitly using FileType::get on source transfer name rather than
    // m_stat->mm_urlcopy.mm_source.mm_type - in this way it should properly
    // fail for tests (stub SRMs return URLs) - and work for real gsiftp addresses
    FileType::Value src_type = FileType::get(file.isSurl ? file.turl : file.name);

    switch (src_type)
    {
    case FileType::TURL:
        _executeGridFTP(filePair, file, timeout, updater, c, m);
        break;
    case FileType::SURL:
    case FileType::URL:
    case FileType::LOCAL:
    case FileType::SFN:
        // Checksum not supported for these file types
        m = "Checksum operation not supported for file type " + boost::lexical_cast<std::string>(
            src_type) + " (file: '" + file.turl + "')";

        c = UrlCopyError::ERROR_CHECKSUM_UNKNOWN;
        break;
    default:
        // Unsupported source file type
        c = UrlCopyError::ERROR_CONSISTENCY;
        m = std::string("Unknown source file type. Invalid stat file? (file: '") + file.turl + "')";
        break;
    }
}

void ChecksumGetter_gridFTP::_executeGridFTP(
    ChecksumCheckerUtils::FilePair& filePair,
    ChecksumCheckerUtils::File& file,
    const unsigned int timeout,
    RefreshTimeUpdater::IUpdateRefreshTime* updater,
    UrlCopyError::Category& c,
    std::string& m)
{
    // Set up the return value (error category / message pair)
    c = UrlCopyError::ERROR_CHECKSUM_UNKNOWN;
    // Obtain the checksum
    std::string algorithm = this->_SelectChecksumAlgorithmForGridFTP(filePair, file);
    std::string cksm = gridftp_checksum(file.turl, algorithm, timeout, c, m, updater);

    if (c == UrlCopyError::SUCCESS) {
        INFO_LOG(ChecksumGetter_gridFTP) << "gridftp_checksum returned '" << algorithm << ":"
                << cksm << "'";

        if (cksm.empty()) {
            c = UrlCopyError::ERROR_CHECKSUM_UNKNOWN;
        } else {
            Checksum c(algorithm, cksm);
            c.copyChecksumAlgorithm(file.algorithm);
            c.copyChecksumValue(file.value);
        }
    }
}

const char* ChecksumGetter_gridFTP::_SelectChecksumAlgorithmForGridFTP(
    const ChecksumCheckerUtils::FilePair& filePair,
    const ChecksumCheckerUtils::File& file)
{
    if (file.algorithm && (strlen(file.algorithm) > 0))
        return file.algorithm;
    else if (filePair.isUserProvidedChecksum)
        return filePair.userAlg;
    else
        return Checksum::DefaultChecksumAlgorithm().c_str();
}

} // namespace urlcopy
} // namespace transfer
} // namespace data
} // namespace glite
