/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: gridftp helper methods 
 *   refactored from previous transfer-url-copy (Peter Kunszt) and lcg_util 
 *   (Jean-Philippe Baud)
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: GridFtpPatterns.cpp,v 1.2 2008/10/22 08:51:37 ptedesco Exp $
 *  Release: $Name:  $
 */

#include "GridFtpPatterns.h" 

using namespace glite::data::agents;
using namespace glite::data::transfer::urlcopy;

/**
 * singleton instance
 */
GridFtpPatterns& GridFtpPatterns::instance(){
    static GridFtpPatterns s_instance;
    return s_instance;
}

/**
 * translateGridFtpError 
 * translate a gridftp error message into an UrlCopyError category and scope
 * @param msg [IN] the error message
 * @param category [OUT] the error category (GRIDFTP if no better match is 
 * found) 
 * @param scope [OUT] the error scope (S_TRANSFER if no better match is 
 * found)
 */
void GridFtpPatterns::translateGridFtpError(
    const std::string&      msg,
    UrlCopyError::Category& category,
    UrlCopyError::Scope&    scope){
    // initialize output values to most generic values
    scope    = UrlCopyError::SCOPE_TRANSFER;
    // search for a match
    const XmlPatternsList::PatternMatch * match = getMatch(msg,category);
    if (0 != match){
        std::string scopeString;
        if (match->getProperty("scope",scopeString)){
            ScopeMap::const_iterator s_it = m_scopes.find(scopeString);
            if (m_scopes.end() != s_it){
                scope = s_it->second;
            }
        }
    }
}

/**
 * categoryFromMessage
 * translate a gridftp error message returning only the corresponding category
 * @param msg [IN] the error message
 * @return the error category (GRIDFTP if no better match is found)
 */
UrlCopyError::Category GridFtpPatterns::categoryFromMessage(
    const std::string& msg){
    UrlCopyError::Category category;
    getMatch(msg,category);
    return category;
}


/**
 * getMatch
 * get the match object corresponding to msg
 * @param msg [IN] the error message
 * @param category [OUT] the error category (GRIDFTP if no better match is 
 * found) 
 * @return PatternMatch if found, null otherwise
 */
const XmlPatternsList::PatternMatch * GridFtpPatterns::getMatch(
    const std::string&      msg,
    UrlCopyError::Category& category){
    // initialize output values to most generic values
    category = UrlCopyError::ERROR_GRIDFTP;
    // search for a match
    for (CategoryMap::const_iterator c_it = m_categories.begin();c_it != m_categories.end();++c_it){
        const XmlPatternsList::PatternMatch * match = m_patterns.getPatternMatch(msg,c_it->first);
        if (0 != match){
            category = c_it->second;
            return match;
        }
    }
    return 0;
}

/**
 * constructor
 */
GridFtpPatterns::GridFtpPatterns():m_patterns("glite-data-transfer-url-copy.patterns.xml"){
    // initialize the categories map
    m_categories["USER"]                    = UrlCopyError::ERROR_USER;
    m_categories["PERMISSION"]              = UrlCopyError::ERROR_PERMISSION;
    m_categories["FILE_EXISTS"]             = UrlCopyError::ERROR_FILE_EXISTS;
    m_categories["INVALID_PATH"]            = UrlCopyError::ERROR_INVALID_PATH;
    m_categories["NO_SPACE_LEFT"]           = UrlCopyError::ERROR_NO_SPACE_LEFT;
    m_categories["GRIDFTP"]                 = UrlCopyError::ERROR_GRIDFTP;
    m_categories["NETWORK"]                 = UrlCopyError::ERROR_NETWORK;
    m_categories["INVALID_SIZE"]            = UrlCopyError::ERROR_INVALID_SIZE;
    m_categories["REQUEST_FAILURE"]         = UrlCopyError::ERROR_REQUEST_FAILURE;
    m_categories["CONNECTION"]              = UrlCopyError::ERROR_CONNECTION;
    m_categories["HTTP_TIMEOUT"]            = UrlCopyError::ERROR_HTTP_TIMEOUT;
    m_categories["REQUEST_TIMEOUT"]         = UrlCopyError::ERROR_REQUEST_TIMEOUT;
    m_categories["TRANSFER_TIMEOUT"]        = UrlCopyError::ERROR_TRANSFER_TIMEOUT;
    m_categories["MARKERS_TIMEOUT"]         = UrlCopyError::ERROR_MARKERS_TIMEOUT;
    m_categories["INTERNAL_ERROR"]          = UrlCopyError::ERROR_INTERNAL_ERROR;
    m_categories["CONSISTENCY"]             = UrlCopyError::ERROR_CONSISTENCY;
    m_categories["ABORTED"]                 = UrlCopyError::ERROR_ABORTED;
    m_categories["GENERAL_FAILURE"]         = UrlCopyError::ERROR_GENERAL_FAILURE;
    m_categories["LOST"]                    = UrlCopyError::ERROR_LOST;
    m_categories["STORAGE_INTERNAL_ERROR"]  = UrlCopyError::ERROR_STORAGE_INTERNAL_ERROR;
    m_categories["SRM_COPY"]                = UrlCopyError::ERROR_SRM_COPY;
    m_categories["LOCALITY"]                = UrlCopyError::ERROR_LOCALITY;
    // initialize the scopes map
    // we only want to set the scope if it's source or destination, otherwise 
    // leave is as it is
    m_scopes["source"]      = UrlCopyError::SCOPE_SOURCE;
    m_scopes["destination"] = UrlCopyError::SCOPE_DESTINATION;
}

/**
 * destructor
 */
GridFtpPatterns::~GridFtpPatterns(){
}


