/*
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Mmap class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *  Version info: $Id: Mmap.cpp,v 1.5 2008/10/20 15:19:46 ptedesco Exp $
 *  Release: $Name:  $
 */
#include "Mmap.h"

#include <glite/data/agents/extra/Handle.h>
#include <glite/data/agents/AgentExceptions.h>

#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <errno.h>

using namespace glite::data::transfer::urlcopy;

using glite::data::agents::Handle;
using glite::data::agents::LogicError;
using glite::data::agents::RuntimeError;
using glite::data::agents::PosixError;

/*
 * Mmap
 *
 * Constructor
 */
Mmap::Mmap(const std::string& file_name):m_fileName(file_name),m_buffer(0),m_size(0){
}

/*
 * ~Mmap
 *
 * Destructor
 */
Mmap::~Mmap(){
    // If the file buffer is open, close it
    this->close(true);
}

/*
 * buffer
 *
 * Return the Mmap buffer
 */
void * Mmap::buffer()  /*throw (LogicError)*/{
    // Check Preconditions
    if(0 == m_buffer){
        // Error
        throw LogicError("not initialized");
    }
    return m_buffer;
}

/*
 * size
 *
 * Return the buffer size
 */
unsigned long Mmap::size()  /*throw (LogicError)*/{
    // Check Preconditions
    if(0 == m_buffer){
        // Error
        throw LogicError("not initialized");
    }
    return m_size;    
}

/*
 * create
 *
 * Create a new file
 */
void * Mmap::create(size_t len, bool shared) /*throw (LogicError, RuntimeError)*/ {
    // Check Preconditions
    if(len <= 0){
        // Error
        throw LogicError("invalid length specified");
    }
    if(0 != m_buffer){
        // Error
        throw LogicError("already open");
    }
    
    m_buffer = 0;
    m_size   = 0;
    
    Handle fh;
    // Create a new file
// TODO: Use SHM    
//#ifdef HAVE_SHM_OPEN  
//    // Check that the files has an initial slash (/) and contain no embedded
//    // slashes
//    fh = ::shm_open(m_fileName.c_str(), O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
//#else
    fh = ::open(m_fileName.c_str(), O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
//#endif
    if(fh.isNull()){
        std::string reason = std::string("error creating file for memmap ") + m_fileName;
        throw PosixError(reason,errno);
    }
    // Set the file size
    int r = ::ftruncate(fh(), len);
    if(0 != r){
        throw PosixError("Error setting file size through ftruncate",errno);
    }
    
    // Open Mmap File
    int flag = (true == shared)?MAP_SHARED:MAP_PRIVATE;
    void * m = ::mmap(0,len, PROT_WRITE | PROT_READ, flag, fh(), 0);
    if(MAP_FAILED == m) {
        this->unlink(true);
        throw PosixError("mmap failed",errno);
    }
    if(0 == m) {
        this->unlink(true);
        throw RuntimeError("null pointer returned by mmap");
    }
    // Set the memory buffer and size
    m_buffer = m;
    m_size   = len;
    
    // Return the buffer
    return m_buffer;
}

/*
 * open
 *
 * Open an existing file
 */
void * Mmap::open(size_t& len, bool read_only, bool shared) /*throw (LogicError, RuntimeError)*/ {
    if(0 != m_buffer){
        // Error
        throw LogicError("already open");
    }
    m_buffer = 0;
    m_size   = 0;

    Handle fh;
    // Create a new file
    int prot_flag = (true == read_only)?O_RDONLY:O_RDWR;
// TODO: Use SHM
//#ifdef HAVE_SHM_OPEN  
//    // Check that the files has an initial slash (/) and contain no embedded
//    // slashes
//    fh = ::shm_open(m_fileName.c_str(), prot_flag);
//#else
    fh = ::open(m_fileName.c_str(), prot_flag);
//#endif
    if(fh.isNull()){
        // Error
        std::string reason = std::string("error creating file for memmap ") + m_fileName;
        throw PosixError(reason,errno);
    }
    // Get the file size
    struct stat stat_buf;
    int r = ::fstat(fh(), &stat_buf);
    if((-1 == r) || (stat_buf.st_size <= 0)){
        throw PosixError("error getting file stats",errno);
    }
    
    // Open Mmap File
    prot_flag = (true == read_only)?PROT_READ:(PROT_WRITE | PROT_READ);
    int flag = (true == shared)?MAP_SHARED:MAP_PRIVATE;
    void * m = ::mmap(0,stat_buf.st_size, prot_flag, flag, fh(), 0);
    if(MAP_FAILED == m) {
        throw PosixError("mmap failed",errno);
    }
    if(0 == m){
        throw RuntimeError("null pointer returned by mmap");
    }
    // Set the memory buffer and size
    m_buffer = m;
    len      = stat_buf.st_size;
    m_size   = len;
    
    // Return the buffer
    return m_buffer;    
}

/*
 * close
 *
 * Close the file buffer
 */
void Mmap::close(bool nothrow) /*throw (LogicError, RuntimeError)*/ {
    // Check if the buffer is set
    if(0 != m_buffer){
        int r = ::munmap(m_buffer,m_size);
        if((0 != r) && (false == nothrow)){
            throw PosixError("munmap failed",errno);
        }
    }
    m_buffer = 0;
    m_size   = 0;
}
    
/*
 * unlink
 *
 * Unlink the file
 */
void Mmap::unlink(bool nothrow) /*throw (LogicError, RuntimeError)*/ {
    int r;
    // Delete the file
// TODO: Use SHM
//#ifdef HAVE_SHM_OPEN  
//    // Check that the files has an initial slash (/) and contain no embedded
//    // slashes
//    r = ::shm_unlink(m_fileName.c_str());
//#else
    r = ::unlink(m_fileName.c_str());
//#endif
    if((0 != r) && (false == nothrow)){
        throw PosixError("failed to unlink the file",errno);
    }
}
