/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: SrmCopyCore class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: SrmCopyCore.cpp,v 1.32 2009/07/30 19:32:55 molnarzs Exp $
 *  Release: $Name:  $
 */
#include "SrmCopyCore.h"
#include "FileType.h"
#include "logger-utility.h"
#include "stat-utility.h"
#include "sighandler-utility.h"
#include "gridftp-utility.h"
#include "ChecksumChecker_SrmCopy.h"
#include "SrmCopyUtils.h"
#include <math.h>
#include <glite/data/srm/util/Context.h>
#include <glite/data/srm/util/SrmPing.h>
#include <glite/data/srm/util/SrmCopy.h>
#include <glite/data/srm/util/SrmStat.h>
#include <glite/data/srm/util/PrepareToGet.h>

#include <glite/data/agents/url-utility.h>
#include <glite/data/agents/surl-utility.h>
#include <glite/data/agents/extra/utility.h>
#include <glite/data/agents/cred/cred-utility.h>

#include <string>
#include <sstream>

#include <string.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <libgen.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <boost/scoped_ptr.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>

using namespace glite::data::transfer::urlcopy;
using namespace glite::data::srm::util;

using glite::data::agents::AgentException;

//------------------------------------------------------------------------------
// Enable this flag in order to set the request lifetimes so the SRM would abort
// the request and do an internal cleanup even in case transfer-url-copy is
// not able to properly complete the clenaup procedue
//#define USE_SAFE_CLEANUP
//------------------------------------------------------------------------------

namespace {
    //------------------------------------------------------------------------------
    // Constants
    //------------------------------------------------------------------------------
    const unsigned int DEFAULT_HTTP_TIMEOUT = 40;
    const unsigned int DEFAULT_COPY_TIMEOUT = 1800;
    const unsigned int DEFAULT_COPYDONE_TIMEOUT = 180;
    const unsigned int DEFAULT_COPY_REFRESH_TIMEOUT = 180;
    const unsigned int DEFAULT_COPY_SLEEP_TIME = 30;
    const unsigned int MAX_COPY_SLEEP_TIME = 30; // bug #39492
    const unsigned int SAFE_CLEANUP_MULIPLY_FACTOR = 3;

    const unsigned int DEFAULT_GET_SLEEP_TIME = 5;

} // end global namespace

//------------------------------------------------------------------------------
// SrmCopyCore Methods
//------------------------------------------------------------------------------

/*
 * SrmCopyCore
 *
 * Constructor
 */
SrmCopyCore::SrmCopyCore(const std::string & id)
: TransferCore(id),
m_stat(0) {
    prepareDependencies_production_.reset(new PrepareDependencies_(*this));
    completeDependencies_production_.reset(new CompleteDependencies_(*this));

}

/*
 * ~SrmCopyCore
 *
 * Destructor
 */
SrmCopyCore::~SrmCopyCore() {

    std::vector<transfer_completed *>::iterator it;
    for (it = tr_completed.begin(); it < tr_completed.end(); it++) {
        if (*it) {
	    msg_ifce::getInstance()->set_tr_timestamp_complete(*it, msg_ifce::getInstance()->getTimestamp());		     
            msg_ifce::getInstance()->SendTransferFinishMessage(*it);
            delete *it;
        }
    }
    tr_completed.clear();
}

/**
 * UpdateRefreshTime
 */
void SrmCopyCore::UpdateRefreshTime() {
    time(&(m_stat->mm_common.mm_last_refresh_time));
}

/*
 * stat
 *
 * Return the Stat Object
 */
SrmCopyStat& SrmCopyCore::stat() /*throw (RuntimeError, LogicError)*/ {
    // Initialize (if not already initialized)
    initialize();
    if (0 == m_stat) {
        throw LogicError("null stat object");
    }
    return *m_stat;
}

/*
 * initialize
 *
 * Initialize the Stat structure from the Mmap file
 */
void SrmCopyCore::initialize() /*throw (RuntimeError, LogicError)*/ {
    if (0 != m_stat) {
        // Already initialized
        return;
    }
    try {
        // Initialize Mmap
        TransferCore::initialize();
        // Get the pointer to the Stat buffer
        m_stat = reinterpret_cast<SrmCopyStat*> (m_mmap->buffer());
        // Check if the Stat id is correct
        if (m_stat->mm_common.mm_id != SRMCOPY_STAT_ID) {
            throw LogicError("Not an SrmCopy transfer request");
        }

        // Set all the timeouts to the default
        // ---------------------------------------------------------------------
        if (m_stat->mm_srmcopy.mm_http_timeout < 0) {
            m_stat->mm_srmcopy.mm_http_timeout = DEFAULT_HTTP_TIMEOUT;
        }
        if (m_stat->mm_srmcopy.mm_copy_timeout < 0) {
            m_stat->mm_srmcopy.mm_copy_timeout = DEFAULT_COPY_TIMEOUT;
        }
        if (m_stat->mm_srmcopy.mm_copy_timeout < 0) {
            m_stat->mm_srmcopy.mm_refresh_timeout = DEFAULT_COPY_REFRESH_TIMEOUT;
        }
    } catch (...) {
        m_mmap.reset();
        m_stat = 0;
        throw;
    }

    // initialize the syslog helper
    ContextLogInfo::user_tag = m_stat->mm_common.mm_user_tag;
    ContextLogInfo::tx_id = id();
}

/*
 * prepare
 *
 * Preapare the source and destination files for the transfer
 */
SrmCopyCore::PrepareDependencies_::PrepareDependencies_(SrmCopyCore& self)
: self_(self) {
    // EMPTY
}

SrmCopyCore::PrepareDependencies_::~PrepareDependencies_() {
    // EMPTY
}

bool SrmCopyCore::PrepareDependencies_::pingSource(UrlCopyError::Category& c, std::string& m) {
    bool res = self_.pingSource(c, m);
    return res;
}

bool SrmCopyCore::PrepareDependencies_::pingDestination(UrlCopyError::Category& c, std::string& m) {
    bool res = self_.pingDestination(c, m);
    return res;
}

void SrmCopyCore::PrepareDependencies_::prepareSource() {
    self_.prepareSource();
}

void SrmCopyCore::PrepareDependencies_::prepareDestination() {
    self_.prepareDestination();
}

void SrmCopyCore::prepare() {
    prepare_(*prepareDependencies_production_);
}

void SrmCopyCore::prepare_(SrmCopyCore::PrepareDependencies_& deps) {
    // Check of the instance is initialized, otherwise initialize it from the
    // status file
    initialize();
    // Get Proxy DN
    std::string user_dn;
    try {
        user_dn = glite::data::agents::cred::get_proxy_dn();
    } catch (const RuntimeError& exc) {
        user_dn = "(not available)";
    }

    char hostname[1024];
    hostname[1023] = '\0';
    gethostname(hostname, 1023);

    info() << "Transfer mode: srmcopy";
    info() << "Transfer ID  : " << this->id();
    info() << "User DN      : " << user_dn;
    info() << "User Descr.  : " << m_stat->mm_common.mm_user_tag;
    unsigned int i = 0;
    info() << "Source SRM [" << m_stat->mm_srmcopy.mm_src_srm_version << "]: " << m_stat->mm_srmcopy.mm_src_srm_endpoint;
    info() << "Dest.  SRM [" << m_stat->mm_srmcopy.mm_dest_srm_version << "]: " << m_stat->mm_srmcopy.mm_dest_srm_endpoint;

    for (; i < m_stat->mm_num_files; ++i) {
        info() << i << " Source     : " << m_stat->mm_files[i].mm_src_name;
        info() << i << " Destination: " << m_stat->mm_files[i].mm_dest_name;
        tr_completed.push_back(new transfer_completed);
        msg_ifce::getInstance()->set_tr_timestamp_start(tr_completed[i], msg_ifce::getInstance()->getTimestamp());	
        msg_ifce::getInstance()->set_agent_fqdn(tr_completed[i], hostname);
        msg_ifce::getInstance()->set_t_channel(tr_completed[i], this->id());
        std::stringstream out;
        out << i;
        std::string temp_id = string(this->id());
        temp_id = temp_id + "_" + out.str();
        msg_ifce::getInstance()->set_transfer_id(tr_completed[i], temp_id);
	msg_ifce::getInstance()->set_channel_type(tr_completed[i], "srmcopy");

        msg_ifce::getInstance()->set_source_srm_version(tr_completed[i], m_stat->mm_srmcopy.mm_src_srm_version);
        msg_ifce::getInstance()->set_destination_srm_version(tr_completed[i], m_stat->mm_srmcopy.mm_dest_srm_version);

        msg_ifce::getInstance()->set_source_url(tr_completed[i], m_stat->mm_files[i].mm_src_name);
        msg_ifce::getInstance()->set_dest_url(tr_completed[i], m_stat->mm_files[i].mm_dest_name);

        msg_ifce::getInstance()->set_source_hostname(tr_completed[i], m_stat->mm_files[i].mm_src_name);
        msg_ifce::getInstance()->set_dest_hostname(tr_completed[i], m_stat->mm_files[i].mm_dest_name);

        msg_ifce::getInstance()->set_vo(tr_completed[i], vo_name, this->id());
        msg_ifce::getInstance()->set_source_site_name(tr_completed[i], src_site);
        msg_ifce::getInstance()->set_dest_site_name(tr_completed[i], dest_site);  	

        // send the transfer started message before the status is set to running
        msg_ifce::getInstance()->SendTransferStartMessage(tr_completed[i]);
    }

    // Precondition checks
    // -------------------------------------------------------------------------
    // Check the number of files to transfer
    if (m_stat->mm_num_files <= 0) {
        std::stringstream ss;
        ss << "prepare called for an srmcopy request with an invalid "
                << " number of files["
                << m_stat->mm_num_files
                << "]";
        std::string reason = ss.str();
        error() << reason;
        tr_completed.push_back(new transfer_completed);
        msg_ifce::getInstance()->set_transfer_error_message(tr_completed[0], reason);
        msg_ifce::getInstance()->set_failure_phase(tr_completed[0], "PREPARATION");
        msg_ifce::getInstance()->set_final_transfer_state(tr_completed[0], "Error"); //---------------OK/Error/Aborted	
        throw LogicError(reason.c_str());
    }
    // Check that the state is IDLE
    if (m_stat->mm_srmcopy.mm_status != UrlCopyStatus::IDLE) {
        // Cannot Call Prepare on an active transfer
        std::stringstream ss;
        ss << "prepare called for a request not in ["
                << UrlCopyStatus::toString(UrlCopyStatus::IDLE)
                << "] state. Current state is ["
                << UrlCopyStatus::toString(
                (UrlCopyStatus::Value)m_stat->mm_srmcopy.mm_status) << "]";
        std::string reason = ss.str();
        error() << reason;
        tr_completed.push_back(new transfer_completed);
        msg_ifce::getInstance()->set_transfer_error_message(tr_completed[0], reason);
        msg_ifce::getInstance()->set_failure_phase(tr_completed[0], "PREPARATION");
        msg_ifce::getInstance()->set_final_transfer_state(tr_completed[0], "Error"); //---------------OK/Error/Aborted	
        throw LogicError(reason.c_str());
    }
    // Check that the source and destination file names are set
    for (i = 0; i < m_stat->mm_num_files; ++i) {
        if (strlen(m_stat->mm_files[i].mm_src_name) == 0) {
            std::stringstream ss;
            ss << "precondition failure: source file [" << i << "] name is empty";
            // Set Source Error
            allFilesError(UrlCopyError::SCOPE_SOURCE, UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_USER, ss.str().c_str());
            // Precondition failure: don't even start the transfer
            finalErrorFromFile(i);
            return;
        }
        if (strlen(m_stat->mm_files[i].mm_dest_name) == 0) {
            std::stringstream ss;
            ss << "precondition failure: destination file [" << i << "] name is empty";
            // Set Destination Error
            allFilesError(UrlCopyError::SCOPE_DESTINATION, UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_USER, ss.str().c_str());
            // Precondition failure: don't even start the transfer
            finalErrorFromFile(i);
            return;
        }
    }
    // The Transfer is preparing
    m_stat->mm_srmcopy.mm_status = UrlCopyStatus::PREPARING;

    for (unsigned int i = 0; i < m_stat->mm_num_files; ++i) {
        msg_ifce::getInstance()->set_time_spent_in_srm_preparation_start(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
    }


    UrlCopyError::Category c;
    std::string m;
    // Ping the Source and destination servers
    if (false == deps.pingSource(c, m)) {
        // Source doesn't respond
        if (0 != m_stat->mm_srmcopy.mm_push_mode) {
            // in push mode, this is a critical error
            allFilesError(UrlCopyError::SCOPE_SOURCE, UrlCopyError::PHASE_PREPARATION, c, m);
            finalError(UrlCopyError::SCOPE_SOURCE, UrlCopyError::PHASE_PREPARATION, c, m);
            return;
        } else {
            warn() << "Cannot contact source SRM: [" << UrlCopyError::categoryToString(c) << "] " << m;
            info() << "Since the transfer is in pull mode, this is not a critical error";
        }
    }
    if (false == deps.pingDestination(c, m)) {
        // destination doesn't respond
        if (0 == m_stat->mm_srmcopy.mm_push_mode) {
            // in pull mode, this is a critical error
            allFilesError(UrlCopyError::SCOPE_DESTINATION, UrlCopyError::PHASE_PREPARATION, c, m);
            finalError(UrlCopyError::SCOPE_DESTINATION, UrlCopyError::PHASE_PREPARATION, c, m);
            return;
        } else {
            warn() << "Cannot contact destination SRM: [" << UrlCopyError::categoryToString(c) << "] " << m;
            info() << "Since the transfer is in push mode, this is not a critical error";
        }
    }
    // Prepare the Source File
    info() << "STATUS:BEGIN:SOURCE - Preparation";
    deps.prepareSource();
    // Check if source preparation is failed
    if (m_stat->mm_srmcopy.mm_error.mm_category != UrlCopyError::SUCCESS) {
        info() << "STATUS:END fail:SOURCE - Preparation";
        return;
    }
    if (finalErrorFromAllFiles(UrlCopyError::SCOPE_SOURCE, UrlCopyError::PHASE_PREPARATION)) {
        return;
    }

    info() << "STATUS:END:SOURCE - Preparation";
    debug() << "Source files are ready";

    // Prepare the Destination File
    info() << "STATUS:BEGIN:DESTINATION - Preparation";
    deps.prepareDestination();
    // Check if destination preparation is failed
    if (m_stat->mm_srmcopy.mm_error.mm_category != UrlCopyError::SUCCESS) {
        info() << "STATUS:END fail:DESTINATION - Preparation";
        return;
    }
    if (finalErrorFromAllFiles(UrlCopyError::SCOPE_DESTINATION, UrlCopyError::PHASE_PREPARATION)) {
        return;
    }
    info() << "STATUS:END:DESTINATION - Preparation";
    debug() << "Destination files are ready";

    // The Transfer is now prepared: update the file Status
    m_stat->mm_srmcopy.mm_status = UrlCopyStatus::READY;
}

/*
 * fileError
 *
 * Set File Error
 */
void SrmCopyCore::fileError(
        unsigned int i,
        UrlCopyError::Scope s,
        UrlCopyError::Phase p,
        UrlCopyError::Category c,
        const std::string& m) throw () {
    if (i < m_stat->mm_num_files) {
        msg_ifce::getInstance()->set_transfer_error_scope(tr_completed[i], UrlCopyError::scopeToString(s));
        msg_ifce::getInstance()->set_transfer_error_message(tr_completed[i], m);
        msg_ifce::getInstance()->set_transfer_error_category(tr_completed[i], UrlCopyError::categoryToString(c));
        msg_ifce::getInstance()->set_failure_phase(tr_completed[i], UrlCopyError::phaseToString(p));
        if (c == UrlCopyError::ERROR_ABORTED)
            msg_ifce::getInstance()->set_final_transfer_state(tr_completed[i], "Aborted"); //---------------OK/Error/Aborted   
        else
            msg_ifce::getInstance()->set_final_transfer_state(tr_completed[i], "Error"); //---------------OK/Error/Aborted   

        SrmCopyUtils::SetFileError(m_stat->mm_files[i], s, p, c, m);
    } else {
        warn() << "file index [" << i << "] out of bound: ignore error";
    }
}

/*
 * allFilesError
 *
 * Set the same Error on all the files
 */
void SrmCopyCore::allFilesError(UrlCopyError::Scope s, UrlCopyError::Phase p, UrlCopyError::Category c, const std::string& m) throw () {
    // Set Error
    for (unsigned int i = 0; i < m_stat->mm_num_files; ++i) {
        fileError(i, s, p, c, m);
    }
}

/**
 * Set final error if all files failed
 * @return true if all files failed
 */
bool SrmCopyCore::finalErrorFromAllFiles(UrlCopyError::Scope s, UrlCopyError::Phase p) throw () {
    typedef std::vector<UrlCopyError::Category> Categories;
    Categories categories;
    for (unsigned int i = 0; i < m_stat->mm_num_files; ++i) {
        UrlCopyError::Category category = (UrlCopyError::Category)m_stat->mm_files[i].mm_error.mm_category;
        if (UrlCopyError::SUCCESS == category) {
            return false;
        }
        if (categories.end() == std::find(categories.begin(), categories.end(), category)) {
            categories.push_back(category);
        }
    }
    if (categories.size() == 0) {
        // should never happen
        finalError(UrlCopyError::SCOPE_AGENT, p, UrlCopyError::ERROR_INTERNAL_ERROR, "All files failed. Error determining category.");
    }
    if (categories.size() == 1) {
        finalError(s, p, categories[0], "All files failed. Check individual file statuses for details.");
    } else {
        finalError(s, p, UrlCopyError::ERROR_GENERAL_FAILURE, "All files failed. Check individual file statuses for details.");
    }
    return true;
}

/*
 * finalError
 *
 * Set Final Error
 */
void SrmCopyCore::finalError(UrlCopyError::Scope s, UrlCopyError::Phase p, UrlCopyError::Category c, const std::string& m) throw () {

    if (m_stat->mm_srmcopy.mm_error.isSet()) {
        info() << "Final error already set.";
        return;
    }

    // Set final Error
    m_stat->mm_srmcopy.mm_error.mm_scope = s;
    m_stat->mm_srmcopy.mm_error.mm_phase = p;
    m_stat->mm_srmcopy.mm_error.mm_category = c;
    copy_message(m_stat->mm_srmcopy.mm_error.mm_message, m.c_str());
    std::string scope = UrlCopyError::scopeToString((UrlCopyError::Scope)m_stat->mm_srmcopy.mm_error.mm_scope);
    error() << "Final error on " << scope << " during " << UrlCopyError::phaseToString((UrlCopyError::Phase)m_stat->mm_srmcopy.mm_error.mm_phase) << " phase: [" << UrlCopyError::categoryToString((UrlCopyError::Category)m_stat->mm_srmcopy.mm_error.mm_category) << "] " << m_stat->mm_srmcopy.mm_error.mm_message;
    boost::to_upper(scope);
    info() << "FINAL:" << scope << ": " << m_stat->mm_srmcopy.mm_error.mm_message;
}

/*
 * finalErrorFromFile
 *
 * Set Final Error from File Error
 */
void SrmCopyCore::finalErrorFromFile(unsigned int i) throw () {
    // Set Error
    if (i < m_stat->mm_num_files) {
        std::stringstream ss;
        ss << "Transfer final error due to file ["
                << i << "] failure:" << m_stat->mm_files[i].mm_error.mm_message;
        finalError((UrlCopyError::Scope)m_stat->mm_files[i].mm_error.mm_scope,
                (UrlCopyError::Phase)m_stat->mm_files[i].mm_error.mm_phase,
                (UrlCopyError::Category)m_stat->mm_files[i].mm_error.mm_category,
                ss.str());
        msg_ifce::getInstance()->set_transfer_error_scope(tr_completed[i], UrlCopyError::scopeToString((UrlCopyError::Scope) m_stat->mm_files[i].mm_error.mm_scope));
        msg_ifce::getInstance()->set_transfer_error_message(tr_completed[i], m_stat->mm_files[i].mm_error.mm_message);
        msg_ifce::getInstance()->set_transfer_error_category(tr_completed[i], UrlCopyError::categoryToString((UrlCopyError::Category)m_stat->mm_files[i].mm_error.mm_category));
        msg_ifce::getInstance()->set_failure_phase(tr_completed[i], UrlCopyError::phaseToString((UrlCopyError::Phase)m_stat->mm_files[i].mm_error.mm_phase));
        msg_ifce::getInstance()->set_final_transfer_state(tr_completed[i], "Error"); //---------------OK/Error/Aborted   

    } else {
        std::stringstream ss;
        ss << "Transfer final error due to an inconnsistency: out of bound "
                "index [" << i << "]";
        finalError(UrlCopyError::SCOPE_AGENT,
                UrlCopyError::PHASE_UNDEF,
                UrlCopyError::ERROR_CONSISTENCY,
                ss.str());
    }
}

/**
 * pingSource
 * Ping Source server
 */
bool SrmCopyCore::pingSource(UrlCopyError::Category& c, std::string& m) /*throw ()*/ {

    if (!m_stat->mm_common.mm_enable_ping_source) {
        info() << "Skipping srmPing on source endpoint.";
        return true;
    }

    bool server_ok = false;
    if ((0 == strlen(m_stat->mm_srmcopy.mm_src_srm_endpoint)) ||
            (0 == strlen(m_stat->mm_srmcopy.mm_src_srm_version))) {
        c = UrlCopyError::ERROR_USER;
        m = "No Source SRM server endpoint specified";
    } else {
        // Create SRM Context
        Context::Ptr ctx = sourceContext();
        // Ping Source SRM
        pingSrm(ctx, c, m, "Source");
        if (c == UrlCopyError::SUCCESS) {
            server_ok = true;
        }
    }
    return server_ok;
}

/**
 * pingDestination
 * Ping Destination server
 */
bool SrmCopyCore::pingDestination(UrlCopyError::Category& c, std::string& m) /*throw ()*/ {

    if (!m_stat->mm_common.mm_enable_ping_dest) {
        info() << "Skipping srmPing on destination endpoint.";
        return true;
    }

    bool server_ok = false;
    if ((0 == strlen(m_stat->mm_srmcopy.mm_dest_srm_endpoint)) ||
            (0 == strlen(m_stat->mm_srmcopy.mm_dest_srm_version))) {
        c = UrlCopyError::ERROR_USER;
        m = "No Destination SRM server endpoint specified";
    } else {
        // Create SRM Context
        Context::Ptr ctx = destinationContext();
        // Ping Source SRM
        pingSrm(ctx, c, m, "Destination");
        if (c == UrlCopyError::SUCCESS) {
            server_ok = true;
        }
    }
    return server_ok;
}

/**
 * prepareSource
 * Preapare the source file for the transfer
 */
void SrmCopyCore::prepareSource() /*throw ()*/ {

    // Utility class to set source preparation times

    class SourcePreparationTimeUpdater {
    public:

        SourcePreparationTimeUpdater(SrmCopyCore * srmCopyCore) : m_SrmCopyCore(srmCopyCore) {
            // Set source preparation start time
            time(&(m_SrmCopyCore->m_stat->mm_srmcopy.mm_src_prep_start_time));
        }

        ~SourcePreparationTimeUpdater() {
            // Set source preparation done time
            time(&(m_SrmCopyCore->m_stat->mm_srmcopy.mm_src_prep_done_time));
        }
    protected:
        SrmCopyCore * m_SrmCopyCore;
    };
    SourcePreparationTimeUpdater sourcePreparationTimeUpdater(this);

    // Create Context
    Context::Ptr ctx = sourceContext();
    ChecksumChecker_SrmCopy checksumChecker(UrlCopyError::PHASE_PREPARATION, this, ctx, true);

    for (unsigned int i = 0; i < m_stat->mm_num_files; ++i) {
        const char * const src_name = m_stat->mm_files[i].mm_src_name;
        UrlCopyError::Category c = UrlCopyError::SUCCESS;
        std::string m;
        // Get the source File Type
        FileType::Value src_type = FileType::get(src_name);
        longlong file_size = 0;
        // Get the source file that should be used for the transfer
        switch (src_type) {
            case FileType::SURL:
            {
                if ((0 == strlen(m_stat->mm_srmcopy.mm_src_srm_endpoint)) ||
                        (0 == strlen(m_stat->mm_srmcopy.mm_src_srm_version))) {
                    fileError(i, UrlCopyError::SCOPE_SOURCE, UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_USER,
                            "No Source SRM server endpoint specified");
                } else {
                    // Check source existence and get the related file size
                    TransferUtils::PathInfo pinfo;
                    bool exists = pathExists(ctx, src_name, c, m, pinfo);
                    if ((false == exists) && (c == UrlCopyError::SUCCESS)) {
                        // Path doesn't exist
                        fileError(i, UrlCopyError::SCOPE_SOURCE, UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_INVALID_PATH,
                                "source file doesn't exist");
                    } else if (c == UrlCopyError::SUCCESS) {
                        // copy file size
                        file_size = pinfo.file_size;

                        msg_ifce::getInstance()->set_timestamp_checksum_source_started(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
                        // Retrieve checksums from SrmLs, and compare if necessary.
                        // If needed, checksums will be retrieved by gridFTP. Collect
                        // those SURL-s here, for later retrieval/comparison.
                        checksumChecker.getSourceChecksumFromSrmLs(pinfo, m_stat->mm_files[i]);

                        msg_ifce::getInstance()->set_timestamp_checksum_source_ended(tr_completed[i], msg_ifce::getInstance()->getTimestamp());


                        // check file locality
                        if (
                                ((FileLocality::NEARLINE == pinfo.file_locality) && (m_stat->mm_common.mm_fail_nearline_files)) ||
                                (FileLocality::LOST == pinfo.file_locality) ||
                                (FileLocality::NONE == pinfo.file_locality) ||
                                (FileLocality::UNAVAILABLE == pinfo.file_locality)
                                ) {
                            c = UrlCopyError::ERROR_LOCALITY;
                            std::stringstream err;
                            err << "Source file [" << src_name << "] locality is " << pinfo.file_locality;
                            fileError(i, UrlCopyError::SCOPE_SOURCE, UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_LOCALITY, err.str());
                        }
                    }
                }
            }
            break;
            case FileType::TURL:
            case FileType::URL:
            {
                // Get GridFtp file size
                file_size = gridftp_getfilesize(src_name, c, m, m_stat->mm_srmcopy.mm_copy_timeout);
            }
            break;
            case FileType::SFN:
                // TODO SFN not supported
            case FileType::LOCAL:
            default:
                // Unsupported source file type
                c = UrlCopyError::ERROR_USER;
                m = "source file type is not supported";
                break;
        }// end switch
        if (c == UrlCopyError::SUCCESS) {
            if (file_size > 0) {
                // Set file size
                m_stat->mm_files[i].mm_src_size = file_size;
            }
        } else {
            // Failed to check file size
            warn() << "Cannot get size for file [" << src_name << "]: [" << UrlCopyError::categoryToString(c) << "] " << m;
            info() << "Not a critical error. The srmcopy transfer request will continue";
        }

        //convert longlong to string
        std::string size_to_string = to_string<longlong > (file_size, std::dec);
        //set the value of file size to the message
        msg_ifce::getInstance()->set_file_size(tr_completed[i], size_to_string.c_str());

    } // End for




    // Get source space token
    if ((0 == strlen(m_stat->mm_common.mm_source_space_token)) && (0 != strlen(m_stat->mm_common.mm_source_token_descr))) {
        debug() << "Resolving source space token from description '" << m_stat->mm_common.mm_source_token_descr << "'";
        unsigned int z = 0;
        for (; z < m_stat->mm_num_files; ++z) {
            msg_ifce::getInstance()->set_srm_space_token_source(tr_completed[z], m_stat->mm_common.mm_source_token_descr);
        }

        if ((0 == strlen(m_stat->mm_srmcopy.mm_src_srm_endpoint)) || (0 == strlen(m_stat->mm_srmcopy.mm_src_srm_version))) {
            std::stringstream ss;
            ss << "Cannot resolve source space token description '" << m_stat->mm_common.mm_source_token_descr <<
                    "' since no source SRM endpoint has been provided";
            error() << ss.str();
            allFilesError(UrlCopyError::SCOPE_SOURCE, UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_USER, ss.str());
            finalError(UrlCopyError::SCOPE_SOURCE, UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_USER, ss.str());
            return;
        }
        UrlCopyError::Category c = UrlCopyError::SUCCESS;
        std::string m;
        std::string space_token = getSpaceToken(ctx, m_stat->mm_common.mm_source_token_descr, c, m, ITokenSelector::Ptr(new SelectFirstToken()));
        if (c != UrlCopyError::SUCCESS) {
            // Fail the preparation
            allFilesError(UrlCopyError::SCOPE_SOURCE, UrlCopyError::PHASE_PREPARATION, c, m);
            finalError(UrlCopyError::SCOPE_SOURCE, UrlCopyError::PHASE_PREPARATION, c, m);
            return;
        }
        if (!space_token.empty()) {
            // Save the token

            copy_token(m_stat->mm_common.mm_source_space_token, space_token.c_str());
        }
    }
    // In SrmCopy case, the TURL-s are not ready in this phase. Get them by a
    // srmPrepareToGet call. Only for the files that require gridFTP checksum
    // operation.
    int start_time = time(NULL);
    bool res = checksumChecker.sendPrepareToGetForMissing(start_time, DEFAULT_GET_TIMEOUT);

    if (res)
        checksumChecker.getMissingChecksumsWithGridFTP(DEFAULT_GET_TIMEOUT);

    checksumChecker.compareSourceChecksums();
}

/*
 * prepareDestination
 *
 * Preapare the destination file for the transfer
 */
void SrmCopyCore::prepareDestination() /*throw ()*/ {

    // Utility class to set source preparation times

    class DestinationPreparationTimeUpdater {
    public:

        DestinationPreparationTimeUpdater(SrmCopyCore * srmCopyCore) : m_SrmCopyCore(srmCopyCore) {
            // Set source preparation start time
            time(&(m_SrmCopyCore->m_stat->mm_srmcopy.mm_dest_prep_start_time));
        }

        ~DestinationPreparationTimeUpdater() {
            // Set source preparation done time
            time(&(m_SrmCopyCore->m_stat->mm_srmcopy.mm_dest_prep_done_time));
        }
    protected:
        SrmCopyCore * m_SrmCopyCore;
    };
    DestinationPreparationTimeUpdater destinationPreparationTimeUpdater(this);

    // Create Context
    Context::Ptr ctx = destinationContext();

    for (unsigned int i = 0; i < m_stat->mm_num_files; ++i) {
        std::string timeout_to_string = to_string<int>(DEFAULT_GET_TIMEOUT, std::dec);
        msg_ifce::getInstance()->set_checksum_timeout(tr_completed[i], timeout_to_string);

        const char * const dest_name = m_stat->mm_files[i].mm_dest_name;
        UrlCopyError::Category c = UrlCopyError::SUCCESS;
        std::string m;
        bool file_exists = false;
        // Get the destination  File Type
        FileType::Value dest_type = FileType::get(dest_name);
        // Check if the dfile already exists
        switch (dest_type) {
            case FileType::SURL:
            {
                if ((0 == strlen(m_stat->mm_srmcopy.mm_dest_srm_endpoint)) ||
                        (0 == strlen(m_stat->mm_srmcopy.mm_dest_srm_version))) {
                    c = UrlCopyError::ERROR_USER;
                    m = "No destination SRM server endpoint specified";
                } else {
                    if (m_stat->mm_common.mm_enable_dir_check) {
                        // Create Destination Directory path
                        try {
                            bool r = prepareDestinationPath(ctx, dest_name);
                            if (false == r) {
                                error() << "Cannot continue preparation for destination file since the destination directory path cannot be created";
                                fileError(i, UrlCopyError::SCOPE_DESTINATION,
                                        UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_INVALID_PATH,
                                        "Cannot create target directory.");
                                continue;
                            }
                        } catch (const UrlCopyError& err) {
                            error() << "Destination path preparation failed: [" << err.description << "]";
                            fileError(i, err.scope, err.phase, err.category, err.description);
                            continue;
                        }
                    } else {
                        info() << "Skipping target directory existence check.";
                    }
                    // Check destination existence
                    file_exists = pathExists(ctx, dest_name, c, m);
                }
            }
            break;
            case FileType::TURL:
            case FileType::URL:
                // Nothing to do
                break;
            case FileType::SFN:
                // TODO SFN not supported
            case FileType::LOCAL:
            default:
                // Unsupported source file type
                c = UrlCopyError::ERROR_USER;
                m = "destination file type is not supported";
                break;
        }// end switch
        if (c == UrlCopyError::SUCCESS) {
            if (true == file_exists && 0 != m_stat->mm_common.mm_overwrite) {
                m_stat->mm_files[i].mm_dest_exists = 0;
                rmDestinationFiles();
            } else {
                m_stat->mm_files[i].mm_dest_exists = 1;
            }
        } else {
            // Failed to check file existence
            warn() << "Cannot check if destination file [" << dest_name << "] exists: [" << UrlCopyError::categoryToString(c) << "] " << m;
            info() << "Not a critical error. The srmcopy transfer request will continue";
        }
        // In any case, set now file to ready
        m_stat->mm_files[i].mm_status = UrlCopyStatus::READY;
    } // End for

    // Get target Space Token
    // -------------------------------------------------------------------------
    if ((0 == strlen(m_stat->mm_common.mm_dest_space_token)) && (0 != strlen(m_stat->mm_common.mm_dest_token_descr))) {

        debug() << "Resolving destination space token from description '" << m_stat->mm_common.mm_dest_token_descr << "'";
        unsigned int z = 0;
        for (; z < m_stat->mm_num_files; ++z) {
            msg_ifce::getInstance()->set_srm_space_token_dest(tr_completed[z], m_stat->mm_common.mm_dest_token_descr);
        }

        if ((0 == strlen(m_stat->mm_srmcopy.mm_dest_srm_endpoint)) || (0 == strlen(m_stat->mm_srmcopy.mm_dest_srm_version))) {
            std::stringstream ss;
            ss << "Cannot resolve destination space token description '" << m_stat->mm_common.mm_dest_token_descr <<
                    "' since no destination SRM endpoint has been provided";
            error() << ss.str();
            allFilesError(UrlCopyError::SCOPE_SOURCE, UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_USER, ss.str());
            finalError(UrlCopyError::SCOPE_SOURCE, UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_USER, ss.str());
            return;
        }

        // Get space token from description
        debug() << "calculating total files size for token selection";
        longlong total_size = 0;
        for (unsigned int i = 0; i < m_stat->mm_num_files; ++i) {
            total_size += m_stat->mm_files[i].mm_src_size;
        }
        debug() << "total size = " << total_size;
        UrlCopyError::Category c = UrlCopyError::SUCCESS;
        std::string m;
        std::string space_token = getSpaceToken(ctx, m_stat->mm_common.mm_dest_token_descr, c, m,
                ITokenSelector::Ptr(new SelectUnusedSpace(total_size, m_stat->mm_num_files)));
        if (c != UrlCopyError::SUCCESS) {
            // Fail the preparation
            allFilesError(UrlCopyError::SCOPE_DESTINATION, UrlCopyError::PHASE_PREPARATION, c, m);
            finalError(UrlCopyError::SCOPE_DESTINATION, UrlCopyError::PHASE_PREPARATION, c, m);
            return;
        }
        if (!space_token.empty()) {
            // Save the token
            copy_token(m_stat->mm_common.mm_dest_space_token, space_token.c_str());
        }
    }

    for (unsigned int i = 0; i < m_stat->mm_num_files; ++i) {
        msg_ifce::getInstance()->set_time_spent_in_srm_preparation_end(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
    }
}

//------------------------------------------------------------------------------
//
// Transfer Methods
//
//------------------------------------------------------------------------------

/**
 * transfer
 * Start the transfer
 */
void SrmCopyCore::transfer() /*throw (RuntimeError, LogicError)*/ {
    // Check of the instance is initialized, otherwise initialize it from the
    // status file
    initialize();

    // Precondition Check
    if (m_stat->mm_srmcopy.mm_status != UrlCopyStatus::READY) {
        // Cannot Call transfer on a request that is not ready
        std::stringstream ss;
        ss << "transfer called for a request not in ["
                << UrlCopyStatus::toString(UrlCopyStatus::READY)
                << "] state. Current state is ["
                << UrlCopyStatus::toString(
                (UrlCopyStatus::Value)m_stat->mm_srmcopy.mm_status) << "]";
        std::string reason = ss.str();
        error() << reason;
        throw LogicError(reason.c_str());
    }

    // Set Transfer start time
    time(&(m_stat->mm_srmcopy.mm_start_time));

    // Set the Transfer Status To Running
    m_stat->mm_srmcopy.mm_status = UrlCopyStatus::RUNNING;
    for (unsigned int i = 0; i < m_stat->mm_num_files; ++i) {
        m_stat->mm_files[i].mm_status = UrlCopyStatus::RUNNING;
    }

    UrlCopyError::Category c;
    std::string m;
    try {
        // Do the Transfer
        info() << "STATUS:BEGIN:TRANSFER";
        info() << "starting srmcopy transfer";

        copyFiles(c, m);


        if (c != UrlCopyError::SUCCESS) {
            info() << "STATUS:END fail:TRANSFER - " << m;
            throw RuntimeError("srmcopy transfer failed");
        }
        info() << "STATUS:END:Transfer";
        info() << "srmcopy transfer completed successfully";

        // Verify if destination file sizes are correct
        debug() << "verify destination file sizes";
        checkDestinationFiles();
        info() << "destination verification completed";

        // The Transfer is completed
        info() << "srmcopy transfer request done";
        m_stat->mm_srmcopy.mm_status = UrlCopyStatus::DONE;
    } catch (const RuntimeError& exc) {
        warn() << exc.what();
        // Set request status
        if (c == UrlCopyError::ERROR_ABORTED) {
            // The Transfer is aborted
            error() << "transfer aborted";
            m_stat->mm_srmcopy.mm_status = UrlCopyStatus::ABORTED;
        } else {
            // The Transfer is failed
            error() << "transfer failed";
            m_stat->mm_srmcopy.mm_status = UrlCopyStatus::FAILED;
        }
    }

    // Set Transfer done time
    time(&(m_stat->mm_srmcopy.mm_done_time));
}

/**
 * copyFiles
 * Start the srmCopy transfer
 */
void SrmCopyCore::copyFiles(UrlCopyError::Category& c, std::string& m) /*throw ()*/ {

    UrlCopyError::Scope s = UrlCopyError::SCOPE_TRANSFER;

    // -------------------------------------------------------------------------
    // Create SrmCopy Request
    // -------------------------------------------------------------------------

    // Create SRM Context
    Context::Ptr ctx(createSrmCtxByCopyMode());

    boost::scoped_ptr<SrmCopy> request;
    try {
        // Create SrmCopy Request
        request.reset(SrmCopy::create(ctx));
    } catch (const LogicError& exc) {
        // SrmCopy method is not supported
        s = UrlCopyError::SCOPE_AGENT;
        c = UrlCopyError::ERROR_INTERNAL_ERROR;
        m = exc.what();
        // Set Final Error
        allFilesError(s, UrlCopyError::PHASE_TRANSFER, c, m);
        finalError(s, UrlCopyError::PHASE_TRANSFER, c, m);
        return;
    }

    time_t current;
    time(&current);

    // Set Request Parameters
    request->files.resize(m_stat->mm_num_files);
    unsigned int i = 0;
    for (; i < m_stat->mm_num_files; ++i) {
        request->files[i].sourceSurl = m_stat->mm_files[i].mm_src_name;
        request->files[i].targetSurl = m_stat->mm_files[i].mm_dest_name;
        // Set Start Time
        m_stat->mm_files[i].mm_start_time = current;
        m_stat->mm_files[i].mm_status = UrlCopyStatus::RUNNING;

        msg_ifce::getInstance()->set_timestamp_transfer_started(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
    }
#ifdef USE_SAFE_CLEANUP
    if (m_stat->mm_srmcopy.mm_copy_timeout > 0) {
        // Set desired Request Time
        request->desiredTotalRequestTime =
                SAFE_CLEANUP_MULIPLY_FACTOR * m_stat->mm_srmcopy.mm_copy_timeout;
    } else if (m_stat->mm_global_timeout > 0) {
        // Set desired Request Time
        request->desiredTotalRequestTime =
                SAFE_CLEANUP_MULIPLY_FACTOR * m_stat->mm_global_timeout;
    }
#endif //USE_SAFE_CLEANUP

    // bug 32007 : desiredTargetSURLLifeTime and targetFileStorageType were
    // specified only if no space token provided; they must be specified
    // always

    // Set desired lifetime and Storage Type
    request->desiredTargetSURLLifeTime = -1;
    request->targetFileStorageType = FileStorageType::PERMANENT;

    if (0 != strlen(m_stat->mm_common.mm_source_space_token)) {
        request->sourceSpaceToken = m_stat->mm_common.mm_source_space_token;
    }
    if (0 != strlen(m_stat->mm_common.mm_dest_space_token)) {
        request->targetSpaceToken = m_stat->mm_common.mm_dest_space_token;
    }


    request->overwriteOption = (0 != m_stat->mm_common.mm_overwrite) ? OverwriteMode::ALWAYS : OverwriteMode::NEVER;
    request->transferParameters.accessPattern = AccessPattern::TRANSFER_MODE;
    request->transferParameters.connectionType = m_stat->mm_common.mm_lan_connection ? ConnectionType::LAN : ConnectionType::WAN;
    if (m_stat->mm_common.mm_copy_pin_lifetime > 0) {
        debug() << "Setting copy pin lifetime in srmCopy request.";
        request->copyPinLifetime = m_stat->mm_common.mm_copy_pin_lifetime;
    }
    // -------------------------------------------------------------------------
    // Send SrmCopy Request
    // -------------------------------------------------------------------------
    try {
        try {
            RefreshTimeUpdater updater(this);
            request->execute();
            info() << "SrmCopy [" << request->token() << "] started";
            log_request(*request);
        } catch (const HttpTimeoutError& exc) {
            error() << "HTTP timeout contacting remote SRM [" << ctx->endpoint << "]: " << exc.what();
            c = UrlCopyError::ERROR_HTTP_TIMEOUT;
            m = exc.what();
            throw;
        } catch (const RemoteError& exc) { // http timeout handled
            error() << "Failed to contact remote SRM [" << ctx->endpoint << "]: " << exc.what();
            c = UrlCopyError::ERROR_CONNECTION;
            m = exc.what();
            throw;
        } catch (const SecurityError& exc) {
            error() << "User is not authorized to contact the remote SRM [" << ctx->endpoint << "]: " << exc.what();
            c = UrlCopyError::ERROR_PERMISSION;
            m = exc.what();
            throw;
        } catch (const SrmError& exc) {
            error() << "Remote method invocation failure on the remote SRM [" << ctx->endpoint << "]: " << exc.what();
            c = UrlCopyError::ERROR_REQUEST_FAILURE;
            m = exc.what();
            throw;
        } catch (const SrmBadResponse& exc) {
            error() << "Invalid result returned by the remote SRM [" << ctx->endpoint << "]: " << exc.what();
            c = UrlCopyError::ERROR_REQUEST_FAILURE;
            m = exc.what();
            throw;
        } catch (const LogicError& exc) {
            error() << "Logic Error invoking SrmCopy: " << exc.what();
            s = UrlCopyError::SCOPE_AGENT;
            c = UrlCopyError::ERROR_INTERNAL_ERROR;
            m = exc.what();
            throw;
        } catch (const AgentException& exc) {
            error() << exc.type() << " invoking SrmCopy: " << exc.what();
            s = UrlCopyError::SCOPE_AGENT;
            c = UrlCopyError::ERROR_INTERNAL_ERROR;
            m = exc.what();
            throw;
        } catch (const std::exception& exc) {
            error() << "Error invoking SrmCopy: " << exc.what();
            s = UrlCopyError::SCOPE_AGENT;
            c = UrlCopyError::ERROR_INTERNAL_ERROR;
            m = "Internal error";
            throw;
        } catch (...) {
            error() << "Unknown error invoking SrmCopy.";
            s = UrlCopyError::SCOPE_AGENT;
            c = UrlCopyError::ERROR_INTERNAL_ERROR;
            m = "Internal error";
            throw;
        }
    } catch (...) {
        // Set Final Error
        allFilesError(s, UrlCopyError::PHASE_TRANSFER, c, m);
        finalError(s, UrlCopyError::PHASE_TRANSFER, c, m);
        return;
    }

    // Save Request Token
    copy_token(m_stat->mm_srmcopy.mm_token, request->token().c_str());

    // -------------------------------------------------------------------------
    // Poll until the copy is completed
    // -------------------------------------------------------------------------
    c = UrlCopyError::ERROR_SRM_COPY;
    m.clear();
    try {
        // Timeout
        int copy_timeout = (m_stat->mm_srmcopy.mm_copy_timeout >= 0) ? m_stat->mm_srmcopy.mm_copy_timeout : DEFAULT_COPY_TIMEOUT;
        // int timeout = copy_timeout * m_stat->mm_num_files;
        glite::data::agents::longlong all_files_size = 0;
        double to_per_mb = fmaxl(0, m_stat->mm_srmcopy.mm_tx_to_per_mb);
        if (to_per_mb > 0) {
            for (unsigned int j = 0; j < m_stat->mm_num_files; ++j) {
                if (m_stat->mm_files[j].mm_src_size > 0) {
                    all_files_size += m_stat->mm_files[j].mm_src_size;
                }
            }
        }
        static const unsigned long MB = 1 << 20;
        double dtimeout = (copy_timeout * m_stat->mm_num_files) + (to_per_mb * all_files_size / MB);
        int timeout = (dtimeout >= INT_MAX) ? 0 : (int) dtimeout;
        info() << "Copy timeout = " << timeout << " = " << copy_timeout << " * " << m_stat->mm_num_files << " + " <<
                to_per_mb << " * " << (all_files_size / MB);

        for (unsigned int z = 0; z < m_stat->mm_num_files; ++z) {
            std::string timeout_to_string = to_string<int>(timeout, std::dec);
            msg_ifce::getInstance()->set_transfer_timeout(tr_completed[z], timeout_to_string.c_str());
        }


        time(&current);
        // Set last refresh time
        m_stat->mm_common.mm_last_refresh_time = current;
        // Loop until the request is completed or failed
        while (((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
                (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)) &&
                (false == TransferState::instance().interrupted())) {
            // Evaluate the last_refresh timeout
            unsigned int last_refresh_timeout =
                    (m_stat->mm_srmcopy.mm_refresh_timeout >= 0) ?
                    m_stat->mm_srmcopy.mm_refresh_timeout : DEFAULT_COPY_REFRESH_TIMEOUT;
            if ((last_refresh_timeout > 0) &&
                    (last_refresh_timeout < (2 * DEFAULT_COPY_SLEEP_TIME))) {
                last_refresh_timeout = 2 * DEFAULT_COPY_SLEEP_TIME;
            }
            if ((timeout > 0) &&
                    (current > (m_stat->mm_srmcopy.mm_start_time + timeout))) {
                // Timeout elapsed
                std::stringstream reason;
                reason << "failed to complete srmcopy transfer request in "
                        << timeout << " seconds";
                c = UrlCopyError::ERROR_REQUEST_TIMEOUT;
                m = reason.str();
                warn() << m;
                break;
            }
            if ((last_refresh_timeout > 0) &&
                    ((unsigned int) current > (m_stat->mm_common.mm_last_refresh_time + last_refresh_timeout))) {
                // Too many seconds elapsed since last refresh
                std::stringstream reason;
                reason << "no status update received since more than "
                        << last_refresh_timeout << " seconds";
                c = UrlCopyError::ERROR_REQUEST_TIMEOUT;
                m = reason.str();
                warn() << m;
                break;
            }
            // Sleep for a while
            unsigned int sleep_time = DEFAULT_COPY_SLEEP_TIME;
            for (i = 0; i < request->files.size(); ++i) {
                if ((request->files[i].extimatedWaitTime > 0) &&
                        (request->files[i].extimatedWaitTime < MAX_COPY_SLEEP_TIME)) {
                    sleep_time = request->files[i].extimatedWaitTime;
                }
            }
            transfer_sleep(sleep_time);

            // Keep polling until the request is completed/failed or the timeout
            // elapsed
            try {
                RefreshTimeUpdater updater(this);
                request->update();
                info() << "Status of SrmCopy [" << request->token() << "] updated";
                log_request(*request);
                time(&current);
                // Check if some file completed
                for (i = 0; i < request->files.size(); ++i) {
                    switch (request->files[i].status.code) {
                        case SrmStatus::SRM_REQUEST_INPROGRESS:
                        case SrmStatus::SRM_REQUEST_QUEUED:
                            // File Still Running
                            break;
                        case SrmStatus::SRM_SUCCESS:
                            // The file completed successfully
                            info() << "File [" << i << "] completed successfully";
                            // Get file size
                            if (request->files[i].size > 0) {
                                m_stat->mm_files[i].mm_size = request->files[i].size;
                            }
                            m_stat->mm_files[i].mm_status = UrlCopyStatus::DONE;
                            m_stat->mm_files[i].mm_done_time = current;
                            msg_ifce::getInstance()->set_final_transfer_state(tr_completed[i], "Ok");

                            break;
                        case SrmStatus::SRM_ABORTED:
                            // The file aborted
                            warn() << "File [" << i << "] aborted";
                            m_stat->mm_files[i].mm_status = UrlCopyStatus::ABORTED;
                            m_stat->mm_files[i].mm_done_time = current;

                            // Set File Error			    			    
                            fileErrorFromSrmCopyStatus(i, UrlCopyError::PHASE_TRANSFER, request->files[i].status);
                            break;
                        default:
                            // The file failed
                            warn() << "File [" << i << "] failed";
                            m_stat->mm_files[i].mm_status = UrlCopyStatus::FAILED;
                            m_stat->mm_files[i].mm_done_time = current;

                            // Set File Error			    
                            fileErrorFromSrmCopyStatus(i, UrlCopyError::PHASE_TRANSFER, request->files[i].status);
                            break;
                    }

                    msg_ifce::getInstance()->set_timestamp_transfer_completed(tr_completed[i], msg_ifce::getInstance()->getTimestamp());

                }
            } catch (const RemoteError& exc) { // don't handle http timeout explicitly here
                warn() << "Failed to contact remote SRM [" << ctx->endpoint << "] in updating the SrmCopy status: " << exc.what();
                info() << "This error is considered transient and the action will be retried";
		for (i = 0; i < request->files.size(); ++i) 
			msg_ifce::getInstance()->set_timestamp_transfer_completed(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
            } catch (const SrmError& exc) {
                warn() << "Remote method invocation failure on the remote SRM [" << ctx->endpoint << "] in updating the SrmCopy status: " << exc.what();
                info() << "This error is considered transient and the action will be retried";
		for (i = 0; i < request->files.size(); ++i) 		
			msg_ifce::getInstance()->set_timestamp_transfer_completed(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
            } catch (const AgentException& exc) {
                error() << exc.type() << " updating SrmCopy status: " << exc.what();
		for (i = 0; i < request->files.size(); ++i) 		
			msg_ifce::getInstance()->set_timestamp_transfer_completed(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
                throw;
            } catch (const std::exception& exc) {
                error() << "Error updating SrmCopy status for remote SRM [" << ctx->endpoint << "]: " << exc.what();
		for (i = 0; i < request->files.size(); ++i) 		
			msg_ifce::getInstance()->set_timestamp_transfer_completed(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
                throw;
            } catch (...) {
                error() << "Unknown error updating SrmCopy status for remote SRM [" << ctx->endpoint << "]";
		for (i = 0; i < request->files.size(); ++i) 		
			msg_ifce::getInstance()->set_timestamp_transfer_completed(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
                throw;
            }
        } // end while
        if (request->status.code == SrmStatus::SRM_SUCCESS) {
            info() << "SrmCopy [" << request->token() << "] completed";
            c = UrlCopyError::SUCCESS;
        } else if (request->status.code == SrmStatus::SRM_PARTIAL_SUCCESS) {
            info() << "SrmCopy [" << request->token() << "] completed with some failures";
            c = UrlCopyError::SUCCESS;
        } else if (true == TransferState::instance().interrupted()) {
            error() << "SrmCopy [" << request->token() << "] should be aborted";
            c = UrlCopyError::ERROR_ABORTED;
            m = "transfer request canceled";
        } else if (c == UrlCopyError::ERROR_REQUEST_TIMEOUT) {
            error() << "SrmCopy [" << request->token() << "] timed out";
        } else {
            error() << "SrmCopy [" << request->token() << "] failed";
            // Decode UrlCopyError From SrmStatus Code
            requestErrorFromSrmCopyStatus(request->status, s, c, m);
        }
    } catch (const SecurityError& exc) {
        error() << "User is not authorized to contact the remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_PERMISSION;
        m = exc.what();
	for (i = 0; i < request->files.size(); ++i) 	
		msg_ifce::getInstance()->set_timestamp_transfer_completed(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
    } catch (const SrmBadResponse& exc) {
        error() << "Invalid result returned by the remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_REQUEST_FAILURE;
        m = exc.what();
	for (i = 0; i < request->files.size(); ++i) 		
		msg_ifce::getInstance()->set_timestamp_transfer_completed(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
    } catch (const LogicError& exc) {
        error() << "Logic Error in invoking SrmCopy: " << exc.what();
        s = UrlCopyError::SCOPE_AGENT;
        c = UrlCopyError::ERROR_INTERNAL_ERROR;
        m = exc.what();
	for (i = 0; i < request->files.size(); ++i) 		
		msg_ifce::getInstance()->set_timestamp_transfer_completed(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
    } catch (const std::exception& exc) {
        error() << "Error invoking SrmCopy: " << exc.what();
	for (i = 0; i < request->files.size(); ++i) 		
		msg_ifce::getInstance()->set_timestamp_transfer_completed(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
        throw;
    } catch (...) {
        error() << "Unknown error invoking SrmCopy.";
	for (i = 0; i < request->files.size(); ++i) 		
		msg_ifce::getInstance()->set_timestamp_transfer_completed(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
        throw;
    }

    // Request is completed: Save File properties
    time(&current);
    for (unsigned int i = 0; i < request->files.size(); ++i) {
        if (UrlCopyError::SUCCESS == c) {
            if (request->files[i].status.code == SrmStatus::SRM_SUCCESS) {
                // File completed successfully
                info() << "File [" << i << "] completed successfully";
                // Get file size
                if ((m_stat->mm_files[i].mm_size <= 0) &&
                        (request->files[i].size > 0)) {
                    m_stat->mm_files[i].mm_size = request->files[i].size;
                }
                // Set Status
                m_stat->mm_files[i].mm_status = UrlCopyStatus::DONE;
            } else {
                // File failed: get error from SrmStatus
                fileErrorFromSrmCopyStatus(i, UrlCopyError::PHASE_TRANSFER, request->files[i].status);
            }
        } else {
            if ((request->files[i].status.code == SrmStatus::SRM_UNDEF) ||
                    (request->files[i].status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
                    (request->files[i].status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
                    (request->files[i].status.code == SrmStatus::SRM_SUCCESS)) {
                // Get Status from final Error
                fileError(i, s, UrlCopyError::PHASE_TRANSFER, c, m);
            } else {
                // File failed: get error from SrmStatus
                fileErrorFromSrmCopyStatus(i, UrlCopyError::PHASE_TRANSFER, request->files[i].status);
            }
        }
        if (m_stat->mm_files[i].mm_done_time <= 0) {
            m_stat->mm_files[i].mm_done_time = current;
        }
	msg_ifce::getInstance()->set_timestamp_transfer_completed(tr_completed[i], msg_ifce::getInstance()->getTimestamp());  
    }
      
    if (UrlCopyError::SUCCESS != c) {
        // Request Failed: it will be aborted on the completion

        // Set Final Error
        finalError(s, UrlCopyError::PHASE_TRANSFER, c, m);
    }
    
}

/*
 * fileErrorFromSrmCopyStatus
 *
 * Get File Error From SrmCopy Status
 */
void SrmCopyCore::fileErrorFromSrmCopyStatus(// [conversion]
        unsigned int i,
        UrlCopyError::Phase p,
        SrmStatus& status) throw () {

    UrlCopyError::Category c;
    UrlCopyError::Scope s;
    std::string m;
    switch (status.code) {
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
            s = UrlCopyError::SCOPE_TRANSFER;
            c = UrlCopyError::ERROR_PERMISSION;
            m = status.message;
            break;
        case SrmStatus::SRM_NO_USER_SPACE:
        case SrmStatus::SRM_NO_FREE_SPACE:
            s = UrlCopyError::SCOPE_DESTINATION;
            c = UrlCopyError::ERROR_NO_SPACE_LEFT;
            m = status.message;
            break;
        case SrmStatus::SRM_FILE_LIFETIME_EXPIRED:
        case SrmStatus::SRM_FILE_LOST:
        case SrmStatus::SRM_FILE_BUSY:
        case SrmStatus::SRM_FILE_UNAVAILABLE:
        case SrmStatus::SRM_INVALID_PATH:
            s = UrlCopyError::SCOPE_SOURCE;
            c = UrlCopyError::ERROR_INVALID_PATH;
            m = status.message;
            break;
        case SrmStatus::SRM_DUPLICATION_ERROR:
            s = UrlCopyError::SCOPE_DESTINATION;
            c = UrlCopyError::ERROR_FILE_EXISTS;
            m = status.message;
            break;
        case SrmStatus::SRM_ABORTED:
            s = UrlCopyError::SCOPE_TRANSFER;
            c = UrlCopyError::ERROR_ABORTED;
            m = status.message;
            break;
        case SrmStatus::SRM_UNDEF:
        case SrmStatus::SRM_SUCCESS:
        {
            // This in principle should not be possible
            s = UrlCopyError::SCOPE_AGENT;
            c = UrlCopyError::ERROR_INTERNAL_ERROR;
            std::stringstream ss;
            ss << "The SRM returned an error, but with an invalid code [" << status.code << "]";
            m = ss.str();
        }
        break;
        default:
            // Set default error
            s = UrlCopyError::SCOPE_TRANSFER;
            c = UrlCopyError::ERROR_SRM_COPY;
            if (false == status.message.empty()) {
                m = status.message;
            } else {
                std::stringstream ss;
                ss << "SrmCopy request failed for that file with error [" << status.code << "]";
                m = ss.str();
            }
            break;
    }
    // Set File Error
    fileError(i, s, p, c, m);
}

/*
 * requestErrorFromSrmCopyStatus
 *
 * Get Request Error From SrmCopy Status
 */
void SrmCopyCore::requestErrorFromSrmCopyStatus(// [conversion]
        const SrmStatus& req_status,
        UrlCopyError::Scope& s,
        UrlCopyError::Category& c,
        std::string& m) throw () {

    switch (req_status.code) {
        case SrmStatus::SRM_AUTHENTICATION_FAILURE:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
            s = UrlCopyError::SCOPE_TRANSFER;
            c = UrlCopyError::ERROR_PERMISSION;
            m = req_status.message;
            break;
        case SrmStatus::SRM_REQUEST_TIMED_OUT:
            s = UrlCopyError::SCOPE_TRANSFER;
            c = UrlCopyError::ERROR_REQUEST_TIMEOUT;
            m = req_status.message;
            break;
        case SrmStatus::SRM_EXCEED_ALLOCATION:
        case SrmStatus::SRM_NO_USER_SPACE:
        case SrmStatus::SRM_NO_FREE_SPACE:
            s = UrlCopyError::SCOPE_DESTINATION;
            c = UrlCopyError::ERROR_NO_SPACE_LEFT;
            m = req_status.message;
            break;
        case SrmStatus::SRM_ABORTED:
            s = UrlCopyError::SCOPE_TRANSFER;
            c = UrlCopyError::ERROR_ABORTED;
            m = req_status.message;
            break;
        case SrmStatus::SRM_INTERNAL_ERROR:
            s = UrlCopyError::SCOPE_TRANSFER;
            c = UrlCopyError::ERROR_STORAGE_INTERNAL_ERROR;
            m = req_status.message;
            break;
        case SrmStatus::SRM_UNDEF:
        case SrmStatus::SRM_PARTIAL_SUCCESS:
        case SrmStatus::SRM_SUCCESS:
        {
            // This in principle should not be possible
            s = UrlCopyError::SCOPE_AGENT;
            c = UrlCopyError::ERROR_INTERNAL_ERROR;
            std::stringstream ss;
            ss << "The SRM returned an error, but with an invalid code [" << req_status.code << "]";
            m = ss.str();
        }
        break;
        default:
            // Set default error
            s = UrlCopyError::SCOPE_TRANSFER;
            c = UrlCopyError::ERROR_SRM_COPY;
            if (false == req_status.message.empty()) {
                m = req_status.message;
            } else {
                std::stringstream ss;
                ss << "SrmCopy request failed for that file with error [" << SrmStatus::toString(req_status.code) << "]";
                m = ss.str();
            }
            break;
    }
}

/*
 * checkDestinationFiles
 *
 * Verify destination files
 */
void SrmCopyCore::checkDestinationFiles() throw () {
    // Create Context
    Context::Ptr ctx = destinationContext();
    for (unsigned int i = 0; i < m_stat->mm_num_files; ++i) {
        const char * const dest_name = m_stat->mm_files[i].mm_dest_name;
        // set now file to DONE
        m_stat->mm_files[i].mm_status = UrlCopyStatus::DONE;
        // Get the destination  File Type
        FileType::Value dest_type = FileType::get(dest_name);
        // Compare source and destination file sizes
        if (dest_type == FileType::SURL) {
            if (m_stat->mm_files[i].mm_error.mm_category == UrlCopyError::SUCCESS) {
                if (0 == m_stat->mm_files[i].mm_src_size) {
                    warn() << "verification not performed for file [" << dest_name << "] since source file size is not available";
                } else {
                    if (0 == m_stat->mm_files[i].mm_size) {
                        // Get File Size From SRM
                        if (0 != ctx.get()) {
                            UrlCopyError::Category c;
                            std::string m;
                            TransferUtils::PathInfo info;
                            bool file_exists = pathExists(ctx, dest_name, c, m, info);
                            if (true == file_exists) {
                                m_stat->mm_files[i].mm_size = info.file_size;
                            } else {
                                warn() << "Cannot get size for file [" << dest_name << "]: " << m;
                            }
                        } else {
                            warn() << "Cannot get size for file [" << dest_name << "] since the target SRM endpoint is not provided";
                        }
                    }
                    if (0 == m_stat->mm_files[i].mm_size) {
                        info() << "Destination file size unset. Skip verification and continue with next file";
                        continue;
                    }
                    if (m_stat->mm_files[i].mm_size == m_stat->mm_files[i].mm_src_size) {
                        info() << "verification for file [" << dest_name << "] succeeded";
                    } else {
                        // Verification failed
                        std::stringstream ss;
                        ss << "destination file size returned by the SRM server ["
                                << m_stat->mm_files[i].mm_size
                                << "] differs from the source one ["
                                << m_stat->mm_files[i].mm_src_size << "]";
                        std::string reason = ss.str();
                        error() << reason;
                        // Set invalid size
                        fileError(i, UrlCopyError::SCOPE_DESTINATION, UrlCopyError::PHASE_TRANSFER, UrlCopyError::ERROR_INVALID_SIZE, reason.c_str());
                    }
                }
            } // end if SUCCESS
        }
    } // End for
}

//------------------------------------------------------------------------------
//
// Transfer Completion
//
//------------------------------------------------------------------------------

SrmCopyCore::CompleteDependencies_::CompleteDependencies_(SrmCopyCore& self)
: self_(self) {
    // EMPTY
}

SrmCopyCore::CompleteDependencies_::~CompleteDependencies_() {
    // EMPTY
}

void SrmCopyCore::CompleteDependencies_::completeCopyRequest() {
    self_.completeCopyRequest();
}

void SrmCopyCore::CompleteDependencies_::abortCopyRequest() {
    self_.abortCopyRequest();
}

void SrmCopyCore::CompleteDependencies_::completeDestinations() {
    self_.completeDestinations();
}

void SrmCopyCore::complete() {
    complete_(*completeDependencies_production_);
}

void SrmCopyCore::complete_(SrmCopyCore::CompleteDependencies_& deps) {
    // Check of the instance is initialized, otherwise initialize it from the
    // status file
    initialize();

    for (unsigned int i = 0; i < m_stat->mm_num_files; ++i) {
        msg_ifce::getInstance()->set_time_spent_in_srm_finalization_start(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
    }

    time(&(m_stat->mm_srmcopy.mm_final_start_time));

    // Depending on the request status, perform the correct cleanup precedure
    // -------------------------------------------------------------------------
    switch (m_stat->mm_srmcopy.mm_status) {
        case UrlCopyStatus::COMPLETED:
            // The request is already completed (either successfully or not):
            // nothing to do
            break;
        case UrlCopyStatus::RUNNING:
            // The transfer is currently running: wait that the transfer is
            // finished or abort it
            throw LogicError("transfer request is running. Please wait that "
                    "the transfer finishes or abort it before calling "
                    "complete");
            break;
        case UrlCopyStatus::DONE:
            // The transfer is completed, complete the SRM requests
            deps.completeCopyRequest();
            break;
        case UrlCopyStatus::IDLE:
        case UrlCopyStatus::PREPARING:
            // The request is idle or preparing, but maybe the source request has been
            // executed: in that case abort it
        case UrlCopyStatus::READY:
            // The request is ready, the file has not been transferred:
            // abort the SRM requests
        case UrlCopyStatus::FAILED:
        case UrlCopyStatus::ABORTED:
            // The transfer has been canceled, abort the SRM requests
        default:
            // Abort Transfer Request
            deps.abortCopyRequest();
            break;
    }// end switch

    // Complete destination Files
    deps.completeDestinations();

    // Set request final done time
    time(&(m_stat->mm_srmcopy.mm_final_done_time));

    if (m_stat->mm_srmcopy.mm_error.mm_category == UrlCopyError::SUCCESS) {
        // Transfer Complete Successfully
        info() << "FINAL:SUCCESS";
    }
    // Now the request has been completed. In case of errors, these are
    // reported on the file attributes (source,destination,tx)
    m_stat->mm_srmcopy.mm_status = UrlCopyStatus::COMPLETED;

    // Set request finish time
    time(&(m_stat->mm_common.mm_finish_time));

    for (unsigned int i = 0; i < m_stat->mm_num_files; ++i) {
        msg_ifce::getInstance()->set_time_spent_in_srm_finalization_end(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
    }

}

/*
 * completeCopyRequest
 *
 * Call Done on the SrmCopy request
 */
void SrmCopyCore::completeCopyRequest() /*throw ()*/ {
    // Precondition check
    if (0 == strlen(m_stat->mm_srmcopy.mm_token)) {
        info() << "No request token provided for the copy request. Assuming SrmCopy request has not been sent";
        return;
    }

    // Create SRM Context
    Context::Ptr ctx(createSrmCtxByCopyMode());

    // -------------------------------------------------------------------------
    // Create SrmCopy Request
    // -------------------------------------------------------------------------
    boost::scoped_ptr<SrmCopy> request;
    try {
        // Create PrepareToPut Request
        request.reset(SrmCopy::create(ctx, m_stat->mm_srmcopy.mm_token));
    } catch (const LogicError& exc) {
        // SrmCopy method is not supported. Since SrmCopy Done should be no-op,
        // ignore this condition
        return;
    }

    // Set Request Parameters
    request->files.resize(m_stat->mm_num_files);
    for (unsigned int i = 0; i < m_stat->mm_num_files; ++i) {
        request->files[i].sourceSurl = m_stat->mm_files[i].mm_src_name;
        request->files[i].targetSurl = m_stat->mm_files[i].mm_dest_name;
    }

    // Call Complete Request
    UrlCopyError::Category c;
    std::string m;
    bool completed = m_srmMethodCaller.completeRequest(ctx, *request, c, m, DEFAULT_COPYDONE_TIMEOUT);
    if (!completed) {
        // The SrmCopy done should be a no-op. this should not be possible
        // Abort the request
        std::stringstream ss;
        ss << "failed to complete SrmCopy request [" << request->token()
                << "] on remote SRM [" << ctx->endpoint
                << "]: [" << request->status.code << "] "
                << request->status.message;
        std::string reason = ss.str();
        warn() << reason;
        info() << "ignore this error since this should not be a real use case";
    } else {
        // Request completed
        info() << "SrmCopy [" << request->token() << "] successfully completed";
        // check pin lifetime and issue an srmBringOnline (addendum to the WLCG
        // usage agreement)
        if (m_stat->mm_common.mm_copy_pin_lifetime > 0) {
            debug() << "Calling srmBringOnline to set copy pin lifetime.";
            std::vector<std::string> surls;
            for (unsigned int i = 0; i < m_stat->mm_num_files; ++i) {
                std::string surl = m_stat->mm_files[i].mm_dest_name;
                if (m_stat->mm_files[i].mm_error.mm_category == UrlCopyError::SUCCESS) {
                    debug() << "Adding SURL to bringOnline request: " << surl;
                    surls.push_back(surl);
                } else {
                    warn() << "Error on file #" << i << " , skipping SURL in bringOnline request: " << surl;
                }
            }
            bringOnline(ctx, surls, m_stat->mm_common.mm_copy_pin_lifetime);
        }

    }
}

/*
 * abortCopyRequest
 *
 * Call AbortRequest on the SrmCopy request
 */
void SrmCopyCore::abortCopyRequest() /*throw ()*/ {
    // Precondition check
    if (0 == strlen(m_stat->mm_srmcopy.mm_token)) {
        info() << "No request token provided for the copy request. Assuming SrmCopy request has not been sent";
        return;
    }

    info() << "aborting SrmCopy [" << m_stat->mm_srmcopy.mm_token << "]";
    // Create SRM Context
    Context::Ptr ctx(createSrmCtxByCopyMode());

    // -------------------------------------------------------------------------
    // Create SrmCopy Request
    // -------------------------------------------------------------------------
    boost::scoped_ptr<SrmCopy> request;
    try {
        // Create SrmCopy Request
        request.reset(SrmCopy::create(ctx, m_stat->mm_srmcopy.mm_token));
    } catch (const LogicError& exc) {
        // SrmCopy method is not supported: this should not be possible
        warn() << exc.what();
        // Remove Destination Files
        rmDestinationFiles();
        return;
    }

    // Set Request Parameters
    request->files.resize(m_stat->mm_num_files);
    unsigned int i = 0;
    for (; i < m_stat->mm_num_files; ++i) {
        request->files[i].sourceSurl = m_stat->mm_files[i].mm_src_name;
        request->files[i].targetSurl = m_stat->mm_files[i].mm_dest_name;
    }

    // -------------------------------------------------------------------------
    // Send Abort Request
    // -------------------------------------------------------------------------
    UrlCopyError::Category c;
    std::string reason;
    if (true == m_srmMethodCaller.abortRequest(ctx, *request, c, reason, -1)) {
        // Request aborted
        info() << "SrmCopy [" << request->token() << "] successfully aborted";
        reason = "Request aborted";

        // Set file status to aborted
        for (i = 0; i < m_stat->mm_num_files; ++i) {
            if (m_stat->mm_files[i].mm_error.mm_category == UrlCopyError::SUCCESS) {
                fileError(i, UrlCopyError::SCOPE_TRANSFER, UrlCopyError::PHASE_FINALIZATION, UrlCopyError::ERROR_ABORTED, reason.c_str());
            }
        }
        if (m_stat->mm_srmcopy.mm_error.mm_category == UrlCopyError::SUCCESS) {
            finalError(UrlCopyError::SCOPE_TRANSFER, UrlCopyError::PHASE_FINALIZATION, UrlCopyError::ERROR_ABORTED, reason.c_str());
        }
        // Remove Destination Files
        rmDestinationFiles();
    } else {
        // Failed to Abort Request
        std::stringstream ss;
        ss << "Failed to abort SrmCopy request [" << request->token()
                << "] on remote SRM [" << ctx->endpoint
                << "]: [" << request->status.code << "] "
                << request->status.message;
        reason = ss.str();
        error() << reason;
    }
}

/**
 * rmDestinationFiles
 * Remove the destination file if they didn't exist before the transfer
 */
void SrmCopyCore::rmDestinationFiles() throw () {
    Context::Ptr ctx = destinationContext();

    if (0 == ctx.get()) {
        warn() << "Cannot enforce deletion of destination file since no information has been provided on the target SRM";
        return;
    }

    // if destination file didn't exist before the transfer, delete it
    for (unsigned int i = 0; i < m_stat->mm_num_files; ++i) {
        if (0 == m_stat->mm_files[i].mm_dest_exists) {
            // Remove destination file
            UrlCopyError::Category c;
            std::string m;
            rmFile(ctx, m_stat->mm_files[i].mm_dest_name, c, m);
        }
    }
}

/**
 * completeDestinations
 * Complete Destination files
 */
void SrmCopyCore::completeDestinations() throw () {
    // Create Context
    Context::Ptr ctx = destinationContext();
    // For each file
    for (unsigned int i = 0; i < m_stat->mm_num_files; ++i) {
        // Check file state
        switch (m_stat->mm_files[i].mm_status) {
            case UrlCopyStatus::DONE:
            {
                // file transfer is completed
                if (m_stat->mm_files[i].mm_error.mm_category == UrlCopyError::SUCCESS) {
                    // Set next status to completed
                    m_stat->mm_files[i].mm_status = UrlCopyStatus::COMPLETED;
                } else {
                    // An error happened during the verification: try to
                    // remove the file
                    if (0 != ctx.get()) {
                        UrlCopyError::Category c;
                        std::string m;
                        rmFile(ctx, m_stat->mm_files[i].mm_dest_name, c, m);
                    } else {
                        warn() << "File [" << i << "] should be removed but the target SRM endpoint is missing";
                    }
                    // Set file status to failed
                    m_stat->mm_files[i].mm_status = UrlCopyStatus::FAILED;
                }
            }
            break;
            case UrlCopyStatus::FAILED:
            {
                // Try to remove the file
                if ((0 != ctx.get()) && (0 == m_stat->mm_files[i].mm_dest_exists)) {
                    UrlCopyError::Category c;
                    std::string m;
                    rmFile(ctx, m_stat->mm_files[i].mm_dest_name, c, m);
                }
            }
            break;
            case UrlCopyStatus::ABORTED:
            case UrlCopyStatus::IDLE:
            case UrlCopyStatus::PREPARING:
            case UrlCopyStatus::READY:
            case UrlCopyStatus::COMPLETED:
                // Nothing to do
                break;
            case UrlCopyStatus::RUNNING:
            default:
                // Invalid file state during completion: this should never happen
                warn() << "File [" << i << "] is not in a valid state [" << UrlCopyStatus::toString((UrlCopyStatus::Value)m_stat->mm_files[i].mm_status) << "] during completion";
                break;
        } // end switch
    } // End for

    verifyDestinationChecksums();
}

void SrmCopyCore::verifyDestinationChecksums() {


    Context::Ptr ctx = destinationContext();
    ChecksumChecker_SrmCopy checksumChecker(UrlCopyError::PHASE_FINALIZATION, this, ctx, false);
    // get checksums with srmLs
    for (unsigned int i = 0; i < m_stat->mm_num_files; ++i) {
        if (m_stat->mm_files[i].mm_status == UrlCopyStatus::COMPLETED) {
            msg_ifce::getInstance()->set_timestamp_checksum_dest_started(tr_completed[i], msg_ifce::getInstance()->getTimestamp());
            checksumChecker.getDestinationChecksumFromSrmLs(m_stat->mm_files[i]);
        } else {
            WARN_LOG(SrmCopyCore) << "Skipping destination checksum manipulation for file '"
                    << m_stat->mm_files[i].mm_dest_name << "', because the transfer is not COMPLETED.";
        }
    }

    bool res = checksumChecker.sendPrepareToGetForMissing(m_stat->mm_srmcopy.mm_final_start_time, DEFAULT_GET_TIMEOUT);

    if (res) {
        checksumChecker.getMissingChecksumsWithGridFTP(DEFAULT_GET_TIMEOUT);
    }

    checksumChecker.removeFailedDestinationFiles(m_stat->mm_srmcopy.mm_copy_timeout);
    checksumChecker.compareDestinationChecksums();
    checksumChecker.removeFailedDestinationFiles(m_stat->mm_srmcopy.mm_copy_timeout);

    for (unsigned int z = 0; z < m_stat->mm_num_files; ++z) {
    	if (m_stat->mm_files[z].mm_status == UrlCopyStatus::COMPLETED) 
        	msg_ifce::getInstance()->set_timestamp_checksum_dest_ended(tr_completed[z], msg_ifce::getInstance()->getTimestamp());
    }
}

/**
 * create srm context based on copy mode
 * @stereotype utility
 */
Context::Ptr SrmCopyCore::createSrmCtxByCopyMode() {
    const char * endpoint;
    const char * version;
    if (0 == m_stat->mm_srmcopy.mm_push_mode) {
        // pull
        endpoint = m_stat->mm_srmcopy.mm_dest_srm_endpoint;
        version = m_stat->mm_srmcopy.mm_dest_srm_version;
    } else {
        // push
        endpoint = m_stat->mm_srmcopy.mm_src_srm_endpoint;
        version = m_stat->mm_srmcopy.mm_src_srm_version;
    }
    return m_srmMethodCaller.createSrmCtx(endpoint, version, m_stat->mm_srmcopy.mm_http_timeout);
}

/**
 * get source context
 */
Context::Ptr SrmCopyCore::sourceContext() {
    return m_srmMethodCaller.createSrmCtx(
            m_stat->mm_srmcopy.mm_src_srm_endpoint,
            m_stat->mm_srmcopy.mm_src_srm_version,
            m_stat->mm_srmcopy.mm_http_timeout);
}

/**
 * get destination context
 */
Context::Ptr SrmCopyCore::destinationContext() {
    return m_srmMethodCaller.createSrmCtx(
            m_stat->mm_srmcopy.mm_dest_srm_endpoint,
            m_stat->mm_srmcopy.mm_dest_srm_version,
            m_stat->mm_srmcopy.mm_http_timeout);
}

SrmCopyStat* SrmCopyCore::getStat() {
    return m_stat;
}

