/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: UrlCopyCore class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: SrmCopyCore.h,v 1.10 2009/07/30 19:32:55 molnarzs Exp $
 *  Release: $Name:  $
 */
#ifndef GLITE_DATA_TRANSFER_URL_COPY_SRMCOPYCORE_H_
#define GLITE_DATA_TRANSFER_URL_COPY_SRMCOPYCORE_H_

#include "SrmCopyStat.h"
#include "TransferCore.h"
#include "msg-ifce.h"
#include <vector>

#include <boost/scoped_ptr.hpp>

namespace glite   {
namespace data    {
namespace transfer{
namespace urlcopy {

/**
 * The helper class that manage the transfer (used in the forked process)
 */
class SrmCopyCore : public TransferCore {
public:
    
     std::string vo_name;
     std::string src_site;
     std::string dest_site;

    /**
     * Construct
     */
    explicit SrmCopyCore(const std::string& id);

    /**
     * Destructor
     */
    virtual ~SrmCopyCore();

    /**
     * UpdateRefreshTime
     */
    virtual void UpdateRefreshTime();

    /**
     * Return the transfer identifier
     */
    const std::string& id() const {return m_id;}

    /**
     * Get UrlCopy Stat object
     */
    SrmCopyStat& stat() /*throw (RuntimeError, LogicError)*/;

    /**
     * Prepare the source and destination files for the transfer
     */
    void prepare();

    /**
     * Start the transfer
     */
    void transfer() /*throw (RuntimeError, LogicError)*/;

    /**
     * Complete the transfer
     */
    void complete() /*throw (RuntimeError, LogicError)*/;

    void FileError(SrmCopy_FileStat& file, UrlCopyError::Scope s, UrlCopyError::Phase p, UrlCopyError::Category c, const std::string& m);

    // External dependencies for "prepare" method. For testability.
    class PrepareDependencies_ {
    public:
        PrepareDependencies_(SrmCopyCore& self);
        virtual ~PrepareDependencies_();

        virtual bool pingSource(UrlCopyError::Category& c, std::string& m);
        virtual bool pingDestination(UrlCopyError::Category& c, std::string& m);
        virtual void prepareSource();
        virtual void prepareDestination();

    protected:
        SrmCopyCore& self_;
    };

    boost::scoped_ptr<PrepareDependencies_> prepareDependencies_production_;

    void prepare_(PrepareDependencies_& deps);
    
    /**
     * Initialize the stat
     */
    void initialize() /*throw (RuntimeError, LogicError)*/;

    /**
     * Set File Error
     */
    void fileError(unsigned int i,UrlCopyError::Scope s,UrlCopyError::Phase p,UrlCopyError::Category c,const std::string& m) throw();

    /**
     * Set the same Error on all the files
     */
    void allFilesError(UrlCopyError::Scope s,UrlCopyError::Phase p,UrlCopyError::Category c,const std::string& m) throw() ;

    /**
     * Set Final Error
     */
    void finalError(UrlCopyError::Scope s,UrlCopyError::Phase p,UrlCopyError::Category c,const std::string& m) throw();

    /**
     * Set Final Error from File Error
     */
    void finalErrorFromFile(unsigned int i) throw();

    /**
     * Set final error if all files failed
     * @return true if all files failed
     */
    bool finalErrorFromAllFiles(UrlCopyError::Scope s,UrlCopyError::Phase p) throw();

    // Request Preparation
    //--------------------------------------------------------------------------

    /**
     * Ping Source server
     */
    bool pingSource(UrlCopyError::Category& c, std::string& m) /*throw ()*/;

    /**
     * Ping Destination server
     */
    bool pingDestination(UrlCopyError::Category& c, std::string& m) /*throw ()*/;

    /**
     * Prepare the source file for the transfer
     */
    void prepareSource() /*throw ()*/;

    /**
     * Prepare the destination file for the transfer
     */
    void prepareDestination() /*throw ()*/;

    // Transfer
    //--------------------------------------------------------------------------

    /**
     * Start the srmCopy transfer
     */
    void copyFiles(UrlCopyError::Category& c, std::string& m) /*throw ()*/;

    /**
     * Verify destination files
     */
    void checkDestinationFiles() throw();

    /**
     * Get File Error From SrmCopy Status
     */
    void fileErrorFromSrmCopyStatus(unsigned int i,UrlCopyError::Phase p,glite::data::srm::util::SrmStatus& status) throw();

    /**
     * Get Request Error From SrmCopy Status
     */
    void requestErrorFromSrmCopyStatus(const glite::data::srm::util::SrmStatus& req_status, UrlCopyError::Scope& s, UrlCopyError::Category& c, std::string& m) throw();

    // Request Completion
    //--------------------------------------------------------------------------

    // External dependencies for "prepare" method. For testability.
    class CompleteDependencies_ {
    public:
        CompleteDependencies_(SrmCopyCore& self);
        virtual ~CompleteDependencies_();

        virtual void completeCopyRequest();
        virtual void abortCopyRequest();
        virtual void completeDestinations();

    protected:
        SrmCopyCore& self_;
    };

    
    SrmCopyStat* getStat();

    boost::scoped_ptr<CompleteDependencies_> completeDependencies_production_;

    void complete_(CompleteDependencies_& deps);

    /**
     * Call Done on the SrmCopy request
     */
    void completeCopyRequest() /*throw ()*/;

    /**
     * Call AbortRequest on the SrmCopy request
     */
    void abortCopyRequest() /*throw ()*/;

    /**
     * Remove the destination file if they didn't exist before the transfer
     */
    void rmDestinationFiles() throw ();

    /**
     * Complete the Destination files
     */
    void completeDestinations() throw ();

    /**
     * Verify checksums of destination files
     */
    void verifyDestinationChecksums();

    /**
     * create srm context based on copy mode
     * @stereotype utility
     */
    glite::data::srm::util::Context::Ptr createSrmCtxByCopyMode();

    /**
     * get source context
     */
    glite::data::srm::util::Context::Ptr sourceContext();

    /**
     * get destination context
     */
    glite::data::srm::util::Context::Ptr destinationContext();

private:

    /**
     * Default Constructor. Not Implemented
     */
    SrmCopyCore();

private:
    /**
     * SrmCopy Stat object
     */
    SrmCopyStat * m_stat;
    std::vector<transfer_completed *> tr_completed;
    
};

} // end namespace urlcopy
} // end namespace transfer
} // end namespace data
} // end namespace glite

#endif //GLITE_DATA_TRANSFER_URL_COPY_SRMCOPYCORE_H_

