/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004-2008.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  Call SRM Methods
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: SrmMethodCaller.cpp,v 1.5 2009/12/03 17:21:44 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#include "SrmMethodCaller.h" 
#include "logger-utility.h"
#include "sighandler-utility.h"
#include <glite/data/config/service/SysLogDaemon.h>
 
using namespace glite::data::transfer::urlcopy;
using namespace glite::data::srm::util;
using glite::data::srm::util::details::SynchRequest;
using glite::data::srm::util::details::AsynchRequest;
using namespace glite::config;

std::string ContextLogInfo::user_tag = "";
std::string ContextLogInfo::tx_id = "";

namespace{

//------------------------------------------------------------------------------
// Constants
//------------------------------------------------------------------------------

const unsigned int DEFAULT_HTTP_TIMEOUT          = 40;
const unsigned int DEFAULT_ABORT_SLEEP_TIME      = 3;
const unsigned int DEFAULT_CONNECTION_SLEEP_TIME = 3;
const unsigned int NUM_RETRY_ABORT               = 3;
const unsigned int NUM_RETRY_CONNECTION          = 3;

// SysLog constants
static const std::string SrmCallEventName = "srmCall";
static const std::string SrmCallMethod = "method";
static const std::string SrmCallIp = "ip";
static const std::string SrmToken = "token";
static const std::string JobId = "jobId";
static const std::string TxId = "txId";

class SrmContext : public Context{
public:
    
    /**
     * Constructor
     */
    SrmContext(){
    }
    
protected:
    
    void add_token(SysLogDaemon& log,const CallInfo * info){
        if (info && !info->token().empty()){
            log.add(SrmToken,info->token());
        }
    }
    
    void log_additional_info(const CallInfo * info,const std::string& log_id){
        if (info){
            CallInfo::AdditionalInfos::const_iterator it;
            const CallInfo::AdditionalInfos& infos = info->additionalInfos();
            for (it = infos.begin();it != infos.end();++it){
                SysLogDaemon(true)
                    .add(LogId::KeyName,log_id)
                    .add(it->key,it->value)
                    .info();
            }
        }
    }
    
    /**
     * before_call
     */
    virtual void before_call(
        const char * method){
        debug() << "SRM > invoking method: " << method;
    }
    
    /**
     * on_success
     */
    virtual void on_success(
        const char * method,
        const char * address,
        const CallInfo * info){
            
        std::string log_id = LogId::instance().get_log_id();
        debug() << "SRM > method " << method << " completed successfully (ip = " << address << ")";
        // bug #35750, also log to SysLog
        SysLogDaemon log(true);
        log.event(SrmCallEventName).success(true)
            .add(LogId::KeyName,log_id)
            .add(JobId,ContextLogInfo::user_tag)
            .add(TxId,ContextLogInfo::tx_id)
            .add(SrmCallMethod,method)
            .add(SrmCallIp,address);
        add_token(log,info);
        log.info();
        log_additional_info(info,log_id);
    }
    
    /**
     * on_failure
     */
    virtual void on_failure(
        const char * method,
        const char * address,
        const CallInfo * info){
            
        std::string log_id = LogId::instance().get_log_id();
        warn() << "SRM > method " << method << " failed (ip = " << address << ")";
        // bug #35750, also log to SysLog
        SysLogDaemon log(true);
        log.event(SrmCallEventName).success(false)
            .add(LogId::KeyName,log_id)
            .add(JobId,ContextLogInfo::user_tag)
            .add(TxId,ContextLogInfo::tx_id)
            .add(SrmCallMethod,method)
            .add(SrmCallIp,address);
        add_token(log,info);
        log.error();
        log_additional_info(info,log_id);
    }
    
    /**
     * on_warning
     */
    virtual void on_warning(
        const char * scope,
        const char * description,
        const char * address){
        warn() << "SRM >  " << scope << " : " << description << " (ip = " << address << ")";    
    }
    
};

} // namespace

/**
 * Constructor
 */
SrmMethodCaller::SrmMethodCaller(
    RefreshTimeUpdater::IUpdateRefreshTime * updateRefreshTime) :
    m_updateRefreshTime(updateRefreshTime){
}

/**
 * Destructor
 */
SrmMethodCaller::~SrmMethodCaller(){
}

/**
 * createSrmCtx
 * Create Srm Util Context
 */
Context::Ptr SrmMethodCaller::createSrmCtx(
    const std::string&  endpoint, 
    const std::string&  version, 
    int                 http_timeout) throw() {
    
    debug() << "Contacting SRM endpoint [" << endpoint << "]. Version is [" << version << "]";
    // Create SRM Context
    Context::Ptr ctx(new SrmContext());
    ctx->endpoint          = endpoint;
    ctx->version           = version;
    // Set timeout
    if(http_timeout > 0){
        ctx->httpTimeout   = http_timeout;
    } else {
        ctx->httpTimeout   = DEFAULT_HTTP_TIMEOUT;
    }
    // Ignore Missing features
    ctx->ignoreMissingFeatures = true;
    return ctx;
}

/**
 * completeRequest
 * Complete an Srm Asynch Request
 */
bool SrmMethodCaller::completeRequest(
    Context::Ptr&           ctx, 
    AsynchRequest&          request, 
    UrlCopyError::Category& c, 
    std::string&            m, 
    int                     timeout) throw() {
        
    bool completed = false;
    c = UrlCopyError::SUCCESS;
    
    // -------------------------------------------------------------------------
    // Send completion Request
    // -------------------------------------------------------------------------
    unsigned int i;
    try{
        time_t current, start;
        time(&current);
        start = current;
        for(i = 0; i < NUM_RETRY_CONNECTION; ++i){
            try{
                RefreshTimeUpdater updater(m_updateRefreshTime);
                request.done();
                completed = true;
                break;
            } catch(const RemoteError& exc){ // don't handle http timeout explicitly here
                warn() << "Failed to contact remote SRM [" << ctx->endpoint << "] for completing the request [" << request.token() << "]: " << exc.what();
            } catch(const SrmError& exc){
                warn() << "Remote method invocation failure on the remote SRM [" << ctx->endpoint << "] in completing the request [" << request.token() << "]: " << exc.what();
            }
            if(i < (NUM_RETRY_CONNECTION-1)){
                if((timeout > 0) && (current > start + timeout)){
                    info() << "This call will not be retried since more that " << timeout << " seconds elapsed";
                    break;
                } else{
                    info() << "This call will be retried";
                    // Sleep for a while
                    sleep(DEFAULT_CONNECTION_SLEEP_TIME);
                    time(&current);
                }
            } else {
                warn() << "This call will not be retried anymore";
            }
        } // end for
    } catch(const SecurityError& exc){
        error() << "User is not authorized to contact the remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_PERMISSION;
        m = exc.what();
        return false;
    } catch(const SrmBadResponse& exc){
        error() << "Invalid result returned by the remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_REQUEST_FAILURE;
        m = exc.what();
        return false;
    } catch(const LogicError& exc){ // TBM ? maybe scope should be set to agent
        error() << "Logic Error in completig request: " << exc.what();
        c = UrlCopyError::ERROR_INTERNAL_ERROR;
        m = exc.what();
        return false;
    } 
    // Check if request has been released 
    if(completed == false){
        // Failed To release Request
        std::stringstream ss;
        ss << "Failed to contact on remote SRM ["
           << ctx->endpoint.c_str() 
           << "] in completing request ["
           << request.token() << "]";
        if((timeout > 0 ) && (i != NUM_RETRY_CONNECTION)){
            ss << ". Givin' up after " << timeout << " seconds";
        } else {
            ss << ". Givin' up after " << NUM_RETRY_CONNECTION << " tries";
        }
        std::string reason = ss.str();
        error() << reason;
        c = UrlCopyError::ERROR_CONNECTION;
        m = reason.c_str();
    }
    return completed;
}

/**
 * abortRequest
 * Abort an Srm Asynch Request
 */
bool SrmMethodCaller::abortRequest(
    Context::Ptr&           ctx, 
    AsynchRequest&          request, 
    UrlCopyError::Category& c,
    std::string&            m, 
    int                     timeout) throw() {
        
    c = UrlCopyError::SUCCESS;
            
    // send abort request
    bool aborted = false;    
    unsigned int i;
    time_t current, start;
    time(&current);
    start = current;
    for(i = 0; i < NUM_RETRY_ABORT; ++i){
        try{
            // Send abort request
            info() << "abort request [" << request.token() << "] invoked";	    
            RefreshTimeUpdater updater(m_updateRefreshTime);
            request.abort();
            info() << "abort request [" << request.token() << "] returned";
            // Ok, request to abort sent
            aborted = true;
            break;
        } catch(const HttpTimeoutError& exc){ 
            if(ctx)
	    	warn() << "HTTP timeout contacting remote SRM [" << ctx->endpoint << "] in aborting request: " << exc.what();
            c = UrlCopyError::ERROR_HTTP_TIMEOUT;
            m = exc.what();        
        } catch(const RemoteError& exc){ // http timeout handled
	    if(ctx)
            	warn() << "Failed to contact remote SRM [" << ctx->endpoint << "] in aborting request: " << exc.what();
            c = UrlCopyError::ERROR_CONNECTION;
            m = exc.what();
        } catch(const SrmError& exc){
	    if(ctx)	
            	warn() << "Remote method invocation failure on the remote SRM [" << ctx->endpoint << "] in aborting request: " << exc.what();
        } catch(const SecurityError& exc){
	    if(ctx)	
            	warn() << "User is not authorized to contact the remote SRM [" << ctx->endpoint << "]: " << exc.what();
            c = UrlCopyError::ERROR_PERMISSION;
            m = exc.what();
        } catch(const SrmBadResponse& exc){
	    if(ctx)	
            	warn() << "Invalid result returned by the remote SRM [" << ctx->endpoint << "]: " << exc.what();
            c = UrlCopyError::ERROR_REQUEST_FAILURE;
            m = exc.what();
        } catch(const LogicError& exc){ // TBM ? maybe scope should be set to agent
            warn() << "Logic Error aborting request: " << exc.what();
            c = UrlCopyError::ERROR_INTERNAL_ERROR;
            m = exc.what();
        } catch(const std::exception& exc){ 
            warn() << "Error aborting request: " << exc.what();
            c = UrlCopyError::ERROR_INTERNAL_ERROR;
            m = exc.what();
        } catch(...){
            // Ignore all errors and retry
            warn() << "Unknown Error aborting request on remote SRM [" << ctx->endpoint << "]";
            c = UrlCopyError::ERROR_INTERNAL_ERROR;
            m = "unknown error";
        }
        if(false == TransferState::instance().interrupted()){
            if(i < (NUM_RETRY_CONNECTION-1)){
                if((timeout > 0) && (current > start + timeout)){
                    info() << "This call will not be retried since more that " << timeout << " seconds elapsed";
                    break;
                } else{
                    info() << "This call will be retried";
                    // Sleep for a while
                    sleep(DEFAULT_ABORT_SLEEP_TIME);
                    time(&current);
                }
            } else {
                warn() << "This call will not be retried anymore";
            }
        } else {
            warn() << "This call will not be retried since the transfer request has been canceled";
            break;
        }
    } // end for
    if(true == aborted){
        // Abort completed
        c = UrlCopyError::SUCCESS;
        m = "";
        info() << "Abort completed for request [" << request.token() << "]";
    } else if(true == TransferState::instance().interrupted()){
        // Failed to Abort Request
        std::stringstream ss;
        ss << "failed to abort request ["
           << request.token()
           << "] on remote SRM. Don't retry since the transfer has been canceled";
        c = UrlCopyError::ERROR_ABORTED;
        m = ss.str();
        error() << m;
    } else if(c == UrlCopyError::ERROR_CONNECTION || c == UrlCopyError::ERROR_HTTP_TIMEOUT){
        // Failed To Abort Request
        std::stringstream ss;
        ss << "failed to abort request ["
           << request.token()
           << "] on remote SRM.";
        if (SrmStatus::SRM_UNDEF != request.status.code){
           ss << " Request status is: [" << request.status.code << "] " << request.status.message;
        }
        m = ss.str();
        error() << "Abort failed: " << m;
    } else {
        // Failed To Abort Request
        std::stringstream ss;
        ss << "failed to abort request ["
           << request.token()
           << "] on remote SRM";
        if((timeout > 0 ) && (i != NUM_RETRY_ABORT)){
            ss << ". Givin' up after " << timeout << " seconds";
        } else {
            ss << ". Givin' up after " << NUM_RETRY_ABORT << " tries";
        }
        c = UrlCopyError::ERROR_GENERAL_FAILURE;
        m = ss.str();
        error() << m;
    }
    return aborted;
}

/**
 * executeRequest
 * Execute an Srm Synch Request, retrying it in case of connection errors.
 * If the method invocation is succesfull, the category is set to SUCCESS,
 * and SrmStatus has to be checked 
 */
SrmStatus SrmMethodCaller::executeRequest(
    Context::Ptr&           ctx, 
    SynchRequest&           request, 
    UrlCopyError::Category& c, 
    std::string&            m) throw() {

    c = UrlCopyError::SUCCESS;
    // Get the pointer to the Stat buffer
    SrmStatus result;
    unsigned int i;
    bool executed = false;
    try{
        for(i = 0; i < NUM_RETRY_CONNECTION; ++i){
            try{
                // Execute Synch Request
                RefreshTimeUpdater updater(m_updateRefreshTime); 
                result = request.execute(); 
                // Method invocation completed
                executed = true;
                break;
            } catch(const HttpTimeoutError& exc){
                warn() << "HTTP timeout contacting remote SRM [" << ctx->endpoint << "]: " << exc.what();
                c = UrlCopyError::ERROR_HTTP_TIMEOUT;
            } catch(const RemoteError& exc){ // http timeout handled
                warn() << "Failed to contact remote SRM [" << ctx->endpoint << "]: " << exc.what();
                c = UrlCopyError::ERROR_CONNECTION;
            } catch(const SrmError& exc){
                warn() << "Remote method invocation failure on the remote SRM [" << ctx->endpoint << "]: " << exc.what();
                c = UrlCopyError::ERROR_REQUEST_FAILURE;
            }
            if(false == TransferState::instance().interrupted()){
                if(i < (NUM_RETRY_CONNECTION-1)){
                    info() << "This call will be retried in " << DEFAULT_CONNECTION_SLEEP_TIME << " seconds";
                    // Sleep for a while
                    sleep(DEFAULT_CONNECTION_SLEEP_TIME);
                } else {
                    warn() << "This call will not be retried anymore";
                }
            } else {
                warn() << "This call will not be retried since the transfer request has been canceled";
                break;
            }
        } // end for
        if(false == executed){
            if(true == TransferState::instance().interrupted()){
                // Failed to Abort Request
                std::stringstream ss;
                ss << "failed to contact on remote SRM ["
                   << ctx->endpoint.c_str() 
                   << "]. Don't retry since the transfer has been canceled";
                c = UrlCopyError::ERROR_ABORTED;
                m = ss.str();
                error() << m;
            } else {
                // Failed to contact the remote SRM
                std::stringstream reason;
                reason << "failed to contact on remote SRM ["
                       << ctx->endpoint.c_str() 
                       << "]. Givin' up after " << i << " tries";
                error() << reason.str();
                if (UrlCopyError::SUCCESS == c){
                    c = UrlCopyError::ERROR_CONNECTION;
                }
                m = reason.str();
            }
        } else {
            // Method invocation completed: return UNDEF for category. 
            // Please note that the method could still be failed: Check the 
            // result SrmStatus 
            c = UrlCopyError::SUCCESS;
        }
    } catch(const SrmBadResponse& exc){
        error() << "Invalid result returned by the remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_REQUEST_FAILURE;
        m = exc.what();
    } catch(const LogicError& exc){ // TBM ? maybe scope should be set to agent
        error() << "Logic Error in invoking GetSpaceTokens on remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_INTERNAL_ERROR;
        m = exc.what();
    } catch(const SecurityError& exc){
        error() << "User is not authorized to contact the remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_PERMISSION;
        m = exc.what();
    }
    return result;
}
 
