/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Common Structure definitions for Stat
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: StatCommon.h,v 1.17 2009/12/03 17:21:44 molnarzs Exp $
 *  Release: $Name:  $
 */
#ifndef GLITE_DATA_TRANSFER_URL_COPY_STAT_COMMON_H_
#define GLITE_DATA_TRANSFER_URL_COPY_STAT_COMMON_H_

#include <glite/data/agents/glite-types.h>
#include <ostream>

namespace glite   {
namespace data    {
namespace transfer{
namespace urlcopy {

/**
 * The version of the stat files. Increase whenever the data structures are
 * modified.
 */
#define TRANSFER_STAT_VERSION     0x00030501

/**
 * UrlCopy Stat Fields size
 */
#define TRANSFER_STAT_URL_LEN       2048
#define TRANSFER_STAT_VERSION_LEN   16
#define TRANSFER_STAT_TOKEN_LEN     256
#define TRANSFER_STAT_MESSAGE_LEN   2048
#define CHECKSUM_ALGORITHM_LEN      32
#define CHECKSUM_VALUE_LEN          256
#define SYSLOG_IDENT_LEN            256
#define SYSLOG_FACILITY_NAME_LEN    256

/**
 * UrlCopy Error Structure
 */
struct ErrorStat {
    // the error scope (source, destination, transfer)
    int          mm_scope;
    // the error phase (preparation, transfer, finalization)
    int          mm_phase;
    // the error category (invalid path, gridftp, markers timeout etc)
    int          mm_category;
    // the error message
    char         mm_message[TRANSFER_STAT_MESSAGE_LEN];

    /**
     * isSet
     * @return true if error information is set for the object.
     */
    bool isSet(){
        return (mm_scope || mm_phase || mm_category);
    }
};

std::ostream& operator << (std::ostream& s, const ErrorStat& stat);

/**
 * Stat Version Structure
 */
struct TransferStat{
    // Stat Version
    unsigned int mm_version;
    // Stat Version extension
    unsigned int mm_extension;
    // Stat Identifier
    unsigned int mm_id;
    // Transfer Request Start Time
    time_t       mm_start_time;
    // Transfer Request Finish Time
    time_t       mm_finish_time;
    // Last Refresh Time
    time_t       mm_last_refresh_time;
    // Id of the transfer process that serves the request
    unsigned int mm_pid;
    // Common properties
#ifdef _USE_GLOBAL_TIMEOUT
    int          mm_global_timeout;
#endif
    unsigned int mm_verbose;
    unsigned int mm_flags;
    unsigned int mm_overwrite;
    // Destination Space Token
    char         mm_dest_space_token[TRANSFER_STAT_TOKEN_LEN];
    // Destination Space Token Descritpion
    char         mm_dest_token_descr[TRANSFER_STAT_MESSAGE_LEN];
    // User Description Tag
    char         mm_user_tag[TRANSFER_STAT_TOKEN_LEN];
    // added in 2_1_0_1
    //--------------------------------------------------------------------------
    // Source Space Token
    char         mm_source_space_token[TRANSFER_STAT_TOKEN_LEN];
    // Source Space Token Descritpion
    char         mm_source_token_descr[TRANSFER_STAT_MESSAGE_LEN];
    // flag to enable/disable checks for existance of target directory
    unsigned int mm_enable_dir_check;
    // flag to enable/disable ping
    unsigned int mm_enable_ping_source;
    unsigned int mm_enable_ping_dest;
    // flag to enable/disable ping
    unsigned int mm_block_after_preparation;
    unsigned int mm_block_after_transfer;
    // added with stat version 3.3.0
    //--------------------------------------------------------------------------
    // lifetime of the copy of the file
    int mm_copy_pin_lifetime;
    // use connection type = LAN for the transfer
    int mm_lan_connection;
    // fail the transfer immediately if the file is nearline
    int mm_fail_nearline_files;

    // Syslog parameters
    int mm_syslog_disable;
    char mm_syslog_ident[SYSLOG_IDENT_LEN];
    int mm_syslog_facility;
    char mm_syslog_facility_name[SYSLOG_FACILITY_NAME_LEN];
};

std::ostream& operator << (std::ostream& s, const TransferStat& stat);

/**
 * return value if not null and not empty, default algorithm otherwise
 */
const char * getChecksumAlgorithm(const char * value);

} // end namespace urlcopy
} // end namespace transfer
} // end namespace data
} // end namespace glite

#endif //GLITE_DATA_TRANSFER_URL_COPY_STAT_COMMON_H_
