/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004-2008.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  Token selection strategies
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: TokenSelector.cpp,v 1.1 2008/10/22 08:51:37 ptedesco Exp $
 *  Release: $Name:  $
 *
 */

#include "TokenSelector.h"
#include "logger-utility.h"
#include <glite/data/srm/util/GetSpaceMetaData.h>
#include <glite/data/srm/util/types.h>
#include <boost/scoped_ptr.hpp>
#include <algorithm>
#include <ostream>
#include <vector>

using namespace glite::data::transfer::urlcopy;
using namespace glite::data::srm::util;
using namespace glite::data::agents;

namespace{
    
UrlCopyError::Category errorFromGetSpaceMetaDataStatus(SrmStatus::Code code){
    switch(code){
    case SrmStatus::SRM_AUTHENTICATION_FAILURE:
    case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        return UrlCopyError::ERROR_PERMISSION;
    case SrmStatus::SRM_TOO_MANY_RESULTS: // Request produced too many results that SRM cannot handle.
    case SrmStatus::SRM_INVALID_REQUEST: // arrayOfSpaceToken is empty. 
    case SrmStatus::SRM_INTERNAL_ERROR: // SRM has an internal transient error, and client may try again.
    case SrmStatus::SRM_FAILURE: // All space requests are failed.
    case SrmStatus::SRM_NOT_SUPPORTED:
        return UrlCopyError::ERROR_STORAGE_INTERNAL_ERROR;
    default:
        warn() << "SRM returned invalid status " << code;
        return UrlCopyError::ERROR_STORAGE_INTERNAL_ERROR;
        break;
    }
    
}

template<typename T>
std::ostream& operator << (std::ostream& stream,const std::vector<T>& v){
    stream << "[";
    bool first = true;
    for (typename std::vector<T>::const_iterator it = v.begin();it != v.end();++it){
        if (!first){
            stream << ", ";
        }
        stream << (*it);
        first = false;
    }
    stream << "]";
    return stream;
}


}

/**-----------------------------------------------------------------------------
 * SelectFirstToken
------------------------------------------------------------------------------*/

/**
 * selectSpaceToken
 */
std::string SelectFirstToken::selectSpaceToken(
    SrmMethodCaller&                caller,
    Context::Ptr                    ctx,
    const std::vector<std::string>& tokens,
    UrlCopyError::Category&         category,
    std::string&                    message){
    debug() << "Select first token strategy";    
    std::string space_token;
    category = UrlCopyError::SUCCESS;
    if (tokens.size() > 0){
        if(tokens.size() > 1){
            warn() << tokens.size() << " tokens found. The first one will be used.";
        }
        space_token = tokens.at(0);
        debug() << "got token [" << space_token << "]";        
    } else {
        category = UrlCopyError::ERROR_USER;
        message = "Empty tokens list";
    }
    return space_token;
}

/**-----------------------------------------------------------------------------
 * SelectUnusedSpace
------------------------------------------------------------------------------*/

const glite::data::agents::longlong SelectUnusedSpace::FileSizeMargin = 1 << 20; // 1MB

/**
 * constructor
 */
SelectUnusedSpace::SelectUnusedSpace(
    glite::data::agents::longlong   fileSize,
    int                             numFiles){
    m_fileSize = fileSize + (numFiles * FileSizeMargin);
    debug() << "Select unused space token strategy: size=" << fileSize << " num=" << 
        numFiles << " total=" << m_fileSize;
}

/**
 * selectSpaceToken
 */
std::string SelectUnusedSpace::selectSpaceToken(
    SrmMethodCaller&                caller,
    Context::Ptr                    ctx,
    const std::vector<std::string>& tokens,
    UrlCopyError::Category&         category,
    std::string&                    message){
        
    std::string space_token;
    category = UrlCopyError::SUCCESS;
    glite::data::agents::longlong bestUnusedSize = 0;
    bool space_found = false;
    if (0 == tokens.size()){
        // empty tokens list
        category = UrlCopyError::ERROR_USER;
        message = "Empty tokens list";
        return "";
    } 
    // iterate through tokens
    boost::scoped_ptr<GetSpaceMetaData> gsmd_request(GetSpaceMetaData::create(ctx));
    gsmd_request->tokens.resize(tokens.size());
    std::copy(tokens.begin(),tokens.end(),gsmd_request->tokens.begin());
    SrmStatus status = caller.executeRequest(ctx,*gsmd_request,category,message);
    if (UrlCopyError::SUCCESS != category){
        // error executing the request
        error() << "Error invoking GetSpaceMetaData: [" << category << "] " << message;
        return "";
    }
    if (SrmStatus::SRM_SUCCESS != status.code &&
        SrmStatus::SRM_PARTIAL_SUCCESS != status.code) {
        // SRM request failed
        std::stringstream ss;
        ss << "GetSpaceMetaData for tokens " << tokens << " failed: " << status;
        error() << ss.str();
        category = errorFromGetSpaceMetaDataStatus(status.code);
        message = ss.str();
        return "";
    }
    // iterate return results
    const std::vector<MetaDataSpace>& spaces = gsmd_request->spaceDetails;
    std::vector<MetaDataSpace>::const_iterator it;
    for (it = spaces.begin();it != spaces.end();++it){
        const MetaDataSpace& space = (*it);
        if (SrmStatus::SRM_SUCCESS != space.status.code){
            warn() << "Getting meta data for space token " << space.spaceToken << " failed: " << space.status;
            continue;
        }
        if (space.unusedSize < m_fileSize){
            debug() << "Space " << space.spaceToken << " has insufficient unusedSize: " << space.unusedSize << " (" << m_fileSize << " required)";
            continue;
        }
        debug() << "Space " << space.spaceToken << " has sufficient unusedSize: " << space.unusedSize << " (" << m_fileSize << " required)";
        space_found = true;
        if (0 == bestUnusedSize){
            bestUnusedSize = space.unusedSize;
        }
        if (bestUnusedSize >= space.unusedSize){
            debug() << "Selecting space " << space.spaceToken << " as best match";
            space_token = space.spaceToken;
        } else {
            debug() << "Space " << space.spaceToken << " is not the best match";
        }
    }
    if (space_found){
        info() << "Selected space token " << space_token;
    } else {
        category = UrlCopyError::ERROR_NO_SPACE_LEFT;
        std::stringstream ss;
        ss << "No space found with at least " << m_fileSize << " bytes of unusedSize";
        error() << ss.str();
        message = ss.str();
    }
    return space_token;
}
