/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004-2008.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  Token selection strategies
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: TokenSelector.h,v 1.1 2008/10/22 08:51:37 ptedesco Exp $
 *  Release: $Name:  $
 *
 */
#ifndef GLITE_DATA_TRANSFER_URL_COPY_TOKENSELECTOR_H_
#define GLITE_DATA_TRANSFER_URL_COPY_TOKENSELECTOR_H_

#include <vector>
#include <string>
#include <boost/shared_ptr.hpp>
#include <glite/data/srm/util/Context.h>
#include <glite/data/transfer/urlcopy/UrlCopyError.h>
#include <glite/data/agents/glite-types.h>

#include "SrmMethodCaller.h"

namespace glite   {
namespace data    {
namespace transfer{
namespace urlcopy {

/**
 * base class for token selection strategies
 */
class ITokenSelector{
public:
    typedef boost::shared_ptr<ITokenSelector> Ptr;

    /**
     * select the space token from the list
     * @param caller [IN] SRM method caller object
     * @param ctx [IN] the SRM call context
     * @param tokens [IN] the list of tokens
     * @param category [OUT] error category
     * @param message [OUT] error message
     * @return token string if successful, empty string otherwise
     */
    virtual std::string selectSpaceToken(
        SrmMethodCaller&                        caller,
        glite::data::srm::util::Context::Ptr    ctx,
        const std::vector<std::string>&         tokens,
        UrlCopyError::Category&                 category,
        std::string&                            message) = 0;
    
protected:
    
    virtual ~ITokenSelector(){} // declare constructor virtual to suppress gcc warning
};

/**
 * select the first token (default behaviour until FTS 2.1)
 */
class SelectFirstToken : public ITokenSelector{
public:
    /**
     * Constructor
     */
    SelectFirstToken(){}
    
    /**
     * Destructor
     */
    ~SelectFirstToken(){}
    
    /**
     * selectSpaceToken
     */
    virtual std::string selectSpaceToken(
        SrmMethodCaller&                        caller,
        glite::data::srm::util::Context::Ptr    ctx,
        const std::vector<std::string>&         tokens,
        UrlCopyError::Category&                 category,
        std::string&                            message);
};

/**
 * select the token with less unused space among those that have
 * enough space for the file size (plus a margin)
 */
class SelectUnusedSpace : public ITokenSelector{
public:
    /**
     * Size margin per file
     */
    static const glite::data::agents::longlong FileSizeMargin;
    
    /**
     * Constructor
     * @param fileSize [IN] total file size
     * @param numFiles [IN] number of files
     */
    SelectUnusedSpace(
        glite::data::agents::longlong fileSize,
        int                           numFiles);
    
    /**
     * Destructor
     */
    ~SelectUnusedSpace(){};
    
    /**
     * selectSpaceToken
     */
    virtual std::string selectSpaceToken(
        SrmMethodCaller&                        caller,
        glite::data::srm::util::Context::Ptr    ctx,
        const std::vector<std::string>&         tokens,
        UrlCopyError::Category&                 category,
        std::string&                            message);
protected:
    glite::data::agents::longlong m_fileSize;
};

} // namespace urlcopy
} // namespace transfer
} // namespace data
} // namespace glite

#endif // GLITE_DATA_TRANSFER_URL_COPY_TOKENSELECTOR_H_
