/*
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: UrlCopy Stat helper function
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *  Version info: $Id: TransferConfig.cpp,v 1.6 2008/10/20 15:19:46 ptedesco Exp $
 *  Release: $Name:  $
 */

#include "TransferConfig.h"
#include <glite/data/agents/AgentExceptions.h>

#include <sstream>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <paths.h>
#include <errno.h>

using namespace glite::data::transfer::urlcopy;

using glite::data::agents::RuntimeError;
using glite::data::agents::PosixError;

namespace {
const char * const URL_COPY_NAME = "glite-url-copy";
const char * const COMPLETED_DIR = "completed";
const char * const FAILED_DIR    = "failed";
const char * const LOST_DIR      = "lost";
const char * const EXT_MEM       = ".mem";
const char * const EXT_LOG       = ".log";
}

/**
 * Return singleton instance
 */
TransferConfig& TransferConfig::instance() throw() {
    static TransferConfig s_instance;
    return s_instance;
}

    
/*
 * name
 *
 * Return the glite-url-copy name
 */
const char * TransferConfig::name() const throw(){
    return URL_COPY_NAME;
}

/*
 * extMem
 *
 * Return the mem extension
 */
const char * TransferConfig::extMem() const throw(){
    return EXT_MEM;
}

/*
 * extLog
 *
 * Return the log extension
 */
const char * TransferConfig::extLog() const throw(){
    return EXT_LOG;
}

/**
 * Return the location where to store the mmap and the log files
 */
std::string TransferConfig::repository() const /*throw(RuntimeError)*/{
// TODO Use SHM ?
//#ifdef HAVE_SHM_OPEN 
    // For Shared Memory, the path should start with / and contain no embedded
    // slashes. 
    // Please note that in case of SHM  I can't use the same dir for log and 
    // mem files, so the cose would have to be changed!
    // return "/";
//#else    
    static std::string s_path;
    // Check it just once, since the process should not change uid
    if(true == s_path.empty()){
        // Get the user account name
        struct passwd *pw = ::getpwuid(getuid());
        int count = 0;
        while((0 == pw) && (count < 20)) {
            // in rare cases quattor updates the pw file just now
            sleep(3);
            pw = ::getpwuid(getuid());
            ++count;
        }
        if(0 == pw){
            throw RuntimeError("failed to get the user's login name");
        }
        
        std::stringstream ss;
        ss << _PATH_VARTMP << URL_COPY_NAME << "-" << pw->pw_name << "/";
        s_path = ss.str();
        
        // Check if path exists
        if(0 != ::access(s_path.c_str(), F_OK)){
            // Create the path
            int r = ::mkdir(s_path.c_str(),0755);
            if(r != 0){
                throw PosixError("cannot create repository",errno);
            }
        }
    }
//#endif HAVE_SHM_OPEN
    return s_path;
}

/*
 * archive
 *
 * Return the location where to archive the mmap and the log files
 */
std::string TransferConfig::archive(bool completed, bool lost, const std::string& prefix) const /*throw(RuntimeError)*/{
    static std::string s_completed_path;
    static std::string s_failed_path;
    static std::string s_lost_path;
    
    // Check it just once, since the process should not change uid
    if(((true  == completed) && (true == s_completed_path.empty())) ||
       ((false == completed) && (false == lost) && (true == s_failed_path.empty())) ||
       ((false == completed) && (true == lost) && (true == s_lost_path.empty())) ){
// TODO Use SHM ?
//#ifdef HAVE_SHM_OPEN 
        // TODO create the directory in case it use SHM
//#else
        std::stringstream ss;    
        ss << this->repository();
        if(false == prefix.empty()){
            ss << prefix;
        }
        if(true == completed){
            ss << COMPLETED_DIR;
        } else {
            if (lost) {
                ss << LOST_DIR;
            }
            else {
                ss << FAILED_DIR;
            }
        }
        ss << "/";
        std::string path = ss.str();
        
        // Check if path exists
        if(0 != ::access(path.c_str(), F_OK)){
            // Create the path
            int r = ::mkdir(path.c_str(),0755);
            if(r != 0){
                throw PosixError("cannot create archive repository",errno);
            }
        }
        if(true == completed){
            s_completed_path = path;
        } else {
            if (false == lost){
                s_failed_path = path;
            }
            else
            {
                s_lost_path = path;
            }
        }
    }
//#endif HAVE_SHM_OPEN
    if(true == completed){
        return s_completed_path;
    } else {
        if (false == lost){
            return s_failed_path;
        }
        else
        {
            return s_lost_path;
        }
    }
}
