/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: TransferCore class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *           Zsolt Molnar <Zsolt.Molnar@cern.ch - www.zsoltmolnar.hu>
 *  Version info: $Id: TransferCore.cpp,v 1.26 2009/07/30 19:32:55 molnarzs Exp $
 *  Release: $Name:  $
 */
#include "TransferCore.h"
#include "logger-utility.h"
#include "stat-utility.h"
#include "sighandler-utility.h"

#include <glite/data/srm/util/Context.h>
#include <glite/data/srm/util/SrmExceptions.h>
#include <glite/data/srm/util/SrmPing.h>
#include <glite/data/srm/util/PrepareToGet.h>
#include <glite/data/srm/util/BringOnline.h>
#include <glite/data/srm/util/SrmStat.h>
#include <glite/data/srm/util/GetSpaceTokens.h>
#include <glite/data/srm/util/SrmRm.h>
#include <glite/data/srm/util/SrmMkDir.h>

#include <glite/data/agents/extra/utility.h>
#include <glite/data/agents/url-utility.h>
#include <glite/data/agents/surl-utility.h>

#include <string>
#include <sstream>

#include <string.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <libgen.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stdlib.h>

#include <boost/scoped_ptr.hpp>
#include <boost/algorithm/string.hpp>

using namespace glite::data::transfer::urlcopy;
using namespace glite::data::srm::util;

using glite::data::srm::util::details::SynchRequest;
using glite::data::srm::util::details::AsynchRequest;
using glite::data::agents::get_dirname;
using glite::data::agents::parse_url;
using glite::data::agents::create_url;
using glite::data::agents::path_from_surl;

namespace {

const char * const SFN_TAG                       = "SFN";

//------------------------------------------------------------------------------
// Helper Methods
//------------------------------------------------------------------------------

/**
 * generate_surl_from_path
 * Generate a surl from a sfn path name
 */
std::string generate_surl_from_path(const std::string& name, const std::string& file_surl) throw() {
    std::string surl;
    // Check if the user provided a SURL is compact or fully qualified format
    try{
        std::string  hostname;
        unsigned int port = 0;
        std::string  protocol;
        std::string  path;
        std::string  query_str;
        // Parse surl
        parse_url(file_surl,hostname,port,protocol,path,query_str);
        // Check query string
        if(query_str.empty()){
            // It's compact
            surl = create_url(protocol,hostname,port,name);
        } else {
            // It's fully qualified
            query_str = (std::string)SFN_TAG + "=" + name;
            surl = create_url(protocol,hostname,port,path,query_str);
        }
    } catch(...){
        // Invalid file SURL
        surl = "";
    }
    return surl;
}

/**
 * get_file_path
 * Extract the file path from a surl
 */
std::string get_file_path(const std::string& surl) throw () {
    std::string file_path;
    try{
        file_path = path_from_surl(surl);
    } catch(...){
        // Invalid SURL
        file_path = "";
    }
    return file_path;
}

} // end global namespace


/**
 * Initialize the stat
 */
void TransferCore::initialize() /*throw (RuntimeError, LogicError)*/{
    if(0 != m_mmap.get()){
        // Already initialized
        return;
    }
    try{
        // Open the Stat file using a Mmap object
        m_mmap = open_stat(m_id,false);
    } catch(...){
        m_mmap.reset();
        throw;
    }
}

/**
 * pingSrm
 * Send Ping Request
 */
void TransferCore::pingSrm(
    Context::Ptr&           ctx,
    UrlCopyError::Category& c,
    std::string&            m,
    const std::string&      name) throw() {
    // Create SrmPing Request
    boost::scoped_ptr<SrmPing> ping_request;
    try{
        // Create SrmPing Request
        ping_request.reset(SrmPing::create(ctx));
    } catch(const LogicError& exc){
        // SrmPing method is not supported
        c = UrlCopyError::ERROR_INTERNAL_ERROR;
        m = exc.what();
        return;
    }

    // Execute SRM Ping
    SrmStatus ping_result = m_srmMethodCaller.executeRequest(ctx,*ping_request,c,m);
    if(c == UrlCopyError::SUCCESS){
        if(ping_result.code == SrmStatus::SRM_SUCCESS){
            // OK
            std::stringstream ss;
            ss << name << " SRM server [" << ctx->endpoint << "] available.";
            for (ExtraInfo::const_iterator it = ping_request->otherInfo.begin();it != ping_request->otherInfo.end();++it){
                ss << " " << it->first << "='" << it->second << "'";
            }
            info() << ss.str();
        } else {
            // SrmPing Failed
            c = UrlCopyError::ERROR_GENERAL_FAILURE;
            if(false == ping_result.message.empty()){
                m = ping_result.message;
            } else {
                std::stringstream ss;
                ss << "SrmPing on " << name << " SRM server [" << ctx->endpoint << "] failed with error [" << ping_result.code << "]";
                m = ss.str();
            }
        }
    }
}

/**
 * pathExists
 */
bool TransferCore::pathExists(
    glite::data::srm::util::Context::Ptr&           ctx,
    const std::string&                              surl,
    UrlCopyError::Category&                         c,
    std::string&                                    m)
{
    TransferUtils::PathInfo info;
    return TransferUtils::PathExists(this, ctx,surl,false,c,m,info);
}

/**
 * pathExists
 */
bool TransferCore::pathExists(
    Context::Ptr&           ctx,
    const std::string&      surl,
    UrlCopyError::Category& c,
    std::string&            m,
    TransferUtils::PathInfo& info)
{
    return TransferUtils::PathExists(this, ctx,surl,true,c,m,info);
}


/**
 * pathExists
 * Check if a path (file or dircetory) already exists
 */
bool TransferCore::pathExists(
    Context::Ptr&                               ctx,
    const std::string&      surl,
    bool                    full_detailed,
    UrlCopyError::Category& c,
    std::string&            m,
    TransferUtils::PathInfo& pinfo)
{
	return TransferUtils::PathExists(this, ctx, surl, full_detailed, c, m, pinfo);
}

/**
 * errorFromRequestStatus
 * Get an UrlCopy error from an SRM request (in case the files did not contain
 * an error code)
 */
void TransferCore::errorFromRequestStatus( // [conversion]
    UrlCopyError::Category&                     c,
    std::string&                                m,
    const glite::data::srm::util::SrmStatus&    req_status) /* throw() */
{
	TransferUtils::ErrorFromRequestStatus(c, m, req_status);
}

/**
 * statErrorFromSrmStatus
 * Get an UrlCopy error from an SRM error for SrmStat
 */
void TransferCore::statErrorFromSrmStatus( // [conversion]
    UrlCopyError::Category&                     c,
    std::string&                                m,
    const glite::data::srm::util::SrmStatus&    req_status,
    const glite::data::srm::util::SrmStatus&    file_status) throw()
{
	TransferUtils::StatErrorFromSrmStatus(c, m, req_status, file_status);
}

/**
 * getSpaceToken
 * Get the space token from the user description
 */
std::string TransferCore::getSpaceToken(
    Context::Ptr&           ctx,
    const std::string&      description,
    UrlCopyError::Category& c,
    std::string&            m,
    ITokenSelector::Ptr     selector) /*throw ()*/ {

    c = UrlCopyError::SUCCESS;
    if (!GetSpaceTokens::isSupported(ctx)){
        warn() << "GetSpaceTokens operation is not supported by endpoint '" << ctx->endpoint << "'. Ignoring token description.";
        return "";
    }

    info() << "getting token for description '" << description << "'";

    // Create the Directory
    boost::scoped_ptr<GetSpaceTokens> gst_request;
    // Create GetSpaceTokens Request
    gst_request.reset(GetSpaceTokens::create(ctx));

    // Set Request Parameters
    gst_request->description = description;

    SrmStatus gst_result = m_srmMethodCaller.executeRequest(ctx,*gst_request,c,m);
    if (c != UrlCopyError::SUCCESS){
        error() << "Failed getting space token for description '" << description << "' : [" << c << "] " << m;
        return "";
    }
    if (gst_result.code != SrmStatus::SRM_SUCCESS){
        error() << "GetSpaceTokens request for description '" << description << "' failed: " << gst_result;
        getSpaceTokensErrorFromSrmStatus(c,m,gst_result);
        return "";
    }
    if(gst_request->tokens.empty()){
        std::stringstream ss;
        ss << "No token found for description '" << description << "'";
        error() << ss.str();
        c = UrlCopyError::ERROR_USER;
        m = ss.str();
        return "";
    }

    std::string space_token = selector->selectSpaceToken(m_srmMethodCaller,ctx,gst_request->tokens,c,m);
    if (UrlCopyError::SUCCESS == c){
        info() << "token found for [" << description << "]: " << space_token;
    } else {
        error() << "error getting token for description (" << UrlCopyError::categoryToString(c) << ") : " << m;
    }
    return space_token;
}

/**
 * getSpaceTokensErrorFromSrmStatus
 * Get an UrlCopy error from an SRM error for GetSpaceTokens
 */
void TransferCore::getSpaceTokensErrorFromSrmStatus( // [conversion]
    UrlCopyError::Category&                     c,
    std::string&                                m,
    const glite::data::srm::util::SrmStatus&    req_status) throw(){

    switch(req_status.code){
        case SrmStatus::SRM_AUTHENTICATION_FAILURE:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
            c = UrlCopyError::ERROR_PERMISSION;
            m = req_status.message;
        break;
        case SrmStatus::SRM_NOT_SUPPORTED:
            // This case should have already been handled
            c = UrlCopyError::ERROR_CONSISTENCY;
            m = req_status.message;
        break;
        case SrmStatus::SRM_INVALID_REQUEST:
            c = UrlCopyError::ERROR_USER;
            m = req_status.message;
        break;
        case SrmStatus::SRM_INTERNAL_ERROR:
            c = UrlCopyError::ERROR_STORAGE_INTERNAL_ERROR;
            m = req_status.message;
        break;
        case SrmStatus::SRM_UNDEF:
        case SrmStatus::SRM_SUCCESS: {
            // The error is neither on the Request status. This in principle
            // should not be possible
            c = UrlCopyError::ERROR_INTERNAL_ERROR;
            std::stringstream ss;
            ss << "The SRM returned an error, but with an invalid code [" << req_status.code << "]";
            m = ss.str();
        } break;
        default:
            // Set default error
            c = UrlCopyError::ERROR_GENERAL_FAILURE;
            if(false == req_status.message.empty()){
                m = req_status.message;
            } else {
                std::stringstream ss;
                ss << "GetSpaceTokens SRM request failed with error [" << req_status.code << "]";
                m = ss.str();
            }
        break;
    }
}

/**
 * rmFile
 * Remove a SURL
 */
bool TransferCore::rmFile(Context::Ptr& ctx, const std::string& surl, UrlCopyError::Category& c, std::string& m) /*throw ()*/
{
	return TransferUtils::RmFile(this, ctx, surl, c, m);
}

/**
 * bringOnline
 * @param ctx [IN]
 * @param surl [IN]
 * @param copyPinLifetime [IN]
 */
void TransferCore::bringOnline(
    Context::Ptr&       ctx,
    const std::string&  surl,
    int                 copyPinLifetime){
    std::vector<std::string> surls(1);
    surls[0] = surl;
    bringOnline(ctx,surls,copyPinLifetime);
}

/**
 * bringOnline
 * @param ctx [IN]
 * @param surl [IN]
 * @param copyPinLifetime [IN]
 */
void TransferCore::bringOnline(
    glite::data::srm::util::Context::Ptr&   ctx,
    const std::vector<std::string>&         surls,
    int                                     copyPinLifetime){
    info() << "Calling srmBringOnline for " << surls.size() << " file(s).";
    boost::scoped_ptr<BringOnline> bringOnlineRequest;
    try{
        // Create PrepareToPut Request
        bringOnlineRequest.reset(BringOnline::create(ctx));
    } catch(const LogicError& exc){
        // method is not supported
        warn() << "LogicError creating BringOnline request: " << exc.what() << ". Ignoring the error.";
        return;
    }
    size_t num_of_files = surls.size();
    bringOnlineRequest->files.resize(num_of_files);
    for (size_t i = 0;i < num_of_files;++i){
        const std::string& surl = surls[i];
        debug() << "Adding SURL to BringOnline request: " << surl;
        bringOnlineRequest->files[i].surl = surl;
    }
    bringOnlineRequest->desiredLifeTime = copyPinLifetime;

    try{
        RefreshTimeUpdater updater(this);
        bringOnlineRequest->execute();
        info() << "BringOnline called, token='" << bringOnlineRequest->token() << "'. Status: " << bringOnlineRequest->status;
    } catch(const HttpTimeoutError& exc){
        warn() << "HTTP timeout contacting remote SRM [" << ctx->endpoint << "]: "
            << exc.what() << ". Ignoring the error.";
        return;
    } catch(const RemoteError& exc){
        warn() << "Failed to contact remote SRM [" << ctx->endpoint << "]: "
            << exc.what() << ". Ignoring the error.";
        return;
    } catch(const SecurityError& exc){
        warn() << "User is not authorized to contact the remote SRM [" << ctx->endpoint << "]: "
            << exc.what() << ". Ignoring the error.";
        return;
    } catch(const SrmError& exc){
        warn() << "Remote method invocation failure on the remote SRM [" << ctx->endpoint << "]: "
            << exc.what() << ". Ignoring the error.";
        return;
    } catch(const SrmBadResponse& exc){
        warn() << "Invalid result returned by the remote SRM [" << ctx->endpoint << "]: "
            << exc.what() << ". Ignoring the error.";
        return;
    } catch(const LogicError& exc){
        warn() << "Logic Error in invoking PrepareToGet on remote SRM [" << ctx->endpoint << "]: "
            << exc.what() << ". Ignoring the error.";
        return;
    }

}

/**
 * rmErrorFromSrmStatus
 * Get an UrlCopy error from an SRM error for SrmRm
 */
void TransferCore::rmErrorFromSrmStatus( // [conversion]
    UrlCopyError::Category&                     c,
    std::string&                                m,
    const glite::data::srm::util::SrmStatus&    req_status,
    const glite::data::srm::util::SrmStatus&    file_status)
{
	TransferUtils::RmErrorFromSrmStatus(c, m, req_status, file_status);
}

/**
 * makeDir
 * Create the parent directory path of the given path
 */
bool TransferCore::makeDir(Context::Ptr& ctx, const std::string& dir_name, UrlCopyError::Category& c, std::string& m) /*throw ()*/ {
    bool directory_created = false;

    // Create the Directory
    boost::scoped_ptr<SrmMkDir> mkdir_request;
    try{
        // Create SrmMkDir Request
        mkdir_request.reset(SrmMkDir::create(ctx));
    } catch(const LogicError& exc){
        // SrmMkDir method is not supported
        c = UrlCopyError::ERROR_INTERNAL_ERROR;
        m = exc.what();
        return false;
    }
    // Set Request Parameters
    mkdir_request->directoryPath = dir_name;

    SrmStatus mkdir_result = m_srmMethodCaller.executeRequest(ctx,*mkdir_request,c,m);
    if(c == UrlCopyError::SUCCESS){
        if(mkdir_result.code == SrmStatus::SRM_SUCCESS){
            // the directory already exists
            info() << "directory [" << dir_name << "] created";
            directory_created = true;
        } else {
            directory_created = false;
            error() << "directory [" << dir_name << "] creation failed. [" << mkdir_result.code << "] : " << mkdir_result.message;
            // Detect the correct UrlCopy error code from SrmStatus
            mkDirErrorFromSrmStatus(c,m,mkdir_result);
        }
    } else {
        directory_created = false;
        error() << "failed to create directory [" << dir_name << "] : " << m;
    }
    return directory_created;
}

/**
 * mkDirErrorFromSrmStatus
 * Get an UrlCopy error from an SRM error for MkDir
 */
void TransferCore::mkDirErrorFromSrmStatus(
    UrlCopyError::Category&                     c,
    std::string&                                m,
    const glite::data::srm::util::SrmStatus&    req_status) /*throw()*/{

    switch(req_status.code){
        case SrmStatus::SRM_AUTHENTICATION_FAILURE:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
            c = UrlCopyError::ERROR_PERMISSION;
            m = req_status.message;
        break;
        case SrmStatus::SRM_INVALID_PATH:
            c = UrlCopyError::ERROR_INVALID_PATH;
            m = req_status.message;
        break;
        case SrmStatus::SRM_DUPLICATION_ERROR:
            c = UrlCopyError::ERROR_FILE_EXISTS;
            m = req_status.message;
        break;
        case SrmStatus::SRM_UNDEF:
        case SrmStatus::SRM_SUCCESS: {
            // The error is neither on the Request status. This in principle
            // should not be possible
            c = UrlCopyError::ERROR_INTERNAL_ERROR;
            std::stringstream ss;
            ss << "The SRM returned an error, but with an invalid code [" << req_status.code << "]";
            m = ss.str();
        } break;
        default:
            // Set default error
            c = UrlCopyError::ERROR_GENERAL_FAILURE;
            if(false == req_status.message.empty()){
                m = req_status.message;
            } else {
                std::stringstream ss;
                ss << "mkDir SRM request failed with error [" << req_status.code << "]";
                m = ss.str();
            }
        break;
    }
}

/**
 * prepareDestinationPath
 *
 * Create the path on the target SRM
 * @param ctx [IN] SRM Context
 * @param destination_name [IN] the destination name as specified by the user
 * @return true if successful, false otherwise
 */
bool TransferCore::prepareDestinationPath(
    Context::Ptr&       ctx,
    const std::string&  destination_name) /*throw ()*/ {
    if(false == SrmMkDir::isSupported(ctx)){
        debug() << "Do not create destination path since MkDir is not supported in SRM version [" << ctx->version << "]";
        return true;
    }
    std::string file_path = get_file_path(destination_name);
    if(true == file_path.empty()){
        std::stringstream reason;
        reason << "User specified the SURL [ "
               << destination_name
               << "] in an invalid format: cannot determine file path";
        throw UrlCopyError(UrlCopyError::SCOPE_DESTINATION,
            UrlCopyError::PHASE_PREPARATION,
            UrlCopyError::ERROR_USER,
            reason.str());
    }
    return prepareParentDir(
        ctx,
        file_path,
        destination_name);
}

/**
 * prepareParentDir
 *
 * Create the parent directory path of the given path
 * @param ctx [IN] SRM Context
 * @param path [IN] SURL of target directory
 * @param destination_name [IN] the destination name as specified by the user
 * @return true if successful, false otherwise
 */
bool TransferCore::prepareParentDir(
    Context::Ptr&       ctx,
    const std::string&  path,
    const std::string&  destination_name) /*throw ()*/ {

    debug() << "checking if the parent directory of [" << path << "] already exists";

    bool path_ready = false;
    // Get the parent directory name
    std::string dir_name = get_dirname(path);
    if(true == dir_name.empty()){
        // In principle this should never happen
        std::stringstream reason;
        reason << "User specified the SURL [ "
               << destination_name
               << "] in an invalid format: cannot determine directory path";
        throw UrlCopyError(UrlCopyError::SCOPE_DESTINATION,
            UrlCopyError::PHASE_PREPARATION,
            UrlCopyError::ERROR_USER,
            reason.str());
    }

    if (path == dir_name){
        debug() << "directory is root, assuming existence.";
        return true;
    }

    debug() << "checking exististence of [" << dir_name << "] in SRM [" << ctx->endpoint << "]";

    // Generate a surl for the dirname
    std::string dir_surl_name =
        generate_surl_from_path(dir_name, destination_name);
    if(true == dir_surl_name.empty()){
        std::stringstream reason;
        reason << "User specified the SURL [ "
               << destination_name
               << "] in an invalid format: cannot create SURL for parent directory path";
        throw UrlCopyError(UrlCopyError::SCOPE_DESTINATION,
            UrlCopyError::PHASE_PREPARATION,
            UrlCopyError::ERROR_USER,
            reason.str());
    }
    debug() << "directory SURL name is [" << dir_surl_name << "]";

    // Check if the directory exists
    UrlCopyError::Category c;
    std::string m;
    path_ready = pathExists(ctx,dir_surl_name,c,m);
    if(true == path_ready){
        // the directory already exists
        debug() << "directory path [" << dir_surl_name << "] exists";
    } else {
        // Check the file status
        if(c == UrlCopyError::SUCCESS){
            info() << "target directory path [" << dir_surl_name << "] does not exist on destination SRM [" << ctx->endpoint << "]";
            // Recursive call
            path_ready = prepareParentDir(ctx,dir_name,destination_name);
            if(true == path_ready){
                // Create Parent Directory
                path_ready = makeDir(ctx,dir_surl_name,c,m);

                // if path_ready is false and c is ERROR_FILE_EXISTS, probably
                // the first time the directory existence was checked the
                // directory did not yet exist, but then another transfer
                // process created it; ignore the error and set path_ready to
                // true
                if((false == path_ready) && (c == UrlCopyError::ERROR_FILE_EXISTS)){
                    path_ready = true;
                }

                if((false == path_ready) && (c != UrlCopyError::ERROR_FILE_EXISTS)){
                    throw UrlCopyError(UrlCopyError::SCOPE_DESTINATION,
                        UrlCopyError::PHASE_PREPARATION,
                        c,m);
                }

            }
        } else {
            path_ready = false;
            throw UrlCopyError(UrlCopyError::SCOPE_DESTINATION,
                UrlCopyError::PHASE_PREPARATION,
                c,m);
        }
    }
    return path_ready;
}

/**
 * Sleep for given amount of time refreshing the last update time periodically
 * @param sleep_time [IN]
 */
void TransferCore::transfer_sleep(
    unsigned int sleep_time){

    debug() << "transfer sleeping for " << sleep_time << " seconds";
    const unsigned int SLEEP_MAX = 5;
    unsigned int slept = 0;
    while (slept < sleep_time){
        RefreshTimeUpdater updater(this);
        unsigned int to_sleep = sleep_time - slept;
        if (to_sleep > SLEEP_MAX){
            to_sleep = SLEEP_MAX;
        }
#ifdef _DEBUG
        debug() << "sleeping for " << to_sleep << " seconds...";
#endif
#ifdef _DEBUG
        // speed up tests
        if (0 == getenv("NO_TRANSFER_SLEEP")){
            sleep(to_sleep);
        }
#else
        sleep(to_sleep);
#endif
        slept += to_sleep;
    }
}

/**
 * send a prepare to get request
 */
bool TransferCore::send_prepare_to_get(
    glite::data::srm::util::Context::Ptr&                       ctx,
    boost::scoped_ptr<glite::data::srm::util::PrepareToGet>&    request,
    UrlCopyError::Category&                                     c,
    std::string&                                                m)
{
	return TransferUtils::SendPrepareToGet(ctx, this, request, c, m);
}

/**
 * poll the status of the PTG request until it reaches a final state or the timeout elapses
 */
bool TransferCore::poll_prepare_to_get(
    Context::Ptr&                           ctx,
    boost::scoped_ptr<PrepareToGet>&        request,
    int                                     timeout,
    int                                     start_time,
    unsigned int                            max_sleep_time,
    UrlCopyError::Category&                 c,
    std::string&                            m)
{
	return TransferUtils::PollPrepareToGet(ctx, this, request, timeout, start_time, max_sleep_time, c, m);
}
