/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: TransferCore class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: TransferCore.h,v 1.16 2009/07/30 19:32:55 molnarzs Exp $
 *  Release: $Name:  $
 */
#ifndef GLITE_DATA_TRANSFER_URL_COPY_TRANSFERCORE_H_
#define GLITE_DATA_TRANSFER_URL_COPY_TRANSFERCORE_H_

#include <glite/data/transfer/urlcopy/UrlCopyError.h>
#include <glite/data/transfer/urlcopy/UrlCopyStatus.h>
#include <glite/data/agents/AgentExceptions.h>
#include <glite/data/srm/util/types.h>
#include <glite/data/srm/util/Context.h>
#include <glite/data/srm/util/details/SynchRequest.h>
#include <glite/data/srm/util/details/AsynchRequest.h>

#include "Mmap.h"
#include "RefreshTimeUpdater.h"
#include "TokenSelector.h"
#include "SrmMethodCaller.h"
#include "TransferUtils.h"

#include <boost/utility.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/scoped_ptr.hpp>

namespace glite{
namespace data{
namespace srm{
namespace util{
    class PrepareToGet;
}
}
}
}

namespace glite   {
namespace data    {
namespace transfer{
namespace urlcopy {

/**
 * Helper class that manages the transfer (used in the forked process)
 */
class TransferCore : boost::noncopyable, public RefreshTimeUpdater::IUpdateRefreshTime {
public:

    /**
     * Destructor
     */
    virtual ~TransferCore(){}

    /**
     * Return the transfer identifier
     */
    const std::string& id() const {return m_id;}

    /**
     * UpdateRefreshTime
     */
    virtual void UpdateRefreshTime() = 0;

protected:
    /**
     * Construct
     */
    explicit TransferCore(const std::string& id):
        m_id(id),
        m_srmMethodCaller(this){}

    /**
     * Initialize the stat
     */
    void initialize() /*throw (RuntimeError, LogicError)*/;

    // Helper Methods
    //--------------------------------------------------------------------------

    /**
     * Send Ping Request to SRM server
     * @param ctx [IN] the SRM context
     * @param c [OUT] error category
     * @param m [OUT] error message
     * @param name [IN] server name, for logging (source, destination)
     */
    void pingSrm(
        glite::data::srm::util::Context::Ptr&   ctx,
        UrlCopyError::Category&                 c,
        std::string&                            m,
        const std::string&                      name) throw();

    /**
     * Check if a path (file or directory) already exists
     * @param ctx [IN] the SRM context
     * @param surl [IN] file or directory SURL
     * @param c [OUT] error category
     * @param m [OUT] error message
     * @param file_size [OUT] pointer to variable that will receive value of file size or NULL
     * @param file_locality [OUT] pointer to variable that will receive value of file locality or NULL
     * @param cksm_algorithm [OUT] the file checksum algorithm
     * @param cksm_value [OUT] the file checksum value
     * @return true if file or directory exists, false otherwise
     */
    bool pathExists(
        glite::data::srm::util::Context::Ptr&           ctx,
        const std::string&                              surl,
        UrlCopyError::Category&                         c,
        std::string&                                    m);

    /**
     * Check if a path (file or directory) already exists
     * @param ctx [IN] the SRM context
     * @param surl [IN] file or directory SURL
     * @param c [OUT] error category
     * @param m [OUT] error message
     * @param info [OUT] structure with output file information
     * @return true if file or directory exists, false otherwise
     */
    bool pathExists(
        glite::data::srm::util::Context::Ptr&   ctx,
        const std::string&                      surl,
        UrlCopyError::Category&                 c,
        std::string&                            m,
        TransferUtils::PathInfo&                info);

    /**
     * Get the space token from the description
     * @param ctx [IN] the SRM context
     * @param description [IN] the token description
     * @param c [OUT] error category
     * @param m [OUT] error message
     * @param selector [IN] token selection strategy
     * @return string with space token name
     */
    std::string getSpaceToken(
        glite::data::srm::util::Context::Ptr&   ctx,
        const std::string&                      description,
        UrlCopyError::Category&                 c,
        std::string&                            m,
        ITokenSelector::Ptr                     selector) /*throw ()*/;

    /**
     * Remove a SURL
     */
    bool rmFile(glite::data::srm::util::Context::Ptr& ctx, const std::string& surl, UrlCopyError::Category& c, std::string& m) /*throw ()*/;

    /**
     * Create the parent directory path of the given path
     */
    bool makeDir(glite::data::srm::util::Context::Ptr& ctx, const std::string& dir_name, UrlCopyError::Category& c, std::string& m) /*throw ()*/;

    /**
     * Create the path on the target SRM
     * @param ctx [IN] SRM Context
     * @param file_path [IN] SURL of target file
     * @param destination_name [IN] the destination name as specified by the user
     * @return true if successful, false otherwise
     */
    bool prepareDestinationPath(
        glite::data::srm::util::Context::Ptr&       ctx,
        const std::string&  destination_name)/*throw ()*/;

    /**
     * Create the parent directory path of the given path
     * @param ctx [IN] SRM Context
     * @param path [IN] SURL of target directory
     * @param destination_name [IN] the destination name as specified by the user
     * @return true if successful, false otherwise
     */
    bool prepareParentDir(
        glite::data::srm::util::Context::Ptr&       ctx,
        const std::string&  path,
        const std::string&  destination_name) /*throw ()*/;

    /**
     * bringOnline
     * @param ctx [IN]
     * @param surl [IN]
     * @param copyPinLifetime [IN]
     */
    void bringOnline(
        glite::data::srm::util::Context::Ptr&   ctx,
        const std::string&                      surl,
        int                                     copyPinLifetime);

    /**
     * bringOnline
     * @param ctx [IN]
     * @param surl [IN]
     * @param copyPinLifetime [IN]
     */
    void bringOnline(
        glite::data::srm::util::Context::Ptr&   ctx,
        const std::vector<std::string>&         surls,
        int                                     copyPinLifetime);

    /**
     * Sleep for given amount of time refreshing the last update time periodically
     * @param sleep_time [IN]
     */
    void transfer_sleep(
        unsigned int sleep_time);

    /**
     * send a prepare to get request
     * @param ctx [IN] the SRM context
     * @param request [IN] the srmPrepareToGet
     * @param c [OUT] error category
     * @param m [OUT] error message
     * @return true if successful, false otherwise
     */
    bool send_prepare_to_get(
        glite::data::srm::util::Context::Ptr&                       ctx,
        boost::scoped_ptr<glite::data::srm::util::PrepareToGet>&    request,
        UrlCopyError::Category&                                     c,
        std::string&                                                m);

    /**
     * poll the status of the PTG request until it reaches a final state or the timeout elapses
     * @param ctx [IN] the SRM context
     * @param request [IN] the srmPrepareToGet
     * @param timeout [IN] the timeout for the operation
     * @param start_time [IN] the time from which timeout computation must be started
     * @param max_sleep_time [IN] maximum interval between two consecutive pollings
     * @param errorFromSrmStatusFunction [IN] error conversion function
     * @param c [OUT] error category
     * @param m [OUT] error message
     */
    bool poll_prepare_to_get(
        glite::data::srm::util::Context::Ptr&                       ctx,
        boost::scoped_ptr<glite::data::srm::util::PrepareToGet>&    request,
        int                                                         timeout,
        int                                                         start_time,
        unsigned int                                                max_sleep_time,
        UrlCopyError::Category&                                     c,
        std::string&                                                m);

private:
    /**
     * Get an UrlCopy error from an SRM error for MkDir
     */
    void mkDirErrorFromSrmStatus(UrlCopyError::Category& c, std::string& m, const glite::data::srm::util::SrmStatus& req_status) /*throw()*/;

    /**
     * Get an UrlCopy error from an SRM error for SrmStat
     */
    void statErrorFromSrmStatus(UrlCopyError::Category& c, std::string& m, const glite::data::srm::util::SrmStatus& req_status,const glite::data::srm::util::SrmStatus& file_status) throw();

    /**
     * Get an UrlCopy error from an SRM error for GetSpaceTokens
     */
    void getSpaceTokensErrorFromSrmStatus(UrlCopyError::Category& c, std::string& m, const glite::data::srm::util::SrmStatus& req_status) throw();

    /**
     * Get an UrlCopy error from an SRM error for SrmRm
     */
    void rmErrorFromSrmStatus(
    	UrlCopyError::Category& c,
    	std::string& m,
    	const glite::data::srm::util::SrmStatus& req_status,
    	const glite::data::srm::util::SrmStatus& file_status);

    /**
     * Get an UrlCopy error from an SRM request (in case the files did not contain
     * an error code)
     */
    void errorFromRequestStatus(
        UrlCopyError::Category&                     c,
        std::string&                                m,
        const glite::data::srm::util::SrmStatus&    req_status);

    /**
     * Check if a path (file or directory) already exists
     * @param ctx [IN] the SRM context
     * @param surl [IN] file or directory SURL
     * @param full_detailed [IN] full detailed listing parameter for srmLs
     * @param c [OUT] error category
     * @param m [OUT] error message
     * @param info [OUT] structure with output information for the file
     * @return true if file or directory exists, false otherwise
     */
    bool pathExists(
        glite::data::srm::util::Context::Ptr&   ctx,
        const std::string&                      surl,
        bool                                    full_detailed,
        UrlCopyError::Category&                 c,
        std::string&                            m,
        TransferUtils::PathInfo&                info);

private:

    /**
     * Default Constructor. Not Implemented
     */
    TransferCore();

protected:

    /**
     * The transfer identifier
     */
    std::string m_id;

    /**
     * The Mmap File object
     */
    boost::shared_ptr<Mmap> m_mmap;

    /**
     * SrmMethodCaller object
     */
    SrmMethodCaller m_srmMethodCaller;

};

} // end namespace urlcopy
} // end namespace transfer
} // end namespace data
} // end namespace glite

#endif //GLITE_DATA_TRANSFER_URL_COPY_TRANSFERCORE_H_


