/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: TransferSrmCopy class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: TransferSrmCopy.cpp,v 1.23 2009/12/03 17:21:44 molnarzs Exp $
 *  Release: $Name:  $
 */
#include <glite/data/transfer/urlcopy/TransferSrmCopy.h>
#include "TransferConfig.h"
#include "SrmCopyCore.h"
#include "SrmCopyStat.h"
#include "transfer-utility.h"
#include "stat-utility.h"
#include "logger-utility.h"
#include "sighandler-utility.h"
#include <glite/data/agents/AgentExceptions.h>

#include <string>
#include <sstream>
#include <iostream>
#include <fstream>
#include "ServiceDiscoveryIfce.h"
#include <map>

static std::map<std::string, std::string> sdStore;

using namespace std;

using namespace glite::data::transfer::urlcopy;
using namespace glite::data::transfer::urlcopy::details;
using namespace glite::data::agents;

namespace {
    const std::string GLITE_SRM_COPY_EXEC = "glite-srm-copy-exec";

    const char * to_bool_str(int value) {
        return (0 != value) ? "true" : "false";
    }

} // end global namespace

/**
 * TransferSrmCopy::Args
 * Constructor
 */
TransferSrmCopy::Args::Args() :
TransferBase::Args(),
pull_mode(true),
copy_timeout(-1),
refresh_timeout(-1) {
}

/**
 * start
 * Factory method to start a transfer
 */
TransferSrmCopy::Ptr TransferSrmCopy::start(
        const TransferSrmCopy::Args& args,
        const std::string& prefix,
        bool blocking) /*throw (LogicError, RuntimeError)*/ {

    // Generate an unique Id for the transfer
    std::string id = generate_request_id(prefix);

    // Create the Repository and Archive directory
    setup_repository(prefix);

    // Create the Transfer Request and store it in the Mmap file
    boost::shared_ptr<Mmap> mmap_ptr = create_stat_SrmCopy(id, args.files.size());

    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *> (mmap_ptr->buffer());
    transferSrmCopyArgs_to_SrmCopyStat(args, *stat);
    // Stat has been created

    // Create a TransferSrmCopy object and return it
    TransferSrmCopy::Ptr tx(new TransferSrmCopy(id));

    std::string source_site;
    std::string destin_site;
    SDException exception;
    char * src_site = NULL;
    char * dst_site = NULL;
    std::string src_srm_endpoint = stat->mm_srmcopy.mm_src_srm_endpoint;
    std::string dst_srm_endpoint = stat->mm_srmcopy.mm_dest_srm_endpoint;

    /*
    do not let the SE/site_name store grow too much
    5000 entries will be ~700kb, so now worries if it is kept in memory
    */

    if( ((int) sdStore.size()) > 5000)
    	sdStore.clear();

    map <string, string> ::const_iterator iter1 = sdStore.find(src_srm_endpoint);
    map <string, string> ::const_iterator iter2 = sdStore.find(dst_srm_endpoint);


    if (iter1 != sdStore.end()) {
        source_site = iter1->second;
    } else {
        src_site = SD_getServiceSite(src_srm_endpoint.c_str(), &exception);
        if (exception.status != SDStatus_SUCCESS)
            SD_freeException(&exception);
        if (src_site) {
            source_site = src_site;
            sdStore.insert(std::make_pair(src_srm_endpoint, source_site));
            free(src_site);
        } else {
            source_site = "";
        }
    }
    if (iter2 != sdStore.end()) {
        destin_site = iter2->second;
    } else {
        dst_site = SD_getServiceSite(dst_srm_endpoint.c_str(), &exception);
        if (exception.status != SDStatus_SUCCESS)
            SD_freeException(&exception);
        if (dst_site) {
            destin_site = dst_site;
            sdStore.insert(std::make_pair(dst_srm_endpoint, destin_site));
            free(dst_site);
        } else {
            destin_site = "";
        }
    }

    std::string temp = stat->mm_common.mm_user_tag;
    char * env = getenv(temp.c_str());
    std::string vo;
    if (env) {
        vo = env;
    } else {
        vo = "";
    }

    // Start the transfer
    execute_transfer(*tx, blocking, args.console_logging, vo, source_site, destin_site);
    unsetenv(temp.c_str());

    return tx;
}

/**
 * get
 * Factory method to get an existing transfer
 */
TransferSrmCopy::Ptr TransferSrmCopy::get(const std::string& id) /*throw (LogicError, RuntimeError)*/ {
    // Check preconditions
    if (true == id.empty()) {
        throw LogicError("empty transfer request id");
    }
    try {
        // Check that the Mmap file exists
        boost::shared_ptr<Mmap> mmap_ptr = open_stat_SrmCopy(id, true);
    } catch (const std::exception& exc) {
        // The mmap file does not exist: probably it's an invalid ID
        std::stringstream reason;
        reason << "no transfer found for the given ID. Details: " << exc.what();
        throw RuntimeError(reason.str());
    }
    // Create a TransferSrmCopy object and return it
    TransferSrmCopy::Ptr tx(new TransferSrmCopy(id));
    return tx;
}

/**
 * Constructor
 */
TransferSrmCopy::TransferSrmCopy(const std::string& id) :
TransferService<TransferSrmCopyStat>(id) {
};

/**
 * Destructor
 */
TransferSrmCopy::~TransferSrmCopy() {
}

/*
 * execName
 *
 * Return the exec name used
 */
const std::string& TransferSrmCopy::execName() const throw () {
    return GLITE_SRM_COPY_EXEC;
}

/*
 * status
 *
 * Retrieve transfer status
 */
TransferSrmCopyStat TransferSrmCopy::status(bool check) /*throw(RuntimeError)*/ {

    // Open the Stat File
    try {
        boost::shared_ptr<Mmap> mmap_ptr = open_stat_SrmCopy(m_id, true);
        // Get the stat object
        SrmCopyStat * stat = reinterpret_cast<SrmCopyStat*> (mmap_ptr->buffer());

        // Compute the timeout for the last refresh
        int refresh_delay = 0;
        switch (stat->mm_srmcopy.mm_status) {
            case UrlCopyStatus::COMPLETED:
                // Don't need to check
                refresh_delay = -1;
                break;
            default:
                // Take the copy timeout
                if (refresh_delay < stat->mm_srmcopy.mm_copy_timeout) {
                    refresh_delay = stat->mm_srmcopy.mm_copy_timeout;
                }
                break;
        } // end switch

        // Check that the transfer is updating the stat periodically
        check_transfer(*this, stat->mm_common, check, stat->mm_srmcopy.mm_status, refresh_delay);

        // Create and Return the Stat Object
        return TransferSrmCopyStat(stat);
    } catch (const LogicError& exc) {
        // LogicError should never happen in this context. Anyway, convert it to
        // a RuntimeError
        throw RuntimeError(exc.what());
    } catch (const RuntimeError& exc) {
        throw;
    }
}

/*
 * stop
 *
 * Stop the transfer
 */
void TransferSrmCopy::stop() /*throw (RuntimeError)*/ {

    // Get Current Time
    time_t current;
    time(&current);

    // Open the Stat File
    try {
        bool abort_request = false;
        bool refresh_mmap = false;
        do {
            boost::shared_ptr<Mmap> mmap_ptr = open_stat_SrmCopy(m_id, true);
            // Get the stat object
            SrmCopyStat * stat = reinterpret_cast<SrmCopyStat*> (mmap_ptr->buffer());
            // Stop the transfer
            int result = stop_transfer(*this, stat->mm_common, stat->mm_srmcopy.mm_status);
            if (result < 0) {
                // Impossible to send stop signal (no proceess served the
                // request). Just call abort
                abort_request = true;
                refresh_mmap = false;
            } else if (result > 0) {
                // sleep for a while and redo the check
                sleep(1);
                refresh_mmap = true;
            } else {
                abort_request = false;
                refresh_mmap = false;
                // Process Stopped
                break;
            }
        } while (refresh_mmap == true); // End while loop
        if (true == abort_request) {
            // abort the request: do it outside the loop in order to
            // refresh mmap
            this->abort();
        }
    } catch (const LogicError& exc) {
        // LogicError should never happen in this context. Anyway, convert it to
        // a RuntimeError
        throw RuntimeError(exc.what());
    } catch (const RuntimeError& exc) {
        throw;
    }
}

/**
 * Continue the transfer after preparation/transfer phase has completed.
 */
pid_t TransferSrmCopy::continueTransfer() /* throw(LogicError) */ {
    throw LogicError("continueTransfer is not supported for SrmCopy transfers.");
}

/**
 * Abort transfer
 */
void TransferSrmCopy::abort() /* throw() */ {

    // Open the Stat File
    try {
        boost::shared_ptr<Mmap> mmap_ptr = open_stat_SrmCopy(m_id, false);
        // Get the stat object
        SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *> (mmap_ptr->buffer());
        // Abort the Transfer
        abort_transfer(*this, stat->mm_common.mm_pid, stat->mm_srmcopy.mm_status, stat->mm_srmcopy.mm_error);
    } catch (...) {
        // Ignore all error (the mmap does not exist)
    }
}

/**
 * Clean the transfer log files
 */
void TransferSrmCopy::clean() /* throw(LogicError) */ {
    bool is_success = false;
    bool is_lost = false;
    // Open the Stat File
    try {
        boost::shared_ptr<Mmap> mmap_ptr = open_stat_SrmCopy(m_id, true);
        // Get the stat object
        SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *> (mmap_ptr->buffer());
        // Check the status
        if (stat->mm_srmcopy.mm_status != UrlCopyStatus::COMPLETED) {
            throw LogicError("cannot clean request since the transfer is still active");
        }
        if (stat->mm_srmcopy.mm_error.mm_category == UrlCopyError::SUCCESS) {
            is_success = true;
        } else {
            is_success = false;
            is_lost = (UrlCopyError::ERROR_LOST == stat->mm_srmcopy.mm_error.mm_category);
        }
    } catch (...) {
        // Ignore all error (the mmap does not exist)
    }

    // Archive the log and mem files
    archive_files(*this, is_success, is_lost);
}

/*
 * exec
 *
 * Execute the transfer and treturn the final result
 */
void TransferSrmCopy::exec(bool enable_console, std::string vo, std::string src_site, std::string dest_site) /*throw(RuntimeError)*/ {

    // Register Signal Handler
    SigAction sa_guard;

    // Open the stat File
    try {
        // bug 30059: it can happen that the process starts after the mem file has
        // already been moved (e.g. the transfer was aborted because it was not
        // served within 5 seconds); the last 'false' parameter specifies not
        // to search subfolders. The subsequent check is used to avoid creating
        // a log file.
        boost::shared_ptr<Mmap> mmap_ptr = open_stat_SrmCopy(m_id, false, false);
        if (0 == mmap_ptr.get()) {
            throw RuntimeError("Could not find MemoryMap file for the request.");
        }

        // Get the stat object
        SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *> (mmap_ptr->buffer());
        // Check that this request is not already taken by another process
        if (0 != stat->mm_common.mm_pid) {
            throw RuntimeError("this request is already served by another process");
        }
        // Write Pid (Reserve request)
        stat->mm_common.mm_pid = getpid();
        // Initialize Logger
        LogLevel level = (stat->mm_common.mm_verbose) ? LOGLEVEL_DEBUG : LOGLEVEL_INFO;

        log_setup(
                m_id.c_str(),
                level,
                (0 != enable_console) ? true : false,
                "glite-srm-copy",
                stat->mm_common.mm_syslog_disable,
                stat->mm_common.mm_syslog_ident,
                stat->mm_common.mm_syslog_facility,
                stat->mm_common.mm_syslog_facility_name);

    } catch (const LogicError& exc) {
        // LogicError should never happen in this context. Anyway, convert it to
        // a RuntimeError
        throw RuntimeError(exc.what());
    } catch (const RuntimeError& exc) {
        throw;
    }

    // Do the transfer
    try {
        // Create Tx Core Object
        SrmCopyCore tx(m_id);
        tx.vo_name = vo;
        tx.src_site = src_site;
        tx.dest_site = dest_site;

        // Get stat objexct
        SrmCopyStat& tx_stat = tx.stat();
        try {
            // Prepare the transfer
            tx.prepare();
            if ((tx_stat.mm_srmcopy.mm_error.mm_category == UrlCopyError::SUCCESS) &&
                    (tx_stat.mm_srmcopy.mm_status == UrlCopyStatus::READY)) {
                // Start the transfer
                tx.transfer();
            }
            // Complete (or Abort) the transfer
            tx.complete();
        } catch (const std::exception& exc) {
            // An error occurred: abort the transfer
            try {
                tx.complete();
            } catch (...) {
            }
        }
    } catch (const LogicError& exc) {
        // LogicError should never happen in this context. Anyway, convert it to
        // a RuntimeError
        throw RuntimeError(exc.what());
    } catch (const RuntimeError& exc) {
        throw;
    }
}

/**
 * dump
 * Dump the content of a stat file into standard output
 */
void TransferSrmCopy::dump(const std::string& file_name) /*throw (LogicError, RuntimeError)*/ {

    // Open stat file
    boost::shared_ptr<Mmap> mmap_ptr = open_stat_file(file_name, true, SRMCOPY_STAT_ID);

    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *> (mmap_ptr->buffer());

    std::cout
            << "--------------------------------------------------------------------------------" << std::endl
            << "--------------------------------------------------------------------------------" << std::endl
            << " Glite Transfer SRM Copy Status" << std::endl
            << "--------------------------------------------------------------------------------" << std::endl
            << "--------------------------------------------------------------------------------" << std::endl
            << "Version             : " << ((stat->mm_common.mm_version >> 16)&0xFFFF) << "." << (stat->mm_common.mm_version & 0xFFFF) << std::endl
            << "Ext. version        : " << ((stat->mm_common.mm_extension >> 16)&0xFFFF) << "." << (stat->mm_common.mm_extension & 0xFFFF) << std::endl
            << "Stat ID             : " << stat->mm_common.mm_id << std::endl
            << "Start Time          : " << time_to_string(stat->mm_common.mm_start_time) << std::endl
            << "Finish Time         : " << time_to_string(stat->mm_common.mm_finish_time) << std::endl
            << "Last Refresh        : " << time_to_string(stat->mm_common.mm_last_refresh_time) << std::endl
            << "Process ID          : " << stat->mm_common.mm_pid << std::endl
            << "Verbose Log         : " << to_bool_str(stat->mm_common.mm_verbose) << std::endl
            << "Ping source         : " << to_bool_str(stat->mm_common.mm_enable_ping_source) << std::endl
            << "Ping destination    : " << to_bool_str(stat->mm_common.mm_enable_ping_dest) << std::endl
            << "Check target dir    : " << to_bool_str(stat->mm_common.mm_enable_dir_check) << std::endl
            << "Flags               : " << stat->mm_common.mm_flags << std::endl
            << "User Description    : " << stat->mm_common.mm_user_tag << std::endl
            << "--------------------------------------------------------------------------------" << std::endl
            << " SRM Copy Request" << std::endl
            << "--------------------------------------------------------------------------------" << std::endl
            << "Request Tx Status.  : " << UrlCopyStatus::toString((UrlCopyStatus::Value)stat->mm_srmcopy.mm_status) << std::endl
            << "Src. SRM Endpoint   : " << stat->mm_srmcopy.mm_src_srm_endpoint << std::endl
            << "Src. SRM Version    : " << stat->mm_srmcopy.mm_src_srm_version << std::endl
            << "Dest. SRM Endpoint  : " << stat->mm_srmcopy.mm_dest_srm_endpoint << std::endl
            << "Dest. SRM Version   : " << stat->mm_srmcopy.mm_dest_srm_version << std::endl
            << "Source Space Token  : " << stat->mm_common.mm_source_space_token << std::endl
            << "Source Token Descr. : " << stat->mm_common.mm_source_token_descr << std::endl
            << "Dest Space Token    : " << stat->mm_common.mm_dest_space_token << std::endl
            << "Dest Token Descr.   : " << stat->mm_common.mm_dest_token_descr << std::endl
            << "Copy pin lifetime   : " << stat->mm_common.mm_copy_pin_lifetime << std::endl
            << "Use LAN connection  : " << to_bool_str(stat->mm_common.mm_lan_connection) << std::endl
            << "Fail nearline files : " << to_bool_str(stat->mm_common.mm_fail_nearline_files) << std::endl
            << "Block after prepare : " << to_bool_str(stat->mm_common.mm_block_after_preparation) << std::endl
            << "Block after tx      : " << to_bool_str(stat->mm_common.mm_block_after_transfer) << std::endl
            << "Request Token       : " << stat->mm_srmcopy.mm_token << std::endl
            << "Mode                : " << ((0 == stat->mm_srmcopy.mm_push_mode) ? "pull" : "push") << std::endl
            << "HTTP Timeout        : " << stat->mm_srmcopy.mm_http_timeout << std::endl
            << "Copy Timeout        : " << stat->mm_srmcopy.mm_copy_timeout << std::endl
            << "Refresh Timeout     : " << stat->mm_srmcopy.mm_refresh_timeout << std::endl
            << "Src. Pr. Start Time : " << time_to_string(stat->mm_srmcopy.mm_src_prep_start_time) << std::endl
            << "Src. Pr. Done Time  : " << time_to_string(stat->mm_srmcopy.mm_src_prep_done_time) << std::endl
            << "Dest. Pr. Start Time: " << time_to_string(stat->mm_srmcopy.mm_dest_prep_start_time) << std::endl
            << "Dest. Pr. Done Time : " << time_to_string(stat->mm_srmcopy.mm_dest_prep_done_time) << std::endl
            << "Copy Start Time     : " << time_to_string(stat->mm_srmcopy.mm_start_time) << std::endl
            << "Copy Done Time      : " << time_to_string(stat->mm_srmcopy.mm_done_time) << std::endl
            << "Final. Start Time   : " << time_to_string(stat->mm_srmcopy.mm_final_start_time) << std::endl
            << "Final. Done Time    : " << time_to_string(stat->mm_srmcopy.mm_final_done_time) << std::endl
            << "Error Scope         : " << UrlCopyError::scopeToString((UrlCopyError::Scope)stat->mm_srmcopy.mm_error.mm_scope) << std::endl
            << "Error Phase         : " << UrlCopyError::phaseToString((UrlCopyError::Phase)stat->mm_srmcopy.mm_error.mm_phase) << std::endl
            << "Error Category      : " << UrlCopyError::categoryToString((UrlCopyError::Category)stat->mm_srmcopy.mm_error.mm_category) << std::endl
            << "Error Message       : " << stat->mm_srmcopy.mm_error.mm_message << std::endl
            << "Number of Files     : " << stat->mm_num_files << std::endl;
    for (unsigned int i = 0; i < stat->mm_num_files; ++i) {
        std::cout
                << "--------------------------------------------------------------------------------" << std::endl
                << " File #" << i << std::endl
                << "--------------------------------------------------------------------------------" << std::endl
                << "Source SURL         : " << stat->mm_files[i].mm_src_name << std::endl
                << "Destination SURL    : " << stat->mm_files[i].mm_dest_name << std::endl
                << "Tx. Status          : " << UrlCopyStatus::toString((UrlCopyStatus::Value)stat->mm_files[i].mm_status) << std::endl
                << "Source Size         : " << stat->mm_files[i].mm_src_size << std::endl
                << "Destination Size    : " << stat->mm_files[i].mm_size << std::endl
                << "Verify checksums    : " << to_bool_str(stat->mm_files[i].mm_compare_checksum) << std::endl
                << "cksm type (user)    : " << stat->mm_files[i].mm_checksum_user_algorithm << std::endl
                << "cksm value (user)   : " << stat->mm_files[i].mm_checksum_user_value << std::endl
                << "Src cksm type       : " << stat->mm_files[i].mm_src_checksum_algorithm << std::endl
                << "Src cksm value      : " << stat->mm_files[i].mm_src_checksum_value << std::endl
                << "Dst cksm type       : " << stat->mm_files[i].mm_dst_checksum_algorithm << std::endl
                << "Dst cksm value      : " << stat->mm_files[i].mm_dst_checksum_value << std::endl
                << "Error Scope         : " << UrlCopyError::scopeToString((UrlCopyError::Scope)stat->mm_files[i].mm_error.mm_scope) << std::endl
                << "Error Phase         : " << UrlCopyError::phaseToString((UrlCopyError::Phase)stat->mm_files[i].mm_error.mm_phase) << std::endl
                << "Error Category      : " << UrlCopyError::categoryToString((UrlCopyError::Category)stat->mm_files[i].mm_error.mm_category) << std::endl
                << "Error Message       : " << stat->mm_files[i].mm_error.mm_message << std::endl;
    }
}
