/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: TransferUrlCopy class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: TransferUrlCopy.cpp,v 1.23 2009/12/03 17:21:44 molnarzs Exp $
 *  Release: $Name:  $
 */
#include <glite/data/transfer/urlcopy/TransferUrlCopy.h>
#include "TransferConfig.h"
#include "FileType.h"
#include "UrlCopyCore.h"
#include "UrlCopyStat.h"
#include "transfer-utility.h"
#include "stat-utility.h"
#include "logger-utility.h"
#include "sighandler-utility.h"
#include <errno.h>
#include <string>
#include <sstream>
#include <iostream>
#include <sys/types.h>
#include <map>
#include <vector>
#include <string>
#include <fstream>
#include <sstream>
#include <iomanip>
#include "ServiceDiscoveryIfce.h"

/*global store to keep cached mapping between srm endpoint and site name*/
static std::map<std::string, std::string> sdStore;

/*global env variable to check/compare previous/current environment*/
static std::string previous_env[2];

extern std::string get_hostname(std::string hostname);

using namespace glite::data::transfer::urlcopy;
using namespace glite::data::transfer::urlcopy::details;
using namespace glite::data::agents;

namespace {
    const std::string GLITE_URL_COPY_EXEC = "glite-url-copy-exec";

    const char * to_bool_str(int value) {
        return (0 != value) ? "true" : "false";
    }

} // end global namespace

/*
 * TransferUrlCopy::Args 
 *
 * Constructor
 */
TransferUrlCopy::Args::Args() :
TransferBase::Args(),
nstreams(0),
tcpbs(0),
put_timeout(-1),
get_timeout(-1),
tx_timeout(-1),
tx_markers_timeout(-1),
putdone_timeout(-1),
getdone_timeout(-1),
first_tx_marker_timeout(-1) {
    // EMPTY
}

/*
 * start
 *
 * Factory Method to start a transfer
 */
TransferUrlCopy::Ptr TransferUrlCopy::start(
        const TransferUrlCopy::Args& args,
        const std::string& prefix,
        bool blocking) /*throw (LogicError, RuntimeError)*/ {
    // Generate an unique Id for the transfer
    std::string id = generate_request_id(prefix);

    // Create the Repository and Archive directory
    setup_repository(prefix);

    // Create the Transfer Request and store it in the Mmap file
    boost::shared_ptr<Mmap> mmap_ptr = create_stat_UrlCopy(id);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *> (mmap_ptr->buffer());
    transferUrlCopyArgs_to_UrlCopyStat(args, *stat);
    // Stat has been created

    // Create a TransferUrlCopy object and return it 
    TransferUrlCopy::Ptr tx(new TransferUrlCopy(id));

    std::string source_site;
    std::string destin_site;
    SDException exception;
    char * src_site = NULL;
    char * dst_site = NULL;
    std::string src_srm_endpoint = stat->mm_urlcopy.mm_source.mm_srm_endpoint;
    std::string dst_srm_endpoint = stat->mm_urlcopy.mm_destination.mm_srm_endpoint;
    
    
    //This is a urlcopy transfer, using gsiftp endpoints directly
    //special "treatment is required to retrieve site_name from cache file or bdii"
    if(src_srm_endpoint.length() == 0){ //it is gsiftp
        src_srm_endpoint = "gsiftp://";
    	src_srm_endpoint.append(get_hostname(stat->mm_urlcopy.mm_source.mm_name));	
	}
    
    if(dst_srm_endpoint.length() == 0){ //it is gsiftp
        dst_srm_endpoint = "gsiftp://";
    	dst_srm_endpoint.append(get_hostname(stat->mm_urlcopy.mm_destination.mm_name));
	}
	    
    /*
    do not let the SE/site_name store grow too much
    5000 entries will be ~700kb, so now worries if it is kept in memory
    */
  if( ((int) sdStore.size()) > 5000)
    	sdStore.clear();

    map <string, string> ::const_iterator iter1 = sdStore.find(src_srm_endpoint);
    map <string, string> ::const_iterator iter2 = sdStore.find(dst_srm_endpoint);

    if (iter1 != sdStore.end()) {
        source_site = iter1->second;
    } else {
        src_site = SD_getServiceSite(src_srm_endpoint.c_str(), &exception);
        if (exception.status != SDStatus_SUCCESS)
            SD_freeException(&exception);
        if (src_site) {
            source_site = src_site;
            sdStore.insert(std::make_pair(src_srm_endpoint, source_site));
            free(src_site);
        } else {
            source_site = "";
        }
    }
    if (iter2 != sdStore.end()) {
        destin_site = iter2->second;
    } else {
        dst_site = SD_getServiceSite(dst_srm_endpoint.c_str(), &exception);
        if (exception.status != SDStatus_SUCCESS)
            SD_freeException(&exception);
        if (dst_site) {
            destin_site = dst_site;
            sdStore.insert(std::make_pair(dst_srm_endpoint, destin_site));
            free(dst_site);
        } else {
            destin_site = "";
        }
    }

    std::string temp = stat->mm_common.mm_user_tag;
    char *env = getenv(temp.c_str());
    std::string vo;

    if (env) {
        vo = env;
        previous_env[0] = temp;
    } else {
        vo = "";
    }


   // Start the transfer
    execute_transfer(*tx, blocking, args.console_logging, vo, source_site, destin_site);
    
    /*
    The following logic is needed because, for url-copy transfer jobs only, the tr_id
    remains tha same for transfer jobs, but multiple transfer submition requests are sent. So, cache it
    and reuse it to speed up things
    */
    if (previous_env[1].length() == 0) {
        previous_env[1] = temp;
    } else {
        if (previous_env[0].compare(previous_env[1]) == 0) {
            previous_env[1] = temp;
        }
        else {
            unsetenv(previous_env[1].c_str());
            previous_env[1] = temp;
        }
    }

    return tx;
}

/**
 * get
 * Factory Method to Get an existing
 */
TransferUrlCopy::Ptr TransferUrlCopy::get(const std::string& id) /*throw (LogicError, RuntimeError)*/ {
    // Check preconditions
    if (true == id.empty()) {
        throw LogicError("empty transfer request id");
    }
    try {
        // Check that the Mmap file exists
        boost::shared_ptr<Mmap> mmap_ptr = open_stat_UrlCopy(id, true);
    } catch (const std::exception& exc) {
        // The mmap file does not exist: probably it's an invalid ID
        std::stringstream reason;
        reason << "no transfer found for the given ID. Details: " << exc.what();
        throw RuntimeError(reason.str());
    }
    // Create a TransferUrlCopy object and return it 
    TransferUrlCopy::Ptr tx(new TransferUrlCopy(id));
    return tx;
}

/*
 * TransferUrlCopy
 *
 * Constructor
 */
TransferUrlCopy::TransferUrlCopy(const std::string& id) :
TransferService<TransferUrlCopyStat>(id) {
};

/*
 * ~TransferUrlCopy
 *
 * Destructor
 */
TransferUrlCopy::~TransferUrlCopy() {
}

/*
 * execName
 *
 * Return the exec name used
 */
const std::string& TransferUrlCopy::execName() const throw () {
    return GLITE_URL_COPY_EXEC;
}

/*
 * status
 *
 * Retrieve transfer status
 */
TransferUrlCopyStat TransferUrlCopy::status(bool check) /*throw(RuntimeError)*/ {

    // Open the Stat File
    try {
        boost::shared_ptr<Mmap> mmap_ptr = open_stat_UrlCopy(m_id, true);
        // Get the stat object
        UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *> (mmap_ptr->buffer());

        // Compute the timeout for the last refresh
        int refresh_delay = 0;
        switch (stat->mm_urlcopy.mm_status) {
            case UrlCopyStatus::COMPLETED:
                // Don't need to check
                refresh_delay = -1;
                break;
            case UrlCopyStatus::IDLE:
            case UrlCopyStatus::PREPARING:
                // get the maximumm between the SRM interactions
                if (refresh_delay < stat->mm_urlcopy.mm_source.mm_prep_timeout) {
                    refresh_delay = stat->mm_urlcopy.mm_source.mm_prep_timeout;
                }
                if (refresh_delay < stat->mm_urlcopy.mm_destination.mm_prep_timeout) {
                    refresh_delay = stat->mm_urlcopy.mm_destination.mm_prep_timeout;
                }
                break;
            case UrlCopyStatus::READY:
            case UrlCopyStatus::RUNNING:
                // The stat should be updated every second. Anyway, take the tx
                // markers timeout
                if (refresh_delay < stat->mm_urlcopy.mm_tx.mm_marker_timeout) {
                    refresh_delay = stat->mm_urlcopy.mm_tx.mm_marker_timeout;
                }
                break;
            case UrlCopyStatus::DONE:
            case UrlCopyStatus::FAILED:
            case UrlCopyStatus::ABORTED:
                // get the maximum between the SRM interactions
                if (refresh_delay < stat->mm_urlcopy.mm_source.mm_final_timeout) {
                    refresh_delay = stat->mm_urlcopy.mm_source.mm_final_timeout;
                }
                if (refresh_delay < stat->mm_urlcopy.mm_destination.mm_final_timeout) {
                    refresh_delay = stat->mm_urlcopy.mm_destination.mm_final_timeout;
                }
                break;
        } // end switch

        // Check that the transfer is updating the stat periodically
        check_transfer(*this, stat->mm_common, check, stat->mm_urlcopy.mm_status, refresh_delay);
        // Create and Return the Stat Object
        return TransferUrlCopyStat(stat);
    } catch (const LogicError& exc) {
        // LogicError should never happen in this context. Anyway, convert it to
        // a RuntimeError
        throw RuntimeError(exc.what());
    } catch (const RuntimeError& exc) {
        throw;
    }
}

/*
 * stop
 *
 * Stop the transfer 
 */
void TransferUrlCopy::stop() /*throw (RuntimeError)*/ {

    // Get Current Time
    time_t current;
    time(&current);

    // Open the Stat File
    try {
        bool abort_request = false;
        bool refresh_mmap = false;
        do {
            boost::shared_ptr<Mmap> mmap_ptr = open_stat_UrlCopy(m_id, true);
            // Get the stat object
            UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *> (mmap_ptr->buffer());
            // Stop the transfer
            int result = stop_transfer(*this, stat->mm_common, stat->mm_urlcopy.mm_status);
            if (result < 0) {
                // Impossible to send stop signal (no proceess served the 
                // request). Just call abort
                abort_request = true;
                refresh_mmap = false;
            } else if (result > 0) {
                // sleep for a while and redo the check
                sleep(1);
                refresh_mmap = true;
            } else {
                abort_request = false;
                refresh_mmap = false;
                // Process Stopped
                break;
            }
        } while (refresh_mmap == true); // End while loop
        if (true == abort_request) {
            // abort the request: do it outside the loop in order to 
            // refresh mmap
            this->abort();
        }
    } catch (const LogicError& exc) {
        // LogicError should never happen in this context. Anyway, convert it to
        // a RuntimeError
        throw RuntimeError(exc.what());
    } catch (const RuntimeError& exc) {
        throw;
    }
}

/**
 * Continue the transfer after preparation/transfer phase has completed.
 */
pid_t TransferUrlCopy::continueTransfer() /* throw(LogicError) */ {
    boost::shared_ptr<Mmap> mmap_ptr = open_stat_UrlCopy(m_id, true);
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *> (mmap_ptr->buffer());
    int result = ::kill(stat->mm_common.mm_pid, SigAction::SIGNAL_CONTINUE);
    if (0 == result) {
        return stat->mm_common.mm_pid;
    } else {
        throw PosixError("continueTransfer", errno);
    }
}

/**
 * Abort transfer
 */
void TransferUrlCopy::abort() /* throw() */ {

    // Open the Stat File
    try {
        boost::shared_ptr<Mmap> mmap_ptr = open_stat_UrlCopy(m_id, false);
        // Get the stat object
        UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *> (mmap_ptr->buffer());
        // Abort the Transfer
        abort_transfer(*this, stat->mm_common.mm_pid, stat->mm_urlcopy.mm_status, stat->mm_urlcopy.mm_error);
    } catch (...) {
        // Ignore all error (the mmap does not exist)
    }
}

/**
 * Clean the transfer log files
 */
void TransferUrlCopy::clean() /* throw(LogicError) */ {
    bool is_success = false;
    bool is_lost = false;
    // Open the Stat File
    try {
        boost::shared_ptr<Mmap> mmap_ptr = open_stat_UrlCopy(m_id, true);
        // Get the stat object
        UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *> (mmap_ptr->buffer());
        // Check the status
        if (stat->mm_urlcopy.mm_status != UrlCopyStatus::COMPLETED) {
            throw LogicError("cannot clean request since the transfer is still active");
        }
        if (stat->mm_urlcopy.mm_error.mm_category == UrlCopyError::SUCCESS) {
            is_success = true;
        } else {
            is_success = false;
            is_lost = (UrlCopyError::ERROR_LOST == stat->mm_urlcopy.mm_error.mm_category);
        }
    } catch (...) {
        // Ignore all error (the mmap does not exist)
    }

    // Archive the log and mem files
    archive_files(*this, is_success, is_lost);
}

/*
 * exec
 *
 * Execute the transfer and return the final result
 */
void TransferUrlCopy::exec(bool enable_console, std::string vo, std::string src_site, std::string dest_site) /*throw(RuntimeError)*/ {

    // Open the stat File
    try {
        // bug 30059: it can happen that the process starts after the mem file has 
        // already been moved (e.g. the transfer was aborted because it was not 
        // served within 5 seconds); the last 'false' parameter specifies not 
        // to search subfolders. The subsequent check is used to avoid creating
        // a log file.
        boost::shared_ptr<Mmap> mmap_ptr = open_stat_UrlCopy(m_id, false, false);
        if (0 == mmap_ptr.get()) {
            throw RuntimeError("Could not find MemoryMap file for the request.");
        }
        // Get the stat object
        UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *> (mmap_ptr->buffer());
        // Check that this request is not already taken by another process
        if (0 != stat->mm_common.mm_pid) {
            throw RuntimeError("this request is already served by another process");
        }
        // Write Pid (Reserve request)
        stat->mm_common.mm_pid = getpid();
        // Initialize Logger
        LogLevel level = (stat->mm_common.mm_verbose) ? LOGLEVEL_DEBUG : LOGLEVEL_INFO;

        log_setup(m_id.c_str(),
                level,
                (0 != enable_console) ? true : false,
                "glite-url-copy",
                stat->mm_common.mm_syslog_disable,
                stat->mm_common.mm_syslog_ident,
                stat->mm_common.mm_syslog_facility,
                stat->mm_common.mm_syslog_facility_name);

    } catch (const LogicError& exc) {
        // LogicError should never happen in this context. Anyway, convert it to
        // a RuntimeError
        throw RuntimeError(exc.what());
    } catch (const RuntimeError& exc) {
        throw;
    }

    // Register signal handlers
    SigAction sa_guard;
    // Do the transfer
    try {

        /* TBA - log exit */

        // Create Tx Core Object
        UrlCopyCore tx(m_id);
        tx.vo_name = vo;
        tx.src_site = src_site;
        tx.dest_site = dest_site;

        // Get stat objexct
        UrlCopyStat& tx_stat = tx.stat();
        try {
            // Prepare the transfer
            tx.prepare();
            if ((tx_stat.mm_urlcopy.mm_error.mm_category == UrlCopyError::SUCCESS) &&
                    (tx_stat.mm_urlcopy.mm_status == UrlCopyStatus::READY)) {
                // Start the transfer
                tx.transfer();
            }

            // Complete (or Abort) the transfer
            tx.complete();
        } catch (const std::exception& exc) {
            // An error occurred: abort the transfer
            try {
                tx.complete();
            } catch (...) {
            }
        }
    } catch (const LogicError& exc) {
        // LogicError should never happen in this context. Anyway, convert it to
        // a RuntimeError
        throw RuntimeError(exc.what());
    } catch (const RuntimeError& exc) {
        throw;
    }
}

/*
 * dump
 *
 * Dump the content of a stat file into standard output
 */
void TransferUrlCopy::dump(const std::string& file_name) /*throw (LogicError, RuntimeError)*/ {

    // Open stat file
    boost::shared_ptr<Mmap> mmap_ptr = open_stat_file(file_name, true, URLCOPY_STAT_ID);

    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *> (mmap_ptr->buffer());
    std::cout
            << "--------------------------------------------------------------------------------" << std::endl
            << "--------------------------------------------------------------------------------" << std::endl
            << " Glite Transfer Url Copy Status" << std::endl
            << "--------------------------------------------------------------------------------" << std::endl
            << "--------------------------------------------------------------------------------" << std::endl
            << "Version             : " << ((stat->mm_common.mm_version >> 16)&0xFFFF) << "." << (stat->mm_common.mm_version & 0xFFFF) << std::endl
            << "Ext. version        : " << ((stat->mm_common.mm_extension >> 16)&0xFFFF) << "." << (stat->mm_common.mm_extension & 0xFFFF) << std::endl
            << "Stat ID             : " << stat->mm_common.mm_id << std::endl
            << "Start Time          : " << time_to_string(stat->mm_common.mm_start_time) << std::endl
            << "Finish Time         : " << time_to_string(stat->mm_common.mm_finish_time) << std::endl
            << "Last Refresh        : " << time_to_string(stat->mm_common.mm_last_refresh_time) << std::endl
            << "Process ID          : " << stat->mm_common.mm_pid << std::endl
            << "Verbose Log         : " << to_bool_str(stat->mm_common.mm_verbose) << std::endl
            << "Ping source         : " << to_bool_str(stat->mm_common.mm_enable_ping_source) << std::endl
            << "Ping destination    : " << to_bool_str(stat->mm_common.mm_enable_ping_dest) << std::endl
            << "Check target dir    : " << to_bool_str(stat->mm_common.mm_enable_dir_check) << std::endl
            << "Flags               : " << stat->mm_common.mm_flags << std::endl
            << "User Description    : " << stat->mm_common.mm_user_tag << std::endl
            << "--------------------------------------------------------------------------------" << std::endl
            << " Url Copy Request" << std::endl
            << "--------------------------------------------------------------------------------" << std::endl
            << "Request Tx Status.  : " << UrlCopyStatus::toString((UrlCopyStatus::Value)stat->mm_urlcopy.mm_status) << std::endl
            << "Dst Space Token     : " << stat->mm_common.mm_dest_space_token << std::endl
            << "Dst Space Token Desc: " << stat->mm_common.mm_dest_token_descr << std::endl
            << "Src Space Token     : " << stat->mm_common.mm_source_space_token << std::endl
            << "Src Space Token Desc: " << stat->mm_common.mm_source_token_descr << std::endl
            << "Copy pin lifetime   : " << stat->mm_common.mm_copy_pin_lifetime << std::endl
            << "Use LAN connection  : " << to_bool_str(stat->mm_common.mm_lan_connection) << std::endl
            << "Fail nearline files : " << to_bool_str(stat->mm_common.mm_fail_nearline_files) << std::endl
            << "Block after prepare : " << to_bool_str(stat->mm_common.mm_block_after_preparation) << std::endl
            << "Block after tx      : " << to_bool_str(stat->mm_common.mm_block_after_transfer) << std::endl
            << "Verify checksums    : " << to_bool_str(stat->mm_urlcopy.mm_compare_checksum) << std::endl
            << "cksm type (user)    : " << stat->mm_urlcopy.mm_checksum_user_algorithm << std::endl
            << "cksm value (user)   : " << stat->mm_urlcopy.mm_checksum_user_value << std::endl
            << "Src cksm type       : " << stat->mm_urlcopy.mm_src_checksum_algorithm << std::endl
            << "Src cksm value      : " << stat->mm_urlcopy.mm_src_checksum_value << std::endl
            << "Dst cksm type       : " << stat->mm_urlcopy.mm_dst_checksum_algorithm << std::endl
            << "Dst cksm value      : " << stat->mm_urlcopy.mm_dst_checksum_value << std::endl
            << "Error Scope         : " << UrlCopyError::scopeToString((UrlCopyError::Scope)stat->mm_urlcopy.mm_error.mm_scope) << std::endl
            << "Error Phase         : " << UrlCopyError::phaseToString((UrlCopyError::Phase)stat->mm_urlcopy.mm_error.mm_phase) << std::endl
            << "Error Category      : " << UrlCopyError::categoryToString((UrlCopyError::Category)stat->mm_urlcopy.mm_error.mm_category) << std::endl
            << "Error Message       : " << stat->mm_urlcopy.mm_error.mm_message << std::endl
            << "--------------------------------------------------------------------------------" << std::endl
            << " Source " << std::endl
            << "--------------------------------------------------------------------------------" << std::endl
            << "Name                : " << stat->mm_urlcopy.mm_source.mm_name << std::endl
            << "Tx. Name            : " << stat->mm_urlcopy.mm_source.mm_tx_name << std::endl
            << "SRM Endpoint        : " << stat->mm_urlcopy.mm_source.mm_srm_endpoint << std::endl
            << "SRM Version         : " << stat->mm_urlcopy.mm_source.mm_srm_version << std::endl
            << "Token               : " << stat->mm_urlcopy.mm_source.mm_token << std::endl
            << "File Exists         : " << to_bool_str(stat->mm_urlcopy.mm_source.mm_exists) << std::endl
            << "File Type           : " << FileType::toString((FileType::Value)stat->mm_urlcopy.mm_source.mm_type) << std::endl
            << "Size                : " << stat->mm_urlcopy.mm_source.mm_size << std::endl
            << "Error Scope         : " << UrlCopyError::scopeToString((UrlCopyError::Scope)stat->mm_urlcopy.mm_source.mm_error.mm_scope) << std::endl
            << "Error Phase         : " << UrlCopyError::phaseToString((UrlCopyError::Phase)stat->mm_urlcopy.mm_source.mm_error.mm_phase) << std::endl
            << "Error Category      : " << UrlCopyError::categoryToString((UrlCopyError::Category)stat->mm_urlcopy.mm_source.mm_error.mm_category) << std::endl
            << "Error Message       : " << stat->mm_urlcopy.mm_source.mm_error.mm_message << std::endl
            << "HTTP Timeout        : " << stat->mm_urlcopy.mm_source.mm_http_timeout << std::endl
            << "Prep. Timeout       : " << stat->mm_urlcopy.mm_source.mm_prep_timeout << std::endl
            << "Final. Timeout      : " << stat->mm_urlcopy.mm_source.mm_final_timeout << std::endl
            << "Pr. Start Time      : " << time_to_string(stat->mm_urlcopy.mm_source.mm_prep_start_time) << std::endl
            << "Pr. Done Time       : " << time_to_string(stat->mm_urlcopy.mm_source.mm_prep_done_time) << std::endl
            << "Final. Start Time   : " << time_to_string(stat->mm_urlcopy.mm_source.mm_final_start_time) << std::endl
            << "Final. Done Time    : " << time_to_string(stat->mm_urlcopy.mm_source.mm_final_done_time) << std::endl
            << "--------------------------------------------------------------------------------" << std::endl
            << " Destination " << std::endl
            << "--------------------------------------------------------------------------------" << std::endl
            << "Name                : " << stat->mm_urlcopy.mm_destination.mm_name << std::endl
            << "Tx. Name            : " << stat->mm_urlcopy.mm_destination.mm_tx_name << std::endl
            << "SRM Endpoint        : " << stat->mm_urlcopy.mm_destination.mm_srm_endpoint << std::endl
            << "SRM Version         : " << stat->mm_urlcopy.mm_destination.mm_srm_version << std::endl
            << "Token               : " << stat->mm_urlcopy.mm_destination.mm_token << std::endl
            << "File Exists         : " << to_bool_str(stat->mm_urlcopy.mm_destination.mm_exists) << std::endl
            << "File Type           : " << FileType::toString((FileType::Value)stat->mm_urlcopy.mm_destination.mm_type) << std::endl
            << "Size                : " << stat->mm_urlcopy.mm_destination.mm_size << std::endl
            << "Error Scope         : " << UrlCopyError::scopeToString((UrlCopyError::Scope)stat->mm_urlcopy.mm_destination.mm_error.mm_scope) << std::endl
            << "Error Phase         : " << UrlCopyError::phaseToString((UrlCopyError::Phase)stat->mm_urlcopy.mm_destination.mm_error.mm_phase) << std::endl
            << "Error Category      : " << UrlCopyError::categoryToString((UrlCopyError::Category)stat->mm_urlcopy.mm_destination.mm_error.mm_category) << std::endl
            << "Error Message       : " << stat->mm_urlcopy.mm_destination.mm_error.mm_message << std::endl
            << "HTTP Timeout        : " << stat->mm_urlcopy.mm_destination.mm_http_timeout << std::endl
            << "Prep. Timeout       : " << stat->mm_urlcopy.mm_destination.mm_prep_timeout << std::endl
            << "Final. Timeout      : " << stat->mm_urlcopy.mm_destination.mm_final_timeout << std::endl
            << "Pr. Start Time      : " << time_to_string(stat->mm_urlcopy.mm_destination.mm_prep_start_time) << std::endl
            << "Pr. Done Time       : " << time_to_string(stat->mm_urlcopy.mm_destination.mm_prep_done_time) << std::endl
            << "Final. Start Time   : " << time_to_string(stat->mm_urlcopy.mm_destination.mm_final_start_time) << std::endl
            << "Final. Done Time    : " << time_to_string(stat->mm_urlcopy.mm_destination.mm_final_done_time) << std::endl
            << "--------------------------------------------------------------------------------" << std::endl
            << " Transfer " << std::endl
            << "--------------------------------------------------------------------------------" << std::endl
            << "Start Time          : " << time_to_string(stat->mm_urlcopy.mm_tx.mm_start_time) << std::endl
            << "Done Time           : " << time_to_string(stat->mm_urlcopy.mm_tx.mm_done_time) << std::endl
            << "First Marker Timest.: " << time_to_string(stat->mm_urlcopy.mm_tx.mm_first_marker) << std::endl
            << "Last Marker Timest. : " << time_to_string(stat->mm_urlcopy.mm_tx.mm_last_marker) << std::endl
            << "Size                : " << stat->mm_urlcopy.mm_tx.mm_size << std::endl
            << "Transferred Bytes   : " << stat->mm_urlcopy.mm_tx.mm_transferred_bytes << std::endl
            << "Avg. Throughput     : " << stat->mm_urlcopy.mm_tx.mm_avg_throughput << std::endl
            << "Instant Throughput  : " << stat->mm_urlcopy.mm_tx.mm_instant_throughput << std::endl
            << "% Done              : " << stat->mm_urlcopy.mm_tx.mm_percent_done << std::endl
            << "Num. Markers        : " << stat->mm_urlcopy.mm_tx.mm_markers << std::endl
            << "Error Scope         : " << UrlCopyError::scopeToString((UrlCopyError::Scope)stat->mm_urlcopy.mm_tx.mm_error.mm_scope) << std::endl
            << "Error Phase         : " << UrlCopyError::phaseToString((UrlCopyError::Phase)stat->mm_urlcopy.mm_tx.mm_error.mm_phase) << std::endl
            << "Error Category      : " << UrlCopyError::categoryToString((UrlCopyError::Category)stat->mm_urlcopy.mm_tx.mm_error.mm_category) << std::endl
            << "Error Message       : " << stat->mm_urlcopy.mm_tx.mm_error.mm_message << std::endl
            << "Num. Streams        : " << stat->mm_urlcopy.mm_tx.mm_streams << std::endl
            << "Tcp BS              : " << stat->mm_urlcopy.mm_tx.mm_tcp_bs << std::endl
            << "Timeout             : " << stat->mm_urlcopy.mm_tx.mm_timeout << std::endl
            << "Marker Timeout      : " << stat->mm_urlcopy.mm_tx.mm_marker_timeout << std::endl
            << "First Marker Timeout: " << stat->mm_urlcopy.mm_tx.mm_first_marker_timeout << std::endl;
}
