/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: TransferUrlCopyStat structure
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: TransferUrlCopyStat.cpp,v 1.10 2009/02/13 15:34:57 ptedesco Exp $
 *  Release: $Name:  $
 */
#include <glite/data/transfer/urlcopy/TransferUrlCopyStat.h>
#include <glite/data/transfer/urlcopy/UrlCopyError.h>
#include "UrlCopyStat.h"

using namespace glite::data::transfer::urlcopy;

namespace {
    
/**
 * copy the data in an ErrorStat object to an UrlCopyError
 * @param src [IN]
 * @param dst [OUT]
 */
void copy_error_info (
    const ErrorStat&    src,
    UrlCopyError&       dst){
    dst.scope       = (UrlCopyError::Scope)src.mm_scope;
    dst.phase       = (UrlCopyError::Phase)src.mm_phase;
    dst.category    = (UrlCopyError::Category)src.mm_category;
    dst.description = src.mm_message;
}

/**
 * copy the data in an UrlCopy_FileRequestStat object to a FileRequest
 * @param src [IN]
 * @param dst [OUT]
 */
void copy_request_info(
    const UrlCopy_FileRequestStat&      src,
    TransferUrlCopyStat::FileRequest&   dst){

    dst.srm_endpoint = src.mm_srm_endpoint;
    dst.srm_version  = src.mm_srm_version;
    dst.name         = src.mm_name;
    dst.tx_name      = src.mm_tx_name;
    dst.token        = src.mm_token;
    dst.exists       = (0 != src.mm_exists) ? true : false;
    dst.size         = src.mm_size;
    
    copy_error_info(src.mm_error,dst.error);        
    
    dst.prep_start_time   = src.mm_prep_start_time;
    dst.prep_done_time    = src.mm_prep_done_time;
    dst.final_start_time  = src.mm_final_start_time;
    dst.final_done_time   = src.mm_final_done_time;

}
    
} // namespace

/**
 * FileRequest
 * Default constructor
 */ 
TransferUrlCopyStat::FileRequest::FileRequest():
    exists(false),
    size(0),
    prep_start_time(0),
    prep_done_time(0),
    final_start_time(0),
    final_done_time(0){
}

/**
 * FileTransfer
 * Default constructor
 */
TransferUrlCopyStat::FileTransfer::FileTransfer(){
}
    
/**
 * TransferUrlCopyStat
 * Default Constructor
 */ 
TransferUrlCopyStat::TransferUrlCopyStat(){
    status            = UrlCopyStatus::IDLE;
    // Request Info
    start_time        = 0;
    finish_time       = 0;
    last_refresh_time = 0;
}

/**
 * TransferUrlCopyStat
 * Constructor
 */ 
TransferUrlCopyStat::TransferUrlCopyStat(const struct UrlCopyStat * const stat){
    if(0 != stat){
        // Status
        status            = (UrlCopyStatus::Value)stat->mm_urlcopy.mm_status;
        // Error
        copy_error_info(stat->mm_urlcopy.mm_error,error);
        
        // Request Info
        start_time        = stat->mm_common.mm_start_time;
        finish_time       = stat->mm_common.mm_finish_time;
        last_refresh_time = stat->mm_common.mm_last_refresh_time;
        
        // checksum
        src_checksum_algorithm  = stat->mm_urlcopy.mm_src_checksum_algorithm;
        src_checksum_value      = stat->mm_urlcopy.mm_src_checksum_value;    
        dst_checksum_algorithm  = stat->mm_urlcopy.mm_dst_checksum_algorithm;
        dst_checksum_value      = stat->mm_urlcopy.mm_dst_checksum_value;    
        
        // Source
        copy_request_info(stat->mm_urlcopy.mm_source,source);
                
        // Destination
        copy_request_info(stat->mm_urlcopy.mm_destination,destination);
        
        destination_space_token = stat->mm_common.mm_dest_space_token;
        user_tag                = stat->mm_common.mm_user_tag;
        
                
        // Transfer 
        tx.start_time         = stat->mm_urlcopy.mm_tx.mm_start_time;
        tx.done_time          = stat->mm_urlcopy.mm_tx.mm_done_time;
        tx.first_marker       = stat->mm_urlcopy.mm_tx.mm_first_marker;
        tx.last_marker        = stat->mm_urlcopy.mm_tx.mm_last_marker;
        tx.size               = stat->mm_urlcopy.mm_tx.mm_size;
        tx.transferred_bytes  = stat->mm_urlcopy.mm_tx.mm_transferred_bytes;
        tx.avg_throughput     = stat->mm_urlcopy.mm_tx.mm_avg_throughput;
        tx.instant_throughput = stat->mm_urlcopy.mm_tx.mm_instant_throughput;
        tx.percent_done       = stat->mm_urlcopy.mm_tx.mm_percent_done;
        tx.markers            = stat->mm_urlcopy.mm_tx.mm_markers;
        // Transfer Error
        copy_error_info(stat->mm_urlcopy.mm_tx.mm_error,tx.error);
        
        
    } else {
        status            = UrlCopyStatus::IDLE;
        // Request Info
        start_time        = 0;
        finish_time       = 0;
        last_refresh_time = 0;
    }
}
