/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Checksum checking logic for Srm/Url copy
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: TransferUtils.cpp,v 1.1 2009/07/30 19:32:55 molnarzs Exp $
 *  Release: $Name:  $
 */
#include "TransferUtils.h"
#include "logger-utility.h"
#include "sighandler-utility.h"
#include "SrmMethodCaller.h"

#include <glite/data/srm/util/SrmStat.h>
#include <glite/data/srm/util/SrmRm.h>
#include <glite/data/srm/util/SrmExceptions.h>

namespace glite   {
namespace data    {
namespace transfer{
namespace urlcopy {

namespace TransferUtils {

using namespace glite::data::srm::util;

bool SendPrepareToGet(
    glite::data::srm::util::Context::Ptr& ctx,
    RefreshTimeUpdater::IUpdateRefreshTime* urt,
    boost::scoped_ptr<glite::data::srm::util::PrepareToGet>& request,
    UrlCopyError::Category& c,
    std::string& m)
{
    try{
        RefreshTimeUpdater updater(urt);
        request->execute();
        info() << "PrepareToGet [" << request->token() << "] started";
        log_request(*request);
        return true;
    } catch(const HttpTimeoutError& exc){
        error() << "HTTP timeout contacting remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_HTTP_TIMEOUT;
        m = exc.what();
    } catch(const RemoteError& exc){ // http timeout handled
        error() << "Failed to contact remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_CONNECTION;
        m = exc.what();
    } catch(const SecurityError& exc){
        error() << "User is not authorized to contact the remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_PERMISSION;
        m = exc.what();
    } catch(const SrmError& exc){
        error() << "Remote method invocation failure on the remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_REQUEST_FAILURE;
        m = exc.what();
    } catch(const SrmBadResponse& exc){
        error() << "Invalid result returned by the remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_REQUEST_FAILURE;
        m = exc.what();
    } catch(const LogicError& exc){
        error() << "Logic Error in invoking PrepareToGet on remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_INTERNAL_ERROR;
        m = exc.what();
    }

    return false;
}

bool PollPrepareToGet (
    Context::Ptr& ctx,
    RefreshTimeUpdater::IUpdateRefreshTime* urt,
    boost::scoped_ptr<PrepareToGet>& request,
    int timeout,
    int start_time,
    unsigned int max_sleep_time,
    UrlCopyError::Category& c,
    std::string& m)
{
    bool ret = false;
    bool isFinished = false;

    glite::data::srm::util::details::IBackoffLogic::Ptr logic(request->getBackoffLogic());
    assert(logic.get());
    logic->setMaxSleepTime(max_sleep_time);
    logic->setTimeout(timeout);
    time_t to = static_cast<time_t>(start_time);
    logic->reset(&to);

    try{
        // Loop until the request is completed or failed
        while ( !isFinished && !TransferState::instance().interrupted() )
        {
        	try {
        		isFinished = request->backoff();
        		RefreshTimeUpdater updater(urt);

        		if (!isFinished) {
        	        info() << "Status of PrepareToGet [" << request->token() << "] updated";
        	        log_request(*request);
        		}
        	} catch(const RemoteError& exc){ // don't handle http timeout explicitly here
                warn() << "Failed to contact remote SRM [" << ctx->endpoint << "] in updating the PrepareToGet status: " << exc.what();
                info() << "This error is considered transient and the action will be retried";
            } catch(const SrmError& exc){
                warn() << "Remote method invocation failure on the remote SRM [" << ctx->endpoint << "] in updating the PrepareToGet status: " << exc.what();
                info() << "This error is considered transient and the action will be retried";
            }
        } // end while
        if(true == TransferState::instance().interrupted()){
            error() << "PrepareToGet [" << request->token() << "] should be aborted";
            c = UrlCopyError::ERROR_ABORTED;
            m = "transfer request canceled";
        } else if(c == UrlCopyError::ERROR_REQUEST_TIMEOUT){
            error() << "PrepareToGet [" << request->token() << "] timed out";
        } else {
            info() << "PrepareToGet [" << request->token() << "] completed with status " << request->status.code;
            c = UrlCopyError::SUCCESS;
            ret = true;
        }
    } catch(const SecurityError& exc){
        error() << "User is not authorized to contact the remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_PERMISSION;
        m = exc.what();
    } catch(const SrmBadResponse& exc){
        error() << "Invalid result returned by the remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_REQUEST_FAILURE;
        m = exc.what();
    } catch(const LogicError& exc){
        error() << "Logic Error in invoking PrepareToGet: " << exc.what();
        c = UrlCopyError::ERROR_INTERNAL_ERROR;
        m = exc.what();
    }
    return ret;
}

void StatErrorFromSrmStatus(
    UrlCopyError::Category&                     c,
    std::string&                                m,
    const glite::data::srm::util::SrmStatus&    req_status,
    const glite::data::srm::util::SrmStatus&    file_status) throw()
{
    // First Check the file status code
    switch(file_status.code){
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
            c = UrlCopyError::ERROR_PERMISSION;
            m = file_status.message;
        break;
        case SrmStatus::SRM_INVALID_PATH:
            c = UrlCopyError::ERROR_INVALID_PATH;
            m = file_status.message;
        break;
        case SrmStatus::SRM_UNDEF:
        case SrmStatus::SRM_SUCCESS:
            // File Status doesn't contain the reason of the error
        break;
        default:
            // Set default error
            c = UrlCopyError::ERROR_GENERAL_FAILURE;
            if(false == file_status.message.empty()){
                m = file_status.message;
            } else {
                std::stringstream ss;
                ss << "Stat failed on the SRM with error [" << file_status.code << "]";
                m = ss.str();
            }
        break;
    }
    // If not yet found, check the Request status
    TRACE("Could not determine failure reason from file status; checking request status");

    if(c == UrlCopyError::SUCCESS)
    	ErrorFromRequestStatus(c, m, req_status);
}

void ErrorFromRequestStatus(
    UrlCopyError::Category&                     c,
    std::string&                                m,
    const glite::data::srm::util::SrmStatus&    req_status)
{
    TRACE(">> errorFromRequestStatus : req_status=[" << req_status << "]");
    switch(req_status.code){
    case SrmStatus::SRM_AUTHENTICATION_FAILURE:
    case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        c = UrlCopyError::ERROR_PERMISSION;
        m = req_status.message;
        break;
    case SrmStatus::SRM_INVALID_REQUEST:
        c = UrlCopyError::ERROR_CONSISTENCY;
        m = req_status.message;
        break;
    case SrmStatus::SRM_INTERNAL_ERROR:
        c = UrlCopyError::ERROR_STORAGE_INTERNAL_ERROR;
        m = req_status.message;
        break;
    case SrmStatus::SRM_UNDEF:
    case SrmStatus::SRM_SUCCESS: {
        // The error is neither on the Request status. This in principle
        // should not be possible
        c = UrlCopyError::ERROR_INTERNAL_ERROR;
        std::stringstream ss;
        ss << "The SRM returned an error, but with an invalid code [" << req_status.code << "]";
        m = ss.str();
        } break;
    default:
        // Set default error
        c = UrlCopyError::ERROR_GENERAL_FAILURE;
        if(false == req_status.message.empty()){
            m = req_status.message;
        } else {
            std::stringstream ss;
            ss << "Stat SRM request failed with error [" << req_status.code << "]";
            m = ss.str();
        }
        break;
    }
    TRACE("<< errorFromRequestStatus : c=[" << c << "] m=[" << m << "]");
}

/**
 * pathExists
 * Check if a path (file or dircetory) already exists
 */
bool PathExists(
	RefreshTimeUpdater::IUpdateRefreshTime * updater,
    Context::Ptr&           ctx,
    const std::string&      surl,
    bool                    full_detailed,
    UrlCopyError::Category& c,
    std::string&            m,
    PathInfo&               pinfo)
{
    TRACE_VALUE(surl);

    bool file_exists = false;

    // Check if the directory exists
    boost::scoped_ptr<SrmStat> stat_request;
    try{
        // Create SrmStat Request
        stat_request.reset(SrmStat::create(ctx));
    } catch(const LogicError& exc){
        // SrmStat method is not supported
        c = UrlCopyError::ERROR_INTERNAL_ERROR;
        m = exc.what();
        return false;
    }
    stat_request->files.resize(1);
    stat_request->files[0].surl = surl;

    stat_request->fullDetailedList = full_detailed;

    // Execute Synch Request
    SrmMethodCaller srmMethodCaller(updater);
    SrmStatus stat_result = srmMethodCaller.executeRequest(ctx, *stat_request, c, m);

    if(c == UrlCopyError::SUCCESS){
        if(stat_result.code == SrmStatus::SRM_SUCCESS){
            // the directory already exists
            debug() << "path [" << surl << "] exists (type = " << stat_request->files[0].fileType << ")";
            file_exists = true;
            pinfo.file_type = stat_request->files[0].fileType;

            if (full_detailed && (pinfo.file_type == FileType::FILE)) {
                pinfo.file_size = stat_request->files[0].size;
                debug() << "file [" << surl << "] size is [" << pinfo.file_size << "]";
                pinfo.file_locality = stat_request->files[0].fileLocality;
                debug() << "file [" << surl << "] locality is [" << pinfo.file_locality << "]";
                pinfo.cksm_algorithm = stat_request->files[0].checkSumType;
                pinfo.cksm_value = stat_request->files[0].checkSumValue;
            }
        } else {
            file_exists = false;
            // Check the file status
            if(stat_request->files[0].status.code == SrmStatus::SRM_INVALID_PATH){
                info() << "path [" << surl << "] does not exist on SRM [" << ctx->endpoint << "]";
                c = UrlCopyError::SUCCESS;
            } else {
                error() << "getting stat for file [" << surl << "] failed: [" << stat_result.code << "] " << stat_result.message;
                // Detect the correct UrlCopy error code from SrmStatus
                StatErrorFromSrmStatus(c, m, stat_result, stat_request->files[0].status);
            }
        }
    } else {
        file_exists = false;
        error() << "failed to get stat for file [" << surl << "] failed : " << m;
    }
    return file_exists;
}

bool RmFile(
	RefreshTimeUpdater::IUpdateRefreshTime * updater,
	Context::Ptr& ctx,
	const std::string& surl,
	UrlCopyError::Category& c,
	std::string& m)
{
    bool file_removed = false;

    // Remove the file
    boost::scoped_ptr<SrmRm> rm_request;
    try{
        // Create SrmRm Request
        rm_request.reset(SrmRm::create(ctx));
    } catch(const LogicError& exc){
        // SrmRm method is not supported
        c = UrlCopyError::ERROR_INTERNAL_ERROR;
        m = exc.what();
        return file_removed;
    }
    // Set Request Parameters
    rm_request->files.resize(1);
    rm_request->files[0].surl = surl;

    // Execute Request
    SrmMethodCaller srmMethodCaller(updater);
    SrmStatus rm_result = srmMethodCaller.executeRequest(ctx, *rm_request, c, m);
    if(c != UrlCopyError::SUCCESS){
        error() << "Cannot Remove file [" << surl << "]: " << m;
        file_removed = false;
    } else {
        if((rm_result.code == SrmStatus::SRM_SUCCESS) ||
           (rm_request->files[0].status.code == SrmStatus::SRM_SUCCESS)){
            info() << "File [" << surl << "] removed";
            file_removed = true;
        } else if(rm_request->files[0].status.code == SrmStatus::SRM_INVALID_PATH){
            info() << "File [" << surl << "] does no longer exist";
            file_removed = true;
        } else {
            file_removed = false;
            // Detect the correct UrlCopy error code from SrmRm
            RmErrorFromSrmStatus(c,m,rm_result,rm_request->files[0].status);
            error() << "Failed to remove file [" << surl << "]: [" << UrlCopyError::categoryToString(c) << "] " << m;
        }
    }
    return file_removed;
}

void RmErrorFromSrmStatus(
    UrlCopyError::Category& c,
    std::string& m,
    const glite::data::srm::util::SrmStatus& req_status,
    const glite::data::srm::util::SrmStatus& file_status)
{
    // First Check if the file status code
    switch(file_status.code){
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
            c = UrlCopyError::ERROR_PERMISSION;
            m = file_status.message;
        break;
        case SrmStatus::SRM_INVALID_PATH:
        case SrmStatus::SRM_FILE_LOST:
        case SrmStatus::SRM_FILE_UNAVAILABLE:
            c = UrlCopyError::ERROR_INVALID_PATH;
            m = file_status.message;
        break;
        case SrmStatus::SRM_UNDEF:
        case SrmStatus::SRM_SUCCESS:
            // File Status doesn't contain the reason of the error
        break;
        default:
            // Set default error
            c = UrlCopyError::ERROR_GENERAL_FAILURE;
            if(false == file_status.message.empty()){
                m = file_status.message;
            } else {
                std::stringstream ss;
                ss << "Stat failed on the SRM with error [" << file_status.code << "]";
                m = ss.str();
            }
        break;
    }
    // If not yet found, check the Request status
    if(c == UrlCopyError::SUCCESS){
        ErrorFromRequestStatus(c, m, req_status);
    }
}

} // namespace TransferUtils

} // namespace urlcopy
} // namespace transfer
} // namespace data
} // namespace glite
