/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: UrlCopyCore class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>,
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *           Zsolt Molnar <Zsolt.Molnar@cern.ch - www.zsoltmolnar.hu>
 *  Version info: $Id: UrlCopyCore.cpp,v 1.54 2009/07/30 19:50:54 molnarzs Exp $
 *  Release: $Name:  $
 */
#include "UrlCopyCore.h"
#include "FileType.h"
#include "logger-utility.h"
#include "stat-utility.h"
#include "sighandler-utility.h"
#include "gridftp-utility.h"
#include "UrlCopyUtils.h"
#include "gridftp-ifce.h"



#include <math.h>
#include <glite/data/srm/util/Context.h>
#include <glite/data/srm/util/SrmExceptions.h>
#include <glite/data/srm/util/SrmPing.h>
#include <glite/data/srm/util/BringOnline.h>
#include <glite/data/srm/util/PrepareToGet.h>
#include <glite/data/srm/util/PrepareToPut.h>
#include <glite/data/srm/util/SrmStat.h>
#include <glite/data/srm/util/SrmMkDir.h>
#include <glite/data/srm/util/GetSpaceTokens.h>
#include <glite/data/srm/util/SrmRm.h>
#include <glite/data/srm/util/details/SynchRequest.h>

#include <glite/data/agents/url-utility.h>
#include <glite/data/agents/surl-utility.h>
#include <glite/data/agents/extra/utility.h>
#include <glite/data/agents/cred/cred-utility.h>

#include <string>
#include <sstream>
#include <iomanip>
#include <fstream>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <libgen.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <boost/scoped_ptr.hpp>
#include <boost/algorithm/string.hpp>

using namespace glite::data::transfer::urlcopy;
using namespace glite::data::srm::util;

using glite::data::srm::util::details::SynchRequest;

using glite::data::agents::parse_query_string;

namespace {

    //------------------------------------------------------------------------------
    // Constants
    //------------------------------------------------------------------------------

    // Timeouts
    const unsigned int DEFAULT_HTTP_TIMEOUT = 40;
    const unsigned int DEFAULT_PUT_TIMEOUT = 180;
    const unsigned int DEFAULT_GETDONE_TIMEOUT = 180;
    const unsigned int DEFAULT_PUTDONE_TIMEOUT = 180;
    const unsigned int DEFAULT_TX_TIMEOUT = 3600;
    const unsigned int DEFAULT_TX_MARKERS_TIMEOUT = 0;
    const unsigned int DEFAULT_FIRST_TX_MARKER_TIMEOUT = 0;

    // Others
    const unsigned int MAX_PUT_SLEEP_TIME = 30; // bug #39492
    const unsigned int DEFAULT_CONNECTION_SLEEP_TIME = 3;
    const unsigned int NUM_RETRY_CONNECTION = 3;
    const unsigned int SAFE_CLEANUP_MULIPLY_FACTOR = 3;

    //------------------------------------------------------------------------------
    // Helper Methods
    //------------------------------------------------------------------------------


    // Compute the transfer timeout

    int compute_timeout(UrlCopyStat* m_stat) {
        int tx_timeout = m_stat->mm_urlcopy.mm_tx.mm_timeout > 0 ? m_stat->mm_urlcopy.mm_tx.mm_timeout : 0;
        double to_per_mb = fmaxl(0, m_stat->mm_urlcopy.mm_tx.mm_tx_to_per_mb);
        static const unsigned long MB = 1 << 20;
        double dtotal_timeout = tx_timeout + ceil(to_per_mb * m_stat->mm_urlcopy.mm_source.mm_size / MB);
        int total_timeout = (dtotal_timeout >= INT_MAX) ? 0 : (int) dtotal_timeout;
        return total_timeout;
    }

    /**
     * tx_callback
     * callback for gridftp copy
     * must return 0 if the transfer should continue, another value otherwise. If an
     * error condition is encountered, the source/transfer/destination Error and
     * finalError functions should be called in order to properly set the error,
     * since the gridftp_copy function cannot tell if the transfer was over because
     * of a gridftp error or because it was interrupted by the callback function.
     */

    int tx_callback(void * arg, PerfMarker& perf) /*throw()*/ {
        static bool first_called = true;
        int retval = 0;
        // Save perf markers
        UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *> (arg);
        if (0 != stat) {

            bool log_as_info = (stat->mm_urlcopy.mm_tx.mm_transferred_bytes != perf.transferred_bytes) || first_called;
            first_called = false;

            time_t current;
            time(&current);
            stat->mm_common.mm_last_refresh_time = current;
            if (0 == stat->mm_urlcopy.mm_tx.mm_first_marker) {
                stat->mm_urlcopy.mm_tx.mm_first_marker = perf.ts.tv_sec;
            }
            if (0 != perf.transferred_bytes && stat->mm_urlcopy.mm_tx.mm_transferred_bytes != perf.transferred_bytes) {
                stat->mm_urlcopy.mm_tx.mm_last_progress_marker = current;
            }


            stat->mm_urlcopy.mm_tx.mm_last_marker = perf.ts.tv_sec;
            stat->mm_urlcopy.mm_tx.mm_transferred_bytes = perf.transferred_bytes;
            stat->mm_urlcopy.mm_tx.mm_avg_throughput = perf.avg_throughput;
            stat->mm_urlcopy.mm_tx.mm_instant_throughput = perf.instant_throughput;
            stat->mm_urlcopy.mm_tx.mm_markers = perf.markers;


            if (0 != stat->mm_urlcopy.mm_tx.mm_size) {
                // verify that transferred_bytes have not exceeded source file size
                if (perf.transferred_bytes > stat->mm_urlcopy.mm_tx.mm_size) {
                    retval = 1;
                    std::stringstream message;
                    message << "Transferred bytes (" << perf.transferred_bytes <<
                            ") exceeded file size (" << stat->mm_urlcopy.mm_tx.mm_size << ")";
                    error() << message.str();
                    UrlCopyUtils::SetScopeError(stat->mm_urlcopy, UrlCopyError::SCOPE_TRANSFER, UrlCopyError::PHASE_TRANSFER, UrlCopyError::ERROR_GRIDFTP, message.str());
                    UrlCopyUtils::SetFinalError(stat->mm_urlcopy, UrlCopyError::SCOPE_TRANSFER);
                }
                stat->mm_urlcopy.mm_tx.mm_percent_done = ((float) perf.transferred_bytes) / stat->mm_urlcopy.mm_tx.mm_size;
            } else {
                stat->mm_urlcopy.mm_tx.mm_percent_done = 0;
            }

            if (log_as_info) {
                info() << "Transferred " << stat->mm_urlcopy.mm_tx.mm_transferred_bytes <<
                        " out of " << stat->mm_urlcopy.mm_tx.mm_size << " bytes [" <<
                        std::showpoint << std::setprecision(2) << std::fixed << (stat->mm_urlcopy.mm_tx.mm_percent_done * 100) << "%]. Inst: " <<
                        std::showpoint << std::setprecision(2) << std::fixed << (stat->mm_urlcopy.mm_tx.mm_instant_throughput / 1024.0) << " KB/s.  Avg: " <<
                        std::showpoint << std::setprecision(2) << std::fixed << (stat->mm_urlcopy.mm_tx.mm_avg_throughput / 1024.0) << " KB/s";
            } else {
                debug() << "Transferred " << stat->mm_urlcopy.mm_tx.mm_transferred_bytes <<
                        " out of " << stat->mm_urlcopy.mm_tx.mm_size << " bytes [" <<
                        std::showpoint << std::setprecision(2) << std::fixed << (stat->mm_urlcopy.mm_tx.mm_percent_done * 100) << "%]. Inst: " <<
                        std::showpoint << std::setprecision(2) << std::fixed << (stat->mm_urlcopy.mm_tx.mm_instant_throughput / 1024.0) << " KB/s.  Avg: " <<
                        std::showpoint << std::setprecision(2) << std::fixed << (stat->mm_urlcopy.mm_tx.mm_avg_throughput / 1024.0) << " KB/s";
            }

            // verify tx markers timeout
            if ((0 == stat->mm_urlcopy.mm_tx.mm_markers) && // no markers received and
                    (stat->mm_urlcopy.mm_tx.mm_first_marker_timeout) && // the first marker timeout is set and
                    (current >= (stat->mm_urlcopy.mm_tx.mm_gridftp_copy_start_time + stat->mm_urlcopy.mm_tx.mm_first_marker_timeout))) { // the first tx marker timeout is elapsed
                retval = 1;
                std::stringstream message;
                message << "First transfer marker not received within " << stat->mm_urlcopy.mm_tx.mm_first_marker_timeout << " seconds";
                error() << message.str();

                UrlCopyUtils::SetScopeError(stat->mm_urlcopy, UrlCopyError::SCOPE_TRANSFER,
                        UrlCopyError::PHASE_TRANSFER, UrlCopyError::ERROR_FIRST_MARKER_TIMEOUT, message.str());

                UrlCopyUtils::SetFinalError(stat->mm_urlcopy, UrlCopyError::SCOPE_TRANSFER);
            }
            if ((0 == perf.transferred_bytes) && // no transfer progress
                    (stat->mm_urlcopy.mm_tx.mm_first_marker_timeout) && // the first marker timeout is set and
                    (current >= (stat->mm_urlcopy.mm_tx.mm_gridftp_copy_start_time + stat->mm_urlcopy.mm_tx.mm_first_marker_timeout))) { // the first tx marker timeout is elapsed
                retval = 1;
                std::stringstream message;
                message << "First non-zero marker not received within " << stat->mm_urlcopy.mm_tx.mm_first_marker_timeout << " seconds";
                error() << message.str();
                UrlCopyUtils::SetScopeError(stat->mm_urlcopy, UrlCopyError::SCOPE_TRANSFER, UrlCopyError::PHASE_TRANSFER, UrlCopyError::ERROR_FIRST_MARKER_TIMEOUT, message.str());
                UrlCopyUtils::SetFinalError(stat->mm_urlcopy, UrlCopyError::SCOPE_TRANSFER);
            }
            if ((0 != stat->mm_urlcopy.mm_tx.mm_markers) && // at least one marker was received and
                    (stat->mm_urlcopy.mm_tx.mm_marker_timeout) && // the markers timeout is set and
                    (current >= (stat->mm_urlcopy.mm_tx.mm_last_marker + stat->mm_urlcopy.mm_tx.mm_marker_timeout))) { // the tx markers timeout is elapsed

                retval = 1;
                std::stringstream message;
                message << "No transfer markers received for more than " << stat->mm_urlcopy.mm_tx.mm_marker_timeout << " seconds";
                error() << message.str();
                UrlCopyUtils::SetScopeError(stat->mm_urlcopy, UrlCopyError::SCOPE_TRANSFER, UrlCopyError::PHASE_TRANSFER, UrlCopyError::ERROR_MARKERS_TIMEOUT, message.str());
                UrlCopyUtils::SetFinalError(stat->mm_urlcopy, UrlCopyError::SCOPE_TRANSFER);
            }
            if ((0 != stat->mm_urlcopy.mm_tx.mm_last_progress_marker) && // at least one marker indicating non-zero progress was received and
                    (stat->mm_urlcopy.mm_tx.mm_no_progress_timeout > 0) && // the no-progress timeout is set
                    (current >= (stat->mm_urlcopy.mm_tx.mm_last_progress_marker + stat->mm_urlcopy.mm_tx.mm_no_progress_timeout))) { // the no-progress timeout is elapsed

                retval = 1;
                std::stringstream message;
                message << "No markers indicating progress received for more than " << stat->mm_urlcopy.mm_tx.mm_no_progress_timeout << " seconds";
                error() << message.str();

                UrlCopyUtils::SetScopeError(stat->mm_urlcopy, UrlCopyError::SCOPE_TRANSFER,
                        UrlCopyError::PHASE_TRANSFER, UrlCopyError::ERROR_NO_PROGRESS, message.str());

                UrlCopyUtils::SetFinalError(stat->mm_urlcopy, UrlCopyError::SCOPE_TRANSFER);
            }
        }



        return retval;
    }

} // end global namespace

//------------------------------------------------------------------------------
// UrlCopyCore Methods
//------------------------------------------------------------------------------

/*
 * UrlCopyCore
 *
 * Constructor
 */
UrlCopyCore::UrlCopyCore(const std::string & id) : TransferCore(id), m_stat(0) {

    char hostname[1024];
    hostname[1023] = '\0';
    gethostname(hostname, 1023);
    msg_ifce::getInstance()->set_tr_timestamp_start(&tr_completed, msg_ifce::getInstance()->getTimestamp());    
    msg_ifce::getInstance()->set_agent_fqdn(&tr_completed, hostname);
    gsiftpOverwrite = false;
}

/*
 * ~UrlCopyCore
 *
 * Destructor
 */
UrlCopyCore::~UrlCopyCore() {
    msg_ifce::getInstance()->set_tr_timestamp_complete(&tr_completed, msg_ifce::getInstance()->getTimestamp());    
    msg_ifce::getInstance()->SendTransferFinishMessage(&tr_completed);
    do_globus_deactivates();
}

/**
 * UpdateRefreshTime
 */
void UrlCopyCore::UpdateRefreshTime() {
    time(&(m_stat->mm_common.mm_last_refresh_time));
}

//------------------------------------------------------------------------------
//
// Transfer Preparation
//
//------------------------------------------------------------------------------

/*
 * stat
 *
 * Return the Stat Object
 */
UrlCopyStat& UrlCopyCore::stat() /*throw (RuntimeError, LogicError)*/ {
    // Initialize (if not already initialized)
    initialize();
    if (0 == m_stat) {
        throw LogicError("null stat object");
    }
    return *m_stat;
}

/*
 * initialize
 *
 * Initialize the Stat structure from the Mmap file
 */
void UrlCopyCore::initialize() /*throw (RuntimeError, LogicError)*/ {
    if (0 != m_stat) {
        // Already initialized
        return;
    }
    try {
        // Initialize Mmap
        TransferCore::initialize();
        // Get the pointer to the Stat buffer
        m_stat = reinterpret_cast<UrlCopyStat*> (m_mmap->buffer());
        // Check if the Stat id is correct
        if (m_stat->mm_common.mm_id != URLCOPY_STAT_ID) {
            throw LogicError("Not an UrlCopy transfer request");
        }
        // Set all the timeouts to the default
        // ---------------------------------------------------------------------

        // Source
        if (m_stat->mm_urlcopy.mm_source.mm_http_timeout < 0) {
            m_stat->mm_urlcopy.mm_source.mm_http_timeout = DEFAULT_HTTP_TIMEOUT;
        }
        if (m_stat->mm_urlcopy.mm_source.mm_prep_timeout < 0) {
            m_stat->mm_urlcopy.mm_source.mm_prep_timeout = DEFAULT_GET_TIMEOUT;
        }
        if (m_stat->mm_urlcopy.mm_source.mm_final_timeout < 0) {
            m_stat->mm_urlcopy.mm_source.mm_final_timeout = DEFAULT_GETDONE_TIMEOUT;
        }
        // Destination
        if (m_stat->mm_urlcopy.mm_destination.mm_http_timeout < 0) {
            m_stat->mm_urlcopy.mm_destination.mm_http_timeout = DEFAULT_HTTP_TIMEOUT;
        }
        if (m_stat->mm_urlcopy.mm_destination.mm_prep_timeout < 0) {
            m_stat->mm_urlcopy.mm_destination.mm_prep_timeout = DEFAULT_PUT_TIMEOUT;
        }
        if (m_stat->mm_urlcopy.mm_destination.mm_final_timeout < 0) {
            m_stat->mm_urlcopy.mm_destination.mm_final_timeout = DEFAULT_PUTDONE_TIMEOUT;
        }
        // Transfer
        if (m_stat->mm_urlcopy.mm_tx.mm_timeout < 0) {
            m_stat->mm_urlcopy.mm_tx.mm_timeout = DEFAULT_TX_TIMEOUT;
        } else if ((m_stat->mm_urlcopy.mm_tx.mm_timeout > 0) && (m_stat->mm_urlcopy.mm_tx.mm_timeout < DEFAULT_TX_TIMEOUT)) {
            m_stat->mm_urlcopy.mm_tx.mm_timeout = DEFAULT_TX_TIMEOUT;
        }

        if (m_stat->mm_urlcopy.mm_tx.mm_marker_timeout < 0) {
            m_stat->mm_urlcopy.mm_tx.mm_marker_timeout = DEFAULT_TX_MARKERS_TIMEOUT;
        }
        if (m_stat->mm_urlcopy.mm_tx.mm_first_marker_timeout < 0) {
            m_stat->mm_urlcopy.mm_tx.mm_first_marker_timeout = DEFAULT_FIRST_TX_MARKER_TIMEOUT;
        }
    } catch (...) {
        m_mmap.reset();
        m_stat = 0;
        throw;
    }

    // initialize the syslog helper
    ContextLogInfo::user_tag = m_stat->mm_common.mm_user_tag;
    ContextLogInfo::tx_id = id();

}

/*
 * prepare
 *
 * Preapare the source and destination files for the transfer
 */
void UrlCopyCore::prepare() /*throw (RuntimeError, LogicError)*/ {

    do_globus_activates();

    // Check of the instance is initialized, otherwise initialize it from the
    // status file
    initialize();

    // Get Proxy DN
    std::string user_dn;
    try {
        user_dn = glite::data::agents::cred::get_proxy_dn();
    } catch (const RuntimeError& exc) {
        user_dn = "(not available)";
    }

    info() << "Transfer mode: urlcopy";
    info() << "Transfer ID  : " << this->id();
    info() << "User DN      : " << user_dn;
    info() << "User Descr.  : " << m_stat->mm_common.mm_user_tag;
    info() << "Source SRM [" << m_stat->mm_urlcopy.mm_source.mm_srm_version << "]: " << m_stat->mm_urlcopy.mm_source.mm_srm_endpoint;
    info() << "Dest.  SRM [" << m_stat->mm_urlcopy.mm_destination.mm_srm_version << "]: " << m_stat->mm_urlcopy.mm_destination.mm_srm_endpoint;
    info() << "Source     : " << m_stat->mm_urlcopy.mm_source.mm_name;
    info() << "Destination: " << m_stat->mm_urlcopy.mm_destination.mm_name;

    msg_ifce::getInstance()->set_t_channel(&tr_completed, this->id());
    msg_ifce::getInstance()->set_transfer_id(&tr_completed, this->id().c_str());
    msg_ifce::getInstance()->set_source_srm_version(&tr_completed, m_stat->mm_urlcopy.mm_source.mm_srm_version);
    msg_ifce::getInstance()->set_destination_srm_version(&tr_completed, m_stat->mm_urlcopy.mm_destination.mm_srm_version);
    msg_ifce::getInstance()->set_source_url(&tr_completed, m_stat->mm_urlcopy.mm_source.mm_name);
    msg_ifce::getInstance()->set_dest_url(&tr_completed, m_stat->mm_urlcopy.mm_destination.mm_name);
    msg_ifce::getInstance()->set_source_hostname(&tr_completed, m_stat->mm_urlcopy.mm_source.mm_name);
    msg_ifce::getInstance()->set_dest_hostname(&tr_completed, m_stat->mm_urlcopy.mm_destination.mm_name);
    msg_ifce::getInstance()->set_channel_type(&tr_completed, "urlcopy");


    info() << "VO: " << vo_name;
    msg_ifce::getInstance()->set_vo(&tr_completed, vo_name, this->id());
    msg_ifce::getInstance()->set_source_site_name(&tr_completed, src_site);
    msg_ifce::getInstance()->set_dest_site_name(&tr_completed, dest_site);    
    

    std::string nstream_to_string = to_string<int>(m_stat->mm_urlcopy.mm_tx.mm_streams, std::dec);
    msg_ifce::getInstance()->set_number_of_streams(&tr_completed, nstream_to_string.c_str());

    std::string tcpbuffer_to_string = to_string<int>(m_stat->mm_urlcopy.mm_tx.mm_tcp_bs, std::dec);
    msg_ifce::getInstance()->set_tcp_buffer_size(&tr_completed, tcpbuffer_to_string.c_str());

    std::string block_to_string = to_string<int>(m_stat->mm_urlcopy.mm_tx.mm_block_size, std::dec);
    msg_ifce::getInstance()->set_block_size(&tr_completed, block_to_string.c_str());

    int total_timeout = compute_timeout(m_stat);
    info() << "Transfer timeout = " << total_timeout;
    std::string timeout_to_string = to_string<int>(total_timeout, std::dec);
    msg_ifce::getInstance()->set_transfer_timeout(&tr_completed, timeout_to_string.c_str());

    msg_ifce::getInstance()->set_srm_space_token_dest(&tr_completed, m_stat->mm_common.mm_dest_token_descr);
    msg_ifce::getInstance()->set_srm_space_token_source(&tr_completed, m_stat->mm_common.mm_source_token_descr);

    //after the precondition check for READY state, and before set it to RUNNING, send the transfer_started message
    //send monitoring transfer started message
    msg_ifce::getInstance()->SendTransferStartMessage(&tr_completed);


    msg_ifce::getInstance()->set_time_spent_in_srm_preparation_start(&tr_completed, msg_ifce::getInstance()->getTimestamp());

    // Precondition checks
    // -------------------------------------------------------------------------
    // Check that the state is IDLE
    if (m_stat->mm_urlcopy.mm_status != UrlCopyStatus::IDLE) {
        // Cannot Call Prepare on an active transfer
        std::stringstream ss;
        ss << "prepare called for a transfer not in ["
                << UrlCopyStatus::toString(UrlCopyStatus::IDLE)
                << "] state. Current state is ["
                << UrlCopyStatus::toString(
                (UrlCopyStatus::Value)m_stat->mm_urlcopy.mm_status) << "]";
        std::string reason = ss.str();
        error() << reason;
        throw LogicError(reason.c_str());
    }
    // Check that the source and destination file names are set
    if (strlen(m_stat->mm_urlcopy.mm_source.mm_name) == 0) {
        // Set Source Error
        sourceError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_USER, "empty source file name");
        finalError(UrlCopyError::SCOPE_SOURCE); // OK
        return;
    }
    if (strlen(m_stat->mm_urlcopy.mm_destination.mm_name) == 0) {
        // Set Destination Error
        destinationError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_USER, "empty destination file name");
        finalError(UrlCopyError::SCOPE_DESTINATION); // OK
        return;
    }

    // Be sure the tx names are empty
    copy_url(m_stat->mm_urlcopy.mm_source.mm_tx_name, "");
    copy_url(m_stat->mm_urlcopy.mm_destination.mm_tx_name, "");

    // Get the source & destination File Type
    FileType::Value src_type = FileType::get(m_stat->mm_urlcopy.mm_source.mm_name);
    FileType::Value dst_type = FileType::get(m_stat->mm_urlcopy.mm_destination.mm_name);
    m_stat->mm_urlcopy.mm_source.mm_type = src_type;
    m_stat->mm_urlcopy.mm_destination.mm_type = dst_type;

    // The Transfer is now preparing
    m_stat->mm_urlcopy.mm_status = UrlCopyStatus::PREPARING;

    // Ping the Source and destination servers
    if (false == pingSource()) {
        // Source doesn't respond
        finalError(UrlCopyError::SCOPE_SOURCE); // OK (pingSource calls sourceError)
        return;
    }
    if (false == pingDestination()) {
        // destination doesn't respond
        finalError(UrlCopyError::SCOPE_DESTINATION); // OK (pingDestination calls sourceError)
        return;
    }

    // Prepare the Source File
    info() << "STATUS:BEGIN:SOURCE - Preparation";

    prepareSource();

    // Check source error
    if (UrlCopyError::SUCCESS != m_stat->mm_urlcopy.mm_source.mm_error.mm_category) {
        info() << "STATUS:END fail:SOURCE - Preparation";
        finalError(UrlCopyError::SCOPE_SOURCE); // OK (just checked)
	msg_ifce::getInstance()->set_time_spent_in_srm_preparation_end(&tr_completed, msg_ifce::getInstance()->getTimestamp());
        return;
    }

    // Check source transfer file name
    if (0 == strlen(m_stat->mm_urlcopy.mm_source.mm_tx_name)) {
        // No Tx file name set: the reason should have already be filled
        info() << "STATUS:END fail:SOURCE - Preparation";
        finalError(UrlCopyError::SCOPE_SOURCE); // should be OK
	msg_ifce::getInstance()->set_time_spent_in_srm_preparation_end(&tr_completed, msg_ifce::getInstance()->getTimestamp());
        return;
    }
    info() << "STATUS:END:SOURCE - Preparation";
    debug() << "Source file is now ready";

    // Prepare the Destination File
    info() << "STATUS:BEGIN:DESTINATION - Preparation";

    prepareDestination();
 
    

    // Check destination error
    if (UrlCopyError::SUCCESS != m_stat->mm_urlcopy.mm_destination.mm_error.mm_category) {
        info() << "STATUS:END fail:DESTINATION - Preparation";
        finalError(UrlCopyError::SCOPE_DESTINATION); // OK (just checked)
	msg_ifce::getInstance()->set_time_spent_in_srm_preparation_end(&tr_completed, msg_ifce::getInstance()->getTimestamp());
        return;
    }
    // Check destination transfer file name
    if (0 == strlen(m_stat->mm_urlcopy.mm_destination.mm_tx_name)) {
        info() << "STATUS:END fail:DESTINATION - Preparation";
        // No Tx file name set: the reason should have already be filled
        finalError(UrlCopyError::SCOPE_DESTINATION); // should be OK
	msg_ifce::getInstance()->set_time_spent_in_srm_preparation_end(&tr_completed, msg_ifce::getInstance()->getTimestamp());
        return;
    }
    info() << "STATUS:END:DESTINATION - Preparation";
    debug() << "Destination file is now ready";

    // The Transfer is now prepared: update the file Status
    m_stat->mm_urlcopy.mm_status = UrlCopyStatus::READY;

    msg_ifce::getInstance()->set_time_spent_in_srm_preparation_end(&tr_completed, msg_ifce::getInstance()->getTimestamp());
 
    // wait
    if (m_stat->mm_common.mm_block_after_preparation) {
        waitContinueSignal();
    }
}

/**
 * wait for signal to continue process
 */
void UrlCopyCore::waitContinueSignal() {
    info() << "Waiting for signal to continue.";
    UpdateRefreshTime();
    while (true) {
        RefreshTimeUpdater updater(this);
        if (TransferState::instance().get_continueTransfer() ||
                TransferState::instance().interrupted()) {
            info() << "Continuing process.";
            TransferState::instance().set_continueTransfer(false);
            break;
        } else {
            debug() << "Waiting for signal to continue.";
            sleep(5);
        }
    }
}

/**
 * pingSource
 * Ping Source server
 */
bool UrlCopyCore::pingSource() /*throw ()*/ {

    if (!m_stat->mm_common.mm_enable_ping_source) {
        info() << "Skipping srmPing on source endpoint.";
        return true;
    }

    bool server_ok = false;
    // Get the source File Type
    FileType::Value src_type = (FileType::Value)m_stat->mm_urlcopy.mm_source.mm_type;

    // Ping the source server
    switch (src_type) {
        case FileType::SURL:
        {
            UrlCopyError::Category c;
            std::string m;
            // Creatre SRM Context
            Context::Ptr ctx = sourceContext();
            // Ping SRM
            pingSrm(ctx, c, m, "Source");
            if (c == UrlCopyError::SUCCESS) {
                server_ok = true;
            } else {
                sourceError(UrlCopyError::PHASE_PREPARATION, c, m);


            }
        }
        break;
        default:
            // Nothing to do
            server_ok = true;
            break;
    }

    return server_ok;
}

/**
 * pingDestination
 * Ping Destination server
 */
bool UrlCopyCore::pingDestination() /*throw ()*/ {

    if (!m_stat->mm_common.mm_enable_ping_dest) {
        info() << "Skipping srmPing on destination endpoint.";
        return true;
    }

    bool server_ok = false;
    // Get the source File Type
    FileType::Value dst_type = (FileType::Value)m_stat->mm_urlcopy.mm_destination.mm_type;

    // Ping the destination server
    switch (dst_type) {
        case FileType::SURL:
        {
            UrlCopyError::Category c;
            std::string m;
            // Create SRM Context
            Context::Ptr ctx = destinationContext();
            // Ping SRM
            pingSrm(ctx, c, m, "Destination");
            if (c == UrlCopyError::SUCCESS) {
                server_ok = true;
            } else {

                destinationError(UrlCopyError::PHASE_PREPARATION, c, m);

            }
        }
        break;
        default:
            // Nothing to do
            server_ok = true;
            break;
    }

    return server_ok;
}

/*
 * prepareSource
 *
 * Preapare the source file for the transfer
 */
void UrlCopyCore::prepareSource() /*throw ()*/ {
    // Get the source File Type
    FileType::Value src_type = (FileType::Value)m_stat->mm_urlcopy.mm_source.mm_type;
    bool doChecksumCheck = false;
    Context::Ptr ctx = sourceContext();
    ChecksumChecker_UrlCopy checksumChecker(UrlCopyError::PHASE_PREPARATION, this, ctx, true);
    // Set source preparation start time
    time(&(m_stat->mm_urlcopy.mm_source.mm_prep_start_time));

    msg_ifce::getInstance()->set_timestamp_checksum_source_started(&tr_completed, msg_ifce::getInstance()->getTimestamp());

    // Get the source file that should be used for the transfer
    switch (src_type) {
        case FileType::SURL:
            // Get TURL From SURL
            doChecksumCheck = prepareSourceTurl(checksumChecker);
            break;
        case FileType::SFN:
            // Convert SFN into a TURL
            translateSourceSfn();
            break;
        case FileType::URL:
        case FileType::LOCAL:
            getGridFtpFileSize_();
            break;
        case FileType::TURL:
            doChecksumCheck = getGridFtpSourceChecksum_(checksumChecker);
            getGridFtpFileSize_();
            break;
        default:
            // Unsupported source file type
            sourceError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_USER, "source file type is not supported");
            break;
    }
    if (doChecksumCheck) {
        // The TURL is ready, we can query for the missing checksum by gridFTP

        //compute timout for checksum 
        int total_timeout = compute_timeout(m_stat);

        std::string timeout_to_string = to_string<int>(total_timeout, std::dec);
        msg_ifce::getInstance()->set_checksum_timeout(&tr_completed, timeout_to_string);

        checksumChecker.getMissingChecksumsWithGridFTP(total_timeout);
        checksumChecker.compareSourceChecksums();

    }

    msg_ifce::getInstance()->set_timestamp_checksum_source_ended(&tr_completed, msg_ifce::getInstance()->getTimestamp());
    // Set source preparation done time
    time(&(m_stat->mm_urlcopy.mm_source.mm_prep_done_time));
}

void UrlCopyCore::getGridFtpFileSize_() {
    UrlCopyError::Category c;
    std::string m;
    longlong file_size = gridftp_getfilesize(m_stat->mm_urlcopy.mm_source.mm_name, c, m,
            m_stat->mm_urlcopy.mm_source.mm_prep_timeout);

    if (c != UrlCopyError::SUCCESS) {
        // Unsupported source file type

        sourceError(UrlCopyError::PHASE_PREPARATION, c, m);
    } else if (0 == file_size) {
        // empty file
        sourceError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_INVALID_SIZE,
                "empty file size returned");
    } else {
        // Set file size
        m_stat->mm_urlcopy.mm_source.mm_size = file_size;
        // Copy the source into the tx_source
        copy_url(m_stat->mm_urlcopy.mm_source.mm_tx_name, m_stat->mm_urlcopy.mm_source.mm_name);
    }
}

bool UrlCopyCore::getGridFtpSourceChecksum_(ChecksumChecker_UrlCopy& checksumChecker) {
    TransferUtils::PathInfo pinfo;
    // Empty pinfo implies that SrmLs did not give checksum. In this case, there is no
    // SrmLs, so it really cannot return checksum...
    checksumChecker.getSourceChecksumFromSrmLs(pinfo, m_stat->mm_urlcopy);
    return true;
}

/*
 * prepareDestination
 *
 * Preapare the destination file for the transfer
 */
void UrlCopyCore::prepareDestination() /*throw ()*/ {
    // Get the destination File Type
    FileType::Value dst_type = (FileType::Value)m_stat->mm_urlcopy.mm_destination.mm_type;

    // Set destination preparation start time
    time(&(m_stat->mm_urlcopy.mm_destination.mm_prep_start_time));
    // Get the destination name that should be used for the transfer
    switch (dst_type) {
        case FileType::SURL:
            // Get TURL From SURL
            prepareDestTurl();
            break;
        case FileType::SFN:
            // Convert SFN into a TURL
            translateDestSfn();
            break;
        case FileType::TURL:
	    if (0 != m_stat->mm_common.mm_overwrite)
	    	gsiftpOverwrite = true;
            copy_url(m_stat->mm_urlcopy.mm_destination.mm_tx_name, m_stat->mm_urlcopy.mm_destination.mm_name);
            break;           		
        case FileType::URL:
        case FileType::LOCAL:
	    // Copy the destination into the tx_destination
            copy_url(m_stat->mm_urlcopy.mm_destination.mm_tx_name, m_stat->mm_urlcopy.mm_destination.mm_name);
            break;
        default:
            // Unsupported destination file type
            destinationError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_USER, "destination file type is not supported");
            return;
            break;
    }
    // Set destination preparation done time
    time(&(m_stat->mm_urlcopy.mm_destination.mm_prep_done_time));
}

/*
 * sourceErrorFromSrmStatus
 *
 * Set a Source Error from the Srm Status Error
 */
void UrlCopyCore::sourceErrorFromSrmStatus(
        const SrmStatus& req_status,
        const SrmStatus& file_status,
        UrlCopyError::Category& c,
        std::string& m) {
    UrlCopyUtils::SourceErrorFromSrmStatus(req_status, file_status, c, m);
    msg_ifce::getInstance()->set_transfer_error_scope(&tr_completed, "SOURCE");
    msg_ifce::getInstance()->set_transfer_error_category(&tr_completed, UrlCopyError::categoryToString(c));
    msg_ifce::getInstance()->set_transfer_error_message(&tr_completed, m);
    msg_ifce::getInstance()->set_failure_phase(&tr_completed, "PHASE_PREPARATION");
    if (c == UrlCopyError::ERROR_ABORTED)
        msg_ifce::getInstance()->set_final_transfer_state(&tr_completed, "Aborted");    
    else
        msg_ifce::getInstance()->set_final_transfer_state(&tr_completed, "Error");        	
}

/*
 * destinationErrorFromSrmStatus
 *
 * Set a Destination Error from the Srm Status Error
 */
void UrlCopyCore::destinationErrorFromSrmStatus(
        const SrmStatus& req_status,
        const SrmStatus& file_status,
        UrlCopyError::Category& c,
        std::string& m) {
    UrlCopyUtils::DestinationErrorFromSrmStatus(req_status, file_status, c, m);
    msg_ifce::getInstance()->set_transfer_error_scope(&tr_completed, "DESTINATION");
    msg_ifce::getInstance()->set_transfer_error_category(&tr_completed, UrlCopyError::categoryToString(c));
    msg_ifce::getInstance()->set_transfer_error_message(&tr_completed, m);
    msg_ifce::getInstance()->set_failure_phase(&tr_completed, "PHASE_PREPARATION");
    if (c == UrlCopyError::ERROR_ABORTED)
        msg_ifce::getInstance()->set_final_transfer_state(&tr_completed, "Aborted");    
    else
        msg_ifce::getInstance()->set_final_transfer_state(&tr_completed, "Error");        		
}

/**
 * sourceError
 * Set a Source Error
 */
void UrlCopyCore::sourceError(UrlCopyError::Phase p, UrlCopyError::Category c, const std::string& m) throw () {
    UrlCopyUtils::SetScopeError(m_stat->mm_urlcopy.mm_source, UrlCopyError::SCOPE_SOURCE, p, c, m);

    msg_ifce::getInstance()->set_transfer_error_scope(&tr_completed, "SOURCE");
    msg_ifce::getInstance()->set_transfer_error_category(&tr_completed, UrlCopyError::categoryToString(c));
    msg_ifce::getInstance()->set_transfer_error_message(&tr_completed, m);
    msg_ifce::getInstance()->set_failure_phase(&tr_completed, UrlCopyError::phaseToString(p));
    if (c == UrlCopyError::ERROR_ABORTED)
        msg_ifce::getInstance()->set_final_transfer_state(&tr_completed, "Aborted");    
    else
        msg_ifce::getInstance()->set_final_transfer_state(&tr_completed, "Error");        
}

/**
 * transferError
 * Set a transfer Error
 */
void UrlCopyCore::transferError(UrlCopyError::Phase p, UrlCopyError::Category c, const std::string& m) throw () {
    UrlCopyUtils::SetScopeError(m_stat->mm_urlcopy.mm_tx, UrlCopyError::SCOPE_TRANSFER, p, c, m);
    msg_ifce::getInstance()->set_transfer_error_message(&tr_completed, m);
    msg_ifce::getInstance()->set_transfer_error_category(&tr_completed, UrlCopyError::categoryToString(c));
    msg_ifce::getInstance()->set_failure_phase(&tr_completed, UrlCopyError::phaseToString(p));
    if (c == UrlCopyError::ERROR_ABORTED)
        msg_ifce::getInstance()->set_final_transfer_state(&tr_completed, "Aborted");    
    else
        msg_ifce::getInstance()->set_final_transfer_state(&tr_completed, "Error");           
}

/**
 * destinationError
 * Set a Destination Error
 */
void UrlCopyCore::destinationError(UrlCopyError::Phase p, UrlCopyError::Category c, const std::string& m) throw () {
    UrlCopyUtils::SetScopeError(m_stat->mm_urlcopy.mm_destination, UrlCopyError::SCOPE_DESTINATION, p, c, m);
    msg_ifce::getInstance()->set_transfer_error_scope(&tr_completed, "DESTINATION");
    msg_ifce::getInstance()->set_transfer_error_message(&tr_completed, m);
    msg_ifce::getInstance()->set_transfer_error_category(&tr_completed, UrlCopyError::categoryToString(c));
    msg_ifce::getInstance()->set_failure_phase(&tr_completed, UrlCopyError::phaseToString(p));
    if (c == UrlCopyError::ERROR_ABORTED)
        msg_ifce::getInstance()->set_final_transfer_state(&tr_completed, "Aborted");    
    else
        msg_ifce::getInstance()->set_final_transfer_state(&tr_completed, "Error");        
}

/**
 * finalError
 * Set a request Error depending on the scope
 */
void UrlCopyCore::finalError(UrlCopyError::Scope s, UrlCopyError::Phase p, UrlCopyError::Category c, const std::string& m) {

    UrlCopyUtils::SetFinalError(m_stat->mm_urlcopy, s, p, c, m);

    msg_ifce::getInstance()->set_transfer_error_scope(&tr_completed, UrlCopyError::scopeToString(s));
    msg_ifce::getInstance()->set_transfer_error_message(&tr_completed, m);
    msg_ifce::getInstance()->set_transfer_error_category(&tr_completed, UrlCopyError::categoryToString(c));
    msg_ifce::getInstance()->set_failure_phase(&tr_completed, UrlCopyError::phaseToString(p));
    if (c == UrlCopyError::ERROR_ABORTED)
        msg_ifce::getInstance()->set_final_transfer_state(&tr_completed, "Aborted");    
    else
        msg_ifce::getInstance()->set_final_transfer_state(&tr_completed, "Error");        
}

/**
 * send a prepare to get request
 */
bool UrlCopyCore::sendPrepareToGet(
        Context::Ptr& ctx,
        boost::scoped_ptr<PrepareToGet>& request,
        UrlCopyCore::ErrorFunction errorFunction,
        UrlCopyError::Phase phase) {
    UrlCopyError::Category c;
    std::string m;
    bool ret = send_prepare_to_get(ctx, request, c, m);
    if (!ret) {
        (this->*errorFunction)(phase, c, m);
    }
    return ret;
}

/**
 * poll the status of the PTG request until it reaches a final state or the timeout elapses
 */
void UrlCopyCore::pollPrepareToGet(
        Context::Ptr& ctx,
        boost::scoped_ptr<PrepareToGet>& request,
        int timeout,
        int start_time,
        unsigned int max_sleep_time,
        UrlCopyCore::ErrorFromSrmStatusFunction errorFromSrmStatusFunction,
        UrlCopyError::Category& c,
        std::string& m) {
    if (poll_prepare_to_get(ctx, request, timeout, start_time, max_sleep_time, c, m)) {
        if (request->status.code == SrmStatus::SRM_SUCCESS && request->files[0].status.code == SrmStatus::SRM_FILE_PINNED) {
            info() << "PrepareToGet [" << request->token() << "] completed successfully";
        } else {
            error() << "PrepareToGet [" << request->token() << "] failed";
            // Decode UrlCopyError From SrmStatus Code
            (this->*errorFromSrmStatusFunction)(request->status, request->files[0].status, c, m);
        }
    }

}

/**
 * prepareSourceTurl
 * Get TURL From SURL (using PrepareToGet)
 */
bool UrlCopyCore::prepareSourceTurl(ChecksumChecker_UrlCopy& checksumChecker) {
    info() << "Getting source from SURL [" << m_stat->mm_urlcopy.mm_source.mm_name << "]";
    // Create SRM Context
    Context::Ptr ctx = sourceContext();

    // Get source Space Token
    // -------------------------------------------------------------------------
    if ((0 == strlen(m_stat->mm_common.mm_source_space_token)) && (0 != strlen(m_stat->mm_common.mm_source_token_descr))) {
        msg_ifce::getInstance()->set_srm_space_token_source(&tr_completed, m_stat->mm_common.mm_source_token_descr);

        debug() << "Resolving source space token from description '" << m_stat->mm_common.mm_source_token_descr << "'";
        UrlCopyError::Category c;
        std::string m;
        // Get space token from description
        std::string space_token = getSpaceToken(ctx, m_stat->mm_common.mm_source_token_descr, c, m, ITokenSelector::Ptr(new SelectFirstToken()));
        if (c != UrlCopyError::SUCCESS) {
            // Fail the preparation
            sourceError(UrlCopyError::PHASE_PREPARATION, c, m);
            return false;
        }
        if (!space_token.empty()) {
            // Save the token
            msg_ifce::getInstance()->set_srm_space_token_source(&tr_completed, space_token);
            copy_token(m_stat->mm_common.mm_source_space_token, space_token.c_str());
        }
    }

    // -------------------------------------------------------------------------
    // stat source file and check locality
    // -------------------------------------------------------------------------
    TransferUtils::PathInfo pinfo;
    UrlCopyError::Category ls_category;
    std::string ls_message;
    bool ls_file_exists = pathExists(ctx, m_stat->mm_urlcopy.mm_source.mm_name, ls_category, ls_message, pinfo);
    if (ls_category != UrlCopyError::SUCCESS) {
        sourceError(UrlCopyError::PHASE_PREPARATION, ls_category, ls_message);
        return false;
    }
    if (!ls_file_exists) {
        sourceError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_USER, "source file doesn't exist");
        return false;
    }

    if (pinfo.file_type != glite::data::srm::util::FileType::FILE) {
        std::stringstream err;
        err << "The SURL [" << m_stat->mm_urlcopy.mm_source.mm_name << "] cannot be copied because its type is "
                << pinfo.file_type << ", should be FILE";
        sourceError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_USER, err.str());
        return false;
    }
    // Get the checksum of the source from the srmLs info first
    checksumChecker.getSourceChecksumFromSrmLs(pinfo, m_stat->mm_urlcopy);

    if (
            ((FileLocality::NEARLINE == pinfo.file_locality) && (m_stat->mm_common.mm_fail_nearline_files)) ||
            (FileLocality::LOST == pinfo.file_locality) ||
            (FileLocality::NONE == pinfo.file_locality) ||
            (FileLocality::UNAVAILABLE == pinfo.file_locality)
            ) {
        std::stringstream err;
        err << "Source file [" << m_stat->mm_urlcopy.mm_source.mm_name << "]: locality is " << pinfo.file_locality;
        sourceError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_LOCALITY, err.str());
        return false;
    }

    // -------------------------------------------------------------------------
    // Create PrepareToGet Request
    // -------------------------------------------------------------------------

    boost::scoped_ptr<PrepareToGet> request;
    try {
        // Create PrepareToGet Request
        request.reset(PrepareToGet::create(ctx));
    } catch (const LogicError& exc) {
        // PrepareToGet method is not supported
        sourceError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_INTERNAL_ERROR, exc.what());
        return false;
    }

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = m_stat->mm_urlcopy.mm_source.mm_name;
#ifdef USE_SAFE_CLEANUP
    if (m_stat->mm_urlcopy.mm_source.mm_prep_timeout > 0) {
        // Set desired Request Time
        request->desiredTotalRequestTime =
                SAFE_CLEANUP_MULIPLY_FACTOR * m_stat->mm_urlcopy.mm_source.mm_prep_timeout;
    }
    if (m_stat->mm_global_timeout > 0) {
        // Set desired Pin Lifetime
        request->desiredPinLifeTime =
                SAFE_CLEANUP_MULIPLY_FACTOR * m_stat->mm_global_timeout;
    }
#endif //USE_SAFE_CLEANUP
    request->transferParameters.accessPattern = AccessPattern::TRANSFER_MODE;
    request->transferParameters.connectionType = m_stat->mm_common.mm_lan_connection ? ConnectionType::LAN : ConnectionType::WAN;
    request->transferParameters.protocol = PROTOCOL_GSIFTP;

    if (0 != strlen(m_stat->mm_common.mm_source_space_token)) {
        request->targetSpaceToken = m_stat->mm_common.mm_source_space_token;
    }

    glite::data::srm::util::details::IBackoffLogic::Ptr logic(request->getBackoffLogic());
    logic->setMaxSleepTime(MAX_GET_SLEEP_TIME);
    logic->reset(&m_stat->mm_urlcopy.mm_source.mm_prep_start_time);
    // -------------------------------------------------------------------------
    // Sent the request and poll until the TURL is available
    // -------------------------------------------------------------------------
    if (!sendPrepareToGet(ctx, request, &UrlCopyCore::sourceError, UrlCopyError::PHASE_PREPARATION)) {
        return false;
    }
    // Save the request token
    copy_token(m_stat->mm_urlcopy.mm_source.mm_token, request->token().c_str());
    UrlCopyError::Category c = UrlCopyError::ERROR_GENERAL_FAILURE;
    std::string m;
    int timeout = (m_stat->mm_urlcopy.mm_source.mm_prep_timeout >= 0) ? m_stat->mm_urlcopy.mm_source.mm_prep_timeout : DEFAULT_GET_TIMEOUT;
    int start_time = m_stat->mm_urlcopy.mm_source.mm_prep_start_time;
    unsigned int max_sleep_time = MAX_GET_SLEEP_TIME;
    ErrorFromSrmStatusFunction errorFromSrmStatusFunction = &UrlCopyCore::sourceErrorFromSrmStatus;
    pollPrepareToGet(ctx, request, timeout, start_time, max_sleep_time, errorFromSrmStatusFunction, c, m);

    if (UrlCopyError::SUCCESS == c) {
        // Request is completed: Save File properties
        copy_url(m_stat->mm_urlcopy.mm_source.mm_tx_name, request->files[0].turl.c_str());
        m_stat->mm_urlcopy.mm_source.mm_exists = 1;

        if (request->files[0].size > 0) {
            m_stat->mm_urlcopy.mm_source.mm_size = request->files[0].size;
        } else {
            // No size returned By PrepareToGet
            if ((true == ls_file_exists) && (pinfo.file_size > 0)) {
                m_stat->mm_urlcopy.mm_source.mm_size = pinfo.file_size;
            } else if ((true == ls_file_exists) && (pinfo.file_size == 0)) {
                std::string reason = "cannot continue since file size is 0 after PrepareToGet or SrmStat";
                error() << reason;
                sourceError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_GENERAL_FAILURE, reason);

            } else {
                std::string reason = "cannot continue since no size has been returned after PrepareToGet or SrmStat";
                error() << reason;
                sourceError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_GENERAL_FAILURE, reason);
            }
        }
    } else {
        // Request will be aborted during completion
        // Set Source Error
        sourceError(UrlCopyError::PHASE_PREPARATION, c, m);
        return false;
    }

    return true;
}

/**
 * prepareDestTurl
 * Get TURL From SURL (using PrepareToPut)
 */
void UrlCopyCore::prepareDestTurl() /*throw()*/ {

    info() << "Getting destination from SURL [" << m_stat->mm_urlcopy.mm_destination.mm_name << "]";
    // Create SRM Context
    Context::Ptr ctx = destinationContext();

    UrlCopyError::Category c;
    std::string m;

    // -------------------------------------------------------------------------
    // Get target Space Token
    // -------------------------------------------------------------------------

    // Get space token
    if ((0 == strlen(m_stat->mm_common.mm_dest_space_token)) && (0 != strlen(m_stat->mm_common.mm_dest_token_descr))) {
        msg_ifce::getInstance()->set_srm_space_token_dest(&tr_completed, m_stat->mm_common.mm_dest_token_descr);
        debug() << "Resolving destination space token from description '" << m_stat->mm_common.mm_dest_token_descr << "'";
        // Get space token from description
        std::string space_token = getSpaceToken(ctx, m_stat->mm_common.mm_dest_token_descr, c, m,
                ITokenSelector::Ptr(new SelectUnusedSpace(m_stat->mm_urlcopy.mm_source.mm_size, 1)));
        if (c != UrlCopyError::SUCCESS) {
            // Fail the preparation
            destinationError(UrlCopyError::PHASE_PREPARATION, c, m);
            return;
        }
        if (!space_token.empty()) {
            // Save the token
            msg_ifce::getInstance()->set_srm_space_token_dest(&tr_completed, space_token);
            copy_token(m_stat->mm_common.mm_dest_space_token, space_token.c_str());
        }
    }

    // -------------------------------------------------------------------------
    // Prepare target directory path
    // -------------------------------------------------------------------------

    // Create Destination Directory path
    if (m_stat->mm_common.mm_enable_dir_check) {
        try {
            bool r = prepareDestinationPath(ctx, m_stat->mm_urlcopy.mm_destination.mm_name);
            if (false == r) {
                error() << "Cannot continue preparation for destination file since the destination directory path cannot be created";
                destinationError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_INVALID_PATH, "Cannot create target directory.");
                return;
            }
        } catch (const UrlCopyError& err) {
            error() << "Destination path preparation failed: [" << err.description << "]";
            destinationError(UrlCopyError::PHASE_PREPARATION, err.category, err.description);
            return;
        }
    } else {
        info() << "Skipping target directory existence check.";
    }

    // -------------------------------------------------------------------------
    // Check if file already exists
    // -------------------------------------------------------------------------

    // Check if file already exists
    bool dest_file_exists = pathExists(ctx, m_stat->mm_urlcopy.mm_destination.mm_name, c, m);
    if (dest_file_exists) {
        m_stat->mm_urlcopy.mm_destination.mm_exists = 1;
        if (0 != m_stat->mm_common.mm_overwrite) {
            std::string reason;
            bool completed = rmFile(ctx, m_stat->mm_urlcopy.mm_destination.mm_name, c, reason);
            if (completed == false)
                warn() << "Failed to delete existing file " << m_stat->mm_urlcopy.mm_destination.mm_name << ": " << reason.c_str();
        }
    } else {
        if (c != UrlCopyError::SUCCESS) {
            warn() << "Cannot determine if destination file existed before the transfer. Error [" << UrlCopyError::categoryToString(c) << "]: " << m;
        }
        m_stat->mm_urlcopy.mm_destination.mm_exists = 0;
    }

    // -------------------------------------------------------------------------
    // Create PrepareToPut Request
    // -------------------------------------------------------------------------

    // Prepare To Put
    boost::scoped_ptr<PrepareToPut> request;
    try {
        // Create PrepareToPut Request
        request.reset(PrepareToPut::create(ctx));
    } catch (const LogicError& exc) {
        // PrepareToPut method is not supported
        destinationError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_INTERNAL_ERROR, exc.what());
        return;
    }

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = m_stat->mm_urlcopy.mm_destination.mm_name;
    request->files[0].size = m_stat->mm_urlcopy.mm_source.mm_size;
#ifdef USE_SAFE_CLEANUP
    if (m_stat->mm_urlcopy.mm_destination.mm_prep_timeout > 0) {
        // Set desired Request Time
        request->desiredTotalRequestTime =
                SAFE_CLEANUP_MULIPLY_FACTOR * m_stat->mm_urlcopy.mm_destination.mm_prep_timeout;
    }
    if (m_stat->mm_global_timeout > 0) {
        // Set desired Pin Lifetime
        request->desiredPinLifeTime =
                SAFE_CLEANUP_MULIPLY_FACTOR * m_stat->mm_global_timeout;
    }
#endif //USE_SAFE_CLEANUP

    // bug 32007 : desiredTargetSURLLifeTime and targetFileStorageType were
    // specified only if no space token provided; they must be specified
    // always

    // Set desired lifetime and Storage Type
    request->desiredFileLifeTime = -1;
    request->desiredFileStorageType = FileStorageType::PERMANENT;

    if (0 != strlen(m_stat->mm_common.mm_dest_space_token)) {
        request->targetSpaceToken = m_stat->mm_common.mm_dest_space_token;
    }
    request->overwriteOption = (0 != m_stat->mm_common.mm_overwrite) ? OverwriteMode::ALWAYS : OverwriteMode::DEFAULT;
    request->transferParameters.accessPattern = AccessPattern::TRANSFER_MODE;
    request->transferParameters.connectionType = m_stat->mm_common.mm_lan_connection ? ConnectionType::LAN : ConnectionType::WAN;
    request->transferParameters.protocol = PROTOCOL_GSIFTP;

    if (m_stat->mm_common.mm_copy_pin_lifetime > 0) {
        debug() << "Setting copy pin lifetime in prepareToPut request.";
        request->copyPinLifetime = m_stat->mm_common.mm_copy_pin_lifetime;
    }

    // -------------------------------------------------------------------------
    // Send PrepareToPut Request
    // -------------------------------------------------------------------------
    try {
        RefreshTimeUpdater updater(this);
        request->execute();
        info() << "PrepareToPut [" << request->token() << "] started";
        log_request(*request);
    } catch (const HttpTimeoutError& exc) { // http timeout handled
        error() << "HTTP timeout contacting remote SRM [" << ctx->endpoint << "]: " << exc.what();
        destinationError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_HTTP_TIMEOUT, exc.what());
        return;
    } catch (const RemoteError& exc) { // http timeout handled
        error() << "Failed to contact remote SRM [" << ctx->endpoint << "]: " << exc.what();
        destinationError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_CONNECTION, exc.what());
        return;
    } catch (const SecurityError& exc) {
        error() << "User is not authorized to contact the remote SRM [" << ctx->endpoint << "]: " << exc.what();
        destinationError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_PERMISSION, exc.what());
        return;
    } catch (const SrmError& exc) {
        error() << "Remote method invocation failure on the remote SRM [" << ctx->endpoint << "]: " << exc.what();
        destinationError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_REQUEST_FAILURE, exc.what());
        return;
    } catch (const SrmBadResponse& exc) {
        error() << "Invalid result returned by the remote SRM [" << ctx->endpoint << "]: " << exc.what();
        destinationError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_REQUEST_FAILURE, exc.what());
        return;
    } catch (const LogicError& exc) {
        error() << "Logic Error in invoking PrepareToPut on remote SRM [" << ctx->endpoint << "]: " << exc.what();
        destinationError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_INTERNAL_ERROR, exc.what());
        return;
    }

    // Save Request Token
    copy_token(m_stat->mm_urlcopy.mm_destination.mm_token, request->token().c_str());

    // -------------------------------------------------------------------------
    // Poll until the TURL is available
    // -------------------------------------------------------------------------
    c = UrlCopyError::ERROR_GENERAL_FAILURE;
    m.clear();
    try {
        // Timeout
        int timeout = (m_stat->mm_urlcopy.mm_destination.mm_prep_timeout >= 0) ?
                m_stat->mm_urlcopy.mm_destination.mm_prep_timeout : DEFAULT_PUT_TIMEOUT;
        bool isFinished = false;
        glite::data::srm::util::details::IBackoffLogic::Ptr logic(request->getBackoffLogic());
        assert(logic.get());
        logic->setMaxSleepTime(MAX_PUT_SLEEP_TIME);
        logic->setTimeout(timeout);
        logic->reset(&m_stat->mm_urlcopy.mm_destination.mm_prep_start_time);

        while (!isFinished && !TransferState::instance().interrupted()) {
            try {
                isFinished = request->backoff();
                RefreshTimeUpdater updater(this);

                if (!isFinished) {
                    info() << "Status of PrepareToPut [" << request->token() << "] updated";
                    log_request(*request);
                }
            } catch (const RemoteError& exc) { // don't handle http timeout explicitly here
                warn() << "Failed to contact remote SRM [" << ctx->endpoint << "] in updating the PrepareToPut status: " << exc.what();
                info() << "This error is considered transient and the action will be retried";
            } catch (const SrmError& exc) {
                warn() << "Remote method invocation failure on the remote SRM [" << ctx->endpoint << "] in updating the PrepareToPut status: %s" << exc.what();
                info() << "This error is considered transient and the action will be retried";
            }
        } // end while
        if ((request->status.code == SrmStatus::SRM_SUCCESS) &&
                (request->files[0].status.code == SrmStatus::SRM_SPACE_AVAILABLE)) {
            info() << "PrepareToPut [" << request->token() << "] completed";
            c = UrlCopyError::SUCCESS;
        } else if (true == TransferState::instance().interrupted()) {
            error() << "PrepareToPut [" << request->token() << "] should be aborted";
            c = UrlCopyError::ERROR_ABORTED;
            m = "transfer request canceled";
        } else if (c == UrlCopyError::ERROR_REQUEST_TIMEOUT) {
            error() << "PrepareToPut [" << request->token() << "] timed out";
        } else {
            error() << "PrepareToPut [" << request->token() << "] failed";
            // Decode UrlCopyError From SrmStatus Code
            destinationErrorFromSrmStatus(request->status, request->files[0].status, c, m);
        }
    } catch (const SecurityError& exc) {
        error() << "User is not authorized to contact the remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_PERMISSION;
        m = exc.what();
    } catch (const SrmBadResponse& exc) {
        error() << "Invalid result returned by the remote SRM [" << ctx->endpoint << "]: " << exc.what();
        c = UrlCopyError::ERROR_REQUEST_FAILURE;
        m = exc.what();
    } catch (const LogicError& exc) {
        error() << "Logic Error in invoking PrepareToPut: " << exc.what();
        c = UrlCopyError::ERROR_INTERNAL_ERROR;
        m = exc.what();
    }
    if (UrlCopyError::SUCCESS == c) {
        // Request is completed: Save File properties
        copy_url(m_stat->mm_urlcopy.mm_destination.mm_tx_name, request->files[0].turl.c_str());
        m_stat->mm_urlcopy.mm_destination.mm_size = request->files[0].size;
    } else {
        // Request will be aborted during completetion

        // Set Destination Error
        destinationError(UrlCopyError::PHASE_PREPARATION, c, m);
    }
}

/*
 * translateSourceSfn
 *
 * Translate the Source SFN into a TURL
 */
void UrlCopyCore::translateSourceSfn() /*throw()*/ {
    // TODO translateSourceSfn
    sourceError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_INTERNAL_ERROR, "SFN not yet supported");
}

/*
 * translateDestSfn
 *
 * Translate the Destination SFN into a TURL
 */
void UrlCopyCore::translateDestSfn() /*throw()*/ {
    // TODO translateDestSfn
    destinationError(UrlCopyError::PHASE_PREPARATION, UrlCopyError::ERROR_INTERNAL_ERROR, "SFN not yet supported");
}

//------------------------------------------------------------------------------
//
// Transfer Completion
//
//------------------------------------------------------------------------------

/*
 * complete
 *
 * Complete the transfer
 */
void UrlCopyCore::complete() /*throw (RuntimeError, LogicError)*/ {


    msg_ifce::getInstance()->set_time_spent_in_srm_finalization_start(&tr_completed, msg_ifce::getInstance()->getTimestamp());


    // Check of the instance is initialized, otherwise initialize it from the
    // status file
    initialize();

    // Depending on the request status, perform the correct cleanup precedure
    // -------------------------------------------------------------------------
    switch (m_stat->mm_urlcopy.mm_status) {
        case UrlCopyStatus::COMPLETED:
            // The request is already completed (either successfully or not):
            // nothing to do
            break;
        case UrlCopyStatus::RUNNING:
            // The transfer is currently running: wait that the transfer is
            // finished or abort it
            throw LogicError("transfer request is running. Please wait that "\
                             "the transfer finishes or abort it before calling "\
                             "complete");
            break;
        case UrlCopyStatus::IDLE:
        case UrlCopyStatus::PREPARING:
            // The request is idle or preparing and maybe the source request
            // has been executed: in that case abort it
        case UrlCopyStatus::READY:
            // The request is ready, the file has not been transferred:
            // abort the SRM requests
        case UrlCopyStatus::FAILED:
        case UrlCopyStatus::ABORTED:
            // The transfer has been canceled, abort the SRM requests
        case UrlCopyStatus::DONE:
            // The transfer is completed, complete the SRM requests
        default:
            // Release (or Abort) source and destination

            completeSource();

            completeDestination();

            break;
    }// end switch

    if (m_stat->mm_urlcopy.mm_error.mm_category == UrlCopyError::SUCCESS) {
        // Transfer Complete Successfully
        info() << "FINAL:SUCCESS";
        msg_ifce::getInstance()->set_final_transfer_state(&tr_completed, "Ok");  
    } else {
        // Transfer Failed
        info() << "FINAL:fail  [" <<
                UrlCopyError::scopeToString((UrlCopyError::Scope)m_stat->mm_urlcopy.mm_error.mm_scope) << "] - [" <<
                UrlCopyError::phaseToString((UrlCopyError::Phase)m_stat->mm_urlcopy.mm_error.mm_phase) << "] - [" <<
                UrlCopyError::categoryToString((UrlCopyError::Category)m_stat->mm_urlcopy.mm_error.mm_category) << "] : '" <<
                m_stat->mm_urlcopy.mm_error.mm_message << "'";

        msg_ifce::getInstance()->set_transfer_error_scope(&tr_completed, UrlCopyError::scopeToString((UrlCopyError::Scope)m_stat->mm_urlcopy.mm_error.mm_scope));
        msg_ifce::getInstance()->set_transfer_error_category(&tr_completed, UrlCopyError::categoryToString((UrlCopyError::Category)m_stat->mm_urlcopy.mm_error.mm_category));
        msg_ifce::getInstance()->set_transfer_error_message(&tr_completed, m_stat->mm_urlcopy.mm_error.mm_message);
        msg_ifce::getInstance()->set_failure_phase(&tr_completed, UrlCopyError::phaseToString((UrlCopyError::Phase)m_stat->mm_urlcopy.mm_error.mm_phase));
        if (m_stat->mm_urlcopy.mm_error.mm_category == UrlCopyError::ERROR_ABORTED)
            msg_ifce::getInstance()->set_final_transfer_state(&tr_completed, "Aborted");    
        else
            msg_ifce::getInstance()->set_final_transfer_state(&tr_completed, "Error");        	    


    }
    // Now the request has been completed. In case of errors, these are
    // reported on the file attributes (source,destination,tx)
    m_stat->mm_urlcopy.mm_status = UrlCopyStatus::COMPLETED;

    // Set request finish time
    time(&(m_stat->mm_common.mm_finish_time));


    msg_ifce::getInstance()->set_time_spent_in_srm_finalization_end(&tr_completed, msg_ifce::getInstance()->getTimestamp());
}

/*
 * completeSource
 *
 * Complete the Source request
 */
void UrlCopyCore::completeSource() /*throw ()*/ {
    // Get the source File Type
    FileType::Value src_type = (FileType::Value)m_stat->mm_urlcopy.mm_source.mm_type;

    // Set source finalization start time
    time(&(m_stat->mm_urlcopy.mm_source.mm_final_start_time));

    info() << "STATUS:BEGIN:SOURCE - Finalization";

    // Complete the source file that should be used for the transfer
    switch (src_type) {
        case FileType::SURL:
            // Always Release TURL
            releaseSourceTurl();
            break;
        case FileType::SFN:
        case FileType::TURL:
        case FileType::URL:
        case FileType::LOCAL:
            // Nothing to do
            break;
        default:
            // Unsupported source file type
            sourceError(UrlCopyError::PHASE_FINALIZATION, UrlCopyError::ERROR_CONSISTENCY,
                    "source file type is not supported. Invalid stat file?");
            break;
    }

    if (m_stat->mm_urlcopy.mm_source.mm_error.mm_phase != UrlCopyError::PHASE_FINALIZATION) {
        info() << "STATUS:END:SOURCE - Finalization";
    } else {
        info() << "STATUS:END fail:SOURCE - Finalization";
    }

    // Set source finalization done time
    time(&(m_stat->mm_urlcopy.mm_source.mm_final_done_time));
}

/**
 * completeDestination
 * Complete the Destination request
 */
void UrlCopyCore::completeDestination() /*throw ()*/ {

    // Get the destination File Type
    FileType::Value dst_type = (FileType::Value)m_stat->mm_urlcopy.mm_destination.mm_type;

    // Set destination finalization start time
    time(&(m_stat->mm_urlcopy.mm_destination.mm_final_start_time));

    info() << "STATUS:BEGIN:DESTINATION - Finalization";
    Context::Ptr ctx = destinationContext();

    // Complete the destination file that should be used for the transfer
    switch (dst_type) {
        case FileType::SURL:
            // Check File Status
            switch (m_stat->mm_urlcopy.mm_status) {
                case UrlCopyStatus::DONE:
                    // Complete TURL
                    completeDestinationTurl();
                    break;
                default:
                    // Abort Destination TURL
                    abortDestinationTurl();
                    break;
            }
            break;
        case FileType::SFN:
        case FileType::TURL: /*do not perform checksum if the source is LOCAL or URL*/
	    if(FileType::LOCAL == (FileType::Value) m_stat->mm_urlcopy.mm_source.mm_type ||
		    FileType::URL == (FileType::Value) m_stat->mm_urlcopy.mm_source.mm_type)
               break;
	     do_globus_deactivates();
	    _verifyDestinationChecksum(ctx);
            break;
        case FileType::URL:
        case FileType::LOCAL:
            // Nothing to do
            break;
        default:
            // Unsupported source file type
            destinationError(UrlCopyError::PHASE_FINALIZATION, UrlCopyError::ERROR_CONSISTENCY,
                    "destination file type is not supported. Invalid stat file?");
            break;
    }
    if (m_stat->mm_urlcopy.mm_destination.mm_error.mm_phase != UrlCopyError::PHASE_FINALIZATION) {
        info() << "STATUS:END:DESTINATION - Finalization";
    } else {
        info() << "STATUS:END fail:DESTINATION - Finalization";
    }

    // Check if the destination failed and the overall error is not set
    if ((m_stat->mm_urlcopy.mm_error.mm_category == UrlCopyError::SUCCESS) &&
            (m_stat->mm_urlcopy.mm_destination.mm_error.mm_category != UrlCopyError::SUCCESS)) {
        // Set final Error
        finalError(UrlCopyError::SCOPE_DESTINATION); // OK
    }
    // Set destination finalization done time
    time(&(m_stat->mm_urlcopy.mm_destination.mm_final_done_time));
}

/**
 * complete a PTG request
 */
void UrlCopyCore::complete_prepare_to_get(
        glite::data::srm::util::Context::Ptr& ctx,
        boost::scoped_ptr<glite::data::srm::util::PrepareToGet>& request,
        int timeout,
        UrlCopyCore::ErrorFunction errorFunction,
        UrlCopyError::Phase phase,
        UrlCopyError::Category& c,
        std::string& m) {

    // Call Complete Request
    bool completed = m_srmMethodCaller.completeRequest(ctx, *request, c, m, timeout);
    if (true == completed) {
        if ((request->status.code == SrmStatus::SRM_SUCCESS) && (request->files[0].status.code == SrmStatus::SRM_RELEASED)) {
            // Request released
            info() << "PrepareToGet [" << request->token() << "] successfully released";
        } else {
            completed = false;
            warn() << "ReleaseFiles for [" << request->token() << "] failed";
        }
    } else {
        if (errorFunction) {
            (this->*errorFunction)(phase, c, m.c_str());
        }
    }
    if (false == completed) {
        warn() << "failed to release PrepareToGet [" << request->token() << "]. Try to abort it";
        SrmStatus release_failure = request->status;
        // Abort the request
        if (false == m_srmMethodCaller.abortRequest(ctx, *request, c, m, m_stat->mm_urlcopy.mm_source.mm_final_timeout)) {
            // Failed To Abort Request
            c = UrlCopyError::ERROR_GENERAL_FAILURE;
            std::stringstream ss;
            ss << "Failed to release PrepareToGet request [" << request->token()
                    << "] on remote SRM [" << ctx->endpoint
                    << "]: [" << release_failure.code << "] "
                    << release_failure.message
                    << ". Attempting to abort the request also failed with: " << m;
            m = ss.str();
            error() << m;
            if (errorFunction) {
                (this->*errorFunction)(phase, c, m.c_str());
            }
        } else {
            // Request Aborted
            info() << "PrepareToGet request [" << request->token() << "] aborted";
        }
    }
}

/**
 * releaseSourceTurl
 * Release the turl reserved with the PrepareToGet request
 */
void UrlCopyCore::releaseSourceTurl() /*throw ()*/ {
    // Precondition check
    if (0 == strlen(m_stat->mm_urlcopy.mm_source.mm_token)) {
        info() << "No request token provided for source file. Assuming PrepareToGet request has not been sent";
        return;
    }

    info() << "Releasing PrepareToGet [" << m_stat->mm_urlcopy.mm_source.mm_token << "] for SURL [" << m_stat->mm_urlcopy.mm_source.mm_name << "]";
    // Create SRM Context
    Context::Ptr ctx = sourceContext();

    // -------------------------------------------------------------------------
    // Create PrepareToGet Request
    // -------------------------------------------------------------------------
    boost::scoped_ptr<PrepareToGet> request;
    try {
        // Create PrepareToPut Request
        request.reset(PrepareToGet::create(ctx, m_stat->mm_urlcopy.mm_source.mm_token));
    } catch (const LogicError& exc) {
        // PrepareToGet method is not supported
        sourceError(UrlCopyError::PHASE_FINALIZATION, UrlCopyError::ERROR_INTERNAL_ERROR, exc.what());
        return;
    }

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = m_stat->mm_urlcopy.mm_source.mm_name;

    UrlCopyError::Category c;
    std::string m;
    int timeout = m_stat->mm_urlcopy.mm_source.mm_final_timeout;
    UrlCopyCore::ErrorFunction errorFunction = &UrlCopyCore::sourceError;
    UrlCopyError::Phase phase = UrlCopyError::PHASE_FINALIZATION;
    complete_prepare_to_get(ctx, request, timeout, errorFunction, phase, c, m);
}

/**
 * completeDestinationTurl
 * Call PutDone on the turl reserved with the PrepareToPut request
 */
void UrlCopyCore::completeDestinationTurl() /*throw ()*/ {

    do_globus_deactivates();
    // Precondition check
    if (0 == strlen(m_stat->mm_urlcopy.mm_destination.mm_token)) {
        info() << "No request token provided for destination file. Assuming PrepareToPut request has not been sent";
        return;
    }

    info() << "completing PutDone [" << m_stat->mm_urlcopy.mm_destination.mm_token << "] for SURL [" << m_stat->mm_urlcopy.mm_destination.mm_name << "]";
    // Create SRM Context
    Context::Ptr ctx = destinationContext();



    // -------------------------------------------------------------------------
    // Create PrepareToPut Request
    // -------------------------------------------------------------------------
    boost::scoped_ptr<PrepareToPut> request;
    try {
        // Create PrepareToPut Request
        request.reset(PrepareToPut::create(ctx, m_stat->mm_urlcopy.mm_destination.mm_token));
    } catch (const LogicError& exc) {
        // PrepareToPut method is not supported
        destinationError(UrlCopyError::PHASE_FINALIZATION, UrlCopyError::ERROR_INTERNAL_ERROR, exc.what());
        return;
    }

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = m_stat->mm_urlcopy.mm_destination.mm_name;
    
    // Call Complete Request
    UrlCopyError::Category c;
    std::string m;
    bool completed = m_srmMethodCaller.completeRequest(ctx, *request, c, m, m_stat->mm_urlcopy.mm_destination.mm_final_timeout);

    if (true == completed) {
        if ((request->status.code == SrmStatus::SRM_SUCCESS) &&
                (request->files[0].status.code == SrmStatus::SRM_SUCCESS)) {
            // Request released
            info() << "PrepareToPut [" << request->token() << "] successfully released";
        } else {
            completed = false;
        }
    } else {
        destinationError(UrlCopyError::PHASE_FINALIZATION, c, m.c_str());
    }

    // Check if request has been released
    if (false == completed) {
        warn() << "Failed to complete PutDone [" << request->token() << "]. Try to abort it";

        SrmStatus status;
        if (SrmStatus::SRM_SUCCESS == request->status.code) {
            warn() << "Request status is " << request->status.code << "; using file status for error message.";
            status = request->files[0].status;
        } else {
            status = request->status;
        }

        // Abort the request
        std::stringstream ss;
        ss << "Failed to complete PutDone request [" << request->token() << "] on remote SRM [" << ctx->endpoint << "]: " << status << ".";
        UrlCopyError::Category c;
        std::string reason;
        if (false == m_srmMethodCaller.abortRequest(ctx, *request, c, reason, m_stat->mm_urlcopy.mm_destination.mm_final_timeout)) {
            // Failed To Abort Request
            ss << "Attempting to abort the request also failed with: " << reason;
        } else {
            ss << "The PutDone request has been successfully aborted";
        }
        reason = ss.str();
        error() << reason;
        destinationError(UrlCopyError::PHASE_FINALIZATION, UrlCopyError::ERROR_GENERAL_FAILURE, reason.c_str());
        // if destination file didn't exist before the transfer, delete it
        if (0 == m_stat->mm_urlcopy.mm_destination.mm_exists) {
            // Remove destination file
            rmFile(ctx, m_stat->mm_urlcopy.mm_destination.mm_name, c, reason);
        }
    } else {
        // prepare to put request has been completed successfully; check pin
        // lifetime and issue an srmBringOnline (addendum to the WLCG usage
        // agreement)
        if (m_stat->mm_common.mm_copy_pin_lifetime > 0) {
            if (FileType::SURL == m_stat->mm_urlcopy.mm_destination.mm_type) {
                debug() << "Calling srmBringOnline to set copy pin lifetime.";
                bringOnline(ctx, m_stat->mm_urlcopy.mm_destination.mm_name, m_stat->mm_common.mm_copy_pin_lifetime);
            } else {
                warn() << "A copy pin lifetime was specified, but the file type is " << (FileType::Value)m_stat->mm_urlcopy.mm_destination.mm_type << ". Skipping request.";
            }
        }

    /*do not perform checksum if the source is LOCAL or URL type*/ 
     if(FileType::LOCAL == (FileType::Value) m_stat->mm_urlcopy.mm_source.mm_type ||
		    FileType::URL == (FileType::Value) m_stat->mm_urlcopy.mm_source.mm_type)
		    return;

		    
        _verifyDestinationChecksum(ctx);

    }
}

void UrlCopyCore::_verifyDestinationChecksum(Context::Ptr& ctx) {

    if (m_stat->mm_urlcopy.mm_error.mm_category != UrlCopyError::SUCCESS) {
        INFO_LOG(UrlCopyCore) << "Skipping destination checksum manipulation because the transfer"
                " is failed.";
        return;
    }

    msg_ifce::getInstance()->set_timestamp_checksum_dest_started(&tr_completed, msg_ifce::getInstance()->getTimestamp());
    do_globus_activates();

    //compute timout for checksum 
    int total_timeout = compute_timeout(m_stat);

    //ensure that the transfer is completed by checking the percent_done(to be 100%), and if the
    //transfered bytes, at the time checksum is requested, is equal to the size of the file
    //after this check, reset the timeout and set it to a computed value.
    if ((m_stat->mm_urlcopy.mm_tx.mm_transferred_bytes == m_stat->mm_urlcopy.mm_tx.mm_size) &&
            (m_stat->mm_urlcopy.mm_tx.mm_percent_done * 100) == 100) {
        info() << "The transfer has reached 100%, checksum timeout is set to " << total_timeout << " secs";
        m_stat->mm_urlcopy.mm_tx.mm_no_progress_timeout = 0;
    }
  
    ChecksumChecker_UrlCopy checksumChecker(UrlCopyError::PHASE_FINALIZATION, this, ctx, false);

    checksumChecker.getDestinationChecksumFromSrmLs(m_stat->mm_urlcopy);

#ifdef USE_SAFE_CLEANUP
    int global_timeout = m_stat->mm_common.mm_global_timeout;
#else
    int global_timeout = 0;
#endif

    bool res = checksumChecker.sendPrepareToGetForMissing(
            m_stat->mm_urlcopy.mm_destination.mm_final_timeout,
            global_timeout,
            m_stat->mm_urlcopy.mm_destination.mm_final_start_time,
            m_stat->mm_common.mm_dest_space_token);


    if (res) {
        checksumChecker.getMissingChecksumsWithGridFTP(total_timeout);
    }

    checksumChecker.removeFailedDestinationFiles(m_stat->mm_urlcopy.mm_destination.mm_final_timeout);

    checksumChecker.compareDestinationChecksums();

    checksumChecker.removeFailedDestinationFiles(m_stat->mm_urlcopy.mm_destination.mm_final_timeout);

    msg_ifce::getInstance()->set_timestamp_checksum_dest_ended(&tr_completed, msg_ifce::getInstance()->getTimestamp());

    do_globus_deactivates();

}

/*
 * abortDestinationTurl
 *
 * Call AbortRequest on the PrepareToPut request
 */
void UrlCopyCore::abortDestinationTurl() /*throw ()*/ {
    // Precondition check
    if (0 == strlen(m_stat->mm_urlcopy.mm_destination.mm_token)) {
        info() << "No request token provided for destination file. Assuming PrepareToPut request has not been sent";
        return;
    }

    info() << "aborting PrepareToPut [" << m_stat->mm_urlcopy.mm_destination.mm_token << "] for SURL [" << m_stat->mm_urlcopy.mm_destination.mm_name << "]";
    // Create SRM Context
    Context::Ptr ctx = destinationContext();

    // -------------------------------------------------------------------------
    // Create PrepareToPut Request
    // -------------------------------------------------------------------------
    boost::scoped_ptr<PrepareToPut> request;
    try {
        // Create PrepareToPut Request
        request.reset(PrepareToPut::create(ctx, m_stat->mm_urlcopy.mm_destination.mm_token));
    } catch (const LogicError& exc) {
        // PrepareToPut method is not supported
        destinationError(UrlCopyError::PHASE_FINALIZATION, UrlCopyError::ERROR_INTERNAL_ERROR, exc.what());
        return;
    }

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = m_stat->mm_urlcopy.mm_destination.mm_name;

    // -------------------------------------------------------------------------
    // Send Abort Request
    // -------------------------------------------------------------------------
    UrlCopyError::Category c;
    std::string reason;
    if (true == m_srmMethodCaller.abortRequest(ctx, *request, c, reason, m_stat->mm_urlcopy.mm_destination.mm_final_timeout)) {
        // Request aborted. Don't set the error message
        info() << "PrepareToPut [" << request->token() << "] successfully aborted";
        reason = "Request aborted";
    }
    if (m_stat->mm_urlcopy.mm_destination.mm_error.mm_category == UrlCopyError::SUCCESS) {
        destinationError(UrlCopyError::PHASE_FINALIZATION, UrlCopyError::ERROR_GENERAL_FAILURE, reason.c_str());
    } else {
        // Don't set the destinationError, otherwise the reason of the put
        // failure may be lost
    }

    // if destination file didn't exist before the transfer, delete it
    if (0 == m_stat->mm_urlcopy.mm_destination.mm_exists) {
        // Remove destination file
        rmFile(ctx, m_stat->mm_urlcopy.mm_destination.mm_name, c, reason);
    }
}

//------------------------------------------------------------------------------
//
// Transfer Methods
//
//------------------------------------------------------------------------------

/*
 * transfer
 *
 * Start the transfer
 */
void UrlCopyCore::transfer() /*throw (RuntimeError, LogicError)*/ {
    // Check of the instance is initialized, otherwise initialize it from the
    // status file
    initialize();

    // Precondition Check
    if (m_stat->mm_urlcopy.mm_status != UrlCopyStatus::READY) {
        // Cannot Call transfer on a request that is not ready
        std::stringstream ss;
        ss << "transfer called for a request not in ["
                << UrlCopyStatus::toString(UrlCopyStatus::READY)
                << "] state. Current state is ["
                << UrlCopyStatus::toString(
                (UrlCopyStatus::Value)m_stat->mm_urlcopy.mm_status) << "]";
        std::string reason = ss.str();
        error() << reason;
        throw LogicError(reason.c_str());
    }



    // Check that the source and destination transfer names are set
    if (strlen(m_stat->mm_urlcopy.mm_source.mm_tx_name) == 0) {
        // Set Source Error
        sourceError(UrlCopyError::PHASE_TRANSFER, UrlCopyError::ERROR_STORAGE_INTERNAL_ERROR, "empty source file transfer name");
        finalError(UrlCopyError::SCOPE_SOURCE); // OK
        return;
    }
    if (strlen(m_stat->mm_urlcopy.mm_destination.mm_tx_name) == 0) {
        // Set Destination Error
        destinationError(UrlCopyError::PHASE_TRANSFER, UrlCopyError::ERROR_STORAGE_INTERNAL_ERROR, "empty destination file transfer name");
        finalError(UrlCopyError::SCOPE_DESTINATION); // OK
        return;
    }

    // Set Transfer start time
    time(&(m_stat->mm_urlcopy.mm_tx.mm_start_time));

    // Set the Transfer Status To Running
    m_stat->mm_urlcopy.mm_status = UrlCopyStatus::RUNNING;

    UrlCopyError::Category c;
    std::string m;

    // TBM? set scope to source or dest here
    bool set_error_info = true;

    try {
      // do_globus_activates();
        // Verify source file size
        debug() << "verify source file size";

        checkSourceFileSize(c, m);
        if (c != UrlCopyError::SUCCESS) {
            set_error_info = false;
            sourceError(UrlCopyError::PHASE_TRANSFER, c, m);
            finalError(UrlCopyError::SCOPE_SOURCE); // OK
            throw RuntimeError("source file verification failed");
        }

        info() << "source verification completed";
        // Do the Transfer
        info() << "STATUS:BEGIN:TRANSFER";
        info() << "starting gsiftp transfer";
        UrlCopyError::Scope s = UrlCopyError::SCOPE_TRANSFER;


        copyFile(c, s, m);
        if (c != UrlCopyError::SUCCESS) {
            // GridFtpPatterns might have set the scope to source or destination
            info() << "STATUS:END fail:TRANSFER - " << m;
            set_error_info = false;
            // bug 38561
            switch (s) {
                case UrlCopyError::SCOPE_SOURCE:
                    sourceError(UrlCopyError::PHASE_TRANSFER, c, m);
                    finalError(UrlCopyError::SCOPE_SOURCE); // OK
                    break;
                case UrlCopyError::SCOPE_DESTINATION:
                    destinationError(UrlCopyError::PHASE_TRANSFER, c, m);
                    finalError(UrlCopyError::SCOPE_DESTINATION); // OK
                    break;
                default:
                    transferError(UrlCopyError::PHASE_TRANSFER, c, m);
                    finalError(UrlCopyError::SCOPE_TRANSFER); // OK
                    break;
            }
            throw RuntimeError("gsiftp transfer failed");
        }
        info() << "STATUS:END:Transfer";
        info() << "gsiftp transfer completed successfully";

        // Verify if destination file size if correct
        debug() << "verify destination file size";
        checkDestinationFileSize(c, m);
        if (c != UrlCopyError::SUCCESS) {
            set_error_info = false;
            destinationError(UrlCopyError::PHASE_TRANSFER, c, m);
            finalError(UrlCopyError::SCOPE_DESTINATION); // OK
            throw RuntimeError("destination file verification failed");
        }
        info() << "destination verification completed";

        // The Transfer is completed
        info() << "transfer request done";
        m_stat->mm_urlcopy.mm_status = UrlCopyStatus::DONE;
    } catch (const RuntimeError& exc) {
        warn() << exc.what();
        if (set_error_info) {
            // Set Error
            transferError(UrlCopyError::PHASE_TRANSFER, c, m);
        }
        // Set final error
        finalError(UrlCopyError::SCOPE_TRANSFER);
        // Set request status
        if (c == UrlCopyError::ERROR_ABORTED) {
            // The Transfer is aborted
            error() << "transfer aborted";
            m_stat->mm_urlcopy.mm_status = UrlCopyStatus::ABORTED;
        } else {
            // The Transfer is failed
            error() << "transfer failed";
            m_stat->mm_urlcopy.mm_status = UrlCopyStatus::FAILED;
        }
    }

    // Set Transfer done time
    time(&(m_stat->mm_urlcopy.mm_tx.mm_done_time));

    // wait
    if (UrlCopyStatus::DONE == m_stat->mm_urlcopy.mm_status) {
        if (m_stat->mm_common.mm_block_after_transfer) {
            waitContinueSignal();
        }
    }
}

/*
 * checkSourceFileSize
 *
 * Check if the source file size of the same as the one returned by the SRM
 */
void UrlCopyCore::checkSourceFileSize(
        UrlCopyError::Category& c,
        std::string& m) /*throw ()*/ {

    c = UrlCopyError::ERROR_INVALID_SIZE;

    longlong file_size = -1;
    // Get the source File Type
    FileType::Value src_type = (FileType::Value)m_stat->mm_urlcopy.mm_source.mm_type;
    switch (src_type) {
        case FileType::SURL:
            // Get GridFtp file size
            file_size = gridftp_getfilesize(m_stat->mm_urlcopy.mm_source.mm_tx_name, c, m, m_stat->mm_urlcopy.mm_tx.mm_timeout);
            info() << "gridftp_getfilesize returned " << file_size;
            if (c != UrlCopyError::SUCCESS) {
                if (true == m.empty()) {
                    m = "failed to get file size from gridftp server";
                }
            } else if (0 == file_size) {
                m = "empty file size returned";
            } else if ((file_size != m_stat->mm_urlcopy.mm_source.mm_size)) {
                c = UrlCopyError::ERROR_INVALID_SIZE;
                m = "source file size differs from the one returned by the SRM";
            } else {
                c = UrlCopyError::SUCCESS;
            }
            break;
        case FileType::TURL:
        case FileType::URL:
        case FileType::LOCAL:
        case FileType::SFN:
            // Nothing to do
            file_size = m_stat->mm_urlcopy.mm_source.mm_size;
            c = UrlCopyError::SUCCESS;
            break;
        default:
            // Unsupported source file type
            c = UrlCopyError::ERROR_CONSISTENCY;
            m = "source file type is not supported. Invalid stat file?";
            break;
    }
    // Set transfer file size
    if (file_size > 0) {
        m_stat->mm_urlcopy.mm_tx.mm_size = file_size;

        //conver longlong to string
        std::string size_to_string = to_string<longlong > (file_size, std::dec);
        //set the value of file size to the message
        msg_ifce::getInstance()->set_file_size(&tr_completed, size_to_string.c_str());
    }
}

/*
 * copyFile
 *
 * Copy the source file into the destination
 */



void UrlCopyCore::copyFile(
        UrlCopyError::Category& c,
        UrlCopyError::Scope& s,
        std::string& m) /*throw ()*/ {
    c = UrlCopyError::ERROR_GENERAL_FAILURE;

    debug() << "calling gridftp_copy";

    time(&(m_stat->mm_urlcopy.mm_tx.mm_gridftp_copy_start_time));

    int total_timeout = compute_timeout(m_stat);
    info() << "Transfer timeout = " << total_timeout;
        
    msg_ifce::getInstance()->set_timestamp_transfer_started(&tr_completed, msg_ifce::getInstance()->getTimestamp());

    // Copy the file
    gridftp_copy(
            m_stat->mm_urlcopy.mm_source.mm_tx_name,
            m_stat->mm_urlcopy.mm_destination.mm_tx_name,
            c, s, m,
            m_stat->mm_urlcopy.mm_tx.mm_streams,
            m_stat->mm_urlcopy.mm_tx.mm_tcp_bs,
            m_stat->mm_urlcopy.mm_tx.mm_block_size,
            total_timeout,
            tx_callback, m_stat, this, gsiftpOverwrite);


    msg_ifce::getInstance()->set_timestamp_transfer_completed(&tr_completed, msg_ifce::getInstance()->getTimestamp());

    if (c != UrlCopyError::SUCCESS) {
        warn() << "gridftp_copy failed: [" << UrlCopyError::categoryToString(c) << "] " << m;
    } else {
        // Copy succeeded: be sure the transferred bytes is correct
        if (m_stat->mm_urlcopy.mm_tx.mm_transferred_bytes < m_stat->mm_urlcopy.mm_tx.mm_size) {
            m_stat->mm_urlcopy.mm_tx.mm_transferred_bytes = m_stat->mm_urlcopy.mm_tx.mm_size;
        }
	    std::string bytes_to_string = to_string<longlong>(m_stat->mm_urlcopy.mm_tx.mm_transferred_bytes, std::dec);
	    msg_ifce::getInstance()->set_total_bytes_transfered(&tr_completed, bytes_to_string.c_str());	
    }
}

/**
 * checkDestinationFileSize
 * Check if the destination file size is the same as the one used as source
 */
void UrlCopyCore::checkDestinationFileSize(
        UrlCopyError::Category& c,
        std::string& m) /*throw ()*/ {
    c = UrlCopyError::ERROR_INVALID_SIZE;

    longlong file_size = -1;

    // Get the destination File Type
    FileType::Value dst_tx_type = FileType::get(m_stat->mm_urlcopy.mm_destination.mm_tx_name);
    switch (dst_tx_type) {
        case FileType::TURL:
        case FileType::LOCAL:
        case FileType::URL:
        {
            for (unsigned int i = 0; i != NUM_RETRY_CONNECTION; ++i) {
                // Get GridFtp file size
                RefreshTimeUpdater updater(this);
                file_size = gridftp_getfilesize(m_stat->mm_urlcopy.mm_destination.mm_tx_name, c, m, m_stat->mm_urlcopy.mm_tx.mm_timeout);
                info() << "gridftp_getfilesize returned " << file_size;
                if (c != UrlCopyError::SUCCESS) {
                    if (true == m.empty()) {
                        m = "failed to get file size from gridftp server";
                    }
                } else if (0 == file_size) {
                    c = UrlCopyError::ERROR_INVALID_SIZE;
                    m = "empty file size returned";
                } else {
                    break;
                }
                if (true == TransferState::instance().interrupted()) {
                    debug() << "interrupt received, do not retry";
                    // Don't retry anymore
                    break;
                } else if (i != (NUM_RETRY_CONNECTION - 1)) {
                    // Sleep for a while and then retry
                    sleep(DEFAULT_CONNECTION_SLEEP_TIME);
                }
            }
        }
        break;
        case FileType::SFN:
        case FileType::SURL:
        default:
            // Unsupported source file type
            c = UrlCopyError::ERROR_CONSISTENCY;
            m = "destination transfer file type is not supported. Invalid stat file?";
            break;
    }
    if ((file_size > 0) &&
            (0 == m_stat->mm_urlcopy.mm_destination.mm_size)) {
        // Copy destination size
        m_stat->mm_urlcopy.mm_destination.mm_size = file_size;
    }
    // Compare Destination size with the transfer one
    if ((c == UrlCopyError::SUCCESS) &&
            (file_size != m_stat->mm_urlcopy.mm_tx.mm_size)) {
        c = UrlCopyError::ERROR_INVALID_SIZE;
        std::stringstream ss;
        ss << "destination file size returned by the gridftp server ["
                << file_size << "] differs from the source one ["
                << m_stat->mm_urlcopy.mm_source.mm_size << "]";
        m = ss.str();
    }
}

/**
 * get source context
 */
Context::Ptr UrlCopyCore::sourceContext() {
    return m_srmMethodCaller.createSrmCtx(
            m_stat->mm_urlcopy.mm_source.mm_srm_endpoint,
            m_stat->mm_urlcopy.mm_source.mm_srm_version,
            m_stat->mm_urlcopy.mm_source.mm_http_timeout);
}

/**
 * get destination context
 */
Context::Ptr UrlCopyCore::destinationContext() {
    return m_srmMethodCaller.createSrmCtx(
            m_stat->mm_urlcopy.mm_destination.mm_srm_endpoint,
            m_stat->mm_urlcopy.mm_destination.mm_srm_version,
            m_stat->mm_urlcopy.mm_destination.mm_http_timeout);
}


