/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: UrlCopyCore class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: UrlCopyCore.h,v 1.18 2009/07/30 19:32:55 molnarzs Exp $
 *  Release: $Name:  $
 */
#ifndef GLITE_DATA_TRANSFER_URL_COPY_URLCOPYCORE_H_
#define GLITE_DATA_TRANSFER_URL_COPY_URLCOPYCORE_H_

#include "UrlCopyStat.h"
#include "TransferCore.h"
#include "TransferUtils.h"
#include "ChecksumChecker_UrlCopy.h"
#include "msg-ifce.h"

#include <boost/scoped_ptr.hpp>
#include <glite/data/srm/util/Context.h>

namespace glite{
namespace data{
namespace srm{
namespace util{
class PrepareToGet;
}
}
}
}

namespace glite   {
namespace data    {
namespace transfer{
namespace urlcopy {

/**
 * The helper class that manage the transfer (used in the forked process)
 */
class UrlCopyCore : public TransferCore {
public:

     std::string vo_name;
     std::string src_site;
     std::string dest_site;

    /**
     * Construct
     */
    explicit UrlCopyCore(const std::string& id);

    /**
     * Destructor
     */
    virtual ~UrlCopyCore();

    /**
     * UpdateRefreshTime
     */
    virtual void UpdateRefreshTime();

    /**
     * Get UrlCopy Stat object
     */
    UrlCopyStat& stat() /*throw (RuntimeError, LogicError)*/;

    /**
     * Prepare the source and destination files for the transfer
     */
    void prepare() /*throw (RuntimeError, LogicError)*/;

    /**
     * Start the transfer
     */
    void transfer() /*throw (RuntimeError, LogicError)*/;

    /**
     * Complete the transfer
     */
    void complete() /*throw (RuntimeError, LogicError)*/;

private:
    /**
     * Initialize the stat
     */
    void initialize() /*throw (RuntimeError, LogicError)*/;

    /**
     * Set a sourceError
     */
    void sourceError(UrlCopyError::Phase p,UrlCopyError::Category c,const std::string& m) throw();
    void sourceErrorFromSrmStatus(const glite::data::srm::util::SrmStatus& req_status,const glite::data::srm::util::SrmStatus& file_status,UrlCopyError::Category& c,std::string& m);
    /**
     * Set a destination Error
     */
    void destinationError(UrlCopyError::Phase p,UrlCopyError::Category c,const std::string& m) throw();
    void destinationErrorFromSrmStatus(const glite::data::srm::util::SrmStatus& req_status,const glite::data::srm::util::SrmStatus& file_status,UrlCopyError::Category& c,std::string& m);

    /**
     * Set a transfer Error
     */
    void transferError(
        UrlCopyError::Phase     p,
        UrlCopyError::Category  c,
        const std::string&      m) throw();

    /**
     * Set a final Error
     */
    void finalError(
        UrlCopyError::Scope     s,
        UrlCopyError::Phase     p = UrlCopyError::PHASE_UNDEF,
        UrlCopyError::Category  c = UrlCopyError::SUCCESS,
        const std::string&      m = "");

    typedef void (UrlCopyCore::*ErrorFunction) (UrlCopyError::Phase,UrlCopyError::Category,const std::string&);

    /**
     * send a prepare to get request
     * @param ctx [IN] the SRM context
     * @param request [IN] the srmPrepareToGet
     * @param errorFunction [IN] pointer to the error function to be called
     * @param phase [IN] the phase argument of the error function
     * @return true if successful, false otherwise
     */
    bool sendPrepareToGet(
        glite::data::srm::util::Context::Ptr&                       ctx,
        boost::scoped_ptr<glite::data::srm::util::PrepareToGet>&    request,
        UrlCopyCore::ErrorFunction                                  errorFunction,
        UrlCopyError::Phase                                         phase);

    /**
     * Function to convert a request/file SRM status to an error category and message
     */
    typedef void (UrlCopyCore::*ErrorFromSrmStatusFunction)(
        const glite::data::srm::util::SrmStatus& request_status,
        const glite::data::srm::util::SrmStatus& file_status,
        UrlCopyError::Category& category,
        std::string& message);

    /**
     * poll the status of the PTG request until it reaches a final state or the timeout elapses
     * @param ctx [IN] the SRM context
     * @param request [IN] the srmPrepareToGet
     * @param timeout [IN] the timeout for the operation
     * @param start_time [IN] the time from which timeout computation must be started
     * @param max_sleep_time [IN] maximum interval between two consecutive pollings
     * @param errorFromSrmStatusFunction [IN] error conversion function
     * @param c [OUT] error category
     * @param m [OUT] error message
     */
    void pollPrepareToGet(
        glite::data::srm::util::Context::Ptr&                       ctx,
        boost::scoped_ptr<glite::data::srm::util::PrepareToGet>&    request,
        int                                                         timeout,
        int                                                         start_time,
        unsigned int                                                max_sleep_time,
        ErrorFromSrmStatusFunction                                  errorFromSrmStatusFunction,
        UrlCopyError::Category&                                     c,
        std::string&                                                m);

    /**
     * complete a PTG request
     */
    void complete_prepare_to_get(
        glite::data::srm::util::Context::Ptr&                       ctx,
        boost::scoped_ptr<glite::data::srm::util::PrepareToGet>&    request,
        int                                                         timeout,
        UrlCopyCore::ErrorFunction                                  errorFunction,
        UrlCopyError::Phase                                         phase,
        UrlCopyError::Category&                                     c,
        std::string&                                                m);

    // Request Preparation
    //--------------------------------------------------------------------------

    /**
     * Ping Source server
     */
    bool pingSource() /*throw ()*/;

    /**
     * Ping Destination server
     */
    bool pingDestination() /*throw ()*/;

    /**
     * Preapare the source file for the transfer
     */
    void prepareSource() /*throw ()*/;

    /**
     * Preapare the destination file for the transfer
     */
    void prepareDestination() /*throw ()*/;

    /**
     * Get TURL From SURL (using PrepareToGet)
     */
    bool prepareSourceTurl(ChecksumChecker_UrlCopy& checksumChecker) /*throw()*/;

    /**
     * Get TURL From SURL (using PrepareToPut)
     */
    void prepareDestTurl() /*throw()*/;

    /**
     * Translate the Source SFN into aprepareDestinationPath TURL
     */
    void translateSourceSfn() /*throw()*/;
    /**
     * Translate the Destination SFN into a TURL
     */
    void translateDestSfn() /*throw()*/;

    // Request Completion
    //--------------------------------------------------------------------------

    /**
     * Complete the Source request
     */
    void completeSource() /*throw ()*/;

    /**
     * Complete the Destination request
     */
    void completeDestination() /*throw ()*/;

    /**
     * Release the turl reserved with the PrepareToGet request
     */
    void releaseSourceTurl() /*throw ()*/;

    /**
     * Call PutDone on the the turl reserved with the PrepareToPut request
     */
    void completeDestinationTurl() /*throw ()*/;

    /**
     * Call AbortRequest on the PrepareToPut request
     */
    void abortDestinationTurl() /*throw ()*/;

    // Transfer
    //--------------------------------------------------------------------------

    /**
     * Check if the source file size of the same as the one returned by the SRM
     */
    void checkSourceFileSize(
        UrlCopyError::Category& c,
        std::string&            m) /*throw ()*/;


    /**
     * Get the checksum of the source file using gridftp
     */
    void getSourceChecksum() /*throw ()*/;

    /**
     * get destination file checksum using gridftp
     */
    bool getDestinationChecksum();

    /**
     * Copy the source file into the destination
     */
    void copyFile(
        UrlCopyError::Category& c,
        UrlCopyError::Scope&    s,
        std::string&            m) /*throw ()*/;

    /**
     * Check if the destination file size is the same as the one used as source
     */
    void checkDestinationFileSize(
        UrlCopyError::Category& c,
        std::string&            m) /*throw ()*/;

    /**
     * wait for signal to continue process
     */
    void waitContinueSignal();

    /**
     * get source context
     */
    glite::data::srm::util::Context::Ptr sourceContext();

    /**
     * get destination context
     */
    glite::data::srm::util::Context::Ptr destinationContext();

private:
    /**
     * Default Constructor. Not Implemented
     */
    UrlCopyCore();
    void _handleSourceChecksum(TransferUtils::PathInfo& pinfo);
    void _verifyDestinationChecksum(glite::data::srm::util::Context::Ptr& ctx);
    void getGridFtpFileSize_();
    bool getGridFtpSourceChecksum_(ChecksumChecker_UrlCopy& checksumChecker);

private:
    /**
     * UrlCopy Stat object
     */
    UrlCopyStat * m_stat;
    transfer_completed tr_completed;
    bool gsiftpOverwrite;
};

} // end namespace urlcopy
} // end namespace transfer
} // end namespace data
} // end namespace glite

#endif //GLITE_DATA_TRANSFER_URL_COPY_URLCOPYCORE_H_
