/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Checksum checking logic for Srm/Url copy
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: UrlCopyUtils.cpp,v 1.1 2009/07/30 19:32:55 molnarzs Exp $
 *  Release: $Name:  $
 */
#include "UrlCopyUtils.h"

#include <boost/algorithm/string.hpp>

namespace glite   {
namespace data    {
namespace transfer{
namespace urlcopy {

namespace UrlCopyUtils {

void SourceErrorFromSrmStatus(
    const SrmStatus&        req_status,
    const SrmStatus&        file_status,
    UrlCopyError::Category& c,
    std::string&            m)
{
    c = UrlCopyError::SUCCESS;

    switch(file_status.code){
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
            c = UrlCopyError::ERROR_PERMISSION;
            m = file_status.message;
        break;
        case SrmStatus::SRM_INVALID_PATH:
        case SrmStatus::SRM_FILE_UNAVAILABLE:
        case SrmStatus::SRM_FILE_LOST:
        case SrmStatus::SRM_FILE_BUSY:
        case SrmStatus::SRM_FILE_LIFETIME_EXPIRED:
            c = UrlCopyError::ERROR_INVALID_PATH;
            m = file_status.message;
        break;
        case SrmStatus::SRM_EXCEED_ALLOCATION:
        case SrmStatus::SRM_NO_USER_SPACE:
        case SrmStatus::SRM_NO_FREE_SPACE:
        case SrmStatus::SRM_SPACE_LIFETIME_EXPIRED:
            c = UrlCopyError::ERROR_NO_SPACE_LEFT;
            m = file_status.message;
        break;
        case SrmStatus::SRM_UNDEF:
        case SrmStatus::SRM_REQUEST_QUEUED:
        case SrmStatus::SRM_REQUEST_INPROGRESS:
        case SrmStatus::SRM_FILE_PINNED:
        case SrmStatus::SRM_SUCCESS:
            // File Status doesn't contain the reason of the error
        break;
        default:
            // Set default error
            c = UrlCopyError::ERROR_GENERAL_FAILURE;
            if(false == file_status.message.empty()){
                m = file_status.message;
            } else {
                std::stringstream ss;
                ss << "source file failed on the SRM with error [" << file_status.code << "]";
                m = ss.str();
            }
        break;
    }
    // In not yet found, check the Request status
    if(c == UrlCopyError::SUCCESS){
        switch(req_status.code){
            case SrmStatus::SRM_AUTHENTICATION_FAILURE:
            case SrmStatus::SRM_AUTHORIZATION_FAILURE:
                c = UrlCopyError::ERROR_PERMISSION;
                m = req_status.message;
            break;
            case SrmStatus::SRM_NO_USER_SPACE:
            case SrmStatus::SRM_NO_FREE_SPACE:
            case SrmStatus::SRM_EXCEED_ALLOCATION:
            case SrmStatus::SRM_SPACE_LIFETIME_EXPIRED:
                c = UrlCopyError::ERROR_NO_SPACE_LEFT;
                m = req_status.message;
            break;
            case SrmStatus::SRM_INVALID_REQUEST:
                c = UrlCopyError::ERROR_CONSISTENCY;
                m = req_status.message;
            break;
            case SrmStatus::SRM_REQUEST_TIMED_OUT:
                c = UrlCopyError::ERROR_REQUEST_TIMEOUT;
                m = req_status.message;
            break;
            case SrmStatus::SRM_INTERNAL_ERROR:
                c = UrlCopyError::ERROR_STORAGE_INTERNAL_ERROR;
                m = req_status.message;
            break;
            case SrmStatus::SRM_UNDEF:
            case SrmStatus::SRM_REQUEST_QUEUED:
            case SrmStatus::SRM_REQUEST_INPROGRESS:
            case SrmStatus::SRM_SUCCESS: {
                // The error is neither on the Request status. This in principle
                // should not be possible
                c = UrlCopyError::ERROR_INTERNAL_ERROR;
                std::stringstream ss;
                ss << "the SRM returned an error, but with an invalid code [" << req_status.code << "]";
                m = ss.str();
            } break;
            default:
                // Set default error
                c = UrlCopyError::ERROR_GENERAL_FAILURE;
                if(false == req_status.message.empty()){
                    m = req_status.message;
                } else {
                    std::stringstream ss;
                    ss << "source SRM request failed with error [" << req_status.code << "]";
                    m = ss.str();
                }
            break;
        }
    }
}

/*
 * destinationErrorFromSrmStatus
 *
 * Set a Destination Error from the Srm Status Error
 */
void DestinationErrorFromSrmStatus(
    const SrmStatus&        req_status,
    const SrmStatus&        file_status,
    UrlCopyError::Category& c,
    std::string&            m)
{
    c = UrlCopyError::SUCCESS;

    switch(file_status.code) {
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
            c = UrlCopyError::ERROR_PERMISSION;
            m = file_status.message;
        break;
        case SrmStatus::SRM_INVALID_PATH:
            c = UrlCopyError::ERROR_INVALID_PATH;
            m = file_status.message;
        break;
        case SrmStatus::SRM_NO_USER_SPACE:
        case SrmStatus::SRM_NO_FREE_SPACE:
        case SrmStatus::SRM_EXCEED_ALLOCATION:
        case SrmStatus::SRM_SPACE_LIFETIME_EXPIRED:
            c = UrlCopyError::ERROR_NO_SPACE_LEFT;
            m = file_status.message;
        break;
        case SrmStatus::SRM_DUPLICATION_ERROR:
        case SrmStatus::SRM_FILE_IN_CACHE:
            c = UrlCopyError::ERROR_FILE_EXISTS;
            m = file_status.message;
        break;
        case SrmStatus::SRM_UNDEF:
        case SrmStatus::SRM_REQUEST_QUEUED:
        case SrmStatus::SRM_REQUEST_INPROGRESS:
        case SrmStatus::SRM_SPACE_AVAILABLE:
        case SrmStatus::SRM_SUCCESS:
            // File Status doesn't contain the reason of the error
        break;
        default:
            // Set default error
            c = UrlCopyError::ERROR_GENERAL_FAILURE;
            if(false == file_status.message.empty()){
                m = file_status.message;
            } else {
                std::stringstream ss;
                ss << "destination file failed on the SRM with error [" << file_status.code << "]";
                m = ss.str();
            }
        break;
    }
    // In not yet found, check the Request status
    if(c == UrlCopyError::SUCCESS){
        switch(req_status.code){
            case SrmStatus::SRM_AUTHENTICATION_FAILURE:
            case SrmStatus::SRM_AUTHORIZATION_FAILURE:
                c = UrlCopyError::ERROR_PERMISSION;
                m = req_status.message;
            break;
            case SrmStatus::SRM_NO_USER_SPACE:
            case SrmStatus::SRM_NO_FREE_SPACE:
            case SrmStatus::SRM_EXCEED_ALLOCATION:
            case SrmStatus::SRM_SPACE_LIFETIME_EXPIRED:
                c = UrlCopyError::ERROR_NO_SPACE_LEFT;
                m = req_status.message;
            break;
            case SrmStatus::SRM_INVALID_REQUEST:
                c = UrlCopyError::ERROR_CONSISTENCY;
                m = req_status.message;
            break;
            case SrmStatus::SRM_REQUEST_TIMED_OUT:
                c = UrlCopyError::ERROR_REQUEST_TIMEOUT;
                m = req_status.message;
            break;
            case SrmStatus::SRM_INTERNAL_ERROR:
                c = UrlCopyError::ERROR_STORAGE_INTERNAL_ERROR;
                m = req_status.message;
            break;
            case SrmStatus::SRM_UNDEF:
            case SrmStatus::SRM_REQUEST_QUEUED:
            case SrmStatus::SRM_REQUEST_INPROGRESS:
            case SrmStatus::SRM_SUCCESS: {
                // The error is neither on the Request status. In principle,
                // this should not be possible
                c = UrlCopyError::ERROR_INTERNAL_ERROR;
                std::stringstream ss;
                ss << "SRM returned an error, but with an invalid code [" << req_status.code << "]";
                m = ss.str();
            } break;
            default:
                // Set default error
                c = UrlCopyError::ERROR_GENERAL_FAILURE;
                if(false == req_status.message.empty()){
                    m = req_status.message;
                } else {
                    std::stringstream ss;
                    ss << "destination SRM request failed with error [" << req_status.code << "]";
                    m = ss.str();
                }
            break;
        }
    }
}

/**
 * finalError
 * Set a request Error depending on the scope
 */
void SetFinalError(
	UrlCopy_FileStat&       stat,
    UrlCopyError::Scope     s,
    UrlCopyError::Phase     p,
    UrlCopyError::Category  c,
    const std::string&      m)
{
    if (stat.mm_error.isSet()){
        info() << "Final error already set.";
        return;
    }

    // debug("Setting final error. [%s] - [%s] - [%s] : '%s'",UrlCopyError::scopeToString(s),UrlCopyError::phaseToString(p),UrlCopyError::categoryToString(c),m.c_str());
    ErrorStat * error_stat = 0;
    switch(s){
        case UrlCopyError::SCOPE_SOURCE:
            debug() << "Setting final error from source.";
            if (stat.mm_source.mm_error.isSet()){
                error_stat = &(stat.mm_source.mm_error);
            } else {
                error() << "Source error not set. Cannot set final error.";
            }
        break;
        case UrlCopyError::SCOPE_DESTINATION:
            debug() << "Setting final error from destination.";
            if (stat.mm_destination.mm_error.isSet()){
                error_stat = &(stat.mm_destination.mm_error);
            } else {
                error() << "Destination error not set. Cannot set final error.";
            }
        break;
        case UrlCopyError::SCOPE_TRANSFER:
            debug() << "Setting final error from transfer.";
            if (stat.mm_tx.mm_error.isSet()){
                error_stat = &(stat.mm_tx.mm_error);
            } else {
                error() << "Transfer error not set. Cannot set final error.";
            }
        break;
        default:
        break;
    }
    if(0 != error_stat) {
        // Set Request Error
        stat.mm_error.mm_scope    = error_stat->mm_scope;
        stat.mm_error.mm_phase    = error_stat->mm_phase;
        stat.mm_error.mm_category = error_stat->mm_category;
        copy_message(stat.mm_error.mm_message,error_stat->mm_message);
    } else {
        // Use the other values
        stat.mm_error.mm_scope    = s;
        stat.mm_error.mm_phase    = p;
        stat.mm_error.mm_category = (UrlCopyError::SUCCESS == c) ? UrlCopyError::ERROR_INTERNAL_ERROR : c;
        copy_message(stat.mm_error.mm_message,m.c_str());
    }
    std::string scope = UrlCopyError::scopeToString(s);
    error() << "Final error on " << scope << " during " <<
        UrlCopyError::phaseToString((UrlCopyError::Phase)stat.mm_error.mm_phase) << " phase: [" <<
        UrlCopyError::categoryToString((UrlCopyError::Category)stat.mm_error.mm_category) << "] " <<
        stat.mm_error.mm_message;
    boost::to_upper(scope);
    info() << "FINAL:" << scope << ": " << stat.mm_error.mm_message;
}

} // namespace UrlCopyUtils

} // namespace urlcopy
} // namespace transfer
} // namespace data
} // namespace glite
