/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: helper methods for copy executables
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: glite-copy-utility.cpp,v 1.2 2008/10/20 15:19:46 ptedesco Exp $
 *  Release: $Name:  $
 */

#include "glite-copy-utility.h"  

#include "ServiceDiscoveryIfce.h"
#include <glite/data/glite-util.h>

#include <glite/data/agents/url-utility.h>
#include <boost/algorithm/string.hpp>

namespace{
const char *  const   SRM_VERSION_1_1     = "1.1";
const char *  const   SRM_VERSION_2_2     = "2.2";
const char *  const   DEFAULT_SRM_VERSION = SRM_VERSION_2_2;
}

using namespace glite::data::agents;

namespace glite{
namespace data{
namespace transfer{
namespace urlcopy{

/**
 * get_endpoint_for_surl
 */
std::string get_endpoint_from_surl(
    const std::string& surl,
    const std::string& endpoint_version) /*throw(RuntimeError)*/ {
        
    std::string endpoint;

    // parse surl
    try{
        std::string  hostname;
        unsigned int port = 0;
        std::string  protocol;
        std::string  path;
        std::string  query_str;
        // Parse surl
        parse_url(surl,hostname,port,protocol,path,query_str);
        // check protocol
        boost::to_lower(protocol);
        if(protocol == "srm"){
            // Check query string
            if(query_str.empty()){
                // It's compact: query service discovery
                SDException exception;
                SDServiceList * s_list = SD_listServicesByHost("SRM", hostname.c_str(), 0, &exception);
                if(0 == s_list){
                    SD_freeException(&exception);
                    throw RuntimeError("Cannot determine SRM endpoint." 
                            " Please check ServiceDiscovery configuration");
                }
                const SDService * svc = 0;
                const SDServiceList * svc_list = 0;
                SDServiceList port_selection = {0,0,0};
                port_selection.services = new SDService*[s_list->numServices];
                memset(port_selection.services,0,s_list->numServices);
                // Check the port
                if(0 == port){
                    svc_list = s_list;
                } else {
                    int i = 0;
                    for(i = 0; i < s_list->numServices;++i){
                        SDService * s = s_list->services[i];
                        // Check if the port is the same
                        std::string  svc_hostname;
                        unsigned int svc_port;
                        if(0 != s->endpoint){
                            parse_endpoint(s->endpoint,svc_hostname,svc_port);
                            if(svc_port == port){
                                // Ok, same port
                                port_selection.services[port_selection.numServices] = s;
                                ++port_selection.numServices;
                            }
                        }
                    }
                    svc_list = &port_selection;
                }
                if(svc_list->numServices > 1){
                    // No port specified: choose based on the version
                    std::string version = /*g_tx_args.src_version*/endpoint_version;
                    if(version.empty() ){
                        // select the default one
                        version = DEFAULT_SRM_VERSION;
                    }
                    // select service by version
                    svc = select_service_by_version(svc_list, version.c_str());
                } else {
                    svc = svc_list->services[0];
                }
                if((svc != 0) && (svc->endpoint != 0)){
                    endpoint = svc->endpoint;
                }
                SD_freeServiceList(s_list);
                if(endpoint.empty()){
                    throw RuntimeError("Cannot determine SRM endpoint." 
                            " Please check ServiceDiscovery configuration");
                }
            } else {
                // It's fully qualified: take it from the SURL
                endpoint  = create_url("httpg",hostname,port,path);
            }    
        } else {
            // Not a SURL
        }
    } catch(const RuntimeError& exc){
        throw;
    } catch(const std::exception& exc){
        // Exception
    }
    return endpoint;    
}
    
/**
 * get_endpoint_version
 */
std::string get_endpoint_version(const std::string& endpoint) /*throw()*/ {
    std::string version;
    // parse endpoint
    try{
        std::string  hostname;
        unsigned int port = 0;
        std::string  protocol;
        std::string  path;
        std::string  query_str;
        // Parse surl
        parse_url(endpoint,hostname,port,protocol,path,query_str);
        // check path
        boost::to_lower(path);
        if(path == "srm/managerv1"){
            // Check query string
            version = SRM_VERSION_1_1;
        } else if(path == "srm/managerv2"){
            // Check query string
            version = SRM_VERSION_2_2;
        } else {
            // query service discovery
            SDException exception;
            SDServiceList * svc_list = SD_listServicesByHost("SRM", hostname.c_str(), 0, &exception);
            if(0 != svc_list){
                int i = 0;
                for(i = 0; i < svc_list->numServices;++i){
                    SDService * s = svc_list->services[i];
                    if((s != 0) && (s->endpoint != 0) &&
                        (endpoint == s->endpoint)){
                        if(0 != s->version){
                            version = s->version; 
                            break;
                        } else {
                            throw RuntimeError("Null version specified in ServiceDiscovery");
                        }
                        }
                }
                if(i == svc_list->numServices){
                    SD_freeServiceList(svc_list);
                    throw RuntimeError("No service specified in ServiceDiscovery for the given endpoint");
                }
                SD_freeServiceList(svc_list);
            } else {
                SD_freeException(&exception);
                throw RuntimeError("Cannot determine SRM endpoint." 
                        " Please check ServiceDiscovery configuration");
            }
        }
    } catch(...){
        // Assign the default one
        if(endpoint.empty()){
            version = "";
        } else {
            version = DEFAULT_SRM_VERSION;
        }
    }
    return version;
}

} // namespace urlcopy
} // namespace transfer
} // namespace data
} // namespace glite


