/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  gLite Transfer Url Copy Clients - glite-srm-copy Tool
 *
 *  Authors: Paolo Badino <Paolo.Badino@cern.ch>
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: glite-srm-copy.cpp,v 1.16 2009/07/02 15:02:09 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include "util-main.h"
#include <glite/data/transfer/urlcopy/TransferSrmCopy.h>
#include "logger-utility.h"
#include "ServiceDiscoveryIfce.h"
#include <glite/data/glite-util.h>

#include <glite/data/agents/AgentExceptions.h>
#include <glite/data/agents/url-utility.h>

#include "glite-copy-utility.h"

#include <boost/algorithm/string.hpp>

using namespace glite::data::agents;
using namespace glite::data::agents::options;
using namespace glite::data::transfer::urlcopy;

//
// Global variables
//
namespace {
const char *          g_prefix          = 0;
TransferSrmCopy::Args g_tx_args;
bool                  g_block_mode      = false;

const char * g_cksm_algorithm = 0;
const char * g_cksm_value = 0;
bool g_compare_checksum = false;

const char * bool_to_str(bool value) {
    return value ? "yes" : "no";
}

const char * pull_mode_to_str(bool pull_mode) {
    return pull_mode ? "pull" : "push";
}

} // namespace

void opt_srmvers(char * arg){
    g_tx_args.src_version = arg;
    g_tx_args.dest_version = arg;
}

void opt_cksm_algorithm(char * arg){
    g_cksm_algorithm = arg;
    g_compare_checksum = true;
}
void opt_cksm_value(char * arg){
    g_cksm_value = arg;
    g_compare_checksum = true;
}


void tool_add_args(Arguments& args){
    // options
    opt(g_tx_args.src_endpoint,args,                '1',"source-srm").arg("URL")        .help("use the SRM service at the specified URL for source.");
    opt(g_tx_args.dest_endpoint,args,               '2',"dest-srm").arg("URL")          .help("use the SRM service at the specified URL for destination.");
    opt(opt_srmvers,args,                           'I',"srm-version").arg("VERSION")   .help("the SRM interface version to be used.");
    opt(g_tx_args.dest_token_descr,args,            'T',"dest-token-desc").arg("DESC")  .help("target space token description.");
    opt(g_tx_args.source_token_descr,args,          'S',"source-token-desc").arg("DESC").help("source space token description.");
    opt(g_block_mode,true,args,                     'b',"blocking")                     .help("run the transfer in blocking mode.");
    opt(g_tx_args.console_logging,true,args,        'C',"console")                      .help("enable console logging.");
    opt(g_tx_args.verbose,true,args,                'd',"debug")                        .help("enable debug logs.");
    opt(g_tx_args.copy_timeout,args,                't',"copy-timeout").arg("N")        .help("set the copy timeout to N seconds.");
    opt(g_tx_args.http_timeout,args,                'H',"http-timeout").arg("N")        .help("set the http timeout to N seconds.");
    opt(g_tx_args.pull_mode,false,args,             'P',"push-mode")                    .help("run the transfer in push mode (default = pull).");
    opt(g_prefix,args,                              'p',"prefix").arg("PREFIX")         .help("transfer id prefix.");
    opt(g_tx_args.enable_ping_source,false,args,    'e',"dont-ping-source")             .help("don't ping source endpoint.");
    opt(g_tx_args.enable_ping_dest,false,args,      'E',"dont-ping-dest")               .help("don't ping destination endpoint.");
    opt(g_tx_args.enable_dir_check,false,args,      'k',"disable-dir-check")            .help("skip target directory existence check and mkdir operations.");
    opt(g_tx_args.block_after_preparation,true,args,'x',"block-after-preparation")      .help("block after preparation phase is complete.");
    opt(g_tx_args.block_after_transfer,true,args,   'X',"block-after-transfer")         .help("block after transfer phase is complete.");
    opt(g_tx_args.copy_pin_lifetime,args,               "copy-pin-lifetime").arg("N")   .help("set the desired copy pin lifetime to N seconds.");
    opt(g_tx_args.lan_connection,true,args,             "lan-connection")               .help("use LAN connection type in get and put operations.");
    opt(g_tx_args.fail_nearline_files,true,args,        "fail-nearline")                .help("fail the transfer immediately if the source file is on tape.");
    opt(g_tx_args.tx_to_per_mb,args,                    "timeout-per-mb").arg("N")      .help("size based transfer timeout, in seconds per MB (float).");
    opt(g_tx_args.user_tag,args,                        "user-tag").arg("ID")           .help("user tag for the transfer.");
    // checksums
    opt(opt_cksm_algorithm,args,                    'a',"algorithm").arg("ALG")         .help("use ALG as checksum algorithm (default = ADLER32).")
                                                                                        .help("    (implies --compare-checksum).");
    opt(opt_cksm_value,args,                        'c',"checksum-value").arg("CKSM")   .help("verify that the checksum of the file is equal to CKSM after transfer.")
                                                                                        .help("    This option is valid only if a single source-destination couple is specified.")
                                                                                        .help("    (implies --compare-checksum).");
    opt(g_compare_checksum,true,args,               'K',"compare-checksum")             .help("verify that the checksum of the file is equal to the source one after transfer.");

    // positional
    args.pos("source")      .help("source");
    args.pos("destination") .help("destination");
    args.pos("[source destination ...]").help("additional source-destination couples.");
}

/**
 * tool_doit_prepare_request
 * Prepare the transfer request
 */
int tool_doit_prepare_request(int argc, char *argv[]){

    if(argc < 2){
        tool_error() << "No files specified" << std::endl;
        return EXIT_FAILURE;
    }
    if(0 != (argc % 2)){
        tool_error() << "source destination mismatch" << std::endl;
        return EXIT_FAILURE;
    }

    int i = 0;
    int num_files = argc / 2;

    if (0 != g_cksm_value){
        tool_error() << "The --checksum-value option is valid only if a single source-destination couple is specified.";
        return EXIT_FAILURE;
    }

    g_tx_args.files.resize(num_files);
    for(; i < num_files; ++i){
        // Set source surl
        g_tx_args.files[i].source = argv[2*i];
        if(g_tx_args.files[i].source.empty()){
            tool_error() << "File " << i << ": empty source name" << std::endl;
            return EXIT_FAILURE;
        }
        // Set destination surl
        g_tx_args.files[i].destination = argv[(2*i)+1];
        if(g_tx_args.files[i].destination.empty()){
            tool_error() << "File " << i << ": empty destination name" << std::endl;
            return EXIT_FAILURE;
        }
        if (g_compare_checksum){
            g_tx_args.files[i].compare_checksum = true;
            if (0 != g_cksm_algorithm){
                g_tx_args.files[i].checksum_algorithm = g_cksm_algorithm;
            }
            if (0 != g_cksm_value){
                g_tx_args.files[i].checksum_value = g_cksm_value;
            }
        }
    }

    // If not set, deduce the endpoints from the surls
    if(true == g_tx_args.src_endpoint.empty()){
        try{
            g_tx_args.src_endpoint = get_endpoint_from_surl(g_tx_args.files[0].source,g_tx_args.src_version);
        } catch(const RuntimeError& exc){
            tool_error() << "cannot find SRM endpoint for SURL [" << g_tx_args.files[0].source << "]: " << exc.what() << std::endl;
            return EXIT_FAILURE;
        }
    }
    if(true == g_tx_args.dest_endpoint.empty()){
        try{
            g_tx_args.dest_endpoint = get_endpoint_from_surl(g_tx_args.files[0].destination,g_tx_args.dest_version);
        } catch(const RuntimeError& exc){
            tool_error() << "cannot find SRM endpoint for SURL [" << g_tx_args.files[0].destination << "]: " << exc.what() << std::endl;
            return EXIT_FAILURE;
        }
    }
    if(true == g_tx_args.src_version.empty()){
        g_tx_args.src_version = get_endpoint_version(g_tx_args.src_endpoint);
    }
    if(true == g_tx_args.dest_version.empty()){
        g_tx_args.dest_version = get_endpoint_version(g_tx_args.dest_endpoint);
    }

    // Print transfer paremeters
    tool_debug() << "# Transfer Request parameters:" << std::endl;
    // Source
    if(false == g_tx_args.src_endpoint.empty()){
    tool_debug() << "# Src SRM Endpoint: " << g_tx_args.src_endpoint << std::endl;
    }
    if(false == g_tx_args.src_version.empty()){
    tool_debug() << "# Src SRM Version : " << g_tx_args.src_version << std::endl;
    }
    // Destination
    if(false == g_tx_args.dest_endpoint.empty()){
    tool_debug() << "# Dst SRM Endpoint: " << g_tx_args.dest_endpoint << std::endl;
    }
    if(false == g_tx_args.dest_version.empty()){
    tool_debug() << "# Dst SRM Version : " << g_tx_args.dest_version << std::endl;
    }
    if(false == g_tx_args.dest_token_descr.empty()){
    tool_debug() << "# Dst Space Descr : " << g_tx_args.dest_token_descr << std::endl;
    }

    for(unsigned i = 0; i < g_tx_args.files.size(); ++i){
    tool_debug() << "# Source      " << i << "  : " << g_tx_args.files[i].source << std::endl;
    tool_debug() << "# Destination " << i << "  : " << g_tx_args.files[i].destination << std::endl;
    }
    // Configuration
    tool_debug() << "# Overwrite       : " << bool_to_str (g_tx_args.overwrite) << std::endl;
    tool_debug() << "# Mode            : " << pull_mode_to_str (g_tx_args.pull_mode) << std::endl;
    // Timeouts
    if(g_tx_args.http_timeout > 0){
    tool_debug() << "# HTTP Timeout    : " << g_tx_args.http_timeout << std::endl;
    }
    if(g_tx_args.copy_timeout > 0){
    tool_debug() << "# Copy Timeout    : " << g_tx_args.copy_timeout << std::endl;
    }
    if(g_tx_args.refresh_timeout > 0){
    tool_debug() << "# Refresh Timeout : " << g_tx_args.refresh_timeout << std::endl;
    }
    return EXIT_SUCCESS;
}

/**
 * tool_doit_transfer
 * Start the transfer
 */
int tool_doit_transfer(){
    int result = EXIT_FAILURE;

    std::string prefix;
    if(0 != g_prefix){
        prefix = g_prefix;
    }
    TransferSrmCopy::Ptr tx;
    if(false == g_block_mode){
        try{
            tx = TransferSrmCopy::start(g_tx_args, prefix, false);
            // Print Transfer ID
            tool_info() << "Transfer Request ID: " << tx->id() << std::endl;
            result = EXIT_SUCCESS;
        } catch(const std::exception& exc){
            tool_error() << "Failed to start transfer: " << exc.what() << std::endl;
            result = EXIT_FAILURE;
        }
    } else {
        try{
            // Run transfer in blocking mode
            tool_debug() << "Starting srmcopy transfer..." << std::endl;
            tx = TransferSrmCopy::start(g_tx_args, prefix, true);
            // Transfer is now completed (successfully or not)
            tool_info() << "Copy request ID: " << tx->id() << std::endl;
            try{
                // Get transfer status
                TransferSrmCopyStat tx_stat = tx->status();
                if(tx_stat.status == UrlCopyStatus::COMPLETED){
                    if(tx_stat.error.category == UrlCopyError::SUCCESS){
                        result = EXIT_SUCCESS;
                        tool_debug() << "Source Prep. Time: " << (tx_stat.src_prep_done_time  - tx_stat.src_prep_start_time) << " secs" << std::endl;
                        tool_debug() << "Dest.  Prep. Time: " << (tx_stat.dest_prep_done_time - tx_stat.dest_prep_start_time) << " secs" << std::endl;
                        tool_debug() << "Copy         Time: " << (tx_stat.copy_done_time      - tx_stat.copy_start_time) << " secs" << std::endl;
                        tool_debug() << "Finalization Time: " << (tx_stat.final_done_time     - tx_stat.final_start_time) << " secs" << std::endl;
                        for(unsigned int i = 0; i < tx_stat.files.size(); ++i){
                            const TransferSrmCopyStat::FileRequest& f = tx_stat.files.at(i);
                            if(f.error.category == UrlCopyError::SUCCESS){
                                tool_info() << "File [" << i << "] completed successfully: " << f.size << " bytes transferred in " << (f.done_time - f.start_time) << " seconds" << std::endl;
                            } else {
                                tool_error() << "File [" << i << "] failed on " << UrlCopyError::scopeToString(f.error.scope) << " during " << UrlCopyError::phaseToString(f.error.phase) << " phase. Reason is: [" << UrlCopyError::categoryToString(f.error.category) << "] " << f.error.description << std::endl;
                            }
                        }
                    } else {
                        result = EXIT_FAILURE;
                        // Print Transfer Failure
                        tool_error() << "Copy request failed on " << UrlCopyError::scopeToString(tx_stat.error.scope) << " during " << UrlCopyError::phaseToString(tx_stat.error.phase) << " phase. Reason is: [" << UrlCopyError::categoryToString(tx_stat.error.category) << "] " << tx_stat.error.description << std::endl;
                    }
                    tool_info() << "Request took " << (tx_stat.finish_time - tx_stat.start_time) << " seconds in total" << std::endl;
                }  else {
                    // This should not happen. Anyway call abort
                    throw LogicError("Transfer didn't complete in blocking mode and will be aborted");
                }
            } catch(const std::exception& exc){
                result = EXIT_FAILURE;
                // This should not happen. Anyway call abort
                tool_error() << "Transfer didn't complete in blocking mode and will be aborted" << std::endl;
                try{
                    tx->abort();
                }catch(...){}
            }
            // clean the log files
            try{
                tx->clean();
            }catch(...){
                tool_error() << "Failed to clean transfer log file" << std::endl;
            }
        } catch(const std::exception& exc){
            tool_error() << "Failed to execute transfer: " << exc.what() << std::endl;
            result = EXIT_FAILURE;
        }
    }
    return result;
}

/*
 * tool_doit
 *
 * Run the Get
 */
int tool_doit(int argc, char *argv[]){
    // Prepare a request
    int result = tool_doit_prepare_request(argc, argv);
    if(result == EXIT_SUCCESS){
        // start the transfer
        result = tool_doit_transfer();
    }
    return result;
}

