/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  gLite Transfer Url Copy Clients - glite-url-copy-status Tool
 *
 *  Authors: Paolo Badino <Paolo.Badino@cern.ch>
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: glite-url-copy-status.cpp,v 1.23 2008/10/20 15:19:46 ptedesco Exp $
 *  Release: $Name:  $
 *
 */

#include "util-main.h"
#include <glite/data/transfer/urlcopy/TransferUrlCopy.h>
#include <glite/data/transfer/urlcopy/TransferSrmCopy.h>
#include "stat-utility.h"
#include "logger-utility.h"

#include <glite/data/agents/AgentExceptions.h>
#include <glite/data/agents/url-utility.h>

#include <boost/algorithm/string.hpp>

using namespace glite::data::agents;
using namespace glite::data::agents::options;
using namespace glite::data::transfer::urlcopy;

/**
 * tool_add_args
 */
void tool_add_args(Arguments& args){
    // positional parameters
    args.pos("request_id").help("the identifier of the transfer request.");
}

/**
 * print_statistics
 * print statistics
 */
void print_statistics(const TransferUrlCopyStat& tx_stat){
    if(tx_stat.status == UrlCopyStatus::COMPLETED){
        if(tx_stat.error.category == UrlCopyError::SUCCESS){
            // Print Transfer details
            tool_info() << tx_stat.tx.size << " bytes transferred in " << (tx_stat.tx.done_time - tx_stat.tx.start_time) << " seconds with an average throughput of " << (tx_stat.tx.avg_throughput / 1024.0) << " KB/s" << std::endl;
            tool_debug() << "Source Prep.  Time: " << (tx_stat.source.prep_done_time - tx_stat.source.prep_start_time) << " secs" << std::endl;
            tool_debug() << "Dest.  Prep.  Time: " << (tx_stat.destination.prep_done_time - tx_stat.destination.prep_start_time) << " secs" << std::endl;
            tool_debug() << "Source Final. Time: " << (tx_stat.source.final_done_time - tx_stat.source.final_start_time) << " secs" << std::endl;
            tool_debug() << "Dest.  Final. Time: " << (tx_stat.destination.final_done_time - tx_stat.destination.final_start_time) << " secs" << std::endl;
        } else {
            // Print Transfer Failure
            tool_error() << "Transfer request failed on " << UrlCopyError::scopeToString(tx_stat.error.scope) << " during " << UrlCopyError::phaseToString(tx_stat.error.phase) << " phase. Reason is: [" << UrlCopyError::categoryToString(tx_stat.error.category) << "] " << tx_stat.error.description << std::endl;
        }
        tool_info() << "Request took " << (tx_stat.finish_time - tx_stat.start_time) << " seconds in total" << std::endl;
    } 
}

/*
 * print_statistics
 * 
 * print statistics
 */
void print_statistics(const TransferSrmCopyStat& tx_stat){
    if(tx_stat.status == UrlCopyStatus::COMPLETED){
        if(tx_stat.error.category == UrlCopyError::SUCCESS){
            tool_debug() << "Source Prep. Time: " << (tx_stat.src_prep_done_time  - tx_stat.src_prep_start_time) << " secs" << std::endl;
            tool_debug() << "Dest.  Prep. Time: " << (tx_stat.dest_prep_done_time - tx_stat.dest_prep_start_time) << " secs" << std::endl;
            tool_debug() << "Copy         Time: " << (tx_stat.copy_done_time      - tx_stat.copy_start_time) << " secs" << std::endl;
            tool_debug() << "Finalization Time: " << (tx_stat.final_done_time     - tx_stat.final_start_time) << " secs" << std::endl;
            for(unsigned int i = 0; i < tx_stat.files.size(); ++i){
                const TransferSrmCopyStat::FileRequest& f = tx_stat.files.at(i);
                if(f.error.category == UrlCopyError::SUCCESS){
                    tool_info() << "File [" << i << "] completed successfully: " << f.size << " bytes transferred in " << (f.done_time - f.start_time) << " seconds" << std::endl;
                } else {
                    tool_error() << "File [" << i << "] failed on " << UrlCopyError::scopeToString(f.error.scope) << " during " << UrlCopyError::phaseToString(f.error.phase) << " phase. Reason is: [" << UrlCopyError::categoryToString(f.error.category) << "] " << f.error.description << std::endl;
                }
            }
        } else {
            // Print Failure
            tool_error() << "Copy request failed on " << UrlCopyError::scopeToString(tx_stat.error.scope) << " during " << UrlCopyError::phaseToString(tx_stat.error.phase) << " phase. Reason is: [" << UrlCopyError::categoryToString(tx_stat.error.category) << "] " << tx_stat.error.description << std::endl;
        }
        tool_info() << "Request took " << (tx_stat.finish_time - tx_stat.start_time) << " seconds in total" << std::endl;
    } 
}

/*
 * transfer_status
 * 
 * Get the transfer's status
 */
template <typename T,typename S>
int transfer_status(T& tx){
    // Get transfer status; do not check if transfer is dead or timed out
    // (otherwise the check method could throw an exception and no status would
    // be printed)
    S tx_stat = tx.status(false);
    
    // print Transfer Information
    tool_info() << "Transfer status is " << UrlCopyStatus::toString(tx_stat.status) << std::endl;
    // If transfer is completed, clean it
    if(tx_stat.status == UrlCopyStatus::COMPLETED){
        // print transfer statistics
        print_statistics(tx_stat);
        try{
            tx.clean();    
        }catch(...){
            tool_error() << "Failed to clean transfer log file" << std::endl;
        }
    }
    return EXIT_SUCCESS;
}

/*
 * tool_doit
 */
int tool_doit(int argc, char *argv[]){
    int result = EXIT_FAILURE;
    
    if((argc < 1) || (0 == argv[0])){
        tool_error() << "No transfer request id specified" << std::endl;
        return EXIT_FAILURE;
    }
    if(argc > 1){
        tool_error() << "Too many request id specified" << std::endl;
        return EXIT_FAILURE;
    }
    std::string request_id = argv[0];    

    try {
        // Get the TransferUrlCopy object
        TransferUrlCopy::Ptr utx = TransferUrlCopy::get(request_id);
        // Get the transfer's status
        result = transfer_status<TransferUrlCopy,TransferUrlCopyStat>(*utx);
    } catch(const std::exception& exc){
        std::string msg = std::string("Not a urlcopy transfer : ") + exc.what();
        tool_debug() << msg << std::endl;
        try{
            // Not an Urlcopy Request: try if it's an SrmCopy one
            TransferSrmCopy::Ptr stx = TransferSrmCopy::get(request_id);
            // Get the transfer's status
            result = transfer_status<TransferSrmCopy,TransferSrmCopyStat>(*stx);
        } catch(const std::exception& exc){
            std::string msg = std::string("Not an srmcopy transfer : ") + exc.what();
            tool_debug() << msg << std::endl;
            tool_error() << "Invalid transfer request id. Not a urlcopy nor a srmcopy valid transfer" << std::endl;
            result = EXIT_FAILURE;
        }
    }
    return result;
}

