/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  gLite Transfer Url Copy Clients - glite-url-copy-stop Tool
 *
 *  Authors: Paolo Badino <Paolo.Badino@cern.ch>
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: glite-url-copy-stop.cpp,v 1.9 2008/10/20 09:01:44 ptedesco Exp $
 *  Release: $Name:  $
 *
 */

#include "util-main.h"
#include <glite/data/transfer/urlcopy/TransferUrlCopy.h>
#include <glite/data/transfer/urlcopy/TransferSrmCopy.h>
#include "stat-utility.h"
#include "logger-utility.h"

#include <glite/data/agents/AgentExceptions.h>
#include <glite/data/agents/url-utility.h>

#include <boost/algorithm/string.hpp>

using namespace glite::data::agents;
using namespace glite::data::agents::options;
using namespace glite::data::transfer::urlcopy;

/**
 * tool_add_args
 */
void tool_add_args(Arguments& args){
    // positional parameters
    args.pos("request_id").help("the identifier of the transfer request.");
}

/*
 * stop_transfer
 * 
 * Stop The Transfer
 */
template <typename T,typename S>
int stop_transfer(T& tx){
    try{
        // Stop the transfer
        tx.stop();
        // Get transfer status
        S tx_stat = tx.status();
        if(tx_stat.status != UrlCopyStatus::COMPLETED){
           // This should not happen. Anyway call abort
           throw LogicError("Transfer didn't complete after a stop. Abort it");
        }
        if(tx_stat.error.category == UrlCopyError::SUCCESS){
            // Print Transfer Failure
            tool_info() << "Transfer completed successfully" << std::endl;
        } else if(tx_stat.error.category == UrlCopyError::ERROR_ABORTED){
            tool_info() << "Transfer aborted" << std::endl;
        } else {
            // Print Transfer Failure
            tool_error() << "Transfer already failed" << std::endl;
        }
    } catch(const RuntimeError& exc){
        // This should not happen. Anyway call abort
        tool_error() << "Transfer didn't stop and will be aborted" << std::endl;
        try{
            tx.abort();
        }catch(...){}
    }
    // If transfer is completed, clean it
    try{
        tx.clean();    
    }catch(...){
        tool_error() << "Failed to clean transfer log file" << std::endl;
    }
    return EXIT_SUCCESS;
}

/*
 * tool_doit
 */
int tool_doit(int argc, char *argv[]){
    int result = EXIT_FAILURE;
    
    if((argc < 1) || (0 == argv[0])){
        tool_error() << "No transfer request id specified" << std::endl;
        return EXIT_FAILURE;
    }
    if(argc > 1){
        tool_error() << "Too many request id specified" << std::endl;
        return EXIT_FAILURE;
    }
    std::string request_id = argv[0];    

    try {
        // Get the TransferUrlCopy objects
        TransferUrlCopy::Ptr utx = TransferUrlCopy::get(request_id);
        // Stop the transfer
        result = stop_transfer<TransferUrlCopy,TransferUrlCopyStat>(*utx);
    } catch(const std::exception& exc){
        try{
            // Not an Urlcopy Request: try if it's an SrmCopy One
            TransferSrmCopy::Ptr stx = TransferSrmCopy::get(request_id);
            // Stop the transfer
            result = stop_transfer<TransferSrmCopy,TransferSrmCopyStat>(*stx);
        } catch(const std::exception& exc){
            tool_error() << "Invalid transfer request id. Not a urlcopy nor a srmcopy valid transfer" << std::endl;
            result = EXIT_FAILURE;
        }
    }
    
    return result;
}

