/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  gLite Transfer Url Copy Clients - glite-url-copy Tool
 *
 *  Authors: Paolo Badino <Paolo.Badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: glite-url-copy.cpp,v 1.26 2009/03/13 09:16:36 ptedesco Exp $
 *  Release: $Name:  $
 *
 */
#include "util-main.h"
#include <glite/data/transfer/urlcopy/TransferUrlCopy.h>
#include "ServiceDiscoveryIfce.h"
#include <glite/data/glite-util.h>

#include <glite/data/agents/AgentExceptions.h>
#include <glite/data/agents/url-utility.h>

#include "glite-copy-utility.h"

#include <boost/algorithm/string.hpp>

#include "util-urlcopy.h"
#include "Instrument.h"

using namespace glite::data::agents;
using namespace glite::data::agents::options;
using namespace glite::data::transfer::urlcopy;

//
// Global variables
//

namespace {
const char *          g_prefix          = 0;
TransferUrlCopy::Args g_tx_args;
bool                  g_block_mode      = false;
} // end global namespace

// Help message
void opt_srmvers(char * arg){
    g_tx_args.src_version = arg;
    g_tx_args.dest_version = arg;
}
void opt_srm_get_to(char * arg){
    g_tx_args.get_timeout = atoi (arg);
    g_tx_args.getdone_timeout = g_tx_args.get_timeout;
}
void opt_srm_put_to(char * arg){
    g_tx_args.put_timeout = atoi (arg);
    g_tx_args.putdone_timeout = g_tx_args.put_timeout;
}

void opt_cksm_algorithm(char * arg){
    g_tx_args.checksum_user_algorithm = arg;
    g_tx_args.compare_checksum = true;
}
void opt_cksm_value(char * arg){
    g_tx_args.checksum_user_value = arg;
    g_tx_args.compare_checksum = true;
}

void opt_instrument(char* arg)
{
    Instrument::Instance().set(arg);
}

void tool_add_args(Arguments& args){
    // transfer options
    opt(g_tx_args.src_endpoint,args,                '1',"source-srm").arg("URL")        .help("use the SRM service at the specified URL for source.");
    opt(g_tx_args.dest_endpoint,args,               '2',"dest-srm").arg("URL")          .help("use the SRM service at the specified URL for destination.");
    opt(opt_srmvers,args,                           'I',"srm-version").arg("VERSION")   .help("the SRM interface version to be used.");
    opt(g_tx_args.dest_token_descr,args,            'T',"dest-token-desc").arg("DESC")  .help("target space token description.");
    opt(g_tx_args.source_token_descr,args,          'S',"source-token-desc").arg("DESC").help("source space token description.");
    opt(g_block_mode,true,args,                     'b',"blocking")                     .help("run the transfer in blocking mode.");
    opt(g_tx_args.nstreams,args,                    'n',"streams").arg("N")             .help("number of streams gridftp parameter.");
    opt(g_tx_args.tcpbs,args,                       's',"tcpbs").arg("N")               .help("tcp buffer size gridftp parameter.");
    opt(g_tx_args.blocksize,args,                   'B',"block-size").arg("N")          .help("tcp block size gridftp parameter.");
    opt(g_prefix,args,                              'p',"prefix").arg("PREFIX")         .help("transfer id prefix.");
    opt(g_tx_args.tx_timeout,args,                  't',"transfer-timeout").arg("N")    .help("set the transfer timeout to N seconds.");
    opt(g_tx_args.tx_markers_timeout,args,          'm',"markers-timeout").arg("N")     .help("set the transfer markers timeout to N seconds.");
    opt(g_tx_args.first_tx_marker_timeout,args,     'f',"first-marker-timeout").arg("N").help("set the first transfer marker timeout to N seconds.");
    opt(opt_srm_get_to,args,                        'g',"srm-get-timeout").arg("N")     .help("set the srm-get timeout to N seconds.");
    opt(opt_srm_put_to,args,                        'u',"srm-put-timeout").arg("N")     .help("set the srm-put timeout to N seconds.");
    opt(g_tx_args.http_timeout,args,                'H',"http-timeout").arg("N")        .help("set the http timeout to N seconds.");
    opt(g_tx_args.console_logging,true,args,        'C',"console")                      .help("enable console logging.");
    opt(g_tx_args.verbose,true,args,                'd',"debug")                        .help("enable debug logs.");
    opt(g_tx_args.enable_ping_source,false,args,    'e',"dont-ping-source")             .help("don't ping source endpoint.");
    opt(g_tx_args.enable_ping_dest,false,args,      'E',"dont-ping-dest")               .help("don't ping destination endpoint.");
    opt(g_tx_args.enable_dir_check,false,args,      'k',"disable-dir-check")            .help("skip target directory existence check and mkdir operations.");
    opt(g_tx_args.overwrite,true,args,              'o',"overwrite")                    .help("Overwrite target file");
    opt(g_tx_args.block_after_preparation,true,args,'x',"block-after-preparation")      .help("block after preparation phase is complete.");
    opt(g_tx_args.block_after_transfer,true,args,   'X',"block-after-transfer")         .help("block after transfer phase is complete.");
    opt(g_tx_args.copy_pin_lifetime,args,               "copy-pin-lifetime").arg("N")   .help("set the desired copy pin lifetime to N seconds.");
    opt(g_tx_args.lan_connection,true,args,             "lan-connection")               .help("use LAN connection type in get and put operations.");
    opt(g_tx_args.fail_nearline_files,true,args,        "fail-nearline")                .help("fail the transfer immediately if the source file is on tape.");
    opt(g_tx_args.tx_to_per_mb,args,                    "timeout-per-mb").arg("N")      .help("size based transfer timeout, in seconds per MB (float).");
    opt(g_tx_args.no_progress_timeout ,args,            "no-progress-timeout").arg("N") .help("abort the transfer if no progress is received for more than N seconds.");
    opt(g_tx_args.user_tag,args,                        "user-tag").arg("ID")           .help("user tag for the transfer.");
    opt(opt_cksm_algorithm,args,                    'a',"algorithm").arg("ALG")         .help("use ALG as checksum algorithm (default = ADLER32).")
                                                                                        .help("    (implies --compare-checksum).");
    opt(opt_cksm_value,args,                        'c',"checksum-value").arg("CKSM")   .help("verify that the checksum of the file is equal to CKSM after transfer.")
                                                                                        .help("    (implies --compare-checksum).");
    opt(g_tx_args.compare_checksum,true,args,       'K',"compare-checksum")             .help("verify that the checksum of the file is equal to the source one after transfer.");
    opt(opt_instrument, args, "instrument").arg("DEBUG STRING").help("Instrumentation for testing/debugging purposes. You neek to know the code to be able to use this option.");

    // positional parameters
    args.pos("source")      .help("source");
    args.pos("destination") .help("destination");
}

/*
 * tool_doit_prepare_request
 * 
 * Prepare the transfer request
 */
int tool_doit_prepare_request(int argc, char *argv[]){
    
    if(argc < 2){
        tool_error() << "No files specified" << std::endl;
        return EXIT_FAILURE;
    }
    if(argc > 2){
        tool_error() << "Too many files specified" << std::endl;
        return EXIT_FAILURE;
    }
    
    // Set the surl 
    g_tx_args.src_name = argv[0];
    if(g_tx_args.src_name.empty()){
        tool_error() << "Empty source name" << std::endl;
        return EXIT_FAILURE;
    }
    
    // Destination
    g_tx_args.dest_name = argv[1];
    if(g_tx_args.dest_name.empty()){
        tool_error() << "Empty destination name" << std::endl;
        return EXIT_FAILURE;
    }

    // If not set, deduce the endpoints from the surls
    if(true == g_tx_args.src_endpoint.empty()){
        try{
            g_tx_args.src_endpoint = get_endpoint_from_surl(g_tx_args.src_name,g_tx_args.src_version);
        } catch(const RuntimeError& exc){
            tool_error() << "cannot find SRM endpoint for SURL [" << g_tx_args.src_name << "]: " << exc.what() << std::endl;
            return EXIT_FAILURE;
        }
    }
    if(true == g_tx_args.dest_endpoint.empty()){
        try{
            g_tx_args.dest_endpoint = get_endpoint_from_surl(g_tx_args.dest_name,g_tx_args.dest_version);
        } catch(const RuntimeError& exc){
            tool_error() << "cannot find SRM endpoint for SURL [" << g_tx_args.dest_name << "]: " << exc.what() << std::endl;
            return EXIT_FAILURE;
        }
    }
    if(true == g_tx_args.src_version.empty()){
        g_tx_args.src_version = get_endpoint_version(g_tx_args.src_endpoint);
    }
    if(true == g_tx_args.dest_version.empty()){
        g_tx_args.dest_version = get_endpoint_version(g_tx_args.dest_endpoint);
    }
    
    // Print transfer paremeters
    print_transfer_parameters(g_tx_args);
    return EXIT_SUCCESS;
}
    
/**
 * tool_doit_transfer
 * Start the transfer 
 */
int tool_doit_transfer(){
    int result = EXIT_FAILURE;
    
    std::string prefix;
    if(0 != g_prefix){
        prefix = g_prefix;
    }
    TransferUrlCopy::Ptr tx;
    if(false == g_block_mode){
        try{
            tx = TransferUrlCopy::start(g_tx_args, prefix, false);
            // Print Transfer ID
            tool_info() << "Transfer Request ID: " << tx->id() << std::endl;
            result = EXIT_SUCCESS;
        } catch(const std::exception& exc){
            tool_error() << "Failed to start transfer: " << exc.what() << std::endl;
            result = EXIT_FAILURE;
        }
    } else {
        try{        
            // Run transfer in blocking mode
            tool_debug() << "Starting urlcopy transfer..." << std::endl;
            tx = TransferUrlCopy::start(g_tx_args, prefix, true);
            // Transfer is now completed (successfully or not)
            tool_info() << "Transfer request ID: " << tx->id() << std::endl;
            try{
                // Get transfer status
                TransferUrlCopyStat tx_stat = tx->status();
                if(tx_stat.status == UrlCopyStatus::COMPLETED){
                    if(tx_stat.error.category == UrlCopyError::SUCCESS){
                        result = EXIT_SUCCESS;
                        // Print Transfer details
                        print_transfer_details(tx_stat);
                        print_preparation_details(tx_stat);
                        print_finalization_details(tx_stat);
                    } else {
                        result = EXIT_FAILURE;
                        // Print Transfer Failure
                        print_transfer_failure(tx_stat);
                    }
                    tool_info() << "Request took " << (tx_stat.finish_time - tx_stat.start_time) << " seconds in total" << std::endl;
                }  else {
                    // This should not happen. Anyway call abort
                    throw LogicError("Transfer didn't complete in blocking mode and will be aborted");
                }
            } catch(const std::exception& exc){
                result = EXIT_FAILURE;
                // This should not happen. Anyway call abort
                tool_error() << "Transfer didn't complete in blocking mode and will be aborted" << std::endl;
                try{
                    tx->abort();
                }catch(...){}
            }
            // clean the log files
            try{
                tx->clean();    
            }catch(...){
                tool_error() << "Failed to clean transfer log file" << std::endl;
            }
        } catch(const std::exception& exc){
            tool_error() << "Failed to execute transfer: " << exc.what() << std::endl;
            result = EXIT_FAILURE;
        }
    }
    return result;
}

/*
 * tool_doit
 * 
 * Run the Get
 */
int tool_doit(int argc, char *argv[]){
    // Prepare a request
    int result = tool_doit_prepare_request(argc, argv);
    if(result == EXIT_SUCCESS){
        // start the transfer
        result = tool_doit_transfer();
    }
    return result;
}

