/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: gridftp helper methods 
 *   refactored from previous transfer-url-copy (Peter Kunszt) and lcg_util 
 *   (Jean-Philippe Baud)
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: globus-utility.cpp,v 1.5 2008/08/06 12:17:35 ptedesco Exp $
 *  Release: $Name:  $
 */

#include "globus-utility.h" 
#include "logger-utility.h" 
  
using namespace glite::data::transfer::urlcopy;

//------------------------------------------------------------------------------
// GlobusObject
//------------------------------------------------------------------------------

/**
 * Constructor
 */
GlobusObject::GlobusObject():m_object(0){
}

/**
 * Constructor
 */
GlobusObject::GlobusObject(globus_object_t * obj):m_object(obj){
}

/**
 * Constructor
 */
GlobusObject::GlobusObject(GlobusObject& obj):m_object(obj.release()){
}

/**
 * Destructor
 */
GlobusObject::~GlobusObject(){
    if(0 != m_object){
        globus_object_free(m_object);
    }
}

/**
 * Assignment Operator
 */
//GlobusObject& 
void GlobusObject::operator=(globus_object_t * obj){
    if(0 != m_object){
        globus_object_free(m_object);
    }
    m_object = obj;
}

/**
 * Assignment Operator
 */
//GlobusObject& 
void GlobusObject::operator=(GlobusObject& obj){
    if(0 != m_object){
        globus_object_free(m_object);
    }
    m_object = obj.release();
}

/**
 * Return the globus Object
 */
globus_object_t * GlobusObject::get() {
    return m_object;
}

/**
 * Release the Object the globus Object
 */
globus_object_t * GlobusObject::release() {
    globus_object_t * obj = m_object;
    m_object = 0; 
    return obj;
}

//------------------------------------------------------------------------------
// GridFtpMonitor
//------------------------------------------------------------------------------

/**
 * Constructor
 */
GridFtpMonitor::GridFtpMonitor():m_done(GLOBUS_FALSE),m_error(GLOBUS_FALSE), m_failed(GLOBUS_FALSE){
    // Initialize Mutex & condition
    globus_mutex_init(&m_mutex, 0);
    globus_cond_init(&m_cond, 0);
}

/**
 * Destructor
 */
GridFtpMonitor::~GridFtpMonitor(){
    // Destory Mutex & Condition
    globus_cond_destroy(&m_cond);
    globus_mutex_destroy(&m_mutex);
}

/**
 * Reset GridFtpMonitor
 */
void GridFtpMonitor::reset(){
    m_done  = GLOBUS_FALSE;
    m_error = GLOBUS_FALSE;
    m_failed = GLOBUS_FALSE;
}

/**
 * Get Done flag
 */
globus_bool_t GridFtpMonitor::isDone() const {
    return m_done;
}
/**
 * Set Done flag
 */
void GridFtpMonitor::done(globus_bool_t b){
    m_done = b;
}
/**
 * Get error flag
 */
globus_bool_t GridFtpMonitor::isFailed() const {
    return m_failed;
}
/**
 * Set error flag
 */
void GridFtpMonitor::failed(globus_bool_t b) {
    m_failed = b;
}

/**
 * Return error object
 */
GlobusObject& GridFtpMonitor::error() {
    return m_error;
}

/**
 * Return the Mutex
 */
globus_mutex_t& GridFtpMonitor::mutex() {
    return m_mutex;
}

/**
 * Return the Condition
 */
globus_cond_t& GridFtpMonitor::cond() {
    return m_cond;
}

//------------------------------------------------------------------------------
// GridFtpMonitor
//------------------------------------------------------------------------------

/**
 * GridFtpCheck
 */
GridFtpCheck::GridFtpCheck(GridFtpMonitor& m, gridftpcopy_callback_t *cb, void * user_arg):
    m_monitor(m),m_callback(cb),m_arg(user_arg),m_done(GLOBUS_FALSE){
    reset();
}

/**
 * Destructor
 */
GridFtpCheck::~GridFtpCheck(){
}

/**
 * Reset GridFtpCheck
 */
void GridFtpCheck::reset(){
    m_done                    = GLOBUS_FALSE;
    m_perf.ts.tv_sec          = 0;
    m_perf.ts.tv_nsec         = 0;
    m_perf.markers            = 0;
    m_perf.transferred_bytes  = 0;
    m_perf.avg_throughput     = 0;
    m_perf.instant_throughput = 0;
}

/**
 * Get Done flag
 */
globus_bool_t GridFtpCheck::isDone() const {
    return m_done;
}
/**
 * Set Done flag
 */
void GridFtpCheck::done(globus_bool_t b){
    m_done = b;
}
/**
 * Return Monitor object
 */
GridFtpMonitor& GridFtpCheck::monitor() {
    return m_monitor;
}

/**
 * call the user's callback
 */
int GridFtpCheck::callback() {
    int timeout = 0;
    if(0 != m_callback){
        try{
            timeout = (*m_callback)(m_arg,m_perf);
        }catch(...){}
    }
    return timeout;
}

/**
 * Return the transfer perf info
 */
PerfMarker& GridFtpCheck::perf() {
    return m_perf;
}

//------------------------------------------------------------------------------
// GridFtpClientHandle
//------------------------------------------------------------------------------

/**
 * Constructor
 */
GridFtpClientHandle::GridFtpClientHandle() : m_own_handle(true){
    globus_ftp_client_handleattr_init(&m_handleAttr);
    globus_ftp_client_handle_init(&m_handle, &m_handleAttr);
    globus_ftp_client_operationattr_init(&m_opAttr);
}

/**
 * Constructor
 */
GridFtpClientHandle::GridFtpClientHandle(globus_ftp_client_handle_t& handle) : 
    m_handle(handle),
    m_own_handle(false){
        
    globus_ftp_client_handleattr_init(&m_handleAttr);
    //globus_ftp_client_handle_init(&m_handle, &m_handleAttr);
    globus_ftp_client_operationattr_init(&m_opAttr);
}

/**
 * Destructor
 */
GridFtpClientHandle::~GridFtpClientHandle(){
    // Release handles
    globus_ftp_client_operationattr_destroy (&m_opAttr);
    if (m_own_handle){
        globus_ftp_client_handle_destroy(&m_handle);
    }
    globus_ftp_client_handleattr_destroy(&m_handleAttr);
}

/**
 * Return GridFtp Client Handle
 */
globus_ftp_client_handle_t& GridFtpClientHandle::handle() {
    return m_handle;
}

/**
 * Return GridFtp Operation Attributes
 */
globus_ftp_client_operationattr_t& GridFtpClientHandle::operationAttr() {
    return m_opAttr;
}

//------------------------------------------------------------------------------
// GridFtpCopyAttr
//------------------------------------------------------------------------------

/**
 * Constructor
 */
GridFtpCopyAttr::GridFtpCopyAttr(){
    globus_gass_copy_attr_init(&attr);
    ftp_attr = 0;
    gass_attr = 0;
}

/**
 * Destructor
 */
GridFtpCopyAttr::~GridFtpCopyAttr(){
    // Release handles
    if(0 != ftp_attr){
        globus_ftp_client_operationattr_destroy(&ftp_attr);
    }
    if(0 != gass_attr){
        globus_gass_transfer_requestattr_destroy(&gass_attr);
    }
}

//------------------------------------------------------------------------------
// GridFtpCopyHandle
//------------------------------------------------------------------------------

/**
 * Constructor
 */
GridFtpCopyHandle::GridFtpCopyHandle() : 
    m_ftpHandle(0){
        
    // bug #33528: must call globus_ftp_client_handleattr_set_cache_all (thanks 
    // to David Smith for finding this)
    globus_gass_copy_handleattr_t gass_copy_handleattr;
    globus_ftp_client_handleattr_t ftp_handleattr;

    globus_gass_copy_handleattr_init(&gass_copy_handleattr);
    globus_ftp_client_handleattr_init(&ftp_handleattr);

    globus_ftp_client_handleattr_set_cache_all(&ftp_handleattr, GLOBUS_TRUE);
    globus_gass_copy_handleattr_set_ftp_attr(&gass_copy_handleattr, &ftp_handleattr);
    globus_gass_copy_handle_init(&m_handle, &gass_copy_handleattr);

    globus_gass_copy_handleattr_destroy(&gass_copy_handleattr);
    globus_ftp_client_handleattr_destroy(&ftp_handleattr);
    
    m_ftpHandle = new GridFtpClientHandle(m_handle.ftp_handle);
    
}

/**
 * Destructor
 */
GridFtpCopyHandle::~GridFtpCopyHandle(){
    // Release handles
    globus_gass_copy_handle_destroy(&m_handle);
    if (m_ftpHandle){
        delete m_ftpHandle;
        m_ftpHandle = 0;
    }
}

/**
 * Return GridFtp Copy Handle
 */
globus_gass_copy_handle_t& GridFtpCopyHandle::handle() {
    return m_handle;
}

/**
 * Return GridFtp Copy Source Handle Attributes
 */
GridFtpCopyAttr& GridFtpCopyHandle::srcAttr() {
    return m_srcAttr;
}

/**
 * Return GridFtp Copy Destination Handle Attributes
 */
GridFtpCopyAttr& GridFtpCopyHandle::destAttr() {
    return m_destAttr;
}

//------------------------------------------------------------------------------
// GlobusActivation
//------------------------------------------------------------------------------

int GlobusActivation::s_activationCount = 0;

/**
 * Constructor
 */
GlobusActivation::GlobusActivation(){
    // initialization
    if(0 == s_activationCount){
        debug() << "activating globus modules";
        (void) globus_module_activate (GLOBUS_GASS_COPY_MODULE);
        (void) globus_module_activate (GLOBUS_FTP_CLIENT_MODULE);
        debug() << "globus modules activated";
    }
    ++s_activationCount;
}

/**
 * Destructor
 */
GlobusActivation::~GlobusActivation(){
    --s_activationCount;
    if(0 == s_activationCount){
        debug() << "deactivating globus modules";
        (void) globus_module_deactivate (GLOBUS_FTP_CLIENT_MODULE);
        (void) globus_module_deactivate (GLOBUS_GASS_COPY_MODULE);
        debug() << "globus modules deactivated";
    }
}

 
