/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: gridftp helper methods 
 *   refactored from previous transfer-url-copy (Peter Kunszt) and lcg_util 
 *   (Jean-Philippe Baud)
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: globus-utility.h,v 1.4 2008/03/11 08:24:53 ptedesco Exp $
 *  Release: $Name:  $
 */
#ifndef GLITE_DATA_TRANSFER_URL_COPY_GLOBUS_UTILITY_H_
#define GLITE_DATA_TRANSFER_URL_COPY_GLOBUS_UTILITY_H_

#include <boost/utility.hpp>
#include "PerfMarker.h"

#ifdef __cplusplus
extern "C"{
#endif //__cplusplus
// Fix Warning
#ifdef IOV_MAX
#undef IOV_MAX
#endif //IOV_MAX
#include "globus_gass_copy.h"
#ifdef __cplusplus
}
#endif //__cplusplus


namespace glite   {
namespace data    {
namespace transfer{
namespace urlcopy {

/**-----------------------------------------------------------------------------
 * GlobusObject class (used for RAII)
------------------------------------------------------------------------------*/    
class GlobusObject: boost::noncopyable {
public:    
    
    /**
     * Constructor
     */
    GlobusObject();
    
    /**
     * Constructor
     */
    GlobusObject(globus_object_t * obj);

    /**
     * Constructor
     */
    GlobusObject(GlobusObject& obj);

    /**
     * Destructor
     */
    ~GlobusObject();

    /**
     * Assignment Operator
     */
    //GlobusObject& 
    void operator=(globus_object_t * obj);

    /**
     * Assignment Operator
     */
    //GlobusObject& 
    void operator=(GlobusObject& obj);

    /**
     * Return the globus Object
     */
    globus_object_t * get();

    /**
     * Release the Object the globus Object
     */
    globus_object_t * release();
    
private:
    
    // The Globus Object
    globus_object_t * m_object;
};

/**-----------------------------------------------------------------------------
 * GridFtpMonitor class (used for RAII)
------------------------------------------------------------------------------*/    
class GridFtpMonitor: boost::noncopyable {
public:
    /**
     * Constructor
     */
    GridFtpMonitor();
    
    /**
     * Destructor
     */
    ~GridFtpMonitor();

    /**
     * Reset GridFtpMonitor
     */
    void reset();
    
    /**
     * Get Done flag
     */
    globus_bool_t isDone() const;
    
    /**
     * Set Done flag
     */
    void done(globus_bool_t b);
    
    /**
     * Get error flag
     */
    globus_bool_t isFailed() const;
    
    /**
     * Set error flag
     */
    void failed(globus_bool_t b);
    
    /**
     * Return error object
     */
    GlobusObject& error();
    
    /**
     * Return the Mutex
     */
    globus_mutex_t& mutex();
    /**
     * Return the Condition
     */
    globus_cond_t& cond();
    
private:
    // The Mutex
    globus_mutex_t m_mutex;
    // The Condition
    globus_cond_t  m_cond;
    // Done Flag
    volatile globus_bool_t m_done;
    // Error Flag
    volatile globus_bool_t m_failed;
    // Error Object
    GlobusObject m_error;
};

/**-----------------------------------------------------------------------------
 * GridFtpCheck class (used for RAII)
------------------------------------------------------------------------------*/    
class GridFtpCheck: boost::noncopyable {
public:
    /**
     * GridFtpCheck
     */
    GridFtpCheck(GridFtpMonitor& m, gridftpcopy_callback_t *cb, void * user_arg);
    
    /**
     * Destructor
     */
    ~GridFtpCheck();

    /**
     * Reset GridFtpCheck
     */
    void reset();
    
    /**
     * Get Done flag
     */
    globus_bool_t isDone() const;
    
    /**
     * Set Done flag
     */
    void done(globus_bool_t b);
    
    /**
     * Return Monitor object
     */
    GridFtpMonitor& monitor();
    
    /**
     * call the user's callback
     */
    int callback();

    /**
     * Return the transfer perf info
     */
    PerfMarker& perf();
    
private:
    // The Mutex
    GridFtpMonitor& m_monitor;
    // The Callback
    gridftpcopy_callback_t* m_callback;
    // The Callback argument
    void * m_arg;
    // Done Flag
    volatile globus_bool_t m_done;
    // The transfer perf info
    PerfMarker m_perf;
};

/**-----------------------------------------------------------------------------
 * GridFtpClientHandle class (used for RAII)
------------------------------------------------------------------------------*/    
class GridFtpClientHandle: boost::noncopyable {
public:    
    
    /**
     * Constructor
     */
    GridFtpClientHandle();
    
    /**
     * Destructor
     */
    ~GridFtpClientHandle();

    /**
     * Return GridFtp Client Handle
     */
    globus_ftp_client_handle_t& handle();
    
    /**
     * Return GridFtp Operation Attributes
     */
    globus_ftp_client_operationattr_t& operationAttr();

    /**
     * Constructor
     */
    GridFtpClientHandle(globus_ftp_client_handle_t& handle);

private:
    
    globus_ftp_client_handle_t        m_handle;
    globus_ftp_client_handleattr_t    m_handleAttr;
    globus_ftp_client_operationattr_t m_opAttr;
    bool m_own_handle;
};


/**-----------------------------------------------------------------------------
 * GridFtpCopyAttributes class (used for RAII)
------------------------------------------------------------------------------*/    
class GridFtpCopyAttr: boost::noncopyable {
public:
    /**
     * Constructor
     */
    GridFtpCopyAttr();

    /**
     * Destructor
     */
    ~GridFtpCopyAttr();

public:
    globus_gass_copy_attr_t            attr;
    // FTP Attributes
    globus_ftp_client_operationattr_t  ftp_attr;
    globus_ftp_control_parallelism_t   parallelism;
    globus_ftp_control_tcpbuffer_t     tcp_buffer;
    globus_ftp_control_dcau_t          dcau;
    // GASS Attributes
    globus_gass_transfer_requestattr_t gass_attr;
};

/**-----------------------------------------------------------------------------
 * GridFtpCopyHandle class (used for RAII)
------------------------------------------------------------------------------*/    
class GridFtpCopyHandle: boost::noncopyable {
public:    
    
    /**
     * Constructor
     */
    GridFtpCopyHandle();

    /**
     * Destructor
     */
    ~GridFtpCopyHandle();

    /**
     * Return GridFtp Copy Handle
     */
    globus_gass_copy_handle_t& handle();
    /**
     * Return GridFtp Copy Source Handle Attributes
     */
    GridFtpCopyAttr& srcAttr();
    /**
     * Return GridFtp Copy Destination Handle Attributes
     */
    GridFtpCopyAttr& destAttr();
    
    /**
     * get the gridftp handle
     * @return GridFtpClientHandle object
     */
    GridFtpClientHandle& ftpHandle(){
        return (*m_ftpHandle);
    }

private:
    globus_gass_copy_handle_t m_handle;
    GridFtpCopyAttr           m_srcAttr;
    GridFtpCopyAttr           m_destAttr;
    // ftp handle
    GridFtpClientHandle * m_ftpHandle;
};


/**-----------------------------------------------------------------------------
 * GlobusActivation class (used for RAII)
------------------------------------------------------------------------------*/    
class GlobusActivation: boost::noncopyable {
public:
    /**
     * Constructor
     */
    GlobusActivation();
    
    /**
     * Destructor
     */
    ~GlobusActivation();
    
private:
    
    static int s_activationCount;
};

} // end namespace urlcopy
} // end namespace transfer
} // end namespace data
} // end namespace glite

#endif // GLITE_DATA_TRANSFER_URL_COPY_GLOBUS_UTILITY_H_



