/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: gridftp helper methods
 *   refactored from previous transfer-url-copy (Peter Kunszt) and lcg_util
 *   (Jean-Philippe Baud)
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: gridftp-utility.cpp,v 1.25 2009/12/03 17:21:44 molnarzs Exp $
 *  Release: $Name:  $
 */
#include "gridftp-utility.h"
#include "sighandler-utility.h"
#include "logger-utility.h"
#include "Checksum.h"
#include <glite/data/transfer/urlcopy/UrlCopyError.h>
#include <glite/data/agents/glite-types.h>
#include <glite/data/agents/AgentExceptions.h>
#include <boost/utility.hpp>
#include <map>
#include <string>
#include "GridFtpPatterns.h"
#include <glite/data/config/service/SysLogDaemon.h>
#include "StatCommon.h" // for CHECKSUM_VALUE_LEN
#include "globus-utility.h"
#include <unistd.h> 
#include <iostream>
#include <fstream>
#include <algorithm>
#include <string>


#ifdef __cplusplus
extern "C" {
#endif //__cplusplus
    // Fix Warning
#ifdef IOV_MAX
#undef IOV_MAX
#endif //IOV_MAX
#include "gridftp-ifce.h"
#ifdef __cplusplus
}
#endif //__cplusplus

using namespace glite::data::transfer::urlcopy;
using namespace glite::config;
using glite::data::agents::XmlPatternsList;
using glite::data::agents::longlong;
using glite::data::agents::RuntimeError;
using glite::data::agents::PosixError;


// Constants
static const int COPY_CANCEL_TIMEOUT = 60;

// Maximum timeout for syncronous operations (should not be greater than there
// MIN_REFRESH_DELAY constant defined in transfer-utility.cpp)
static const int MAX_SYNCH_TIMEOUT = 180;

// -----------------------------------------------------------------------------
// Helper Methods
// -----------------------------------------------------------------------------

/*
 * str_from_error
 *
 * Return a string from a Globus Error object
 */
std::string str_from_error(globus_object_t * errobj) throw () {
    std::string msg;
    if (0 != errobj) {
#ifdef GLOBUS_ERROR_TYPE_MULTIPLE
        char * p = globus_error_print_friendly(errobj);
#else
        char * p = globus_object_printable_to_string(errobj);
#endif
        if (0 != p) {
            msg = p;
            globus_libc_free(p);
        }
    }
    // add some explanations
    if (std::string::npos != msg.find("an end-of-file was reached")) {
        // this error is probably returned also when the destination disk is full
        msg.append(" (possibly the destination disk is full)");
    }
    return msg;
}

/**
 * category_from_errno
 * Return the category from an error number
 */
UrlCopyError::Category category_from_errno(int err) throw () {
    UrlCopyError::Category c;

    switch (err) {
        case 0:
            c = UrlCopyError::SUCCESS;
            break;
        case ENOENT:
            c = UrlCopyError::ERROR_INVALID_PATH;
            break;
        case EACCES:
            c = UrlCopyError::ERROR_PERMISSION;
            break;
        case EEXIST:
            c = UrlCopyError::ERROR_FILE_EXISTS;
            break;
        case ETIMEDOUT:
            c = UrlCopyError::ERROR_TRANSFER_TIMEOUT;
            break;
        case ECANCELED:
            c = UrlCopyError::ERROR_ABORTED;
            break;
        default:
            c = UrlCopyError::ERROR_GRIDFTP;
            break;
    }
    return c;
}

/*
 * timeout_callback
 *
 * Callback used to evaluate if the transfer timed out
 */
void timeout_callback(void * user_arg) {
    // Get Check object

    if (user_arg) {
        GridFtpCheck * check = reinterpret_cast<GridFtpCheck *> (user_arg);

        if (check) {
            GridFtpMonitor& monitor = check->monitor();


            // Lock Mutex
            globus_mutex_lock(&(monitor.mutex()));
            {
                // set the delay
                globus_reltime_t delay;
                GlobusTimeReltimeSet(delay, 1, 0);

                // Call user callback
                int istimeout = check->callback();
                if (0 != istimeout) {
                    // Use defined Time out
                    TransferState::instance().timeout();
                }
                if ((GLOBUS_FALSE == monitor.isDone()) &&
                        (false == TransferState::instance().timedout()) &&
                        (false == TransferState::instance().interrupted())) {
                    // Register the callback again
                    globus_callback_register_oneshot(0, &delay, timeout_callback, check);
                } else {
                    // Check completed
                    check->done(GLOBUS_TRUE);
                }
            }
            // Signal Condition
            globus_cond_broadcast(&(monitor.cond()));
            globus_mutex_unlock(&(monitor.mutex()));
        }
    }
}

/*
 * ggc_callback
 *
 * callback used to notify that a gridftp copy has been cancelled
 */
void ggc_callback(void *callback_arg, globus_gass_copy_handle_t *handle, globus_object_t *error) {


    if (callback_arg) {
        // Get Monitor
        GridFtpMonitor * monitor = reinterpret_cast<GridFtpMonitor*> (callback_arg);
        if (monitor) {
            // Acquire Lock 
            globus_mutex_lock(&(monitor->mutex()));

            if (error != GLOBUS_SUCCESS) {
                monitor->failed(GLOBUS_TRUE);
                monitor->error() = globus_object_copy(error);
            }

            monitor->done(GLOBUS_TRUE);

            // Signal condition
            globus_cond_broadcast(&(monitor->cond()));
            globus_mutex_unlock(&(monitor->mutex()));
        }
    }
}

/*
 * gridftp_copy_performance
 *
 * Callback for performance markers
 */
void gridftp_copy_performance(void * user_arg, globus_gass_copy_handle_t * handle, globus_off_t total_bytes, float instantaneous_throughput, float avg_throughput) {
    GridFtpCheck * check = reinterpret_cast<GridFtpCheck*> (user_arg);
    if (0 != check) {
        check->perf().ts.tv_sec = time(0);
        check->perf().ts.tv_nsec = 0;
        check->perf().transferred_bytes = total_bytes;
        check->perf().instant_throughput = instantaneous_throughput;
        check->perf().avg_throughput = avg_throughput;
        ++(check->perf().markers);
    }
}

/*
 * gridftp_copy_wait
 *
 * wait until the condition is set. Return the error in case of failure
 */
int gridftp_copy_wait(void* _monitor, void* _check, globus_gass_copy_handle_t *gfhp, int const timeout, void * _iupdater) {

    int error = 0;
    GridFtpMonitor* monitor = (GridFtpMonitor*) _monitor;
    GridFtpCheck* check = (GridFtpCheck*) _check;
    RefreshTimeUpdater::IUpdateRefreshTime * iupdater = (RefreshTimeUpdater::IUpdateRefreshTime *) _iupdater;
    if (monitor && check && iupdater) {
        // Lock the mutex
        globus_mutex_lock(&(monitor->mutex()));

        // Register timeout callback
        globus_callback_register_oneshot(0, 0, timeout_callback, check);
        if (timeout > 0) {
            // Use timeout
            struct timespec ts;
            ts.tv_sec = time(0) + timeout;
            ts.tv_nsec = 0;

            while ((GLOBUS_FALSE == monitor->isDone()) &&
                    (0 == error) &&
                    (false == TransferState::instance().timedout())) {

                if ((true == TransferState::instance().interrupted()) &&
                        (false == TransferState::instance().interruptProcessed())) {
                    // Cancel transfer
                    warn() << "received interrupt: canceling transfer";
                    globus_gass_copy_cancel(gfhp, &ggc_callback, monitor);
                    TransferState::instance().processInterrupt();
                }

                // wait for condition	
                error = globus_cond_timedwait(&(monitor->cond()), &(monitor->mutex()), &ts);
                int istimeout = check->callback();

                if (error == 110 && istimeout == 0) {
                    ts.tv_sec = time(0) + 1000;
                    error = 0;
                }

                if (EINTR == error) {
                    error = 0;
                }
            }
        } else {
            // Do not use timeout
            while ((GLOBUS_FALSE == monitor->isDone()) &&
                    (0 == error) &&
                    (false == TransferState::instance().timedout())) {

                if ((true == TransferState::instance().interrupted()) &&
                        (false == TransferState::instance().interruptProcessed())) {
                    // Cancel transfer
                    warn() << "received interrupt: canceling transfer";
                    globus_gass_copy_cancel(gfhp, &ggc_callback, monitor);
                    TransferState::instance().processInterrupt();
                }
                // wait for condition
                error = globus_cond_wait(&(monitor->cond()), &(monitor->mutex()));
                if (EINTR == error) {
                    // Continue
                    error = 0;
                }
            }
        }
        if (error == ETIMEDOUT) {
            // Set that the transfer is timed out
            TransferState::instance().timeout();
        }

        iupdater->UpdateRefreshTime();

        // Wait transfer to complete
        if (GLOBUS_FALSE == monitor->isDone()) {
            warn() << "canceling transfer";
            int ret;
            globus_gass_copy_cancel(gfhp, &ggc_callback, monitor);
            // wait for the cancel to complete before returning to user
            do {
                struct timespec ts;
                ts.tv_sec = time(0) + COPY_CANCEL_TIMEOUT;
                ts.tv_nsec = 0;
                ret = globus_cond_timedwait(&(monitor->cond()), &(monitor->mutex()), &ts);
                if (EINTR == ret) {
                    // Continue
                    ret = 0;
                }
            } while ((GLOBUS_FALSE == monitor->isDone()) && (0 == ret));
            if ((0 != ret) && (0 == error)) {
                error = ret;
            }
        }

        iupdater->UpdateRefreshTime();

        // Wait timeout_callback to complete
        if (GLOBUS_FALSE == check->isDone()) {
            info() << "waiting timeout_callback to complete";
            int ret;
            do {
                ret = globus_cond_wait(&(monitor->cond()), &(monitor->mutex()));
                if (EINTR == ret) {
                    // Continue
                    ret = 0;
                }
            } while ((GLOBUS_FALSE == check->isDone()) && (0 == ret));
            if ((0 != ret) && (0 == error)) {
                error = ret;
            }
        }
        // Unlock Mutex
        globus_mutex_unlock(&(monitor->mutex()));
    }
    return error;
}



// for system log
static const std::string GriftpCall = "gridftpCall";
static const std::string MethodName = "method";
static const std::string Turl = "turl";
static const std::string SourceTurl = "sourceTurl";
static const std::string DestTurl = "destTurl";



namespace glite {
    namespace data {
        namespace transfer {
            namespace urlcopy {

                /**
                 * gridftp_getfilesize
                 * Get the size of a file using gsiftp
                 */
                longlong gridftp_getfilesize(
                        const std::string& name,
                        UrlCopyError::Category& c,
                        std::string& m,
                        int timeout) /* throw () */ {

                    std::string posixError("");


                    if (timeout > MAX_SYNCH_TIMEOUT) {
                        timeout = MAX_SYNCH_TIMEOUT;
                    }

                    int errbufsz = ERROR_BUFFER_SIZE;
                    char errbuf[ERROR_BUFFER_SIZE] = {0};
                    globus_off_t size = -1;

                    c = UrlCopyError::ERROR_INVALID_SIZE;

                    try {

                        globus_result_t gresult = dmutils_gridftp_get_filesize(
                                name.c_str(),
                                &size,
                                errbuf,
                                errbufsz,
                                timeout);

                        if (gresult != GLOBUS_TRUE) {
                            SysLogDaemon(true).event(GriftpCall).success(false).add(MethodName, "globus_ftp_client_size").add(Turl, name).error();

                            if (errno != 0) {
                                posixError = "globus_ftp_client_size failed";
                            }
                            throw RuntimeError(errbuf);
                        }

                        SysLogDaemon(true).event(GriftpCall).success(true).add(MethodName, "globus_ftp_client_size").add(Turl, name).info();
                        //went well, return success
                        c = UrlCopyError::SUCCESS;
                    } catch (const RuntimeError& exc) {
                        // Scan error string and find the correct category
                        m = exc.what();
                        if (m.length() > 0) {
                            replace(m.begin(), m.end(), '\n', ' ');
                            error() << m;
                            UrlCopyError::Scope scope;
                            GridFtpPatterns::instance().translateGridFtpError(m, c, scope);
                        } else if (errno != 0 && m.length() == 0) {
                            m = posixError;
                            c = category_from_errno(errno);
                        } else {
                            error() << "globus_ftp_client_size failed";
                            m = "globus_ftp_client_size failed";
                            c = UrlCopyError::ERROR_GRIDFTP;
                        }
                    }
                    return size;
                }

                void gridftp_check_exists(
                        const std::string& turl,
                        int timeout) {
                    int errbufsz = ERROR_BUFFER_SIZE;
                    char errbuf[ERROR_BUFFER_SIZE] = {0};
                    std::string posixError("");
                    std::string message("");


                    info() << "check if dest TURL exists " << turl;
                    globus_result_t result = check_if_exists_delete(
                            turl.c_str(),
                            errbuf,
                            errbufsz,
                            timeout);

                    if (GLOBUS_TRUE != result) {
                        warn() << "gsiftp dest file exists but cannot be deleted before overwrite:" << errbuf;
                    }


                }

                /**
                 * gridftp_copy
                 * Copy a file using gsiftp
                 */
                void gridftp_copy(
                        const std::string& src_file,
                        const std::string& dest_file,
                        UrlCopyError::Category& c,
                        UrlCopyError::Scope& s,
                        std::string& m,
                        int const nbstreams,
                        int const tcpbs,
                        int const block_size,
                        int const timeout,
                        gridftpcopy_callback_t cp_callback,
                        void * user_arg,
                        RefreshTimeUpdater::IUpdateRefreshTime * iupdater,
                        bool gsiftpOverwrite) {

                    std::string posixError("");

                    globus_result_t gresult;
                    int errbufsz = ERROR_BUFFER_SIZE;
                    char errbuf[ERROR_BUFFER_SIZE] = {0};
                    TransferState::instance().reset();

                    try {
                        if (gsiftpOverwrite == true) {
                            gridftp_check_exists(
                                    dest_file.c_str(),
                                    timeout);
                        }

                        // Initialize Monitor (for conditional wait)
                        GridFtpMonitor monitor;

                        // Initialize Check
                        GridFtpCheck check(monitor, &cp_callback, user_arg);

                        s = UrlCopyError::SCOPE_TRANSFER;

                        // Reset Monitor & Check
                        monitor.reset();
                        check.reset();

                        info() << "starting url-to-url copy transfer";
                        gresult = dmutils_gridftp_copy(
                                src_file.c_str(),
                                dest_file.c_str(),
                                errbuf,
                                errbufsz,
                                timeout,
                                nbstreams,
                                tcpbs,
                                block_size,
                                0,
                                NULL,
                                &ggc_callback,
                                &monitor,
                                gridftp_copy_performance,
                                &check,
                                gridftp_copy_wait,
                                (void *) iupdater);

                        std::string log_id = LogId::instance().get_log_id();
                        SysLogDaemon log(true);
                        log.event(GriftpCall)
                                .add(LogId::KeyName, log_id)
                                .add(MethodName, "globus_gass_copy_register_url_to_url");

                        if (GLOBUS_TRUE != gresult) {
                            log.success(false).error();
                            SysLogDaemon(true)
                                    .add(LogId::KeyName, log_id)
                                    .add(SourceTurl, src_file)
                                    .info();
                            SysLogDaemon(true)
                                    .add(LogId::KeyName, log_id)
                                    .add(DestTurl, dest_file)
                                    .info();
                            monitor.failed(GLOBUS_TRUE);
                            if (0 != errno) {
                                warn() << "transfer failed with error [" << errno << "]";
                            }

                            if (true == TransferState::instance().timedout()) {
                                warn() << "transfer timed out";
                                // Transfer timed out
                                posixError = "globus_gass_copy_register_url_to_url transfer timed out";
                                errno = ETIMEDOUT;
                                throw RuntimeError(posixError);
                            }

                            throw RuntimeError(errbuf);

                        }
                        // Check if the transfer has been aborted or if it timed out
                        if (true == TransferState::instance().interruptProcessed()) {
                            warn() << "transfer aborted";
                            // Transfer aborted
                            posixError = "globus_gass_copy_register_url_to_url transfer aborted";
                            errno = ECANCELED;
                            throw RuntimeError(posixError);
                        } else if (GLOBUS_TRUE == monitor.isFailed()) {
                            // Transfer failed
                            std::string reason = str_from_error(monitor.error().get());
                            replace(reason.begin(), reason.end(), '\n', ' ');
                            throw RuntimeError(reason);

                        }

                        // Transfer completed
                        log.success(true).info();
                        SysLogDaemon(true)
                                .add(LogId::KeyName, log_id)
                                .add(SourceTurl, src_file)
                                .info();
                        SysLogDaemon(true)
                                .add(LogId::KeyName, log_id)
                                .add(DestTurl, dest_file)
                                .info();
                        c = UrlCopyError::SUCCESS;

                    } catch (const RuntimeError& exc) {
                        m = exc.what();
                        if (m.length() > 0) {
                            GridFtpPatterns::instance().translateGridFtpError(m, c, s);
                        } else if (errno != 0 && m.length() == 0) {
                            m = posixError;
                            c = category_from_errno(errno);
                        } else {
                            m = "globus_gass_copy_register_url_to_url transfer failed";
                            c = category_from_errno(errno);
                        }
                    }

                }

                /**
                 * gridftp_checksum
                 * Get a file checksum
                 * @param turl [IN] the TURL of the file
                 * @param algorithm [IN] the checksum algorithm
                 * @param timeout [IN] timeout for the operation
                 * @param category [OUT] error category
                 * @param message [OUT] error message
                 * @param iupdater [IN] pointer to IUpdateRefreshTime
                 * @return file checksum if successful, empty string otherwise
                 */
                std::string gridftp_checksum(
                        const std::string& turl,
                        const std::string& algorithm,
                        int timeout,
                        UrlCopyError::Category& category,
                        std::string& message,
                        RefreshTimeUpdater::IUpdateRefreshTime * iupdater) {

                    std::string ret("");
                    std::string posixError("");
                    int errbufsz = ERROR_BUFFER_SIZE;
                    char errbuf[ERROR_BUFFER_SIZE] = {0};

                    try {
                        info() << "getting checksum for TURL " << turl;
                        std::string log_id = LogId::instance().get_log_id();
                        SysLogDaemon log(true);
                        log.event(GriftpCall)
                                .add(LogId::KeyName, log_id)
                                .add(MethodName, "globus_ftp_client_cksm");

                        char checksum[CHECKSUM_VALUE_LEN];

                        globus_result_t result = dmutils_gridftp_checksum(
                                turl.c_str(),
                                checksum,
                                algorithm.c_str(),
                                errbuf,
                                errbufsz,
                                timeout);

                        if (GLOBUS_TRUE != result) {
                            log.success(false).error();
                            SysLogDaemon(true)
                                    .add(LogId::KeyName, log_id)
                                    .add(Turl, turl)
                                    .info();
                            if (0 != errno) {
                                posixError = "globus_ftp_client_cksm failed";
                            }

                            throw RuntimeError(errbuf);
                        }

                        log.success(true).info();
                        SysLogDaemon(true)
                                .add(LogId::KeyName, log_id)
                                .add(Turl, turl)
                                .info();
                        ret = checksum;
                        info() << "got checksum: " << ret;
                        category = UrlCopyError::SUCCESS;
                    } catch (const RuntimeError& exc) {
                        // Scan error string and find the correct category
                        message = exc.what();
                        if (message.length() > 0) {
                            error() << message;
                            UrlCopyError::Scope scope;
                            GridFtpPatterns::instance().translateGridFtpError(message, category, scope);
                        } else if (errno != 0 && message.length() == 0) {
                            message = posixError;
                            category = category_from_errno(errno);
                        } else {
                            error() << "globus_ftp_client_cksm failed";
                            message = "globus_ftp_client_cksm failed";
                            category = UrlCopyError::ERROR_GRIDFTP;
                        }
                    }
                    return ret;
                }

                /**
                 * gridftp_rmfile
                 * Delete a file
                 */
                void gridftp_rmfile(
                        const std::string& turl,
                        int timeout,
                        UrlCopyError::Category& category,
                        std::string& message,
                        RefreshTimeUpdater::IUpdateRefreshTime * iupdater) {


                    int errbufsz = ERROR_BUFFER_SIZE;
                    char errbuf[ERROR_BUFFER_SIZE] = {0};
                    std::string posixError("");

                    try {
                        info() << "removing TURL " << turl;
                        std::string log_id = LogId::instance().get_log_id();
                        SysLogDaemon log(true);
                        log.event(GriftpCall)
                                .add(LogId::KeyName, log_id)
                                .add(MethodName, "globus_ftp_client_delete");

                        globus_result_t result = dmutils_gridftp_delete(
                                turl.c_str(),
                                errbuf,
                                errbufsz,
                                timeout);

                        if (GLOBUS_TRUE != result) {
                            log.success(false).error();
                            SysLogDaemon(true)
                                    .add(LogId::KeyName, log_id)
                                    .add(Turl, turl)
                                    .info();

                            if (0 != errno) {
                                posixError = "globus_ftp_client_delete failed";

                            }
                            throw RuntimeError(errbuf);
                        }

                        log.success(true).info();
                        SysLogDaemon(true)
                                .add(LogId::KeyName, log_id)
                                .add(Turl, turl)
                                .info();
                        category = UrlCopyError::SUCCESS;
                    } catch (const RuntimeError& exc) {
                        // Scan error string and find the correct category
                        message = exc.what();
                        if (message.length() > 0) {
                            error() << message;
                            UrlCopyError::Scope scope;
                            GridFtpPatterns::instance().translateGridFtpError(message, category, scope);
                        } else if (errno != 0 && message.length() == 0) {
                            message = posixError;
                            category = category_from_errno(errno);
                        } else {
                            error() << "globus_ftp_client_delete failed";
                            message = "globus_ftp_client_delete failed";
                            category = UrlCopyError::ERROR_GRIDFTP;
                        }

                    }
                }

            } // end namespace urlcopy
        } // end namespace transfer
    } // end namespace data
} // end namespace glite
