/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Logger helper methods
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>,
 *           Paolo Tedesco <paolo.tedesco@cern.ch>,
 *           Zsolt Molnar <Zsolt.Molnar@cern.ch - www.zsoltmolnar.hu>
 *  Version info: $Id: logger-utility.cpp,v 1.17 2009/12/11 17:55:51 molnarzs Exp $
 *  Release: $Name:  $
 */

#include "logger-utility.h"
#include "TransferConfig.h"

#include <log4cpp/FileAppender.hh>
#include <log4cpp/OstreamAppender.hh>
#include <log4cpp/PatternLayout.hh>

#include <boost/lexical_cast.hpp>
#include <boost/filesystem/path.hpp>

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <unistd.h>
#include <execinfo.h>

#include <glite/data/config/service/SysLogDaemon.h>
#include <glite/data/config/service/SysLogInfo.h>

#include <glite/data/srm/util/PrepareToGet.h>
#include <glite/data/srm/util/PrepareToPut.h>
#include <glite/data/srm/util/SrmCopy.h>

using namespace log4cpp;
using namespace glite::config;
using namespace glite::data::srm::util;

namespace {

const char * const PATTERN_LAYOUT  = "%d [%-5p] - %m%n";

/**
 * Logger
 * singleton class to handle logging initialization and finalization.
 */
class Logger{
public:

    /**
     * instance
     * @return singleton instance
     */
    static Logger& instance();

    /**
     * initialize
     * initialize logging (create appenders)
     * @param logFileName [IN] log file name
     * @param priority [IN] log priority
     * @param enableConsole [IN] if true, enable console logging
     * @param syslogDisable [IN] true: disable logging to syslog
     * @param syslogIdent identification string for syslog
     * @param syslogFacility Facility number for syslog
     * @param syslogFacilityName Facility name for syslog
     */
    void initialize(
        const char *    logFileName,
        Priority::Value priority,
        bool            enableConsole,
        const char *    exec_name,
        const bool syslogDisable,
        const std::string& syslogIdent,
        const int syslogFacility,
        const std::string& syslogFacilityName);

    /**
     * Get the wrapped Category object
     * @return Category
     */
    Category& category(){
        return m_category;
    }

    /**
     * destructor
     */
    ~Logger();

private:

    /**
     * constructor
     */
    Logger();

    // Category
    Category& m_category;
};

Logger& Logger::instance(){
    static Logger s_instance;
    return s_instance;
}

/**
 * initialize
 * initialize logging (create appenders)
 * @param logFileName [IN] log file name
 * @param priority [IN] log priority
 * @param enableConsole [IN] if true, enable console logging
 */
void Logger::initialize(
    const char *    logFileName,
    Priority::Value priority,
    bool            enableConsole,
    const char *    exec_name,
    const bool syslogDisable,
    const std::string& syslogIdent,
    const int syslogFacility,
    const std::string& syslogFacilityName)
{
    // Set Log Level
    m_category.setPriority(priority);

    try {
        m_category.removeAllAppenders();
        // create appenders
        if (enableConsole){
            m_category.addAppender(new OstreamAppender("ConsoleAppender",&std::cout));
        }
        // file appender
        if (logFileName){
            m_category.addAppender(new FileAppender("FileAppender",logFileName));
        }
        // set layout for each appender
        AppenderSet appenders = m_category.getAllAppenders();
        for (AppenderSet::iterator it = appenders.begin();it != appenders.end();++it){
            Appender * appender = (*it);
            PatternLayout * layout = new PatternLayout();
            layout->setConversionPattern(PATTERN_LAYOUT);
            appender->setLayout(layout);
        }
        // re-configure syslog appenders
        SysLogInfo::instance().reconfigure(
            exec_name, syslogIdent, syslogFacility, syslogFacilityName);

        if (!syslogDisable)
            SysLogDaemon::EnableSystemLog();

    } catch (const std::exception& exc){
        Category::getRoot().error("Exception initializing logger : %s",exc.what());
    }
}

/**
 * constructor
 */
Logger::Logger() : m_category(Category::getInstance("transfer-url-copy")){
}

/**
 * destructor
 */
Logger::~Logger(){
    // finalize logging
    Category::shutdown();
}


} // End global namespace

namespace glite   {
namespace data    {
namespace transfer{
namespace urlcopy {

std::string time_to_string(time_t tt){
    std::string s;
    if((0 == tt) || ((time_t)-1 == tt)){
        s = "(unset)";
    } else {
        char ts[32];
        struct tm * tm = gmtime(&tt);
        strftime(ts,32,"%Y-%m-%d %H:%M:%S +00:00",tm);
        s = ts;
    }
    return s;
}


/**
 * Setup Logger
 */
void log_setup(
    const char * name,
    LogLevel logLevel,
    bool enableConsole,
    const char * exec_name,
    const bool syslogDisable,
    const std::string& syslogIdent,
    const int syslogFacility,
    const std::string& syslogFacilityName) throw()
{
    Priority::Value priority = Priority::INFO;
    if (LOGLEVEL_DEBUG == logLevel){
        priority = Priority::DEBUG;
    } else if (LOGLEVEL_INFO == logLevel) {
        priority = Priority::INFO;
    } else if (LOGLEVEL_WARN == logLevel) {
        priority = Priority::WARN;
    } else if (LOGLEVEL_ERROR == logLevel) {
        priority = Priority::ERROR;
    }

    std::string logFileName;
    const char * logFileName_cstr = NULL;

    if (name) {
        logFileName = TransferConfig::instance().repository() + name + TransferConfig::instance().extLog();
        logFileName_cstr = logFileName.c_str();
    }

    Logger::instance().initialize(
        logFileName_cstr, priority, enableConsole, exec_name,
        syslogDisable, syslogIdent, syslogFacility, syslogFacilityName);
}

/**
 * get the debug stream
 */
log4cpp::CategoryStream debug(){
    return Logger::instance().category().debugStream();
}

/**
 * get the info stream
 */
log4cpp::CategoryStream info(){
    return Logger::instance().category().infoStream();
}

/**
 * get the warn stream
 */
log4cpp::CategoryStream warn(){
    return Logger::instance().category().warnStream();
}

/**
 * get the error stream
 */
log4cpp::CategoryStream error(){
    return Logger::instance().category().errorStream();
}

/**
 * log_stack
 * log program stack as warnings
 */
void log_stack(){
    const int stack_size = 25;
    void * array[stack_size];
    int nSize = backtrace(array, stack_size);
    char ** symbols = backtrace_symbols(array, nSize);
    for (int i = 0; i < nSize; ++i){
        warn() << symbols[i];
    }
    free(symbols);
}

//------------------------------------------------------------------------------
// SignalLogger
//------------------------------------------------------------------------------

/**
 * registerSignal
 * register a handler to log given signal
 * @param signum [IN] signal number
 * @param signame [IN] signal name
 */
void SignalLogger::registerSignal(const int signum,const std::string& signame){
    SignalInfoMap::const_iterator it = m_map.find(signum);
    if (m_map.end() == it){
        m_map.insert(std::make_pair<int,SignalInfo *>(signum,new SignalInfo(signum,signame)));
    }
}

/**
 * destructor
 */
SignalLogger::~SignalLogger(){
    for (SignalInfoMap::iterator it = m_map.begin();it != m_map.end();++it){
        SignalInfo * info = it->second;
        delete info;
    }
    m_map.clear();
}

/**
 * constructor
 * @param signum [IN] signal number
 * @param signame [IN] signal name
 */
SignalLogger::SignalInfo::SignalInfo(int signum,const std::string& signame) :
    m_signum(signum),
    m_signame(signame),
    m_set(true){
    struct sigaction sa;
    sa.sa_handler = SignalLogger::handleSignal;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    sigaction(m_signum, &sa, &m_old);
}

/**
 * deregister
 * deregister handler
 */
void SignalLogger::SignalInfo::deregister(){
    if (m_set){
        m_set = false;
        sigaction(m_signum, &m_old, 0);
    }
}

/**
 * destructor
 */
SignalLogger::SignalInfo::~SignalInfo(){
    deregister();
}

/**
 * handleSignal
 * actual signal handler function (registered by SignalInfo objects)
 */
void SignalLogger::handleSignal(int signum){
    instance().logSignal(signum);
}

/**
 * logSignal
 * actual signal logging
 */
void SignalLogger::logSignal(int signum){
    SignalInfoMap::iterator it = m_map.find(signum);
    if (m_map.end() != it){
        SignalInfo * info = it->second;
        info->deregister();
        warn () << "received signal " << info->signame();
        log_stack();
    }
}

//------------------------------------------------------------------------------
// LogId
//------------------------------------------------------------------------------

const std::string LogId::KeyName = "logId";

LogId& LogId::instance(){
    static LogId s_instance;
    return s_instance;
}
std::string LogId::get_log_id(){
    ++m_id;
    std::stringstream ss;
    ss << m_id;
    return ss.str();
}
LogId::LogId() : m_id(0){
}

//------------------------------------------------------------------------------
// Requests
//------------------------------------------------------------------------------

namespace {
std::string _format_wait_time(const unsigned long t)
{
	return ((unsigned long)-1) == t ?
				"UNKNOWN (-1)" :
				boost::lexical_cast<std::string>(t);
}

} // namespace

/**
 * log_request
 * log a PrepareToGet Request
 */
void log_request(const PrepareToGet& request) /*throw ()*/{
    try{
        // Log Request Detail
        debug() << "> -------------------------------------------------";
        info() << "Token            : " << request.token();
        info() << "Status           : " << request.status.code;
        info() << "Message          : " << request.status.message;
        debug() << "> -------------------------------------------------";
        // Log File Detail
        std::vector<GetFileRequest>::const_iterator it;
        for(it = request.files.begin(); it != request.files.end(); ++it){
            info() << "> File          : " << (*it).surl;
            info() << "> Status        : " << (*it).status.code;
            info() << "> Message       : " << (*it).status.message;
            info() << "> Size          : " << (*it).size;
            info() << "> TURL          : " << (*it).turl;
            debug() << "> Ext.Wait Time : " << _format_wait_time((*it).extimatedWaitTime);
            debug() << "> Rem.Pin Time  : " << _format_wait_time((*it).remainingPinTime);
            debug() << "> -------------------------------------------------";
        }
    } catch(...){
    }
}

/**
 * log_request
 * log a PrepareToPut Request
 */
void log_request(const PrepareToPut& request) /*throw ()*/{
    try{
        // Log Request Detail
        debug() << "> -------------------------------------------------";
        info() << "Token           : " << request.token();
        info() << "Space Token     : " << request.targetSpaceToken;
        info() << "Status          : " << request.status.code;
        info() << "Message         : " << request.status.message;
        debug() << "Storage Type    : " << FileStorageType::toString(request.desiredFileStorageType);
        debug() << "Overwrite       : " << OverwriteMode::toString(request.overwriteOption);
        debug() << "> -------------------------------------------------";

        // Log File Detail
        std::vector<PutFileRequest>::const_iterator it;
        for(it = request.files.begin(); it != request.files.end(); ++it){
            info() << "> File          : " << (*it).surl;
            info() << "> Status        : " << (*it).status.code;
            info() << "> Message       : " << (*it).status.message;
            info() << "> Size          : " << (*it).size;
            info() << "> TURL          : " << (*it).turl;
            debug() << "> Ext.Wait Time : " << _format_wait_time((*it).extimatedWaitTime);
            debug() << "> Rem.Pin Time  : " << _format_wait_time((*it).remainingPinTime);
            debug() << "> Rem.File Time : " << _format_wait_time((*it).remainingFileTime);
            debug() << "> -------------------------------------------------";
        }
    } catch(...){
    }
}

/**
 * log_request
 * log a SrmCopy Request
 */
void log_request(const SrmCopy& request) /*throw ()*/{
    try{
        // Log Request Detail
        debug() << "> -------------------------------------------------";
        info() << "Token           : " << request.token();
        info() << "Space Token     : " << request.targetSpaceToken;
        info() << "Status          : " << request.status.code;
        info() << "Message         : " << request.status.message;
        debug() << "Storage Type    : " << FileStorageType::toString(request.targetFileStorageType);
        debug() << "Overwrite       : " << OverwriteMode::toString(request.overwriteOption);
        debug() << "> -------------------------------------------------";

        // Log File Detail
        std::vector<CopyFileRequest>::const_iterator it;
        for(it = request.files.begin(); it != request.files.end(); ++it){
            info() << "> Source File   : " << (*it).sourceSurl;
            info() << "> Target File   : " << (*it).targetSurl;
            info() << "> Status        : " << (*it).status.code;
            info() << "> Message       : " << (*it).status.message;
            info() << "> Size          : " << (*it).size;
            debug() << "> Ext.Wait Time : " << _format_wait_time((*it).extimatedWaitTime);
            debug() << "> Rem.File Time : " << _format_wait_time((*it).remainingFileTime);
            debug() << "> -------------------------------------------------";
        }
    } catch(...){
    }
}

std::string GetFileName(const char* f)
{
    boost::filesystem::path p(f);
    return p.leaf();
}

} // end namespace urlcopy
} // end namespace transfer
} // end namespace data
} // end namespace glite
