/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Logger helper methods
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: logger-utility.h,v 1.16 2009/12/03 17:21:44 molnarzs Exp $
 *  Release: $Name:  $
 */
#ifndef GLITE_DATA_TRANSFER_URL_COPY_LOGGER_UTILITY_H_
#define GLITE_DATA_TRANSFER_URL_COPY_LOGGER_UTILITY_H_

#include <string>
#include <map>
#include <sys/types.h>
#include <syslog.h>

#include <signal.h>
#include <log4cpp/Category.hh>

// forward declarations for requests
namespace glite{
namespace data{
namespace srm{
namespace util{
    class PrepareToGet;
    class PrepareToPut;
    class SrmCopy;
}
}
}
}

namespace glite   {
namespace data    {
namespace transfer{
namespace urlcopy {

//------------------------------------------------------------------------------
// requests logging utilities
//------------------------------------------------------------------------------
void log_request(const glite::data::srm::util::PrepareToGet& request) /*throw ()*/;
void log_request(const glite::data::srm::util::PrepareToPut& request) /*throw ()*/;
void log_request(const glite::data::srm::util::SrmCopy& request) /*throw ()*/;
void log_stack();

/**
 * Convert a time_t into a string
 */
std::string time_to_string(time_t tt);

/**
 * LogLevel enum
 */
typedef enum {
    LOGLEVEL_DEBUG,
    LOGLEVEL_INFO,
    LOGLEVEL_WARN,
    LOGLEVEL_ERROR
} LogLevel;

/**
 * Setup Logger
 */
void log_setup(
    const char * name,
    LogLevel l,
    bool enable_console,
    const char * exec_name,
    const bool syslogDisable,
    const std::string& syslogIdent,
    const int syslogFacility,
    const std::string& syslogFacilityName) throw();

/**
 * get the debug stream
 */
log4cpp::CategoryStream debug();

/**
 * get the info stream
 */
log4cpp::CategoryStream info();

/**
 * get the warn stream
 */
log4cpp::CategoryStream warn();

/**
 * get the error stream
 */
log4cpp::CategoryStream error();

std::string GetFileName(const char* f);

#ifdef _DEBUG
	#define DEBUG_LOG(component) debug() << #component << " " << GetFileName(__FILE__) << ":" << __LINE__ << " "
	#define ERROR_LOG(component) error() << #component << " " << GetFileName(__FILE__) << ":" << __LINE__ << " "
	#define WARN_LOG(component) warn() << #component << " " <<  GetFileName(__FILE__) << ":" << __LINE__ << " "
	#define INFO_LOG(component) info() << #component << " " << GetFileName(__FILE__) << ":" << __LINE__ << " "
#else // _DEBUG
	#define DEBUG_LOG(component) debug() << #component << " "
	#define ERROR_LOG(component) error() << #component << " "
	#define WARN_LOG(component) warn() << #component << " "
	#define INFO_LOG(component) info() << #component << " "
#endif // _DEBUG

// helper macros

#ifdef _DEBUG
#define TRACE(msg) debug() << msg
#define TRACE_VALUE(var) debug() << __FUNCTION__ << " - " << #var << " = " <<  var
#else
#define TRACE(msg)
#define TRACE_VALUE(msg)
#endif

// -----------------------------------------------------------------------------
// Signal logging
// -----------------------------------------------------------------------------
#define REGISTER_SIGNAL(signum) SignalLogger::instance().registerSignal(signum,#signum)

/**
 * SignalLogger class
 */
class SignalLogger {
public:

    /**
     * instance
     * return singleton instance
     */
    static SignalLogger& instance(){
        static SignalLogger s_instance;
        return s_instance;
    }

    /**
     * registerSignal
     * register a handler to log given signal
     * @param signum [IN] signal number
     * @param signame [IN] signal name
     */
    void registerSignal(const int signum,const std::string& signame);

    /**
     * destructor
     */
    ~SignalLogger();

private:

    /**
     * SignalInfo
     * helper class to register and deregister signal handlers
     */
    class SignalInfo {
    public:
        /**
         * constructor
         * @param signum [IN] signal number
         * @param signame [IN] signal name
         */
        SignalInfo(int signum,const std::string& signame);
        /**
         * deregister
         * deregister handler
         */
        void deregister();
        /**
         * signame
         * @return signal name
         */
        const std::string& signame() const{
            return m_signame;
        }
        /**
         * destructor
         */
        ~SignalInfo();
    private:
        // signal number
        int m_signum;
        // signal name
        std::string m_signame;
        // flag indicating if signal handler is registered
        bool m_set;
        // old sigaction struct
        struct sigaction m_old;
    };

    /**
     * handleSignal
     * actual signal handler function (registered by SignalInfo objects)
     */
    static void handleSignal(int signum);

    /**
     * logSignal
     * actual signal logging
     */
    void logSignal(int signum);

    /**
     * constructor
     */
    SignalLogger(){
    }

    // map of SignalInfo objects
    typedef std::map<int,SignalInfo *> SignalInfoMap;
    SignalInfoMap m_map;
};

/**
 * utility class to get a unique log id for system logs
 */
class LogId{
public:
    static LogId& instance();
    std::string get_log_id();
    static const std::string KeyName;
private:
    LogId();
    unsigned long m_id;
};


} // end namespace urlcopy
} // end namespace transfer
} // end namespace data
} // end namespace glite

#endif //GLITE_DATA_TRANSFER_URL_COPY_LOGGER_UTILITY_H_
