/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Transfer signal Handler helper methods
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: sighandler-utility.h,v 1.4 2008/04/23 11:57:25 ptedesco Exp $
 *  Release: $Name:  $
 */
#ifndef GLITE_DATA_TRANSFER_URL_SIGHANDLER_UTILITY_H_
#define GLITE_DATA_TRANSFER_URL_SIGHANDLER_UTILITY_H_

#include <signal.h>
#include <boost/utility.hpp>

namespace glite   {
namespace data    {
namespace transfer{
namespace urlcopy {

/**
 * SigAction class (register a signal handler and restore it when it ends)
 */    
class SigAction: boost::noncopyable {
public:
    
    /**
     * Constructor
     */
    SigAction();

    /**
     * Destructor
     */
    ~SigAction();
    
    /**
     * Signal to stop the transfer
     */
    static const int SIGNAL_STOP;

    /**
     * Signal to continue the transfer
     */
    static const int SIGNAL_CONTINUE;
    
private:
    // int m_signum;
    struct sigaction m_sigstop;
    struct sigaction m_sigstop_old;
    
    struct sigaction m_sigcontinue;
    struct sigaction m_sigcontinue_old;
};
    
/**
 * TransferState class: used to keep track of the transfer status
 * @stereotype singleton
 */
class TransferState: boost::noncopyable {
public:     
    /**
     * Return singleton instance
     */
    static TransferState& instance() {
        static TransferState s_instance;
        return s_instance;
    }
    /**
     * Set timedout On
     */
    void timeout(){ m_timedout = true;}
    /**
     * Get timedout Flag
     */
    bool timedout() const { return m_timedout;}

    /**
     * Set Interrupt On
     */
    void interrupt(){ m_interrupted = true;}
    /**
     * Get Interrupt Flag
     */
    bool interrupted() const { return m_interrupted;}

    /**
     * Set Interrupt On
     */
    void processInterrupt(){ m_intProcessed = true;}
    /**
     * Get Interrupt Flag
     */
    bool interruptProcessed() const { return m_intProcessed;}
    
    void set_continueTransfer(const bool value){
        m_continueTransfer = value;
    }
    
    bool get_continueTransfer(){
        return m_continueTransfer;
    }
    
    /**
     * Reset flags
     */
    void reset(){ 
        m_timedout = false; 
        m_interrupted = false; 
        m_intProcessed = false; 
        m_continueTransfer = false;
    }

private:
    /**
     * Constructor
     */
    TransferState():
        m_timedout(false),
        m_interrupted(false),
        m_intProcessed(false),
        m_continueTransfer(false){
    }
    /**
     * Destructor
     */
    ~TransferState(){}
private:
    volatile bool m_timedout;
    volatile bool m_interrupted;
    volatile bool m_intProcessed;
    volatile bool m_continueTransfer;
};

} // end namespace urlcopy
} // end namespace transfer
} // end namespace data
} // end namespace glite

#endif //GLITE_DATA_TRANSFER_URL_SIGHANDLER_UTILITY_H_
