/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: UrlCopy Stat helper function
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: stat-utility.cpp,v 1.18 2009/12/03 17:21:44 molnarzs Exp $
 *  Release: $Name:  $
 */

#include "stat-utility.h"
#include "TransferConfig.h"
#include <glite/data/agents/AgentExceptions.h>
#include <sstream>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <paths.h>

#include <boost/algorithm/string/find.hpp>

using glite::data::agents::RuntimeError;
using glite::data::agents::PosixError;
using glite::data::agents::LogicError;


namespace {

namespace tuc = glite::data::transfer::urlcopy;

/**
 * Copy common arguments
 * NOTE : does NOT set timeouts
 */
void transferBaseArgs_to_TransferStat(
    const tuc::details::TransferBase::Args& args,
    tuc::TransferStat& common){

    common.mm_verbose   = (true == args.verbose)?1:0;
    common.mm_overwrite = (true == args.overwrite)?1:0;

    // Write Request Start Time
    time(&(common.mm_start_time));
    common.mm_last_refresh_time = common.mm_start_time;
    // Write User Description
    tuc::copy_token(common.mm_user_tag,args.user_tag.c_str());

    // new with struct version 2.1.0
    common.mm_enable_dir_check = (true == args.enable_dir_check)?1:0;
    common.mm_enable_ping_source = (true == args.enable_ping_source)?1:0;
    common.mm_enable_ping_dest = (true == args.enable_ping_dest)?1:0;

    // Source and destination space tokens
    tuc::copy_token(common.mm_source_space_token,args.source_space_token.c_str());
    tuc::copy_token(common.mm_source_token_descr,args.source_token_descr.c_str());

    tuc::copy_token(common.mm_dest_space_token,args.dest_space_token.c_str());
    tuc::copy_token(common.mm_dest_token_descr,args.dest_token_descr.c_str());

    // new with 3.2.0
    common.mm_block_after_preparation = args.block_after_preparation ? 1 : 0;
    common.mm_block_after_transfer = args.block_after_transfer ? 1 : 0;

    // new with 3.3.0
    common.mm_copy_pin_lifetime = args.copy_pin_lifetime;
    common.mm_lan_connection = args.lan_connection ? 1 : 0;
    common.mm_fail_nearline_files = args.fail_nearline_files ? 1 : 0;

    common.mm_syslog_disable = args.syslog_disable ? 1 : 0;
    common.mm_syslog_facility = args.syslog_facility;

    tuc::copy_syslog_facility_name(
        common.mm_syslog_facility_name, args.syslog_facility_name.c_str());

    tuc::copy_syslog_ident(common.mm_syslog_ident, args.syslog_ident.c_str());
}

} // namespace

namespace glite   {
namespace data    {
namespace transfer{
namespace urlcopy {

/*------------------------------------------------------------------------------
                            Conversion Utilities
------------------------------------------------------------------------------*/

/**
 * Convert TransferUrlCopy::Args to UrlCopyStat
 * @param args [IN]
 * @param stat [OUT]
 */
void transferUrlCopyArgs_to_UrlCopyStat(
    const TransferUrlCopy::Args&    args,
    UrlCopyStat&                    stat){
    // Write Configuration
    transferBaseArgs_to_TransferStat(args,stat.mm_common);

    // Source File
    copy_url(stat.mm_urlcopy.mm_source.mm_srm_endpoint,args.src_endpoint.c_str());
    copy_version(stat.mm_urlcopy.mm_source.mm_srm_version,args.src_version.c_str());
    copy_url(stat.mm_urlcopy.mm_source.mm_name,args.src_name.c_str());

    stat.mm_urlcopy.mm_source.mm_http_timeout  = args.http_timeout;
    stat.mm_urlcopy.mm_source.mm_prep_timeout  = args.get_timeout;
    stat.mm_urlcopy.mm_source.mm_final_timeout = args.getdone_timeout;

    // Destination File
    copy_url(stat.mm_urlcopy.mm_destination.mm_srm_endpoint,args.dest_endpoint.c_str());
    copy_version(stat.mm_urlcopy.mm_destination.mm_srm_version,args.dest_version.c_str());
    copy_url(stat.mm_urlcopy.mm_destination.mm_name,args.dest_name.c_str());

    stat.mm_urlcopy.mm_destination.mm_http_timeout  = args.http_timeout;
    stat.mm_urlcopy.mm_destination.mm_prep_timeout  = args.put_timeout;
    stat.mm_urlcopy.mm_destination.mm_final_timeout = args.putdone_timeout;

    // Transfer
    stat.mm_urlcopy.mm_tx.mm_streams        = args.nstreams;
    stat.mm_urlcopy.mm_tx.mm_tcp_bs         = args.tcpbs;
    stat.mm_urlcopy.mm_tx.mm_block_size     = args.blocksize;
    stat.mm_urlcopy.mm_tx.mm_timeout        = args.tx_timeout;
    stat.mm_urlcopy.mm_tx.mm_marker_timeout = args.tx_markers_timeout;

    stat.mm_urlcopy.mm_tx.mm_first_marker_timeout = args.first_tx_marker_timeout;
    stat.mm_urlcopy.mm_tx.mm_no_progress_timeout = args.no_progress_timeout;
    stat.mm_urlcopy.mm_tx.mm_tx_to_per_mb = args.tx_to_per_mb;

    // Checksum
    stat.mm_urlcopy.mm_compare_checksum = args.compare_checksum ? 1 : 0;
    if (args.checksum_user_algorithm.length()){
        copy_checksum_algorithm(stat.mm_urlcopy.mm_checksum_user_algorithm,args.checksum_user_algorithm.c_str());
    }
    if (args.checksum_user_value.length()){
        copy_checksum_value(stat.mm_urlcopy.mm_checksum_user_value,args.checksum_user_value.c_str());
    }

#ifdef _USE_GLOBAL_TIMEOUT
    // compute global timeout: try to estimate it.
    stat.mm_common.mm_global_timeout  = 0;
    if(args.tx_timeout > 0){
        // Add tx_timeout
        stat.mm_common.mm_global_timeout = args.tx_timeout;
        if(((args.get_timeout > 0 ) || (args.getdone_timeout > 0 )) &&
            ((args.put_timeout > 0 ) || (args.putdone_timeout > 0 ))){
            // Add srm interaction
            if(args.get_timeout > 0 ){
                stat.mm_common.mm_global_timeout += args.get_timeout;
            } else {
                stat.mm_common.mm_global_timeout += args.getdone_timeout;
            }
            if(args.getdone_timeout > 0 ){
                stat.mm_common.mm_global_timeout += args.getdone_timeout;
            } else {
                stat.mm_common.mm_global_timeout += args.get_timeout;
            }
            if(args.put_timeout > 0 ){
                stat.mm_common.mm_global_timeout += args.put_timeout;
            } else {
                stat.mm_common.mm_global_timeout += args.putdone_timeout;
            }
            if(args.putdone_timeout > 0 ){
                stat.mm_common.mm_global_timeout += args.putdone_timeout;
            } else {
                stat.mm_common.mm_global_timeout += args.put_timeout;
            }
        } else {
            // Estimate srm interaction 1/4 of the transfer
            stat.mm_common.mm_global_timeout += (args.tx_timeout/4);
        }
        // over estimation
        stat.mm_common.mm_global_timeout *= 2;
    }
#endif

}

/**
 * Convert TransferSrmCopy::Args to SrmCopyStat
 * @param args [IN]
 * @param stat [OUT]
 */
void transferSrmCopyArgs_to_SrmCopyStat(
    const TransferSrmCopy::Args&    args,
    SrmCopyStat&                    stat){

    // Write Configuration
    transferBaseArgs_to_TransferStat(args,stat.mm_common);

    stat.mm_srmcopy.mm_push_mode = (true == args.pull_mode)?0:1;
    // Timeouts
    stat.mm_srmcopy.mm_http_timeout    = args.http_timeout;
    stat.mm_srmcopy.mm_copy_timeout    = args.copy_timeout;
    stat.mm_srmcopy.mm_refresh_timeout = args.refresh_timeout;
    // Source SRM
    copy_url(stat.mm_srmcopy.mm_src_srm_endpoint,args.src_endpoint.c_str());
    copy_version(stat.mm_srmcopy.mm_src_srm_version,args.src_version.c_str());
    // Destination SRM
    copy_url(stat.mm_srmcopy.mm_dest_srm_endpoint,args.dest_endpoint.c_str());
    copy_version(stat.mm_srmcopy.mm_dest_srm_version,args.dest_version.c_str());
#ifdef _USE_GLOBAL_TIMEOUT
    // compute global timeout: try to estimate it.
    stat.mm_common.mm_global_timeout  = 0;
    if(args.copy_timeout > 0){
        // Add copy_timeout
        stat.mm_common.mm_global_timeout = args.copy_timeout;
        // over estimation
        stat.mm_common.mm_global_timeout *= 2;
        // Multiply for the number of files
        stat.mm_common.mm_global_timeout *= stat.mm_num_files;
    }
#endif
    // Copy Files info
    for(unsigned int i = 0; i < args.files.size(); ++i){
        copy_url(stat.mm_files[i].mm_src_name , args.files[i].source.c_str());
        copy_url(stat.mm_files[i].mm_dest_name, args.files[i].destination.c_str());
        stat.mm_files[i].mm_compare_checksum = args.files[i].compare_checksum ? 1 : 0;
        copy_checksum_algorithm(stat.mm_files[i].mm_checksum_user_algorithm,args.files[i].checksum_algorithm.c_str());
        copy_checksum_value(stat.mm_files[i].mm_checksum_user_value,args.files[i].checksum_value.c_str());
    }
    stat.mm_srmcopy.mm_tx_to_per_mb = args.tx_to_per_mb;
}

/*
 * create_stat_UrlCopy
 *
 * Create a stat File for UrlCopy and return the Mmap object for it
 */
boost::shared_ptr<Mmap> create_stat_UrlCopy(const std::string& name) /*throw (RuntimeError, LogicError)*/{
    // Precondition check
    if(true == name.empty()){
        throw LogicError("invalid UrlCopyStat file name: empty string");
    }
    // MMap File Name
    std::string mmap_file_name = TransferConfig::instance().repository() + name
        + TransferConfig::instance().extMem();

    // Create the Mmap object
    boost::shared_ptr<Mmap> mmap_ptr(new Mmap(mmap_file_name));

    // Map the file to memory
    try{
        size_t size = sizeof(UrlCopyStat);
        void * buffer = mmap_ptr->create(size, true);
        // Set all the values to 0
        memset(buffer,0,size);
        // Cast the buffer and check the version
        UrlCopyStat * stat = reinterpret_cast<UrlCopyStat*>(buffer);
        // Initialize values
        stat->mm_common.mm_version   = TRANSFER_STAT_VERSION;
        stat->mm_common.mm_extension = URLCOPY_STAT_EXTENSION;
        stat->mm_common.mm_id        = URLCOPY_STAT_ID;
        stat->mm_urlcopy.mm_status   = UrlCopyStatus::IDLE;
    }catch(...){
        mmap_ptr.reset();
        throw;
    }
    return mmap_ptr;
}

/*
 * create_stat_SrmCopy
 *
 * Create a stat File for SrmCopy and return the Mmap object for it
 */
boost::shared_ptr<Mmap> create_stat_SrmCopy(const std::string& name, unsigned int num_files) /*throw (RuntimeError, LogicError)*/{
    // Precondition check
    if(true == name.empty()){
        throw LogicError("invalid SrmCopyStat file name: empty string");
    }
    if(0 == num_files){
        throw LogicError("invalid SrmCopyStat: no files specified");
    }
    // MMap File Name
    std::string mmap_file_name = TransferConfig::instance().repository() + name
        + TransferConfig::instance().extMem();

    // Create the Mmap object
    boost::shared_ptr<Mmap> mmap_ptr(new Mmap(mmap_file_name));

    // Map the file to memory
    try{
        size_t size = sizeof(SrmCopyStat) + (num_files)*sizeof(SrmCopy_FileStat);
        void * buffer = mmap_ptr->create(size, true);
        // Set all the values to 0
        memset(buffer,0,size);
        // Cast the buffer and check the version
        SrmCopyStat * stat = reinterpret_cast<SrmCopyStat*>(buffer);
        // Initialize values
        stat->mm_common.mm_version   = TRANSFER_STAT_VERSION;
        stat->mm_common.mm_extension = SRMCOPY_STAT_EXTENSION;
        stat->mm_common.mm_id        = SRMCOPY_STAT_ID;
        stat->mm_srmcopy.mm_status   = UrlCopyStatus::IDLE;
        stat->mm_num_files           = num_files;
    }catch(...){
        mmap_ptr.reset();
        throw;
    }
    return mmap_ptr;
}


/*
 * open_stat_file
 *
 * Open a stat File, guess its type and check it's validity and return the Mmap object for it
 */
boost::shared_ptr<Mmap> open_stat_file(const std::string& file_name, bool read_only, unsigned int requested_stat_id) /*throw (RuntimeError, LogicError)*/{

    // Precondition check
    if(true == file_name.empty()){
        throw LogicError("invalid Stat file name: empty string");
    }

    // Create the Mmap object
    boost::shared_ptr<Mmap> mmap_ptr(new Mmap(file_name));

    // Map the file to memory
    try{
        size_t size;
        void * buffer = mmap_ptr->open(size, read_only,true);
        if(size <= sizeof(TransferStat)){
            // Invalid Size
            throw RuntimeError("invalid Stat file: invalid size. Is the version correct?");
        }
        // Cast the buffer and check the version
        TransferStat * tstat = reinterpret_cast<TransferStat*>(buffer);
        if(tstat->mm_version   != TRANSFER_STAT_VERSION){
            // Invalid Version
            throw RuntimeError("invalid Stat file: invalid version");
        }
        // Check the ID
        switch(tstat->mm_id){
            case URLCOPY_STAT_ID:{
                // Check the size
                if(size != sizeof(UrlCopyStat)){
                    // Invalid Size
                    throw RuntimeError("invalid UrlCopyStat file: invalid size. Is the version correct?");
                }
                // Cast the buffer and check the version
                UrlCopyStat * stat = reinterpret_cast<UrlCopyStat*>(buffer);
                if(stat->mm_common.mm_extension != URLCOPY_STAT_EXTENSION){
                    // Invalid Version
                    throw RuntimeError("invalid UrlCopyStat file: invalid version");
                }
            } break;
            case SRMCOPY_STAT_ID:{
                // Check the size
                if(size <= sizeof(SrmCopyStat)){
                    // Invalid Size
                    throw RuntimeError("invalid SrmCopyStat file: invalid size. Is the version correct?");
                }
                // Cast the buffer and check the version
                SrmCopyStat * stat = reinterpret_cast<SrmCopyStat*>(buffer);
                if(stat->mm_common.mm_extension != SRMCOPY_STAT_EXTENSION){
                    // Invalid Version
                    throw RuntimeError("invalid SrmCopyStat file: invalid version");
                }
                // Check the size
                if(size != (sizeof(SrmCopyStat) + (stat->mm_num_files * sizeof(SrmCopy_FileStat)))){
                    // Invalid Size
                    throw RuntimeError("invalid SrmCopyStat file: invalid size. Is the version correct?");
                }
            } break;
            default:
                // Invalid Stat ID
                throw RuntimeError("invalid Stat file: unknwon Stat ID");
        }// End swicth
        if((0 != requested_stat_id) && (tstat->mm_id != requested_stat_id)){
            throw LogicError("the Stat type differs from the requested one");
        }
    }catch(...){
        mmap_ptr.reset();
        throw;
    }
    return mmap_ptr;
}

/*
 * open_stat
 *
 * Open a stat File, guess its type and check it's validity and return the Mmap object for it
 */
boost::shared_ptr<Mmap> open_stat(
    const std::string&  name,
    bool                read_only,
    unsigned int        requested_stat_id,
    bool                search_subfolders) /*throw (RuntimeError, LogicError)*/{
    // Precondition check
    if(true == name.empty()){
        throw LogicError("invalid Stat file name: empty string");
    }
    // MMap File Name
    std::string mmap_file_name = TransferConfig::instance().repository() + name
        + TransferConfig::instance().extMem();
    // Check if file exists
    if ((0 != access(mmap_file_name.c_str(), F_OK)) &&
        search_subfolders){
        // Extract prefix
        // TODO : move this function somewhere else, the prefix is added
        // in transfer-utility !
        std::string prefix;
        boost::iterator_range<std::string::const_iterator> res = boost::find_last(name,"__");
        if(res.begin() != name.end()){
            int pos = std::distance(name.begin(),res.begin());
            prefix = name.substr(0,pos);
        }
        // Look in the completed
        std::string completed_mmap_file_name = TransferConfig::instance().archive(true,false,prefix) + name + TransferConfig::instance().extMem();
        if(0 == access(completed_mmap_file_name.c_str(), F_OK)){
            mmap_file_name = completed_mmap_file_name;
        } else {
            // Look in the failed
            std::string failed_mmap_file_name = TransferConfig::instance().archive(false,false,prefix) + name + TransferConfig::instance().extMem();
            if(0 == access(failed_mmap_file_name.c_str(), F_OK)){
                mmap_file_name = failed_mmap_file_name;
            }
            else{
                // Look in lost
                std::string lost_mmap_file_name = TransferConfig::instance().archive(false,true,prefix) + name + TransferConfig::instance().extMem();
                if(0 == access(lost_mmap_file_name.c_str(), F_OK)){
                    mmap_file_name = lost_mmap_file_name;
                }
            }
        }
    }
    // Open file
    return open_stat_file(mmap_file_name, read_only, requested_stat_id);
}

/*
 * open_stat_UrlCopy
 *
 * Open a stat File for UrlCopy and check it's validity and return the Mmap object for it
 */
boost::shared_ptr<Mmap> open_stat_UrlCopy(
    const std::string&  name,
    bool                read_only,
    bool                search_subfolders) /*throw (RuntimeError, LogicError)*/{
    // Open the file and perform the consistency checks
    return open_stat(name,read_only,URLCOPY_STAT_ID,search_subfolders);
}

/*
 * open_stat_SrmCopy
 *
 * Open a stat File for SrmCopy and check it's validity and return the Mmap object for it
 */
boost::shared_ptr<Mmap> open_stat_SrmCopy(
    const std::string&  name,
    bool                read_only,
    bool                search_subfolders) /*throw (RuntimeError, LogicError)*/{
    // Open the file and perform the consistency checks
    return open_stat(name,read_only,SRMCOPY_STAT_ID,search_subfolders);
}

} // end namespace urlcopy
} // end namespace transfer
} // end namespace data
} // end namespace glite


