/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004-2008.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  Get a file checksum using gridftp
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: checksum-main.cpp,v 1.3 2009/12/03 17:21:44 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#include <glite/data/agents/extra/options.h>
#include <gridftp-utility.h>
#include <logger-utility.h>
using namespace glite::data::agents::options;
using namespace glite::data::transfer::urlcopy;

namespace{

/**
 * tool_info
 */
std::ostream& tool_info(){
    return std::cout;
}

/**
 * tool_error
 */
std::ostream& tool_error(){
    return std::cerr;
}

class DummyRefreshTimeUpdater : public RefreshTimeUpdater::IUpdateRefreshTime{
public:
    virtual void UpdateRefreshTime(){
        tool_info() << "Updating refresh time." << std::endl;
    }   
};

}

int main(int argc,char * argv[]){
    // arguments parser
    Arguments args(argc,argv,tool_info(),tool_error());
    bool show_help = false;
    bool show_version = false;
    int timeout = -1;
    bool debug = false;
    std::string algorithm = "ADLER32"; 
    try{
        // add common options
        opt(algorithm,          args,'a',"algorithm")   .arg("VALUE")
                                                        .help("Checksum algorithm (default = ADLER32).");
        opt(timeout,            args,'t',"timeout")     .arg("N")
                                                        .help("Operation timeout in seconds (default = no timeout).");
        opt(debug,true,         args,'d',"debug")       .help("Enable debug messages.");
        opt(show_help,true,     args,'h',"help")        .help("Print this help text and exit.");
        opt(show_version,true,  args,'V',"version")     .help("Print the version number and exit.");
        args.pos("TURL")                                .help("The TURL of the file.");
    } catch(const std::exception& exc){
        tool_error() << exc.what() << std::endl;
        return EXIT_FAILURE;
    }
    
    if (!args.parse()){
        tool_error() << "Failed parsing arguments." << std::endl;
        args.showHelp();
        return EXIT_FAILURE;
    }
    if (show_help){
        args.showHelp();
        return EXIT_SUCCESS;
    }
    if (show_version){
        tool_info() << MODULE_VERSION << std::endl;
        return EXIT_SUCCESS;
    }

    argc -= optind;
    argv += optind;
    
    if((argc < 1) || (0 == argv[0])){
        tool_error() << "No TURL specified." << std::endl;
        args.showHelp();
        return EXIT_FAILURE;
    }
    if(argc > 1){
        tool_error() << "Too many TURLs specified." << std::endl;
        args.showHelp();
        return EXIT_FAILURE;
    }
    std::string turl = argv[0];
        
    // Initialize logging
    log_setup(0,
        debug ? LOGLEVEL_DEBUG : LOGLEVEL_INFO,
        true,"get-checksum",
        false, "checksum", LOG_USER, "LOG_USER");
    
    UrlCopyError::Category error_category = UrlCopyError::SUCCESS;
    std::string error_message;
    DummyRefreshTimeUpdater updater;
    std::string checksum = gridftp_checksum(
        turl,
        algorithm,
        timeout,
        error_category,
        error_message,
        &updater);
    
    int ret = 0;
    if (UrlCopyError::SUCCESS == error_category){
        tool_info() << checksum << std::endl;
    } else {
        ret = -1;
        tool_info() << error_category << ": " << error_message << std::endl;
    }
            
    return ret;

}

