/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Transfer helpers class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: transfer-utility.cpp,v 1.18 2009/08/07 11:28:52 molnarzs Exp $
 *  Release: $Name:  $
 */
#include "transfer-utility.h"
#include <glite/data/transfer/urlcopy/UrlCopyStatus.h>
#include <glite/data/transfer/urlcopy/UrlCopyError.h>
#include "TransferConfig.h"
#include "stat-utility.h"
#include "logger-utility.h"
#include "sighandler-utility.h"
#include <glite/data/agents/AgentExceptions.h>

#include "msg-ifce.h"
#include <vector>
#include <string>
#include <fstream>
#include <sstream>
#include <iomanip>



#include <stdio.h>
#include <signal.h>
#include <paths.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <errno.h>


#include <boost/algorithm/string/predicate.hpp>
#include <boost/algorithm/string/find.hpp>

using namespace glite::data::transfer::urlcopy;
using namespace glite::data::transfer::urlcopy::details;
using namespace glite::data::agents;

namespace {

const char * const GLITE_LOCATION_ENV     = "GLITE_LOCATION";
const char * const LIBEXEC_PATH           = "libexec/";
const char * const DEFAULT_GLITE_LOCATION = "/opt/glite/";
#ifndef _PATH_PROC
const char * const _PATH_PROC             = "/proc/";
#endif //_PATH_PROC
const char * const CMD_LINE               = "cmdline";

const char CMD_LINE_SEPARATOR             = '\0';

const char * const EXEC_TEST_OPTIONS = " -V > " _PATH_DEVNULL " 2>&1";

const unsigned int MAX_START_DELAY        = 5;
// minimum value for the refresh delay
const int MIN_REFRESH_DELAY      = 180;
const int STOP_WAIT_TIME         = 120;

// consider process dead if not running and last update time was more than
// MAX_NOT_RUNNING_TIME seconds ago (avoid case in which process has just
// completed but the stat file was read before)
const unsigned int MAX_NOT_RUNNING_TIME   = 10;

/**
 * file_exists
 * @param path [IN]
 * @return true if file exists, false otherwise
 */
bool file_exists(const std::string& path){
    struct stat s;
    int r = stat(path.c_str(),&s);
    return (0 == r);
}

/**
 * get_exec_path
 * Get the name of the executable program.
 * @param name [IN] the executable name
 * @return the executable path
 * @throw LogicError if the file path does not exist
 */
std::string get_exec_path(const std::string& name) /*throw(LogicError)*/ {
    bool using_default_location = false;
    const char * location = getenv(GLITE_LOCATION_ENV);
    if(0 == location){
        // No Glite Location set: try to guess it
        location = DEFAULT_GLITE_LOCATION;
        using_default_location = true;
    }
    std::string exec_name = location;
    if(*(exec_name.rbegin()) != '/'){
        exec_name += '/';
    }
    exec_name += LIBEXEC_PATH;
    exec_name += name;

    if (!file_exists(exec_name)){
        std::stringstream reason;
        reason << "File " << name << " not found under $GLITE_LOCATION/" << LIBEXEC_PATH;
        if (using_default_location) {
            reason << " (GLITE_LOCATION not set, using default " << location << ").";
        } else {
            reason << " (GLITE_LOCATION = " << location << ").";
        }
        throw LogicError(reason.str());
    }
    return exec_name;
}

/**
 * get_test_exec_command
 * @param name [IN] the executable name
 * @return true if command was executed successfully, false otherwise
 */
bool execute_test_command(const std::string& path) /*throw(LogicError)*/ {
    std::string cmd = path + EXEC_TEST_OPTIONS;
    int result = system(cmd.c_str());
    return (0 == result);
}

} // end global namespace


namespace glite   {
namespace data    {
namespace transfer{
namespace urlcopy {

/*
 * setupRepository
 *
 * Create the Repository and Archive directory
 */
void setup_repository(const std::string& prefix) /*throw(RuntimeError)*/ {
    // Create the Repository and Archive directory
    try{
        TransferConfig::instance().repository();
        TransferConfig::instance().archive(true,false,prefix);
        TransferConfig::instance().archive(false,false,prefix);
        TransferConfig::instance().archive(false,true,prefix);
    } catch(const RuntimeError& exc){
        // Failed to create the one of the needed directories: don't even start
        // the transfer
        throw;
    }
}

/*
 * generateId
 *
 * Generate a new (unique) request id
 */
std::string generate_request_id(const std::string& prefix) throw() {

    std::string new_name;

    // Get current time
    time_t current;
    time(&current);
    struct tm * date = gmtime(&current);

    // Create template
    std::stringstream ss;
    if(false == prefix.empty()){
        ss << prefix;
    }
    ss << std::setfill('0');
    ss << "__" << std::setw(4) << (date->tm_year+1900)
       <<  "-" << std::setw(2) << (date->tm_mon+1)
       <<  "-" << std::setw(2) << (date->tm_mday)
       <<  "-" << std::setw(2) << (date->tm_hour)
               << std::setw(2) << (date->tm_min)
       << "_XXXXXX";

    // Generate File name
    new_name = ss.str();
    char* ret = mktemp(&(*(new_name.begin())));

    if (ret == NULL)
        throw PosixError("mktemp", errno);

    return new_name;
}

/**
 * execute_transfer
 * Execute the transfer
 * @param tx [IN] the transfer object
 * @param blocking [IN]
 * @param console_logging [IN]
 */
void execute_transfer(
    TransferBase&   tx,
    bool            blocking,
    bool            console_logging, 
    std::string vo, 
    std::string src_site, 
    std::string dst_site) /* throw (RuntimeError) */{

    REGISTER_SIGNAL(SIGABRT);
    REGISTER_SIGNAL(SIGSEGV);
    
 

    if(false == blocking){
        // Asynchnonous mode: double fork

        // Fork a child process
        pid_t pid_1 = fork();
        if(-1 == pid_1){
            throw PosixError("fork",errno);
        } else if(0 != pid_1){
            // Parent process: wait for the first child
            int status;
            ::wait(&status);
            if(0 == WEXITSTATUS(status)){
                // OK
            } else if(1 == WEXITSTATUS(status)){
                throw RuntimeError("second fork failed");
            } else if(2 == WEXITSTATUS(status)){
                // try to reproduce the error.
                try {
                    get_exec_path(tx.execName());
                } catch (const LogicError& exc){
                    throw RuntimeError(exc.what());
                }
                // the error was not reproducible, raise a generic one
                // (this should never happen)
                std::stringstream reason;
                reason << tx.execName() << " not found.";
                throw RuntimeError(reason.str());
            } else if (3 == WEXITSTATUS(status)){
                std::stringstream reason;
                reason << "Error executing '" << tx.execName() << "'";
                throw RuntimeError(reason.str());
            }
            else {
                throw RuntimeError("Unknown error starting the transfer");
            }
        } else {
            // Child process

            // Be sure exec is in the path
            std::string exec_path;
            try{
                exec_path = get_exec_path(tx.execName());
            } catch(const LogicError&){
                exit(2);
            }
            // execute the test command
            if (!execute_test_command(exec_path)){
                // test command execution failed
                exit(3);
            }
            // Ignore SIGCLD: Don't wait for the child to complete
            signal(SIGCLD, SIG_IGN);
            pid_t pid_2 = fork();
            if(-1 == pid_2){
                // Failed to do the second fork
                exit(1);
            } else if(0 != pid_2){
                // Parent process exit 0
                exit(0);
            } else {
                // Detach from parent
                setsid();
                // Set working directory
                int ret = chdir(_PATH_TMP);

                if (ret == -1)
                    throw PosixError("chdir", errno);

                // Close stdin and out
                int fd;
                if ((fd = ::open(_PATH_DEVNULL, O_RDWR, 0)) != -1) {
                    ::dup2(fd, STDIN_FILENO);
                    ::dup2(fd, STDOUT_FILENO);
                    ::dup2(fd, STDERR_FILENO);
                    if (fd > 2){
                       close(fd);
                    }
                }
          
                int r = execlp(exec_path.c_str(), tx.execName().c_str(), tx.id().c_str(), vo.c_str(), src_site.c_str(), dst_site.c_str(), NULL);
                if(-1 == r){
                    // Process Execution failed: exit the forked process
                    // The status will detect if the request has not been
                    // assigned
                    exit(1);
                }
            }
        }
    } else {
        // Synchonous mode
        tx.exec(console_logging, vo, src_site, dst_site);
    }
}

/*
 * archive_files
 *
 * Move log and mem files to completed/failed/lost directories
 */
void archive_files(const TransferBase& tx, bool is_success,bool is_lost) throw () {

    // Get the prefix from the id
    const std::string& id = tx.id();
    std::string prefix;
    boost::iterator_range<std::string::const_iterator> res = boost::find_last(id,"__");
    if(res.begin() != id.end()){
        int pos = std::distance(id.begin(),res.begin());
        prefix = id.substr(0,pos);
    }
    // Archive the log and mem files
    try{
        // Get Repository
        std::string rep_dir = TransferConfig::instance().repository();
        // Get Archive
        std::string archive_dir = TransferConfig::instance().archive(is_success,is_lost,prefix);

        // Get File names
        std::string mem_file = rep_dir + id + TransferConfig::instance().extMem();
        std::string log_file = rep_dir + id + TransferConfig::instance().extLog();

        // Get New File names
        std::string archived_mem_file = archive_dir + id +
            TransferConfig::instance().extMem();
        std::string archived_log_file = archive_dir + id +
            TransferConfig::instance().extLog();

        // Move the files
        if(0 == access(mem_file.c_str(), F_OK)) {
            int result = rename(mem_file.c_str(), archived_mem_file.c_str());
            if(0 != result){
                // Ignore error => a script would check this offline
            }
        }
        if(0 == access(log_file.c_str(), F_OK)) {
            int result = rename(log_file.c_str(), archived_log_file.c_str());
            if(0 != result){
                // Ignore error => a script would check this offline
            }
        }
    } catch (const RuntimeError& exc){
        // This should never happen (the repositories are created when
        // submitting the request): Ignore error => a script would check this
        // offline
    } catch (...){
        // Ignore all other errors
    }
}

/**
 * Verify that given command line file contains the given executable name and
 * transfer id.
 * @param file_path [IN] the path of the command line file to open
 * @param exec_name [IN] the executable name
 * @param tx_id [IN] the transfer id
 * @param reason [OUT] the reason for which the check failed
 */
bool check_cmdline_file(
    const std::string& file_path,
    const std::string& exec_name,
    const std::string& tx_id,
    std::string& reason){
    bool result = false;
    try{
        // open the cmdline file
        std::ifstream proc_file;
        proc_file.open(file_path.c_str());
        if(true == proc_file.is_open()){
            // Get the process name
            std::stringbuf exec_name_buf;
            proc_file.get(exec_name_buf,CMD_LINE_SEPARATOR);
            proc_file.seekg((std::streamoff)1,std::ios::cur);
            // Check that the process name is correct
            if(true == boost::ends_with(exec_name_buf.str(),exec_name)){
                if (proc_file.eof()) // should never happen
                {
                    return false;
                }
                // Get the request id
                std::stringbuf request_id_buf;
                proc_file.get(request_id_buf,CMD_LINE_SEPARATOR);
                // Check that the request id is correct
                if(request_id_buf.str() == tx_id){
                    result = true;
                }
                else{
                    std::stringstream ss;
                    ss << "different tx ids (" << request_id_buf.str() << " / "
                       << tx_id << ")";
                    reason = ss.str();
                }
            }
            else{
                std::stringstream ss;
                ss << "different exec names (" << exec_name_buf.str() << " / "
                   << file_path << ")";
                reason = ss.str();
            }
        }
        else{
            std::stringstream ss;
            ss << "could not open file " << file_path;
            reason = ss.str();
        }
    } catch(...){
        reason = "an error occurred";
        result = false;
    }
    return result;
}


/*
 * is_running
 *
 * Check that the transfer process service the request is still active
 */
bool is_running(const TransferBase& tx, pid_t pid) throw(){
    std::string reason;
    return is_running_with_reason(tx,pid,reason);
}

/**
 * Check that the transfer process serving the request is still active and
 * provide a description why the check failed
 */
bool is_running_with_reason(const details::TransferBase& tx, pid_t pid,std::string& reason) throw()
{
    std::stringstream ss;
    ss << _PATH_PROC << pid << "/" << CMD_LINE;
    return check_cmdline_file(ss.str(),tx.execName(),tx.id(),reason);
}

/**
 * check_transfer
 *
 * Check that the transfer process is updating the stat file periodically
 * @param tx [IN] the TransferBase object
 * @param stat [IN] the TransferStat object
 * @param check [IN] if true, the last update time will be checked
 * @param status [IN] the UrlCopyStatus for the transfer
 * @param refresh_delay [IN] the maximum amount of time from the last update
 * time refresh after which the transfer is considered not responding. If set to
 * -1 the refresh delay is not checked
 */
void check_transfer(
    const TransferBase& tx,
    const TransferStat& stat,
    bool check,
    unsigned int status,
    int refresh_delay) /*throw(RuntimeError)*/{

    time_t current;
    time(&current);

    // Check that the transfer pid is set
    if((0 == stat.mm_pid) &&
       (current > (stat.mm_start_time + MAX_START_DELAY))){
        // The request has not been processed. This may happen only in some
        // weird circumstances (exec failed, mmap file corruption etc).
        // throw an exception so the user can abort the transfer
        std::stringstream reason;
        reason << "Failed to process the transfer in [" << MAX_START_DELAY << "] seconds.";
        throw RuntimeError(reason.str());
    }
    // Check the status: The timeout are overestimated in order to minimize
    // the risk of a race condition
    if((true == check) && (status != UrlCopyStatus::COMPLETED)){

        // Verify if the process is running
        std::string running_reason;
        bool tx_is_running = is_running_with_reason(tx,stat.mm_pid,running_reason);
        if ((false == tx_is_running) &&
            (current > (stat.mm_last_refresh_time + MAX_NOT_RUNNING_TIME))){
            // the process is dead, not completed and the refresh time was not
            // updated for a while.
            std::stringstream reason;
            reason << "The process serving the transfer (status = "
                   << UrlCopyStatus::toString((UrlCopyStatus::Value)status)
                   << ") is no longer active (" << running_reason << ")";
            throw RuntimeError(reason.str());
        }
#ifdef _USE_GLOBAL_TIMEOUT
        if((stat.mm_global_timeout > 0) &&
           ((current > (stat.mm_start_time + stat.mm_global_timeout)))){
            // request still active after the global timeout
            std::stringstream reason;
            if(true == tx_is_running){
                reason << "The transfer process [" << stat.mm_pid
                       << "] is still running after the global timeout ["
                       << stat.mm_global_timeout << "] elapsed. The "
                          "transfer should be aborted";
            } else {
                reason << "The process serving the transfer is no longer"
                          " active (" << running_reason
                       << ") but the request has not been completed"
                          " (still active after global timeout, Status = "
                       << UrlCopyStatus::toString((UrlCopyStatus::Value)status)
                       << ").";
            }
            throw RuntimeError(reason.str());
        }
#endif
        if(-1 != refresh_delay){
            // Check that the stats have been updated
            if((0 == refresh_delay) || (refresh_delay < MIN_REFRESH_DELAY)){
                refresh_delay = MIN_REFRESH_DELAY;
            }
            // overestimation
            refresh_delay *= 2;
            // Check if the process is stuck or if it's still updating the process
            if((current > (stat.mm_last_refresh_time + refresh_delay))){
                // No update received since a while
                std::stringstream reason;
                if(true == tx_is_running){
                    reason << "No status updates received since more than ["
                           << refresh_delay << "] seconds. Probably the process"
                              " serving the transfer is stuck";
                } else {
                    reason << "The process serving the transfer [" << stat.mm_pid
                           << "] is no longer active (" << running_reason
                           << ") but the request has not"
                              " been completed (no updates received in "
                           << refresh_delay << " seconds, Status = "
                           << UrlCopyStatus::toString((UrlCopyStatus::Value)status)
                           << ").";
                }
                throw RuntimeError(reason.str());
            }
        }
    } // if((true == check) && (status != UrlCopyStatus::COMPLETED))
}

/*
 * stop_transfer
 *
 * Stop the running transfer. Return 0 is success, -1 if failed and and 1 if the
 * stat should be refreshed
 */
int stop_transfer(const TransferBase& tx, const TransferStat& stat, unsigned int status) /*throw(RuntimeError)*/{
    int stopped = -1;

    time_t current;
    time(&current);

    // Check the status
    if(status == UrlCopyStatus::COMPLETED){
        // Already completed
        stopped = 0;
    } else {
        // The transfer request is still active: Check the process id
        if(0 == stat.mm_pid){
            if(current > (stat.mm_start_time + MAX_START_DELAY)){
                // No process served the request: abort it
                stopped = -1;
            } else {
                stopped = 1;
            }
        } else {
            // Check if the process is running
            if(true == is_running(tx,stat.mm_pid)){
                // Send a term signal
                int result = ::kill(stat.mm_pid, SigAction::SIGNAL_STOP);
                if(0 == result){
                    // Wait the process is stopped
                    int count = 0;
                    do{
                        // Sleep for a while
                        sleep(1);
                        ++count;
                    } while((true == is_running(tx,stat.mm_pid)) &&
                            (count < STOP_WAIT_TIME ));
                    // Check if the process is still running
                    if(true == is_running(tx,stat.mm_pid)){
                        std::stringstream reason;
                        reason << "failed to stop transfer process ["
                               << stat.mm_pid << "] in ["
                               << STOP_WAIT_TIME << "] seconds";
                        throw RuntimeError(reason.str());
                    } else {
                        stopped = 0;
                    }
                } else if(ESRCH == errno){
                    // OK: process already gone
                } else {
                    throw PosixError("kill", errno);
                }
            }
        }
    }

    return stopped;
}

/*
 * abort_transfer
 *
 * Abort a running transfer
 */
void abort_transfer(const TransferBase& tx, pid_t pid, unsigned int& status, ErrorStat& error_stat) /*throw()*/{
    // Check the status
    if(UrlCopyStatus::COMPLETED == status){
        // Transfer request completed: nothing to do
    } else {
        // The transfer request is still active: Check the process id
        if(0 == pid){
            // Update the error
            error_stat.mm_scope    = UrlCopyError::SCOPE_AGENT;
            error_stat.mm_phase    = UrlCopyError::PHASE_PREPARATION;
            error_stat.mm_category = UrlCopyError::ERROR_INTERNAL_ERROR;
            copy_message(error_stat.mm_message,"no process served the transfer request");
        } else {
            // Check if the process is running
            bool tx_running = is_running(tx,pid);
            if(true == tx_running){
                // Send a KILL signal
                ::kill(pid, SIGKILL);
            }
            // Update the error
            error_stat.mm_scope = UrlCopyError::SCOPE_AGENT;
            switch(status){
                case UrlCopyStatus::IDLE:
                case UrlCopyStatus::PREPARING:
                    error_stat.mm_phase    = UrlCopyError::PHASE_PREPARATION;
                break;
                case UrlCopyStatus::READY:
                case UrlCopyStatus::RUNNING:
                    error_stat.mm_phase    = UrlCopyError::PHASE_TRANSFER;
                break;
                case UrlCopyStatus::DONE:
                case UrlCopyStatus::FAILED:
                case UrlCopyStatus::ABORTED:
                    error_stat.mm_phase    = UrlCopyError::PHASE_FINALIZATION;
                break;
                default:
                    error_stat.mm_phase    = UrlCopyError::PHASE_UNDEF;
                break;
            }
            error_stat.mm_category = UrlCopyError::ERROR_LOST;
            if(false == tx_running){
                copy_message(error_stat.mm_message,"the process that served the request died");
            } else {
                copy_message(error_stat.mm_message,"the process that served the request has been killed");
            }
        }
        // Update the status
        status = UrlCopyStatus::COMPLETED;
    }
}


} // namespace urlcopy
} // namespace transfer
} // namespace data
} // namespace glite
