/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  gLite Transfer Url Copy Clients - utility functions for urlcopy transfers
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: util-urlcopy.cpp,v 1.3 2008/09/09 08:32:49 ptedesco Exp $
 *  Release: $Name:  $
 *
 */

#include "util-urlcopy.h"
#include "util-main.h"
#include <glite/data/transfer/urlcopy/UrlCopyError.h>

using namespace glite::data::transfer::urlcopy;

namespace {
/**
 * return a string representation for a boolean value
 * @param value 
 * @return 
 */
template<typename T>
const char * bool2str(T value){
    return value ? "yes" : "no";
}
} // namespace

/**
 * print transfer parameters
 * @param tx_stat [IN]
 */
void print_transfer_parameters(const TransferUrlCopy::Args& tx_args){
    tool_debug() << "# Transfer Request parameters:" << std::endl;
    // Source
    tool_debug() << "# Source          : " << tx_args.src_name << std::endl;
    if(false == tx_args.src_endpoint.empty()){
        tool_debug() << "# Src SRM Endpoint: " << tx_args.src_endpoint << std::endl;
    }
    if(false == tx_args.src_version.empty()){
        tool_debug() << "# Src SRM Version : " << tx_args.src_version << std::endl;
    }
    // Destination
    tool_debug() << "# Destination     : " << tx_args.dest_name << std::endl;
    if(false == tx_args.dest_endpoint.empty()){
        tool_debug() << "# Dst SRM Endpoint: " << tx_args.dest_endpoint << std::endl;
    }
    if(false == tx_args.dest_version.empty()){
        tool_debug() << "# Dst SRM Version : " << tx_args.dest_version << std::endl;
    }
    if(false == tx_args.dest_token_descr.empty()){
        tool_debug() << "# Dst Space Descr : " << tx_args.dest_token_descr << std::endl;
    }
    // Configuration
    tool_debug() << "# Overwrite       : " << bool2str(tx_args.overwrite) << std::endl;
    if(tx_args.nstreams > 0){
        tool_debug() << "# Num streams     : " << tx_args.nstreams << std::endl;
    }
    if(tx_args.tcpbs > 0){
        tool_debug() << "# Tcp BS          : " << tx_args.tcpbs << std::endl;
    }
    if(tx_args.blocksize > 0){
        tool_debug() << "# Block Size      : " << tx_args.blocksize << std::endl;
    }
    // Timeouts
    if(tx_args.http_timeout > 0){
        tool_debug() << "# HTTP Timeout    : " << tx_args.http_timeout << std::endl;
    }
    if(tx_args.put_timeout > 0){
        tool_debug() << "# Put Timeout     : " << tx_args.put_timeout << std::endl;
    }
    if(tx_args.get_timeout > 0){
        tool_debug() << "# Get Timeout     : " << tx_args.get_timeout << std::endl;
    }
    if(tx_args.tx_timeout > 0){
        tool_debug() << "# Transfer Timeout: " << tx_args.tx_timeout << std::endl;
    }
    if(tx_args.tx_markers_timeout > 0){
        tool_debug() << "# Markers Timeout : " << tx_args.tx_markers_timeout << std::endl;
    }
    if(tx_args.first_tx_marker_timeout > 0){
        tool_debug() << "# First Marker Timeout : " << tx_args.first_tx_marker_timeout << std::endl;
    }
    if(tx_args.putdone_timeout > 0){
        tool_debug() << "# Put Done Timeout: " << tx_args.putdone_timeout << std::endl;
    }
    if(tx_args.getdone_timeout > 0){
        tool_debug() << "# Get Done Timeout: " << tx_args.getdone_timeout << std::endl;
    }
    tool_debug() << "# Block after preparation: " << bool2str(tx_args.block_after_preparation) << std::endl;
    tool_debug() << "# Block after transfer: " << bool2str(tx_args.block_after_transfer) << std::endl;

}

/**
 * print transfer details
 * @param stat [IN]
 */
void print_transfer_details(const TransferUrlCopyStat& tx_stat){
    tool_info() << tx_stat.tx.size << " bytes transferred in " << (tx_stat.tx.done_time - tx_stat.tx.start_time) << " seconds with an average throughput of " << (tx_stat.tx.avg_throughput / 1024.0) << " KB/s" << std::endl;
}

/**
 * print transfer failure
 * @param tx_stat [IN]
 */
void print_transfer_failure(const TransferUrlCopyStat& tx_stat){
    tool_error() << "Transfer request failed on " << UrlCopyError::scopeToString(tx_stat.error.scope) << " during " << UrlCopyError::phaseToString(tx_stat.error.phase) << " phase. Reason is: [" << UrlCopyError::categoryToString(tx_stat.error.category) << "] " << tx_stat.error.description << std::endl;
}

/**
 * print details about preparation phase
 * @param tx_stat [IN]
 */
void print_preparation_details(const TransferUrlCopyStat& tx_stat){
    tool_info() << "Source Prep.  Time: " << (tx_stat.source.prep_done_time - tx_stat.source.prep_start_time) << " secs" << std::endl;
    tool_info() << "Dest.  Prep.  Time: " << (tx_stat.destination.prep_done_time - tx_stat.destination.prep_start_time) << " secs" << std::endl;
}
        
/**
 * print details about finalization phase
 * @param tx_stat [IN]
 */
void print_finalization_details(const TransferUrlCopyStat& tx_stat){
    tool_info() << "Source Final. Time: " << (tx_stat.source.final_done_time - tx_stat.source.final_start_time) << " secs" << std::endl;
    tool_info() << "Dest.  Final. Time: " << (tx_stat.destination.final_done_time - tx_stat.destination.final_start_time) << " secs" << std::endl;
}



