#! /usr/bin/env python
#
#  Copyright (c) Members of the EGEE Collaboration. 2004.
#  See http://www.eu-egee.org/partners/ for details on the copyright
#  holders.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  transfer-url-copy test script
#
#  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
#  Version info: $Id: CgsiTraceReader.py,v 1.4 2008/12/08 12:38:10 ptedesco Exp $
#  Release: $Name:  $

import os
import sys
import re

from Common import *

from xml.dom.minidom import parseString
from xml.dom import Node

# CgsiTraceReader class
#-------------------------------------------------------------------------------
class CgsiTraceReader:
    """Utility to extract xml messages from cgsi trace files"""
    
    def __init__(self):
    
        # find xml messages
        self.xmlre = re.compile(r"""
            <\?[^<>]*\?>\s*
            (?P<xml><(?P<envelope>[^<>\ ]+)[^<>]*>\s*
            .*?
            </(?P=envelope)>)
            """,re.VERBOSE | re.DOTALL)

        # expression to remove unuseful lines from the log
        self.to_remove = re.compile(r"""
            (?:[a-fA-F0-9]{2}\ ){3,16}|
            Iteration:<\d+>|
            ===.*|
            <(?:Recving|Sending)\ SOAP\ Packet>.*
            """,re.VERBOSE)
        
        # expression to search blank space before pretty-printing xml
        self.remove_blanks = re.compile(r"(?P<close>>\s+)|(?P<open>\s+<)")

    def substitute_blanks(self,match):
        """Remove blank spaces at beginning and end of xml text."""
        if not match.group("close") is None:
            return ">"
        if not match.group("open") is None:
            return "<"
        return match.group(0)


    def GetMessages(self,contents,callback = None):
        """Extract xml messages from given string
            contents: cgsi trace file contents
            callback: function accepting: message_num,method name,xml string
                if None results are printed on standard output
        """
        
        def get_first_node(parent):
            """Get the first non null child node"""
            if parent is None:
                return None
            if not parent.hasChildNodes():
                return None
            for node in parent.childNodes:
                if not node is None and node.nodeType == Node.ELEMENT_NODE:
                    return node
            return None
        
        start = 0
        message_num = 1
        
        while True:
            match = self.xmlre.search(contents,start)
            if match is None:
                break
            # set next search to start from end of current match
            start = match.end() 
            # get xml string
            xml = match.group("xml")
            # remove extra blanks
            xml = self.remove_blanks.sub(self.substitute_blanks,xml)
            # parse the xml string into a dom and get a pretty formatted xml
            method_name = ""
            try:
                dom = parseString(xml)
                xml = dom.toprettyxml("  ") # use two spaces to indent
                c1 = get_first_node(dom)
                c2 = get_first_node(c1)
                c3 = get_first_node(c2)
                if not c3 is None:
                    method_name = c3.localName
            except Exception,exc:
                pass
            if callback is None:
                print "[%d] : %s" % (message_num,method_name)
                print xml
                print "-" * 80
            else:
                callback(message_num,method_name,xml)
            message_num = message_num + 1
            
    def GetMessagesFromFile(self,file_name,callback = None):
        """Extract xml messages from given string
            file_name: the cgsi trace file to open
            callback: function accepting: message_num,method name,xml string
                if None results are printed on standard output
        """

        file = open(file_name)
        file_cleaned_name = file_name + ".clean"
        file_cleaned = open(file_cleaned_name,"w")
        try:
            for line in file:
                try:
                    if None == self.to_remove.match(line):
                        file_cleaned.write(line)
                except Exception,exc:
                    log.error ("Error matching '%s': %s",line,str(exc))
                    file_cleaned.write(line)
            file.close()
            file = None
            file_cleaned.close()
            file_cleaned = None
            file = open(file_cleaned_name)
            contents = file.read()
            file.close()
            file = None
            try:
                os.remove(file_cleaned_name)
            except:
                pass
            self.GetMessages(contents,callback)
        finally:
            if not file is None:
                file.close()
            if not file_cleaned is None:
                file_cleaned.close()            

# main
#-------------------------------------------------------------------------------
def main (argv = None):
    
    # initialize logging to console only
    InitLogging(None)

    if argv is None:
        argv = sys.argv
    if (1 == len(argv)):
        print "specify input file"
        return 1
        
    file_name = argv[1]
    if not os.path.exists(file_name):
        log.error("file '%s' not found",file_name)
        return 1
    
    try:
        cgsi_trace_reader = CgsiTraceReader()
        cgsi_trace_reader.GetMessagesFromFile(file_name)
    except Exception,exc:
        log.error ("error getting messages: %s",str(exc))
        return 1
        
    return 0

# run main
#-------------------------------------------------------------------------------
if __name__ == "__main__":
    sys.exit (main())
