#!/usr/bin/env python
#
# Copyright (c) Members of the EGEE Collaboration. 2004.
# See http://www.eu-egee.org/partners/ for details on the copyright
# holders.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# transfer-url-copy test script configuration
#
# Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
# Version info: $Id: Configuration.py,v 1.8 2009/05/04 12:30:18 molnarzs Exp $
# Release: $Name:  $

import os
import time
from os.path import *
import re
import GLite
import shlex

from Common import *

class Configuration:
    """The tests configuration."""
    
    # default values
    DefaultTitle            = "Results at %{TestYear} %{TestMonth} %{TestDay}, v. %{TucVersion}"
    DefaultMaxPingAttempts  = 3
    DefaultFileSize         = 50
    DefaultMaxTxRunning     = 4
    DefaultMaxTxPerEp       = 2
    DefaultAfsPath          = "/afs/cern.ch/project/gd/www/ftssrm2/"
    DefaultUrlCopyOptions   = "-p TEST_URL_COPY"
    DefaultSrmCopyOptions   = "-p TEST_SRM_COPY"
    DefaultUrlCopyChecksumOptions   = "-p TEST_URL_COPY_WITH_CHECKSUM -K"
    
    def __init__(self):
    
        # test date and time
        self.TestDateYear = time.strftime("%Y")
        self.TestDateMonth = time.strftime("%m")
        self.TestDateDay = time.strftime("%d")
        self.TestDate = self.TestDateYear + self.TestDateMonth + self.TestDateDay
        self.TestTime = time.strftime("%H%M%S")
        
        # transfer url copy version
        self.TucVersion = GLite.transfer_url_copy_version()
        
        # paths for log files
        self.__logDirPath = "/var/tmp/transfer-url-copy-${USER}/%{TestDate}/%{TestTime}"
        self.TestsLogFileName = "tests-log"
        
        # base dir for endpoints
        self.SrmBaseDir = "${USER}/%{TestDate}/%{TestTime}"
        
        # options for copy processes
        self.UrlCopyOptions = Configuration.DefaultUrlCopyOptions
        self.SrmCopyOptions = Configuration.DefaultSrmCopyOptions
        self.UrlCopyChecksumOptions = Configuration.DefaultUrlCopyChecksumOptions
        # attempts to ping an endpoint before excluding it
        self.MaxPingAttempts = Configuration.DefaultMaxPingAttempts

        # size of test file in MB
        self.FileSize = Configuration.DefaultFileSize

        # maximum transfers running
        self.MaxTxRunning = Configuration.DefaultMaxTxRunning
        # maximum transfers per endpoint
        self.MaxTxPerEp   = Configuration.DefaultMaxTxPerEp
        
        self.Title = Configuration.DefaultTitle
              
        self.Endpoints = None

        self.AfsPath = Configuration.DefaultAfsPath
        self.init_params()
    
    def GetEndpoints(self):
        if self.Endpoints is None:
            return None
        return self.Endpoints.split()    
        
    def GetUrlCopyOptions(self):
        return shlex.split(self.UrlCopyOptions)

    def GetUrlCopyChecksumOptions(self):
        return shlex.split(self.UrlCopyChecksumOptions)

    def GetSrmCopyOptions(self):
        return shlex.split(self.SrmCopyOptions)

    def init_params(self):
        # private members for pattern substitution
        self.__regexp = re.compile('\${(?P<env>\w+)}|%{(?P<par>\w+)}')
        self.__parameters = {
            "TestYear"      : self.TestDateYear,
            "TestMonth"     : self.TestDateMonth,
            "TestDay"       : self.TestDateDay,
            "TestDate"      : self.TestDate,
            "TestTime"      : self.TestTime,
            "TucVersion"    : self.TucVersion}

    def GetBaseDir(self):
        return self.__expand(self.SrmBaseDir)
    
    def GetTestDateStr(self):
        return "%s %s %s" % (self.TestDateYear,self.TestDateMonth,self.TestDateDay)
 
    def GetPathForIndex(self):
        return "%s/%s" % (self.TestDate,self.TestTime)

    def LogOptions(self): # WARNING
        log.debug("Log directory              : %s",self.GetLogDirPath())
        log.debug("transfer-url-copy version  : %s",self.TucVersion)
        log.debug("Url copy options           : %s",self.UrlCopyOptions)
        log.debug("Srm copy options           : %s",self.SrmCopyOptions)
        log.debug("Max ping attempts          : %d",self.MaxPingAttempts)
        log.debug("File size                  : %d",self.FileSize)
        log.debug("Max running transfers      : %d",self.MaxTxRunning)
        log.debug("Max transfers per endpoint : %d",self.MaxTxPerEp)

    def PrintOptions(self): # WARNING
        print "Log directory              : %s" % self.GetLogDirPath()
        print "transfer-url-copy version  : %s" % self.TucVersion
        print "Url copy options           : %s" % self.UrlCopyOptions
        print "Srm copy options           : %s" % self.SrmCopyOptions
        print "Max ping attempts          : %d" % self.MaxPingAttempts
        print "File size                  : %d" % self.FileSize
        print "Max running transfers      : %d" % self.MaxTxRunning
        print "Max transfers per endpoint : %d" % self.MaxTxPerEp

    def GetLogDirPath(self):
        return self.__expand(self.__logDirPath)
    def GetTestsLogFileName(self):
        return self.__expand(self.TestsLogFileName)
    def GetTestsLogFilePath(self):
        return Join(self.GetLogDirPath(),self.GetTestsLogFileName())
    
    def GetPath(self,name):
        return Join(self.GetLogDirPath(),name)
   
    def GetAfsDestDir(self):
        dest_dir = Join(self.AfsPath,self.TestDate)
        dest_dir = Join(dest_dir,self.TestTime)
        return dest_dir
    
    def GetTitle(self):
        return self.__expand(self.Title) 
        
    def __expand(self,value):
        return self.__regexp.sub(self.__replace,value)
        
    def __replace(self,match):
        if match.group("env"):
            value = match.group("env")
            return os.environ[value]
        elif match.group("par"):
            value = match.group("par")
            return self.__parameters[value]
        else:
            raise ScriptException("Error in patterns substitution")

# Utilities            
def from_file(path):
    """Create a configuration object and initialize it from given file."""
    cfg = Configuration()
    globals()["options"] = cfg
    execfile(path)
    del globals()["options"]
    return cfg
            
            