#! /usr/bin/env python
#
#  Copyright (c) Members of the EGEE Collaboration. 2004.
#  See http://www.eu-egee.org/partners/ for details on the copyright
#  holders.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  transfer-url-copy test script
#
#  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
#  Version info: $Id: Endpoints.py,v 1.11 2009/05/04 12:30:18 molnarzs Exp $
#  Release: $Name:  $
from Common import *
import re
import yaml

# transfer statistics for an endpoint
class TxStats:
    """Container for transfer statistics."""
    
    def __init__(self):
        self.Reset()
        
    def AddFrom(self,success):
        """Add information for a transfer where the endpoint is the source.
            success: True if the transfer was successful
        """
        self.TxFromTotal = self.TxFromTotal + 1
        if success:
            self.TxFromSucceeded = self.TxFromSucceeded + 1
            
    def AddTo(self,success):
        """Add information for a transfer where the endpoint is the destination.
            success: True if the transfer was successful
        """
        self.TxToTotal = self.TxToTotal + 1
        if success:
            self.TxToSucceeded = self.TxToSucceeded + 1
    
    def Reset(self):
        """Clear all values."""
        self.TxFromSucceeded = 0
        self.TxFromTotal     = 0
        self.TxToSucceeded   = 0
        self.TxToTotal       = 0
    
# Endpoint class
class Endpoint:
    """Container for endpoint data"""
    
    def __init__(self,name,version,host,port,servicePath,sapath,basedir,calculateChecksum=False,doTransfer=True,space=None):
        """Create a new 'Endpoint' object"""
        self.Name = name
        self.Version = version
        self.Host = host
        self.Port = port
        self.ServicePath = servicePath
        self.SaPath = sapath
        self.BaseDir = basedir
        # set srm copy options
        self.SupportsSrmCopyPush = False
        self.SupportsSrmCopyPull = False
        # number of transfers running on endpoint (used by transfer class) 
        self.TxRunning       = 0
        # transfer statistics
        self.UrlCopyStats = TxStats() # url copy transfer
        self.SrmCopyStats = TxStats() # srm copy transfer
        # exclusion flag and reason
        self.Excluded = False
        self.ExcludeReason = ""
        # target space token or source space token
        if space is None:
            space = ""
        self.SpaceToken = space
        # Should we test checksum calculation or not
        self.CalculateChecksum = calculateChecksum
        self.DoTransfer = doTransfer
        # array of files successfully transferred to this endpoint (useful for 
        # transfers: if transfer from local fails, any of these can be used)
        self.TransferredFiles = []
    
    def __cmp__(self,other):
        return cmp(self.Name,other.Name)    
        
    def SetSpaceToken(self,sp):
        """Set the space token to use in copy operations.

            returns the endpoint object"""
        if sp is None:
            sp = ""
        self.SpaceToken = sp
        return self

    def SrmCopy(self,push,pull):
        """Set srmcopy support flags for endpoint.
        
            returns the endpoint object"""
        self.SupportsSrmCopyPush = push
        self.SupportsSrmCopyPull = pull
        return self
    
    def Exclude(self,reason):
        """Set Excluded flag and provide a reason to use in reports.
        
            returns the endpoint object"""
        self.Excluded = True
        self.ExcludeReason = reason
        return self
    
    def Address(self):
        """Get the endpoint address"""
        return "httpg://" + self.__common_addr()
    
    def Surl(self,path):
        """Get the SURL for file with given path"""
        sfn = Join(self.SaPath,self.BaseDir)
        sfn = Join(sfn,path)
        
        return "srm://" + self.__common_addr() + "?SFN=" + sfn 
    
    # utilities
    def __common_addr(self):
        return self.Host + ":" + self.Port + self.ServicePath

# Local class
class Local(Endpoint):
    """Represents the local machine for transfers"""

    def __init__(self):
        Endpoint.__init__(self,"local","","","","","","")

    def SrmCopy(self,push,pull):
        return self
    
    def Address(self):
        return None
        
    def AddAddress(self,args,argname):
        pass
    
    def Surl(self,path):
        return "file://" + path

        
# load endpoints list
#-------------------------------------------------------------------------------

__address_re = re.compile("(?P<host>[^:]+):(?P<port>\d+)(?P<path>.*)")
def _parse_address(address):
    """Parse an address in the form host:port/path
    address: the address to parse
    returns host,port,path
    """
    try:
        match = __address_re.match(address)
        host = match.group("host")
        port = match.group("port")
        path = match.group("path")
        return host,port,path
    except Exception,exc:
        raise ScriptException("Failed parsing endpoint address '%s': %s" % (address,str(exc)))
            
def _load_endpoints(yaml_obj,basedir):
    """create a list of endpoints from yaml object
    """
    _key_endpoint       = "endpoint"
    _key_path           = "path"
    _key_version        = "version"
    _key_token          = "token"
    _key_srmcopy        = "srmcopy"
    _key_srmcopy_push   = "srmcopy-push"
    _key_srmcopy_pull   = "srmcopy-pull"
    _key_exclude        = "exclude"
    _key_checksum       = "checksum"
    _key_transfer       = "transfer"
     
    ret = []
    for name,ep in yaml_obj.items():
        if not ep.has_key(_key_endpoint):
            raise ScriptException("Missing key '%s' for endpoint '%s'" % (_key_endpoint,name))
        if not ep.has_key(_key_path):
            raise ScriptException("Missing key '%s' for endpoint '%s'" % (_key_path,name))
        host,port,servicePath = _parse_address(ep[_key_endpoint])
        sapath = ep[_key_path]
        version = ep.get(_key_version,"2.2")
        token = ep.get(_key_token,"")
        srmcopy = ep.get(_key_srmcopy,False)
        srmcopy_push = ep.get(_key_srmcopy,srmcopy)
        srmcopy_pull = ep.get(_key_srmcopy,srmcopy)
        calculateChecksum = ep.get(_key_checksum,False)
        doTransfer = ep.get(_key_transfer,True)
        endpoint = Endpoint(name,str(version),host,port,servicePath,sapath,basedir,calculateChecksum,doTransfer,token)
        endpoint.SrmCopy(srmcopy_push,srmcopy_pull)
        if ep.has_key(_key_exclude):
            endpoint.Exclude(ep[_key_exclude])
        ret.append(endpoint)
    ret.sort()
    return ret
        
def load_endpoints(path,basedir):
    """load endpoints definitions from file.
    path: path of file with endpoints definitions
    basedir: value for the 'BaseDir' property of the endpoint objects
    return list of endpoints
    """
    yaml_obj = yaml.safe_load(open(path))
    return _load_endpoints(yaml_obj,basedir)
    
# Utilities
#-------------------------------------------------------------------------------

# PrintEndpoints
def PrintEndpoints(list):
    """Print a list of endpoint definitions"""

    def __format_line(index,name,version,address,sapath,surl,space):
        """Format a line of data for the endpoint list"""
        s = " - "
        line = str(index).rjust(3) + s + name.ljust(ml_name) + s + \
            version + s + address.ljust(ml_address) + s + \
            sapath.ljust(ml_sapath) + s + surl.ljust(ml_testsurl) + \
            s + space
        return line
    
    ml_name = len ("name")
    ml_address = len ("address")
    ml_sapath = len ("sapath")
    ml_testsurl = 0
    for endpoint in list:
        ml_name     = max (ml_name,     len (endpoint.Name) )
        ml_address  = max (ml_address,  len (endpoint.Address()) )
        ml_sapath   = max (ml_sapath,   len (endpoint.SaPath) )
        ml_testsurl = max (ml_testsurl, len (endpoint.Surl("/test") ) )
    header = __format_line("  #","name","ver","address","SAPath", \
        "Surl of 'test' file", "Space Token")
    separator = "-" * len(header)
    log.info (separator)
    log.info (header)
    log.info (separator)
    i = 0
    for endpoint in list:
        i = i + 1
        line = __format_line(i,endpoint.Name,endpoint.Version, \
            endpoint.Address(),endpoint.SaPath,endpoint.Surl("/test"), endpoint.SpaceToken)
        log.info (line)
    log.info (separator)

# PrintTxStats
def PrintTxStats(list):
    """Print a list of endpoint transfer statistics"""
    
    ml_name = len ("name")
    for e in list:
        ml_name = max (ml_name,len (e.Name) )
    header = "name".ljust(ml_name) + " - urlcopy from - urlcopy to - srmcopy from - srmcopy to "
    separator = "-" * len(header)
    log.info (separator)
    log.info (header)
    log.info (separator)
    i = 0
    for e in list:
        i = i + 1
        line = e.Name.ljust(ml_name) + (" -   %3d/%3d    -   %3d/%3d  -    %3d/%3d   -   %3d/%3d " % ( \
            e.UrlCopyStats.TxFromSucceeded,e.UrlCopyStats.TxFromTotal,e.UrlCopyStats.TxToSucceeded,e.UrlCopyStats.TxToTotal, \
            e.SrmCopyStats.TxFromSucceeded,e.SrmCopyStats.TxFromTotal,e.SrmCopyStats.TxToSucceeded,e.SrmCopyStats.TxToTotal)) 
        log.info (line)
    log.info (separator)

# Tests
#-------------------------------------------------------------------------------
if __name__ == "__main__": 
    import unittest
    # TestEndpoints
    class TestEndpoints(unittest.TestCase):
        
        def test_parse_address_01(self):
            host,port,path = _parse_address("srm-dteam.cern.ch:8443/srm/managerv2")
            assert host == "srm-dteam.cern.ch"
            assert port == "8443"
            assert path == "/srm/managerv2"

        def test_parse_address_02(self):
            self.assertRaises(ScriptException,_parse_address,"srm-dteam.cern.ch/srm/managerv2")
                        
        def test_load_01(self):
            yaml_obj = yaml.load("""
            castor-cern:
                endpoint:   srm-dteam.cern.ch:8443/srm/managerv2
                path:       /castor/cern.ch/grid/dteam
                token:      srm2_d0t1
            castor-pps:
                endpoint:   srm-pps.cern.ch:8443/srm/managerv1
                path:       /castor/cern.ch/grid/dteam
                version:    1.1
                srmcopy:    True
            """)
            endpoints = _load_endpoints(yaml_obj,"testdir")
            assert len(endpoints) == 2
            ep0 = None
            ep1 = None
            if endpoints[0].Name == "castor-cern":
                ep0 = endpoints[0]
                ep1 = endpoints[1]
            else:
                ep0 = endpoints[1]
                ep1 = endpoints[0]
            assert "castor-cern" == ep0.Name
            assert "2.2" == ep0.Version
            assert "srm-dteam.cern.ch" == ep0.Host
            assert "8443" == ep0.Port
            assert "/srm/managerv2" == ep0.ServicePath
            assert "/castor/cern.ch/grid/dteam" == ep0.SaPath
            assert "testdir" == ep0.BaseDir
            assert False == ep0.SupportsSrmCopyPush
            assert False == ep0.SupportsSrmCopyPull
            assert "srm2_d0t1" == ep0.SpaceToken
            assert "castor-pps" == ep1.Name
            assert "1.1" == ep1.Version
            assert "srm-pps.cern.ch" == ep1.Host
            assert "8443" == ep1.Port
            assert "/srm/managerv1" == ep1.ServicePath
            assert "/castor/cern.ch/grid/dteam" == ep1.SaPath
            assert "testdir" == ep1.BaseDir
            assert True == ep1.SupportsSrmCopyPush
            assert True == ep1.SupportsSrmCopyPull
            assert "" == ep1.SpaceToken
            
        def test_load_02(self):
            yaml_obj = yaml.load("""
            castor-cern:
                path:       /castor/cern.ch/grid/dteam
                token:      srm2_d0t1
            """)
            self.assertRaises(ScriptException,_load_endpoints,yaml_obj,"testdir")
            
        def test_load_03(self):
            yaml_obj = yaml.load("""
            castor-cern:
                endpoint:   srm-dteam.cern.ch:8443/srm/managerv2
            """)
            self.assertRaises(ScriptException,_load_endpoints,yaml_obj,"testdir")
            
    unittest.main()
            

