#! /usr/bin/env python
#
#  Copyright (c) Members of the EGEE Collaboration. 2004.
#  See http://www.eu-egee.org/partners/ for details on the copyright
#  holders.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  transfer-url-copy test script
#
#  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
#  Version info: $Id: GLite.py,v 1.7 2008/12/10 16:13:55 ptedesco Exp $
#  Release: $Name:  $

import os
import sys
import getopt
import time
import subprocess

from Common import *
from Endpoints import *
import CgsiTraceReader

class GLiteProcess:
    """An helper class for process and log files handling."""

    CgsiExt = ".cgsi"

    _cgsiTraceReader = CgsiTraceReader.CgsiTraceReader()

    def __init__(self,args,logFilePath):
        self.LogFilePath = logFilePath
        self.CgsiFilePath = logFilePath + GLiteProcess.CgsiExt
        self._log_file = open(self.LogFilePath,'w')
        self.ExitCode = None
        log.debug(">>>" + " ".join(args))
        self._proc = subprocess.Popen(args,stdout=self._log_file,stderr=subprocess.STDOUT,env=self._getEnv())
        # LogRelDir is used to store the directory where the log file was moved after completion
        self.LogRelDir = None

    def Wait(self):
        self.ExitCode = self._proc.wait()
        self._close()
        return self.ExitCode
    
    def Poll(self):
        self.ExitCode = self._proc.poll()
        if self.ExitCode != None:
            self._close()
        return self.ExitCode

    def _close(self):
        """Close the log file handle, concatenate cgsi messages and cleanup."""
        self._log_file.close()
        if os.path.exists(self.CgsiFilePath):
            try:
                self._log_file = open(self.LogFilePath,'a')
                self._log_file.write("\n\n--- SOAP MESSAGES ---\n\n")
                def append_to_log_file(message_num,method_name,xml_string):
                    self._log_file.write("\n[%d] : %s\n" % (message_num,method_name))
                    self._log_file.write(xml_string)
                GLiteProcess._cgsiTraceReader.GetMessagesFromFile(self.CgsiFilePath,append_to_log_file)
                RemoveFile(self.CgsiFilePath)
            except Exception, exc:
                log.error ("Error getting cgsi messages : " + str(exc))
            finally:
                self._log_file.close()
        base_dir = os.path.dirname(self.LogFilePath)
        file_name = os.path.basename(self.LogFilePath)
        if self.ExitCode == 0:
            self.LogRelDir = "completed"
        else:
            self.LogRelDir = "failed"
        dest_path = Join(base_dir,self.LogRelDir)
        dest_path = Join(dest_path,file_name)
        self.LogFilePath = MoveFile(self.LogFilePath,dest_path)

    def _getEnv(self):
        """Get the environment to pass to processes to enable cgsi trace."""
        env = os.environ.copy()
        env["CGSI_TRACE"]     = "1"
        env["CGSI_TRACEFILE"] = self.CgsiFilePath
        return env

def GLiteUrlCopy(source,destination,sourceFileName,destFileName,logFilePath,options):
    """Invoke glite-url-copy.
        source : source endpoint object.
        destination : destination endpoint object.
        sourceFileName : name (relative path) of source file in endpoint.
        destFileName : name (relative path) of destination file in endpoint.
        logFilePath : path of log file for copy process.
        options : array of options to append to command line.
    """
    # create the arguments list
    args = [glite_bin_path ("glite-url-copy"),"-b","-d","-C"]
    if not destination.SpaceToken is None and not destination.SpaceToken == "":
        args[len(args):] = ["-T",destination.SpaceToken]
    if not source.SpaceToken is None and not source.SpaceToken == "":
        args[len(args):]= ["-S",source.SpaceToken]
    if options != None:
        args[len(args):] = options
    args.append(source.Surl(sourceFileName))
    args.append(destination.Surl(destFileName))

    # return the process wrapper
    return GLiteProcess(args,logFilePath)

def GLiteSrmCopy(source,destination,sourceFileName,destFileName,push,logFilePath,options):
    """Invoke glite-srm-copy.
        source : source endpoint object.
        destination : destination endpoint object.
        sourceFileName : name (relative path) of source file in endpoint.
        destFileName : name (relative path) of destination file in endpoint.
        push : true for srm copy in push mode.
        logFilePath : path of log file for copy process.
        options : array of options to append to command line.
    """
    # create the arguments list
    args = [glite_bin_path ("glite-srm-copy"),"-b","-d","-C"]
    if push:
        args.append("-P")
    if not destination.SpaceToken is None and not destination.SpaceToken == "":
        args[len(args):] = ["-T",destination.SpaceToken]
    if not source.SpaceToken is None and not source.SpaceToken == "":
        args[len(args):]= ["-S",source.SpaceToken]
    if options != None:
        args[len(args):] = options
    args.append(source.Surl(sourceFileName))
    args.append(destination.Surl(destFileName))
    
    # return the process wrapper
    return GLiteProcess(args,logFilePath)
    
def GLiteRemove(endpoint,file,logFilePath):
    """Remove a file."""
    args = [glite_libexec_path("glite-srm-util-rm"),"-I",endpoint.Version,"-s",endpoint.Address(),endpoint.Surl(file)]
    proc = GLiteProcess(args,logFilePath)
    return proc.Wait()

def GLitePing (endpoint,logFilePath):
    """Ping given endpoint."""
    args = [glite_libexec_path("glite-srm-util-ping"),"-I",endpoint.Version,"-s",endpoint.Address()]
    proc = GLiteProcess(args,logFilePath)
    return proc.Wait()

# utilities

def glite_bin_path(name):
    value = Join(os.environ["GLITE_LOCATION"],"bin")
    return Join(value,name)

def glite_libexec_path(name):
    value = Join(os.environ["GLITE_LOCATION"],"libexec")
    return Join(value,name)

def transfer_url_copy_version():
    cmd = [glite_bin_path("glite-url-copy"),"-V"]
    tuc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    code = tuc.wait()
    output = tuc.stdout.readlines()[0]
    ol = len(output)
    if output[ol - 1] == '\n':
        output = output[0 : ol-1]
    return output


