#! /usr/bin/env python
#
#  Copyright (c) Members of the EGEE Collaboration. 2004.
#  See http://www.eu-egee.org/partners/ for details on the copyright
#  holders.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  transfer-url-copy test script
#
#  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
#  Version info: $Id: HtmlReport.py,v 1.8 2008/12/10 16:13:55 ptedesco Exp $
#  Release: $Name:  $

import os
import xml.dom.minidom
from Endpoints import *
from Common import *

"""Module to generate html reports with test results."""

# Tag
#-------------------------------------------------------------------------------
class Tag:
    """An html tag."""
    
    def __init__(self,name,text=None):
        self.Name = name
        #self.doc = doc
        self.attributes = {}
        self.children = []
        self.text = text
        
    def getnode(self,doc):
        """Get an xml node representing the tag.
            doc: xml document.
        """
        node = doc.createElement(self.Name)
        for key,value in self.attributes.iteritems():
            node.setAttribute(key,value)
        for child in self.children:
            node.appendChild(child.getnode(doc))
        if not self.text is None:
            text_node = doc.createTextNode(self.text)
            node.appendChild(text_node)
        return node


# TH
#-------------------------------------------------------------------------------
class TH(Tag):
    """A table th element."""
    
    def __init__(self,text=None):
        Tag.__init__(self,"th",text)


# TD
#-------------------------------------------------------------------------------
class TD(Tag):
    """A table td element."""
    
    def __init__(self,text=None):
        Tag.__init__(self,"td",text)


# TR
#-------------------------------------------------------------------------------
class TR(Tag):
    """A table tr element."""
    
    def __init__(self,elements=None,headers=None):
        """Create a new tr element.
            elements: list of strings to be added as TDs.
            headers: list of strings to be added as THs.
        """
        Tag.__init__(self,"tr")
        if not elements is None:
            for element in elements:
                self.children.append(TD(element))
        if not headers is None:
            for header in headers:
                self.children.append(TH(header))


# Table
#-------------------------------------------------------------------------------
class Table(Tag):
    """An html table."""
    
    def __init__(self,headers=None,caption=None,border=1,cellspacing=1):
        """Create a new table.
            headers: list of strings to be added as headers.
            caption: table caption
            border: table border
            cellspacing: table cellspacing
        """
        Tag.__init__(self,"table")
        self.attributes["border"] = str(border)
        self.attributes["cellspacing"] = str(cellspacing)
        self.caption = None
        if not caption is None:
            self.caption = Tag("caption")
            self.caption.text = caption
            self.children.append(self.caption)
        self.headers = None
        if not headers is None:
            self.headers = TR(headers=headers)
            self.children.append(self.headers)

# CssLink
#-------------------------------------------------------------------------------
class CssLink(Tag):
    """The page stylesheet"""
    def __init__(self):
        Tag.__init__(self,"link")
        self.attributes["rel"] = "stylesheet"
        self.attributes["type"] = "text/css"
        self.attributes["href"] = "http://test-fts-srm2.web.cern.ch/test-fts-srm2/style.css"
        
# Head
#-------------------------------------------------------------------------------
class Head(Tag):
    """The page head element."""
    
    def __init__(self):
        Tag.__init__(self,"head")
        self.title = Tag("title")
        self.children.append(self.title)
        self.children.append(CssLink())

# Page
#-------------------------------------------------------------------------------
class Page(Tag):
    """An html page."""
    
    def __init__(self):
        Tag.__init__(self,"html")
        self.head = Head()
        self.children.append(self.head)
        self.body = Tag("body")
        self.children.append(self.body)
        
    def __str__(self):
        """Return html string representation of page."""
        doc = xml.dom.minidom.Document()
        doc.appendChild(self.getnode(doc))
        return doc.toprettyxml()

def set_info(tag):
    #tag.attributes["bgcolor"] = "LightBlue"
    tag.attributes["class"] = "info"
def set_ok(tag):
    #tag.attributes["bgcolor"] = "LightGreen"
    tag.attributes["class"] = "ok"
def set_warn(tag):
    #tag.attributes["bgcolor"] = "LightYellow"
    tag.attributes["class"] = "warn"
def set_err(tag):
    #tag.attributes["bgcolor"] = "Crimson"
    tag.attributes["class"] = "err"
    
# HtmlReport
#-------------------------------------------------------------------------------
class HtmlReport:
    """Transfer tests report page generator."""
   
    def __init__(self,cfg,date,endpoints,tx_map):
        """Constructor.
        NOTE: the date is passed explicitly (and not taken from the 
        configuration object) to allow results recovery."""
        self.page = Page()
        title = "FTS - SRM2.2 transfer tests - %s" % (date)
        self.page.head.title.text = title
        self.page.body.children.append(Tag("h1",title))
        
        self.page.body.children.append(Tag("h2","Test configuration"))
        config_table = Table()
        config_table.children.append(TR(["transfer-url-copy library version",cfg.TucVersion]))
        config_table.children.append(TR(["Url copy options",cfg.UrlCopyOptions]))
        config_table.children.append(TR(["Srm copy options",cfg.SrmCopyOptions]))
        config_table.children.append(TR(["File size",str(cfg.FileSize)]))
        self.page.body.children.append(config_table)
                
        self.page.body.children.append(Tag("h2","Endpoint definitions"))
        
        self.page.body.children.append(HtmlReport.create_endpoints_table(endpoints))
        
        dest_ep_names = []
        #srm_ep_names = []

        pull_source_ep_names = []
        pull_dest_ep_names = []
        push_source_ep_names = []
        push_dest_ep_names = []

        for ep in endpoints:
            if "local" == ep.Name:
                continue
            if not ep.Excluded:
                dest_ep_names.append(ep.Name)
                if ep.SupportsSrmCopyPull:
                    pull_dest_ep_names.append(ep.Name)
                pull_source_ep_names.append(ep.Name)
                if ep.SupportsSrmCopyPush:
                    push_source_ep_names.append(ep.Name)
                push_dest_ep_names.append(ep.Name)

                #if ep.SupportsSrmCopyPull or ep.SupportsSrmCopyPush:
                #    srm_ep_names.append(ep.Name)

        source_ep_names = ["local"] + dest_ep_names
        
        self.page.body.children.append(Tag("h2","URL copy transfers"))
        self.page.body.children.append(\
            self.create_tx_table(endpoints,tx_map,"url",source_ep_names,dest_ep_names))

        self.page.body.children.append(Tag("h2","SRM copy PULL transfers"))
        self.page.body.children.append(\
            self.create_tx_table(endpoints,tx_map,"srm-pull",pull_source_ep_names,pull_dest_ep_names))

        self.page.body.children.append(Tag("h2","SRM copy PUSH transfers"))
        self.page.body.children.append(\
            self.create_tx_table(endpoints,tx_map,"srm-push",push_source_ep_names,push_dest_ep_names))
    
    @staticmethod 
    def CreateEndpointsPage(endpoints):
        page = Page()
        title = "FTS - SRM2.2 transfer tests - endpoint definitions"
        page.head.title.text = title
        page.body.children.append(Tag("h1",title))
        page.body.children.append(HtmlReport.create_endpoints_table(endpoints))
        return page

    @staticmethod
    def create_endpoints_table(endpoints):
        """Create a table with endpoint definitions."""
        ep_table = Table(["Name","Host","Port","Service path","SApath","Space","SRMCopy push","SRMCopy pull","Status"])
        set_info(ep_table.headers)
        def s_bool(b):
            if b: return "yes"
            return "no"
        def s_status(b):
            if b == "": return "ok"
            return b
        for ep in endpoints:
            if "local" == ep.Name:
                continue
            erow = TR( [ep.Name,ep.Host,ep.Port,ep.ServicePath,ep.SaPath, \
                ep.SpaceToken, s_bool(ep.SupportsSrmCopyPush),\
                s_bool(ep.SupportsSrmCopyPull), s_status(ep.ExcludeReason)])
            if ep.Excluded:
                set_warn(erow)
            else:
                set_ok(erow)
            ep_table.children.append(erow)
        return ep_table
    
    def create_tx_table(self,endpoints,tx_map,mode,source_ep_names,dest_ep_names):
        """Create a table with transfer results."""
        tx_table_headers = ["From/To"] + dest_ep_names
        tx_table = Table(tx_table_headers)
        set_info(tx_table.headers)
        # url copy transfers
        for source in source_ep_names:
            tr = TR()
            tx_table.children.append(tr)
            source_td = TD(source)
            set_info(source_td)
            tr.children.append(source_td)
            for dest in dest_ep_names:
                tx_name = mode + "-" + source + "-to-" + dest
                if tx_name in tx_map:
                    tx = tx_map[tx_name]
                    td = TD()
                    tr.children.append(td)
                    log_path = ""
                    log_caption = ""
                    log_path = os.path.basename(tx.GetLogFilePath())
                    log_path = Join(tx.GetLogFileRelDir(),log_path)
                    if tx.Succeeded:
                        if None == tx.FailureReason:
                            set_ok(td)
                            log_caption = "ok"
                        else:
                            set_warn(td)
                            log_caption = tx.FailureReason
                    else:
                        set_err(td)
                        if None == tx.FailureReason:
                            log_caption = "failure"
                        else:
                            log_caption = tx.FailureReason
                    href = Tag("a",log_caption)
                    href.attributes["href"]=log_path
                    td.children.append(href)
                else:
                    td = TD("-")
                    set_warn(td)
                    tr.children.append(td)
        return tx_table
    
    def write(self,path):
        """Write results to file."""
        file = open(path,"w")
        file.write(str(self.page))
        file.close()
