#!/usr/bin/env python
#
# Copyright (c) Members of the EGEE Collaboration. 2004.
# See http://www.eu-egee.org/partners/ for details on the copyright
# holders.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# transfer-url-copy test script
#
# Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
#          Zsolt Molnar <Zsolt.Molnar@cern.ch - www.zsoltmolnar.hu>
# Version info: $Id: TestSet.py,v 1.8 2009/05/04 12:30:18 molnarzs Exp $
# Release: $Name:  $

import random
import os
import zlib
import ctypes

from Common import *
from Endpoints import *
from Configuration import *
from GLite import *
from Transfers import *
from md5 import md5

class TestSet:
    
    TmpFileName = "tmp.dat"
    TmpChecksumFileName = "tmpchksum.dat"
    
    def __init__(self,cfg):

        # endpoints
        # list of endpoints configured to be tested
        self.Endpoints = []
        # list of available (not excluded and successfully pinged) endpoints
        self.AvailableEndpoints = []
        # fake endpoint for local machine
        self.Local = Local()
        
        # the configuration object
        self.Cfg = cfg
        
        # init random
        random.seed()
        
        # transfer lists
        self._queued_transfers = [] # transfers not yet started
        self._current_transfers = [] # transfers currently running
        self._finished_transfers = [] # finished transfers
        
    def initialize(self,endpointsFile = "endpoints.yaml"):
        """Initialize options and enpoint list reading the configuration file.
            Should not be used directly by the transfer test program, but only
            from the class itself and by the main method to print the endpoints
            list.
        """
        self.Endpoints = load_endpoints(endpointsFile,self.Cfg.GetBaseDir())
        if not self.Cfg.Endpoints is None:
            enabled_endpoints = self.Cfg.GetEndpoints()
            for ep in self.Endpoints:
                if not ep.Name in enabled_endpoints:
                    ep.Exclude("Disabled by tests configuration.")
        
    def RunStart(self):
        self.initialize()
        self.Cfg.LogOptions()

        # create tests root dir
        MakeDir(self.Cfg.GetLogDirPath())
        
        # run the tests
        log.debug("Starting tests")
        PrintEndpoints(self.Endpoints)

        self._temp_file_path = self.Cfg.GetPath(TestSet.TmpFileName)
        self.create_temp_file(self._temp_file_path, self.Cfg.FileSize)
        self._temp_checksum_file_path = self.Cfg.GetPath(TestSet.TmpChecksumFileName)
        self.create_temp_file(self._temp_checksum_file_path, self.Cfg.ChecksumFileSize)
        
        self._temp_checksum_adler32 = zlib.adler32(file(self._temp_checksum_file_path).read())
        self._temp_checksum_adler32 = "%.8x" % self._temp_checksum_adler32
        log.debug("Adler32 checksum of test file: " + self._temp_checksum_adler32)
        
        a = ctypes.c_uint(0)
        a.value = zlib.crc32(file(self._temp_checksum_file_path).read())
        self._temp_checksum_crc32 = "%.8x" % a.value
        log.debug("CRC32 checksum of test file: " + self._temp_checksum_crc32)
        
        block_size = 0x10000
        def upd(m, data):
            m.update(data)
            return m
        
        fd = open(self._temp_checksum_file_path, "rb")
        try:
            contents = iter(lambda: fd.read(block_size), "")
            m = reduce(upd, contents, md5())
            self._temp_checksum_md5 = "%s" % m.hexdigest()
            log.debug("MD5 checksum of test file: " + self._temp_checksum_md5)
        finally:
            fd.close()
        
        

        log.debug("Determining available endpoints")
        self.get_available_endpoints()
        
        random.shuffle(self.AvailableEndpoints)     
        
    def Run(self):
        self.RunStart()
        # create a local - endpoint transfer object and add it to list of
        # queued transfers
        for endpoint in self.AvailableEndpoints:
            if endpoint.DoTransfer:
                transfer = UrlCopyTransfer(self.Local,endpoint,self._temp_file_path,options=self.Cfg.GetUrlCopyOptions())
                self._queued_transfers.append(transfer)
        
            if endpoint.CalculateChecksum:
                # Use case 2 - user does provides checksum
                opt = self.Cfg.GetUrlCopyChecksumOptions();
                opt.extend(['-a', 'ADLER32', '-c', self._temp_checksum_adler32])
                transfer = UrlCopyTransfer(self.Local,endpoint,self._temp_checksum_file_path,options=opt, mode="url-uc2-adler32", isChkSum=True)
                self._queued_transfers.append(transfer)
               
                opt = self.Cfg.GetUrlCopyChecksumOptions();
                opt.extend(['-a', 'CRC32', '-c', self._temp_checksum_crc32])
                transfer = UrlCopyTransfer(self.Local,endpoint,self._temp_checksum_file_path,options=opt, mode="url-uc2-crc32", isChkSum=True)
                self._queued_transfers.append(transfer)
        
                opt = self.Cfg.GetUrlCopyChecksumOptions();
                opt.extend(['-a', 'MD5', '-c', self._temp_checksum_md5])
                transfer = UrlCopyTransfer(self.Local,endpoint,self._temp_checksum_file_path,options=opt, mode="url-uc2-md5", isChkSum=True)
                self._queued_transfers.append(transfer)

        self.RunFinish()
        
    def RunFinish(self):
        try:
            last_running_len = 0
            last_queued_len  = 0

            while True:
                
                # search finished transfers
                self.check_finished_transfers()

                # start new transfers
                self.start_new_transfers()

                if (0 == len (self._current_transfers)) and (0 == len (self._queued_transfers)):
                    # all transfers completed
                    break
                # log progress
                current_running_len = len(self._current_transfers)
                current_queued_len = len(self._queued_transfers)
                if not last_running_len == current_running_len or not last_queued_len == current_queued_len:
                    last_running_len = current_running_len
                    last_queued_len  = current_queued_len
                    log.debug ("running = %d - queued = %d" % (current_running_len,current_queued_len) )
                
                time.sleep(10)
        except ScriptException, exc:
            log.error ("script error: " + str(exc))
        except Exception, exc:
            log.error ("Exception : " + str(type(exc)) + " args = " + str(exc.args))
            raise
        finally:
            # remove temporary file
            RemoveFile(self._temp_file_path)
            RemoveFile(self._temp_checksum_file_path)
            # remove transferred files
            for transfer in self._finished_transfers:
                log_file_name = "rm-" + transfer.Name
                log_file_path = self.Cfg.GetPath(log_file_name)
                transfer.RemoveDestinationFile(log_file_path)
                
                
    def GetTxMap(self):
        tx_map = {}
        for transfer in self._finished_transfers:
            tx_map[transfer.Name] = transfer
        return tx_map
    
    def GetAllEndpoints(self):
        return [self.Local] + self.Endpoints

    def create_temp_file(self, path, fileSizeInMbytes):
        log.debug("Creating temporary file " + path)
        tsize = int(1048576 * fileSizeInMbytes)
        # verify if tmp file exists
        if (os.path.exists(path)):
            size = os.stat(path)[ST_SIZE]
            if (size == tsize):
                log.debug("temporary file exists")
                return
            else:
                log.debug("temporary file exists but has wrong size (%d - %d)" % (size,tsize))
        cmd = "dd if=/dev/zero of=" + path + " bs=1048576 count=" + str(fileSizeInMbytes) + " > /dev/null 2>&1"
        res = os.system(cmd)
        if not 0 == res:
            raise ScriptException("Could not create temporary file " + path + " with size " + str(tsize)) 


    def get_available_endpoints(self):
        self.AvailableEndpoints = []
        
        for endpoint in self.Endpoints:
            if endpoint.Excluded:
                log.info ("excluding endpoint " + endpoint.Name + " : " + endpoint.ExcludeReason)
                continue
            
            base_log_file_name = "ping-" + endpoint.Name
            attempt = 1
            result = 0
            while attempt <= self.Cfg.MaxPingAttempts:
                log.debug ("pinging endpoint " + endpoint.Name + " (" + str(attempt) + ")" )
                log_file_path = self.Cfg.GetPath(base_log_file_name + "-" + str(attempt))
                result = GLitePing(endpoint,log_file_path)
                if (0 == result):
                    break
                attempt = attempt + 1
            if (0 == result):
                log.info ("endpoint " + endpoint.Name + " ping succeeded")
                self.AvailableEndpoints.append(endpoint)
            else:
                log.warning ("could not ping endpoint " + endpoint.Name + " (attempts = " + str(self.Cfg.MaxPingAttempts) + "), excluding from tests")
                endpoint.Exclude("Ping failed")
    
    def check_finished_transfers(self):
        # double loop is necessary because on_tx_finished modifies
        # transfer lists
        tx_ended = True
        while (len (self._current_transfers) > 0) and tx_ended:
            tx_ended = False
            for tx in self._current_transfers:
                if not tx.IsRunning():
                    tx_ended = True
                    self.on_tx_finished(tx)
                    break

    def on_tx_finished(self,tx):
        # tx is over, remove it from running txs...
        self._current_transfers.remove(tx)
        self._finished_transfers.append(tx)
        if (1 == len(tx.Destination.TransferredFiles)) and (tx.Succeeded):
            # this is the first successful transfer for the destination; add 
            # transfers with destination as source
            log.debug ("adding transfers from : " + tx.Destination.Name)
            # use destination as source for new transfers
            source = tx.Destination
            # the transfer was a local->endpoint tx; add endpoint->other endpoints
            # transfers to list of transfers to perform
            for destination in self.AvailableEndpoints:
                # always add url copy mode transfer
                if tx.IsChecksumTest:
                    # This code checks Checksum Use Case 1
                    opts = self.Cfg.GetUrlCopyChecksumOptions()
                else:
                    opts = self.Cfg.GetUrlCopyOptions()
                
                new_tx = UrlCopyTransfer(source,destination,options=opts,mode=tx.Mode)
                log.debug ("adding transfer " + new_tx.Name)
                self._queued_transfers.append(new_tx)
                # add srm copy transfers checking if endpoints support srm copy
                if destination.SupportsSrmCopyPull:
                    pull_tx = SrmCopyTransfer(source,destination,False,options=self.Cfg.GetSrmCopyOptions())
                    self._queued_transfers.append(pull_tx)
                    
                if source.SupportsSrmCopyPush:
                    push_tx = SrmCopyTransfer(source,destination,True,options=self.Cfg.GetSrmCopyOptions())
                    self._queued_transfers.append(push_tx)
                    
            random.shuffle(self.AvailableEndpoints)    

    def start_new_transfers(self):
        tx_added = True
        while (len (self._current_transfers) < self.Cfg.MaxTxRunning) and tx_added:
            tx_added = False
            for tx in self._queued_transfers:
                if (tx.Source.TxRunning < self.Cfg.MaxTxPerEp) and (tx.Destination.TxRunning < self.Cfg.MaxTxPerEp):
                    tx_added = True
                    self._queued_transfers.remove(tx)
                    self._current_transfers.append(tx)
                    tx.Start(self.Cfg.GetPath(tx.Name))
                    break


def print_endpoints(list):
    def format_line(name,version,address,sapath):
        s = " - "
        line = name.ljust(ml_name) + s + version + s + address.ljust(ml_address) + s + sapath.ljust(ml_sapath)
        return line
    ml_name = len ("name")
    ml_address = len ("address")
    ml_sapath = len ("sapath")
    ml_testsurl = 0
    for endpoint in list:
        ml_name     = max (ml_name,     len (endpoint.Name) )
        ml_address  = max (ml_address,  len (endpoint.Address()) )
        ml_sapath   = max (ml_sapath,   len (endpoint.SaPath) )
    header = format_line("name","ver","address","SAPath")
    separator = "-" * len(header)
    print (separator)
    print (header)
    print (separator)
    i = 0
    for endpoint in list:
        i = i + 1
        line = format_line(endpoint.Name,endpoint.Version,endpoint.Address(),endpoint.SaPath)
        print (line)
    print (separator)

# main (test)
#-------------------------------------------------------------------------------
def main (argv = None):
    if argv is None:
        argv = sys.argv
    
    t = TestSet()
    t.initialize()
    print_endpoints(t.Endpoints)

    return 0

# run main
#-------------------------------------------------------------------------------
if __name__ == "__main__":
    sys.exit (main())
