#! /usr/bin/env python
#
#  Copyright (c) Members of the EGEE Collaboration. 2004.
#  See http://www.eu-egee.org/partners/ for details on the copyright
#  holders.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  transfer-url-copy test script
#
#  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
#  Version info: $Id: Transfers.py,v 1.11 2009/05/04 12:30:18 molnarzs Exp $
#  Release: $Name:  $

import os
import sys
import time
import re

from Common import *
from Endpoints import *
from GLite import *

# Transfer
#-------------------------------------------------------------------------------
class Transfer:
    """Base class for transfers"""
    
    def __init__(self,mode,source,destination,sourceName = None,destName = None,options = None, isChkSum = False):
        self.Source = source
        self.Destination = destination
        self.Mode = mode
        self.Name = self.Mode + "-" + self.Source.Name + "-to-" + self.Destination.Name
        if sourceName is None:
            if len(self.Source.TransferredFiles) > 0:
                sourceName = self.Source.TransferredFiles[0]
            else:
                raise ScriptException("cannot determine source file name")
                # by default, use the file copied with local transfer
                sourceName = "url-local-to-" + self.Source.Name + ".dat"
        if destName is None:
            destName = self.Mode + "-" + self.Source.Name + "-to-" + self.Destination.Name + ".dat"
        self.Started = False
        self.Succeeded = False
        self.Finished = False
        self.SourceName = sourceName
        self.DestName = destName
        self._glite_proc = None
        self.FailureReason = None
        self.Options = options
        self.IsChecksumTest = isChkSum

    def Start(self,logFilePath):
        if self.Started:
            raise ScriptException ("transfer already started")
        log.info ("starting tranfer " + self.Name)
        self.Started = True
        self.Source.TxRunning = self.Source.TxRunning + 1
        self.Destination.TxRunning = self.Destination.TxRunning + 1
        self._glite_proc = self._start_copy(logFilePath)
    
    def _start_copy(self,logFilePath):
        raise ScriptException("_start_copy called on base class")
    
    def IsRunning(self):
        if (not self.Started) or (self.Finished):
            return False
        self._exitcode = self._glite_proc.Poll()
        if self._exitcode is None:
            return True
        else:
            self.__on_end_tx()
            return False
    
    def Wait(self):
        if not self.Started:
            raise ScriptException ("transfer not started")
        self._exitcode = self._glite_proc.Wait()
        self.__on_end_tx()
    
    def RemoveDestinationFile(self,logFilePath):
        if (not self.Succeeded):
            log.debug ("Transfer " + self.Name + " failed, no cleanup needed")
            return
        log.info ("removing " + self.DestName + " from " + self.Destination.Name)
        result = GLiteRemove(self.Destination,self.DestName,logFilePath)
        if (0 == result):
            log.info (self.DestName + " removed successfully from " + self.Destination.Name)
        else:
            log.error ("removing " + self.DestName + " from " + self.Destination.Name + " failed")
    
    def GetLogFilePath(self):
        return self._glite_proc.LogFilePath
        
    def __on_end_tx(self):
        self.Finished = True
        self.Succeeded = (0 == self._exitcode)
        self.FailureReason = self.GetFailureReason(self._glite_proc.LogFilePath)
        if (self.Succeeded):
            if self.FailureReason != None:
                log.warning("transfer " + self.Name + " succeeded, but failure reason was found in log file.")
            else:
                log.info ("tranfer " + self.Name + " succeeded")
        else:
            if self.FailureReason != None:
                log.warning ("tranfer " + self.Name + " failed : " + self.FailureReason)
            else:
                log.warning("transfer " + self.Name + " failed, but failure reason was not found in log file.")
        self.Source.TxRunning = self.Source.TxRunning - 1
        self.Destination.TxRunning = self.Destination.TxRunning - 1
        if self.Succeeded:
            self.Destination.TransferredFiles.append(self.DestName)
        self._update_stats()
    
    def GetLogFileRelDir(self):
        # for links in html reports
        return self._glite_proc.LogRelDir

    @classmethod
    def GetFailureReason(cls,path):
        f = None
        reason = None
        try:
            log.debug("opening " + path)
            f = open(path)
            for line in f:
                reason = cls._parse_failure(line)
                if not reason is None:
                    break
        except Exception,exc:
            log.error ("failed getting failure reason : " + str(exc))
        finally:
            if not f is None:
                f.close()
        return reason
    
    @classmethod
    def _parse_failure(self,line):
        raise ScriptException("_parse_failure called on base class")
    
    def _update_stats(self):
        raise ScriptException("_update_stats called on base class")

# UrlCopyTransfer
#-------------------------------------------------------------------------------
class UrlCopyTransfer(Transfer):
    """A tranfer in UrlCopy mode"""
    
    failure_re = re.compile(r'failed on (?P<scope>\w*) during (?P<phase>\w*) phase\. Reason is: \[(?P<category>\w*)\] (?P<reason>.*)')
    
    def __init__(self,source,destination,sourceName = None,destName = None,options = None,mode = "url", isChkSum = False):
        Transfer.__init__(self,mode,source,destination,sourceName,destName,options,isChkSum)
        
    def _update_stats(self):
        self.Source.UrlCopyStats.AddFrom(self.Succeeded)
        self.Destination.UrlCopyStats.AddTo(self.Succeeded)
    
    def _start_copy(self,logFilePath):
        return GLiteUrlCopy(self.Source,self.Destination,self.SourceName,self.DestName,logFilePath,self.Options)

    @classmethod
    def _parse_failure(cls,line):
        m = UrlCopyTransfer.failure_re.search(line)
        if not m is None:
            return "%s - %s - %s - %s" % (m.group("scope"),m.group("phase"),m.group("category"),m.group("reason"))
        return None

# SrmCopyTransfer
#-------------------------------------------------------------------------------
class SrmCopyTransfer(Transfer):
    """A tranfer in SrmCopy mode"""
    
    failure_re = re.compile(r'failed on (?P<scope>\w*) during (?P<phase>\w*) phase\. Reason is: \[(?P<category>\w*)\] (?P<reason>.*)')

    def __init__(self,source,destination,push,sourceName = None,destName = None,options = None):
        self.Push = push
        mode = ""
        if (self.Push):
            mode = "srm-push"
        else:
            mode = "srm-pull"
        Transfer.__init__(self,mode,source,destination,sourceName,destName,options)

    def _update_stats(self):
        self.Source.SrmCopyStats.AddFrom(self.Succeeded)
        self.Destination.SrmCopyStats.AddTo(self.Succeeded)
    
    def _start_copy(self,logFilePath):
        return GLiteSrmCopy(self.Source,self.Destination,self.SourceName,self.DestName,self.Push,logFilePath,self.Options)

    @classmethod
    def _parse_failure(cls,line):
        m = SrmCopyTransfer.failure_re.search(line)
        if not m is None:
            return "%s - %s - %s - %s" % (m.group("scope"),m.group("phase"),m.group("category"),m.group("reason"))
        return None


# main (test)
#-------------------------------------------------------------------------------
def main (argv = None):
    if argv is None:
        argv = sys.argv
    
    if len(argv) != 3:
        print "usage : %s mode path" % (argv[0])
        print " mode : u|s (urlcopy|srmcopy)"
        print " path : log file path"
        return 1
    
    file_path = argv[2]
    if not os.path.exists(file_path):
        print "file %s not found" % (file_path)
        return 1
    
    reason = None
    if "u" == argv[1]:
        print "Getting urlcopy error message from file %s" % (file_path)
        reason = UrlCopyTransfer.GetFailureReason(file_path)
    elif "s" == argv[1]:
        print "Getting srmcopy error message from file %s" % (file_path)
        reason = SrmCopyTransfer.GetFailureReason(file_path)
    else:
        print "invalid mode specified"
        return 1

    if None == reason:
        print "No reason found; transfer succeeded"
    else:
        print reason
    return 0

# run main
#-------------------------------------------------------------------------------
if __name__ == "__main__":
    sys.exit (main())
