#!/bin/bash
#
#  Copyright (c) Members of the EGEE Collaboration. 2004.
#  See http://www.eu-egee.org/partners/ for details on the copyright
#  holders.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  transfer-url-copy test script utilities
#
#  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
#  Version info: $Id: transfer-url-copy-test-utils.sh,v 1.7 2007/04/18 07:59:47 ptedesco Exp $
#  Release: $Name:  $

# name and version of the script
scriptname=${0##*/}
scriptVersion="1.0.2"

# script return values
E_OK=0          # success
E_SCRIPTERR=1   # script error
E_BADOPTS=2     # bad command line options
E_ENV=3         # environment check failed (GLITE_LOCATION not set, etc)
E_TESTSFAILED=4 # one or more tests failed

# command line parameters
tempFileSize=50

# test date and time
testDate=`date +%Y%m%d`
testTime=`date +%H%M%S`

# log and temporary files
log_files_path="/var/tmp/transfer-url-copy-test-${USER}/${testDate}/${testTime}"

#-------------------------------------------------------------------------------
# ShowHelp
# show help message
#-------------------------------------------------------------------------------
function ShowHelp ()
{
	echo "Usage : ${scriptname} [OPTIONS]"
	echo "Options:"
	echo " -e file:    use endpoints list contained in file"
	echo " -d:         enable script debug messages"
	echo " -c:         enable CGSI_TRACE messages"
	echo " -s:         use stub srm services"
	echo " -f size:    temporary file size in MB (default=50)"
	echo " -p:         ping endpoints and exit"
	echo " -o options: additional glite-url-copy options"
	echo " -S:         use SRM copy (default=URL copy)"
	echo " -x prefix:  prefix to use in calls to glite-url-copy (default=TEST_URL_COPY / TEST_SRM_COPY)"
	echo ' -l dir:     log files directory (default=/var/tmp/transfer-url-copy-test-${USER}/${testDate}/${testTime})'
	echo " -t:         keep local transfer files"
	echo " -v:         print version and exit"
	echo " -h:         print this help message and exit"
}

# process command line parameters
while getopts ":e:dcsf:po:Sx:l:tvh" option
do
	case ${option} in
	e )
		endpointsListFile="${OPTARG}"
		echo "using endpoints list from file ${endpointsListFile}"
		;;
	d )
		debug=0
		echo "enabling debug messages"
		;;
	c ) 
		cgsi_trace=1
		echo "enabling CGSI_TRACE"
		;;
	s )
		stub=0
		echo "using stub services"
		;;
	f )
		tempFileSize="${OPTARG}"
		echo "setting temporary file size to ${tempFileSize} MBs"
		;;
	p )
		ping=0
		echo "setting ping option"
		;;
	o )
		additional_options="${OPTARG}" 
		echo "adding options \"${additional_options}\""
		;;
	S ) 
	    use_srm_copy=0
		echo "using SRM copy"
		;;
	x )
		prefix="${OPTARG}"
		echo "using prefix ${prefix}"
		;;
	l ) 
		log_files_path="${OPTARG}"
		echo "setting log files directory to ${log_files_path}"
		;;
	t )
		keep_transfer_files=0
		echo "keeping transfer files"
		;;
	v )
		echo "${scriptname} - version ${scriptVersion}"
		exit ${E_OK}
		;;
	h )
		ShowHelp
		exit ${E_OK}
		;;
	* )
		echo "invalid options"
		ShowHelp
		exit ${E_BADOPTS}
		;;
	esac
done
shift $(($OPTIND - 1)) # decrement argument pointer so that $1 now points to 
	# first non option parameter (if any)

if [ ${use_srm_copy} ]; then
	if [ ! ${prefix} ]; then 
		prefix="TEST_SRM_COPY"
	fi
	#copy_exe="glite-srm-copy"
else
	if [ ! ${prefix} ]; then 
		prefix="TEST_URL_COPY"
	fi
	#copy_exe="glite-url-copy"
fi

# directories
logs_completed="${log_files_path}/completed"
logs_failed="${log_files_path}/failed"
tmp_dir="${log_files_path}/dat"


echo "creating log files directory : ${log_files_path}"
mkdir -p ${log_files_path} || {
	echo "could not create directory ${log_files_path}"
	exit ${E_SCRIPTERR}
}
if [ ! ${ping} ]; then
	mkdir -p ${tmp_dir} || {
    	echo "could not create directory ${tmp_dir}"
    	exit ${E_SCRIPTERR}
	}
fi

if [ ${cgsi_trace} ]; then
	export CGSI_TRACE=1
	export CGSI_TRACEFILE="${log_files_path}/cgsi_tracefile"
fi

# script log file (used by Info, Debug and Error functions)
script_log="${log_files_path}/tests_log"

#-------------------------------------------------------------------------------
# Debug
# print a debug message
#-------------------------------------------------------------------------------
function Debug ()
{
	if [ ${debug} ]; then
		echo -e "\e[33m[DEBUG] ${*}\e[0m" 
		echo "[DEBUG] ${*}" >> ${script_log}
	else
		true
	fi
}

#-------------------------------------------------------------------------------
# Info
# print an informational message
#-------------------------------------------------------------------------------
function Info ()
{
	echo -e "\e[34m[INFO] ${*}\e[0m" 
	echo "[INFO] ${*}" >> ${script_log}
}

#-------------------------------------------------------------------------------
# Error
# print an error message
#-------------------------------------------------------------------------------
function Error ()
{
	echo -e "\e[31m[ERROR] ${*}\e[0m" 
	echo "[ERROR] ${*}" >> ${script_log}
}

#-------------------------------------------------------------------------------
# MoveLogFile
# remove a log file
#-------------------------------------------------------------------------------
# $1 = log file path
# $2 = execution result
function MoveLogFile()
{
	local log_file_path=${1}
	local result=${2}
	if [ 0 -eq ${result} ]; then
		mv ${log_file_path} ${logs_completed}
	else
		mv ${log_file_path} ${logs_failed}
	fi
}

#-------------------------------------------------------------------------------
# RemoveTransferFile
# remove a local transfer file
#-------------------------------------------------------------------------------
# $1 = transfer file path
function RemoveTransferFile()
{
	if [ ${keep_transfer_files} ]; then
		true
	else
		rm -rf "${1}"
	fi
}

#-------------------------------------------------------------------------------
# CheckParams
# checks function parameters
#-------------------------------------------------------------------------------
# $1 = function name
# $2 = expected number of parameters
# $3 = actual number of parameters
# $4 [optional] = comparison operator (default = -ne)
function CheckParams ()
{
	if [ ${debug} ]; then
		op="-ne"
		if [ 4 -eq ${#} ]; then
			op=${4}
		fi
		if [ ${2} ${op} ${3} ]; then
			Error "Script error: function ${1} invoked with ${3} parameters"
			exit ${E_SCRIPTERR}
		fi
	else
		true
	fi
}

#-------------------------------------------------------------------------------
# CheckVarRef
# checks that the variable referenced by given name is defined and not empty
#-------------------------------------------------------------------------------
# $1 = variable name
function CheckVarRef ()
{
	if [ -z "${!1}" ] || [ "" == "${!1}" ]; then
		return 1
	else
		return 0	
	fi
}

#-------------------------------------------------------------------------------
# CheckVar
# checks that the variable is defined and not empty
#-------------------------------------------------------------------------------
# $1 = variable 
function CheckVar ()
{
	if [ -z "${1}" ] || [ "" == "${1}" ]; then
		return 1
	else
		return 0	
	fi
}

#-------------------------------------------------------------------------------
# CheckEnvVar
# checks that an environment variable is defined
#-------------------------------------------------------------------------------
# $1 = variable name
function CheckEnvVar ()
{
    if [ -z ${!1} ] || [ "" == ${!1} ] ; then
        Error "${1} is not defined"
		exit ${E_ENV}
    fi
}

#-------------------------------------------------------------------------------
# CheckFile
# checks file existence
#-------------------------------------------------------------------------------
# $1 = file path
function CheckFile ()
{
	if [ ! -e ${1} ]; then
		Error "cannot find file ${1}"
		exit ${E_ENV}
	fi
}

#-------------------------------------------------------------------------------
# CreateTempFile
# creates a local temporary file
#-------------------------------------------------------------------------------
# $1 = file path
function CreateTempFile ()
{
	CheckParams "CreateTempFile" 1 ${#}
	local tmp_file_dir=`dirname ${1}`
	mkdir -p ${tmp_file_dir}
	dd if=/dev/zero of=${1} bs=1048576 count=${tempFileSize} > /dev/null 2>&1
	if [ 0 -ne "${?}" ]; then
		Error "could not create temporary file"
		exit ${E_SCRIPTERR}
	fi
}


