#!/bin/bash
#
#  Copyright (c) Members of the EGEE Collaboration. 2004.
#  See http://www.eu-egee.org/partners/ for details on the copyright
#  holders.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  transfer-url-copy test script
#
#  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
#  Version info: $Id: transfer-url-copy-test.sh,v 1.8 2007/04/18 07:59:47 ptedesco Exp $
#  Release: $Name:  $

# - script version 1.0.1 - 15/02/2007
# fixed bug in transfers result check
# fixed bug in surl creation
# modified endpoint definitions
# added options to specify logs directory, keep logs and transfer files
# - script version 1.0.0 - 31/01/2007
# creation


# include utility functions and command line parsing 
. transfer-url-copy-test-utils.sh

# program paths
glite_url_copy_path="${GLITE_LOCATION}/bin/"
#glite_url_copy="glite-url-copy"
glite_srm_util_path="${GLITE_LOCATION}/libexec/"
glite_srm_util_rm="glite-srm-util-rm"
glite_srm_util_ping="glite-srm-util-ping"

# include endpoints definitions
. transfer-url-copy-test-endpoints.sh

# counters for test
testsSucceeded=0
testsFailed=0
testsSkipped=0

#-------------------------------------------------------------------------------
# ExecCopy
# invokes glite-url-copy or glite-srm-copy with common options
#-------------------------------------------------------------------------------
# $1 = log file name
# $2 = source name
# $3 = destination name
# remaining parameters are appended to command line
#-------------------------------------------------------------------------------
function ExecCopy ()
{
	CheckParams "UrlCopy" 5 ${#} "-gt"
	if [ ${cgsi_trace} ]; then
		# clean cgsi trace file
		> ${CGSI_TRACEFILE}
	fi
	local command="${glite_copy_exe} -b -d -C -p ${prefix}"
	if [ "${additional_options}" ] && [ 0 -ne ${use_additional_options} ]; then
		command="${command} ${additional_options}"
	fi
	local log_file=${log_files_path}/${1}
	local argc=${#}
	local i=0
	for ((i=4;i<=argc;i++)); do
		command="${command} ${!i}"
	done
	Debug "running ${command} -> ${log_file}"
	local time1=`date +%s`
	${command} > ${log_file} 2>&1
	local result=${?}
	local time2=`date +%s`
	local dt=0
	let dt=${time2}-${time1}
	if [ 0 -eq ${result} ]; then
		Info "Transfer ${2} to ${3} succeeded (${dt} sec)"
		let testsSucceeded=${testsSucceeded}+1
	else
		Error "Transfer ${2} to ${3} failed (${dt} sec)"
		testsFailedSummary[testsFailed]="${2} to ${3}"
		let testsFailed=${testsFailed}+1
	fi
	if [ ${cgsi_trace} ]; then
		echo >> ${log_file}
		echo "-------------------------- CGSI TRACE --------------------------" >> ${log_file}
		echo >> ${log_file}
		# exclude some hex sequences
		cat ${CGSI_TRACEFILE} | grep -v -e "\([0-9a-fA-F]\{2\} \)\{3,16\}" >> ${log_file}
	fi
	MoveLogFile ${log_file} ${result}
	return ${result}
}

#-------------------------------------------------------------------------------
# UrlCopy
# invokes glite-url-copy 
#-------------------------------------------------------------------------------
# $1 = log file name
# $2 = source name
# $3 = destination name
# remaining parameters are appended to command line
#-------------------------------------------------------------------------------
function UrlCopy()
{
	glite_copy_exe="glite-url-copy"
	if [ ${use_srm_copy} ]; then
		use_additional_options=0
	else
		use_additional_options=1
	fi
	ExecCopy ${*}
}

#-------------------------------------------------------------------------------
# Copy
# invokes glite-url-copy or glite-srm-copy
#-------------------------------------------------------------------------------
# $1 = log file name
# $2 = source name
# $3 = destination name
# remaining parameters are appended to command line
#-------------------------------------------------------------------------------
function Copy()
{
	if [ ${use_srm_copy} ]; then
		glite_copy_exe="glite-srm-copy"
	else
		glite_copy_exe="glite-url-copy"
	fi
	use_additional_options=1
	ExecCopy ${*}
}

#-------------------------------------------------------------------------------
# UrlRemove
# invokes glite-srm-util-rm
#-------------------------------------------------------------------------------
# $1 = endpoint SURL
# $2 = endpoint version
# $3 = path of file to delete
# $4 = log file name
#-------------------------------------------------------------------------------
function UrlRemove ()
{
	CheckParams "UrlRemove" 4 ${#}
	local command="${glite_srm_util_rm} -s ${1} -I ${2} ${3}"
	Debug "running ${command}"
	local log_file=${log_files_path}/${4}
	${command} > ${log_file} 2>&1
	local result=${?}
	if [ 0 -eq ${result} ]; then
		Info "Removed file ${3} from ${1}"
		rm -f ${log_file}
	else
		Error "Could not remove file ${3} from ${1}" 
	fi
	return ${result}
}

#-------------------------------------------------------------------------------
# LogFileName
#-------------------------------------------------------------------------------
# $1 : source name
# $2 : destination name
function LogFileName ()
{
	CheckParams "LogFileName" 2 ${#}
	echo "${1}-to-${2}"
}

#-------------------------------------------------------------------------------
# DestinationFileName
#-------------------------------------------------------------------------------
# $1 : source name
# $2 : destination name
function DestinationFileName ()
{
	CheckParams "DestinationFileName" 2 ${#}
	echo "${USER}/tests/transfer-url-copy/${testDate}/${testTime}/${1}-to-${2}.dat"
}

#-------------------------------------------------------------------------------
# main
#-------------------------------------------------------------------------------

# check environment
CheckEnvVar "GLITE_LOCATION"
CheckFile "${glite_url_copy_path}/${glite_url_copy}"
CheckFile "${glite_srm_util_path}/${glite_srm_util_rm}"
CheckFile "${glite_srm_util_path}/${glite_srm_util_ping}"

testGridProxyInit=`grid-proxy-info | grep "timeleft : 0:00:00"`
if [ ! -z "${testGridProxyInit}" ] && [ "" != "${testGridProxyInit}" ]; then
	Error "grid-proxy-init not run"
	exit ${E_ENV}
fi

# add programs paths to PATH
PATH="${glite_url_copy_path}:${glite_srm_util_path}:${PATH}"

if [ ${ping} ]; then
	Info "pinging endpoints..."
	for ((i=0;i<totalEndpoints;i++)); do
		command="${glite_srm_util_ping} -s `EndpointAddress ${i}` -I `EndpointVersion ${i}`"
		Info "pinging `EndpointName ${i}`: ${command}"
		${command} 2>&1 | tee -a ${script_log}
	done
	exit
fi

# create temporary file
Info "creating temporary file..."
tmp_file=${tmp_dir}/tmp.dat

mkdir -p ${logs_completed}
mkdir -p ${logs_failed}

CreateTempFile ${tmp_file} 

# transfers local -> endpoint

Debug "--- transfers local -> endpoint ---"

for ((i=0;i<totalEndpoints;i++)); do
	endpointName=`EndpointName ${i}`
	Info "transferring: local to ${endpointName}"
	# now invoke url_copy; no -1 (local)
	destinationFileFromLocal=`DestinationFileName local ${endpointName}`
	log=`LogFileName "local" "${endpointName}"` 
	UrlCopy ${log} "local" "${endpointName}" -2 `EndpointAddress ${i}` "file://${tmp_file}" `Surl ${i} ${destinationFileFromLocal}`
	result=${?}
	localToEndpointOk[${i}]=${result}
done

# transfers endpoint -> endpoint

Debug "--- transfers endpoint -> endpoint ---"

for ((i=0;i<totalEndpoints;i++)); do
	ok=${localToEndpointOk[${i}]}
	sourceEP=`EndpointName ${i}`
	if [ "0" -ne ${ok} ]; then
		Info "skipping transfers from ${sourceEP}"
		let testsSkipped=${testsSkipped}+${totalEndpoints}-1
		continue
	fi
	destinationFileFromLocal=`DestinationFileName local ${sourceEP}`
	surl1=`Surl ${i} ${destinationFileFromLocal}`
	for ((j=0;j<totalEndpoints;j++)); do
		if [ ${i} -eq ${j} ]; then
			continue
		fi
		destEP=`EndpointName ${j}`
		destinationFileName=`DestinationFileName ${sourceEP} ${destEP}`
		Info "transferring: ${sourceEP} to ${destEP}"
		surl2=`Surl ${j} ${destinationFileName}`
		log=`LogFileName "${sourceEP}" "${destEP}"`
		Copy ${log} "${sourceEP}" "${destEP}" -1 `EndpointAddress ${i}` -2 `EndpointAddress ${j}` "${surl1}" "${surl2}"
		result=${?}
		if [ 0 -eq ${result} ];	then
			# remove destination file
			UrlRemove `EndpointAddress ${j}` `EndpointVersion ${j}` "${surl2}" "rm_${logFileName}"
		fi
	done
done

# transfers endpoint -> local

Debug "--- transfers endpoint -> local ---"

for ((i=0;i<totalEndpoints;i++)); do
	ok=${localToEndpointOk[${i}]}
	if [ "0" -ne ${ok} ]; then
		Debug "skipping transfer from ${endpointName} to local"
		let testsSkipped=${testsSkipped}+1
		continue
	fi
	endpointName=`EndpointName ${i}`
	Info "transferring: ${endpointName} to local"
	# now invoke url_copy; no -2 (local)
	destinationFileToLocal="from_${endpointName}" # *** Change this value to test local directory creation
	destinationFileFromLocal=`DestinationFileName local ${endpointName}`
	log=`LogFileName "${endpointName}" "local"`
	UrlCopy ${log} "${endpointName}" "local" -1 `EndpointAddress ${i}` `Surl ${i} ${destinationFileFromLocal}` "file://${tmp_dir}/${destinationFileToLocal}" 
	result=${?}
	if [ 0 -eq ${result} ]
	then
		RemoveTransferFile "${tmp_dir}/${destinationFileToLocal}"
	fi
done

# remove files copied from local host

Debug "--- removing files ---"

for ((i=0;i<totalEndpoints;i++)); do
	if [ "0" -ne ${localToEndpointOk[${i}]} ];	then
		continue
	fi
	endpointName=`EndpointName ${i}`
	destinationFileFromLocal=`DestinationFileName local ${endpointName}`
	logFileName=`LogFileName local ${endpointName}`
	UrlRemove `EndpointAddress ${i}` `EndpointVersion ${i}` `Surl ${i} ${destinationFileFromLocal}` "rm_${logFileName}"
done

# remove temporary file
RemoveTransferFile ${tmp_file}
RemoveTransferFile ${tmp_dir}

if [ ${cgsi_trace} ]; then
	rm -rf ${CGSI_TRACEFILE}
fi

Info "Tests succeeded: ${testsSucceeded}"
Info "Tests failed: ${testsFailed}"
Info "Tests skipped: ${testsSkipped}"

if ((0 != testsFailed)); then
	Error "Failed tests:"
	for ((i=0;i<testsFailed;i++)); do
		Error "${testsFailedSummary[i]}"
	done
fi

Info "Log directory: ${log_files_path}"

if ((0==testsFailed)); then
	exit ${E_OK}
else
	exit ${E_TESTSFAILED}
fi
