/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: SrmCopyCoreTokens Test Case
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: SrmCopyCoreCheckSumTestCase.cpp,v 1.10 2009/12/11 17:53:22 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include "ChecksumCheckerUtils.h"
#include "ChecksumChecker.h"

#include "SrmCopyCoreCheckSumTestCase.h"

/*
 * The SRM endpoints
 */
extern std::string g_srm_11_endpoint;
extern std::string g_srm_22_endpoint;

namespace {
const char SAMPLE_TEXT[] = "Hello World! This is a test\nWrite some data for Transfer Srm Copy tests\n";
const char EMPTY_FILE[] = "";
static const char * V = "2.2";

static const char * algorithm   = "ADLER32";
static const char * value_01    = "A0A0A0A0A0A0";
static const char * value_02    = "B1B1B1B1B1B1";

std::string src_01 = make_cksm_file("srmcopy-src01.dat",algorithm,value_01);
std::string src_02 = make_cksm_file("srmcopy-src02.dat",algorithm,value_02);

std::string src_no_cksm_01 = "srmcopy-no-cksm-src01.dat";
std::string src_no_cksm_02 = "srmcopy-no-cksm-src02.dat";

}

void SrmCopyCoreCheckSumTestCase::test_setup(){
    putFile(src_01,g_srm_22_endpoint,V,SAMPLE_TEXT);
    putFile(src_02,g_srm_22_endpoint,V,SAMPLE_TEXT);
    putFile(src_no_cksm_01,g_srm_22_endpoint,V,SAMPLE_TEXT);
    putFile(src_no_cksm_02,g_srm_22_endpoint,V,SAMPLE_TEXT);
}

void SrmCopyCoreCheckSumTestCase::testUseCase0()
{
	__TIMER;
    static const char * test_name   = "testUseCase0";
    std::string dst_01 = make_cksm_file("srmcopy01-dst01.dat",algorithm,value_01);
    std::string dst_02 = make_cksm_file("srmcopy01-dst02.dat",algorithm,value_02);

    statSetUpCksm(
        test_name,src_01,src_02,dst_01,dst_02,g_srm_22_endpoint,V,
        "", "", "", "", false);

    prepare(test_name);
    transfer(test_name);
    complete(test_name);
    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name, true);
    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == SRMCOPY_STAT_ID);
    // Check that checksum data is written in output structures
    CHECK_STRING(stat->mm_files[0].mm_src_checksum_algorithm, "");
    CHECK_STRING(stat->mm_files[0].mm_src_checksum_value, "");
    CHECK_STRING(stat->mm_files[0].mm_dst_checksum_algorithm, "");
    CHECK_STRING(stat->mm_files[0].mm_dst_checksum_value, "");
    CPPUNIT_ASSERT_MESSAGE("Destination file was deleted.", fileExists(dst_01,g_srm_22_endpoint,V));

    CHECK_STRING(stat->mm_files[1].mm_src_checksum_algorithm, "");
    CHECK_STRING(stat->mm_files[1].mm_src_checksum_value, "");
    CHECK_STRING(stat->mm_files[1].mm_dst_checksum_algorithm, "");
    CHECK_STRING(stat->mm_files[1].mm_dst_checksum_value, "");
    CPPUNIT_ASSERT_MESSAGE("Destination file was deleted.", fileExists(dst_02,g_srm_22_endpoint,V));
}

void SrmCopyCoreCheckSumTestCase::testUseCase1_src_dst_checksumMatch()
{
	__TIMER;
    static const char * test_name   = "testUseCase1_src_checksumMatch";
    std::string dst_01 = make_cksm_file("srmcopy02-dst01.dat",algorithm,value_01);
    std::string dst_02 = make_cksm_file("srmcopy02-dst02.dat",algorithm,value_02);

    statSetUpCksm(
        test_name,src_01,src_02,dst_01,dst_02,g_srm_22_endpoint,V,
        algorithm, value_01, algorithm, value_02, false);

    prepare(test_name);
    transfer(test_name);
    complete(test_name);
    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name, true);
    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == SRMCOPY_STAT_ID);
    // Check that checksum data is written in output structures
    CHECK_STRING(stat->mm_files[0].mm_src_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_files[0].mm_src_checksum_value, value_01);
    CHECK_STRING(stat->mm_files[0].mm_dst_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_files[0].mm_dst_checksum_value, value_01);
    CHECK_STRING(stat->mm_files[1].mm_src_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_files[1].mm_src_checksum_value, value_02);
    CHECK_STRING(stat->mm_files[1].mm_dst_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_files[1].mm_dst_checksum_value, value_02);
}

void SrmCopyCoreCheckSumTestCase::testUseCase1_src_checksumNotAvailable()
{
	__TIMER;
    static const char * test_name   = "testUseCase1_src_checksumNotAvailable";
    std::string dst_01 = make_cksm_file("srmcopy03-dst01.dat",algorithm,value_01);
    std::string dst_02 = make_cksm_file("srmcopy03-dst02.dat",algorithm,value_02);

    statSetUpCksm(
        test_name,src_no_cksm_01,src_no_cksm_02,dst_01,dst_02,g_srm_22_endpoint,V,
        algorithm, value_01, algorithm, value_02, false);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP(algorithm, value_01);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);

    prepare(test_name);
    transfer(test_name);
    complete(test_name);
    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name, true);
    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == SRMCOPY_STAT_ID);
    // Check that checksum data is written in output structures
    CHECK_STRING(stat->mm_files[0].mm_src_checksum_algorithm, "");
    CHECK_STRING(stat->mm_files[0].mm_src_checksum_value, "");
    CHECK_STRING(stat->mm_files[0].mm_dst_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_files[0].mm_dst_checksum_value, value_01);
    CHECK_STRING(stat->mm_files[1].mm_src_checksum_algorithm, "");
    CHECK_STRING(stat->mm_files[1].mm_src_checksum_value, "");
    CHECK_STRING(stat->mm_files[1].mm_dst_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_files[1].mm_dst_checksum_value, value_02);
    CPPUNIT_ASSERT(!getter->fallbackHappened());
}

void SrmCopyCoreCheckSumTestCase::testUseCase1_src_checksumDifferent()
{
	__TIMER;
    static const char * test_name   = "testUseCase1_src_checksumDifferent";
    static const char * u_value     = "dfdfdfdfdfdfdfdf";
    std::string dst_01 = make_cksm_file("srmcopy04-dst01.dat",algorithm,value_01);
    std::string dst_02 = make_cksm_file("srmcopy04-dst02.dat",algorithm,value_02);

    statSetUpCksm(
        test_name,src_01,src_02,dst_01,dst_02,g_srm_22_endpoint,V,
        algorithm, u_value, algorithm, u_value, false);

    {
        SrmCopyCore tx(test_name);
        tx.prepare();
    }

    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == SRMCOPY_STAT_ID);

    check_status("Invalid status (request)",stat->mm_srmcopy.mm_status,UrlCopyStatus::PREPARING);
    check_status("Invalid status (file 0)",stat->mm_files[0].mm_status,UrlCopyStatus::FAILED);
    check_status("Invalid status (file 1)",stat->mm_files[1].mm_status,UrlCopyStatus::FAILED);

    check_scope("Invalid scope (file 0)",       stat->mm_files[0].mm_error.mm_scope,    UrlCopyError::SCOPE_SOURCE);
    check_phase("Invalid phase (file 0)",       stat->mm_files[0].mm_error.mm_phase,    UrlCopyError::PHASE_PREPARATION);
    check_category("Invalid category (file 0)", stat->mm_files[0].mm_error.mm_category, UrlCopyError::ERROR_CHECKSUM_MISMATCH);

    check_scope("Invalid scope (file 1)",       stat->mm_files[1].mm_error.mm_scope,    UrlCopyError::SCOPE_SOURCE);
    check_phase("Invalid phase (file 1)",       stat->mm_files[1].mm_error.mm_phase,    UrlCopyError::PHASE_PREPARATION);
    check_category("Invalid category (file 1)", stat->mm_files[1].mm_error.mm_category, UrlCopyError::ERROR_CHECKSUM_MISMATCH);

    CHECK_STRING(stat->mm_files[0].mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_files[0].mm_src_checksum_value,    value_01);
    CHECK_STRING(stat->mm_files[0].mm_dst_checksum_algorithm,"");
    CHECK_STRING(stat->mm_files[0].mm_dst_checksum_value,    "");

    CHECK_STRING(stat->mm_files[1].mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_files[1].mm_src_checksum_value,    value_02);
    CHECK_STRING(stat->mm_files[1].mm_dst_checksum_algorithm,"");
    CHECK_STRING(stat->mm_files[1].mm_dst_checksum_value,    "");
}

void SrmCopyCoreCheckSumTestCase::testUseCase1_dst_gridFTPfallback()
{
	__TIMER;
    static const char* test_name   = "testUseCase1_dst_gridFTPfallback";
    std::string dst_01 = make_cksm_file("srmcopy05-dst01.dat",algorithm,value_01);
    std::string dst_02 = "srmcopy05-dst02.dat";

    statSetUpCksm(
        test_name,src_01,src_02,dst_01,dst_02,g_srm_22_endpoint,V,
        algorithm, value_01, algorithm, value_02, false);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP(algorithm, value_02);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);

    prepare(test_name);
    transfer(test_name);
    {
        SrmCopyCore tx(test_name);
        tx.complete();
    }

    CPPUNIT_ASSERT(getter->fallbackHappened());
    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT(getter->fallbackHappened());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == SRMCOPY_STAT_ID);

    CHECK_STRING(stat->mm_files[0].mm_src_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_files[0].mm_src_checksum_value, value_01);
    CHECK_STRING(stat->mm_files[0].mm_dst_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_files[0].mm_dst_checksum_value, value_01);
    CPPUNIT_ASSERT_MESSAGE("Destination file was deleted.", fileExists(dst_01,g_srm_22_endpoint,V));

    CHECK_STRING(stat->mm_files[1].mm_src_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_files[1].mm_src_checksum_value, value_02);
    CHECK_STRING(stat->mm_files[1].mm_dst_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_files[1].mm_dst_checksum_value, value_02);
    CPPUNIT_ASSERT_MESSAGE("Destination file was deleted.", fileExists(dst_02,g_srm_22_endpoint,V));
}

void SrmCopyCoreCheckSumTestCase::testUseCase1_dst_checksumNotAvailable()
{
	__TIMER;
    static const char* test_name   = "testUseCase1_dst_checksumNotAvailable";
    std::string dst_01 = "srmcopy06-dst01.dat";
    std::string dst_02 = "srmcopy06-dst02.dat";

    statSetUpCksm(
        test_name,src_01,src_02,dst_01,dst_02,g_srm_22_endpoint,V,
        algorithm, value_01, algorithm, value_02, false);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP("", "", UrlCopyError::ERROR_CHECKSUM_UNKNOWN);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);

    prepare(test_name);
    transfer(test_name);
    {
        SrmCopyCore tx(test_name);
        tx.complete();
    }

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT(getter->fallbackHappened());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == SRMCOPY_STAT_ID);

    CPPUNIT_ASSERT(getter->fallbackHappened());
    check_status("Invalid status (request)",stat->mm_srmcopy.mm_status,UrlCopyStatus::COMPLETED);
    check_status("Invalid status (file 0)",stat->mm_files[0].mm_status,UrlCopyStatus::FAILED);
    check_status("Invalid status (file 1)",stat->mm_files[1].mm_status,UrlCopyStatus::FAILED);

    check_scope("Invalid scope (file 0)",       stat->mm_files[0].mm_error.mm_scope,    UrlCopyError::SCOPE_DESTINATION);
    check_phase("Invalid phase (file 0)",       stat->mm_files[0].mm_error.mm_phase,    UrlCopyError::PHASE_FINALIZATION);
    check_category("Invalid category (file 0)", stat->mm_files[0].mm_error.mm_category, UrlCopyError::ERROR_CHECKSUM_UNKNOWN);

    check_scope("Invalid scope (file 1)",       stat->mm_files[1].mm_error.mm_scope,    UrlCopyError::SCOPE_DESTINATION);
    check_phase("Invalid phase (file 1)",       stat->mm_files[1].mm_error.mm_phase,    UrlCopyError::PHASE_FINALIZATION);
    check_category("Invalid category (file 1)", stat->mm_files[1].mm_error.mm_category, UrlCopyError::ERROR_CHECKSUM_UNKNOWN);

    CHECK_STRING(stat->mm_files[0].mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_files[0].mm_src_checksum_value,    value_01);
    CHECK_STRING(stat->mm_files[0].mm_dst_checksum_algorithm,"");
    CHECK_STRING(stat->mm_files[0].mm_dst_checksum_value,    "");

    CHECK_STRING(stat->mm_files[1].mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_files[1].mm_src_checksum_value,    value_02);
    CHECK_STRING(stat->mm_files[1].mm_dst_checksum_algorithm,"");
    CHECK_STRING(stat->mm_files[1].mm_dst_checksum_value,    "");

    CPPUNIT_ASSERT_MESSAGE("Destination file was not deleted.",!fileExists(dst_02,g_srm_22_endpoint,V));
    CPPUNIT_ASSERT_MESSAGE("Destination file was not deleted.",!fileExists(dst_01,g_srm_22_endpoint,V));
}

void SrmCopyCoreCheckSumTestCase::testUseCase1_dst_checksumDifferent()
{
	__TIMER;
    static const char * test_name   = "testUseCase1_dst_checksumDifferent";

    static const char * d_value     = "2";
    std::string dst_01 = make_cksm_file("srmcopy07-dst01.dat",algorithm,d_value);
    std::string dst_02 = make_cksm_file("srmcopy07-dst02.dat",algorithm,d_value);

    statSetUpCksm(
        test_name,src_01,src_02,dst_01,dst_02,g_srm_22_endpoint,V,
        algorithm, value_01, algorithm, value_02, false);

    prepare(test_name);
    transfer(test_name);
    {
        SrmCopyCore tx(test_name);
        tx.complete();
    }

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == SRMCOPY_STAT_ID);

    check_status("Invalid status (request)",stat->mm_srmcopy.mm_status,UrlCopyStatus::COMPLETED);
    check_status("Invalid status (file 0)",stat->mm_files[0].mm_status,UrlCopyStatus::FAILED);
    check_status("Invalid status (file 1)",stat->mm_files[1].mm_status,UrlCopyStatus::FAILED);

    check_scope("Invalid scope (file 0)",       stat->mm_files[0].mm_error.mm_scope,    UrlCopyError::SCOPE_DESTINATION);
    check_phase("Invalid phase (file 0)",       stat->mm_files[0].mm_error.mm_phase,    UrlCopyError::PHASE_FINALIZATION);
    check_category("Invalid category (file 0)", stat->mm_files[0].mm_error.mm_category, UrlCopyError::ERROR_CHECKSUM_MISMATCH);

    check_scope("Invalid scope (file 1)",       stat->mm_files[1].mm_error.mm_scope,    UrlCopyError::SCOPE_DESTINATION);
    check_phase("Invalid phase (file 1)",       stat->mm_files[1].mm_error.mm_phase,    UrlCopyError::PHASE_FINALIZATION);
    check_category("Invalid category (file 1)", stat->mm_files[1].mm_error.mm_category, UrlCopyError::ERROR_CHECKSUM_MISMATCH);

    CHECK_STRING(stat->mm_files[0].mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_files[0].mm_src_checksum_value,    value_01);
    CHECK_STRING(stat->mm_files[0].mm_dst_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_files[0].mm_dst_checksum_value,    d_value);
    CPPUNIT_ASSERT_MESSAGE("Destination file was not deleted.",!fileExists(dst_01,g_srm_22_endpoint,V));

    CHECK_STRING(stat->mm_files[1].mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_files[1].mm_src_checksum_value,    value_02);
    CHECK_STRING(stat->mm_files[1].mm_dst_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_files[1].mm_dst_checksum_value,    d_value);
    CPPUNIT_ASSERT_MESSAGE("Destination file was not deleted.",!fileExists(dst_02,g_srm_22_endpoint,V));
}
