/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: SrmCopyCore Test Case
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: SrmCopyCoreTestCase_Common.cpp,v 1.6 2009/07/30 19:32:55 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#include "SrmCopyCore.h"
#include "test-utility.h"
#include "Mmap.h"
#include "stat-utility.h"
#include "TransferConfig.h"

#include <glite/data/agents/AgentExceptions.h>
#include <glite/data/srm/util/Context.h>
#include <glite/data/srm/util/PrepareToPut.h>
#include <glite/data/srm/util/SrmMkDir.h>

#include <memory>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <boost/shared_ptr.hpp>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

#include "SrmCopyCoreTestCase_Common.h"

using namespace glite::data::agents;
using namespace glite::data::srm::util;
using namespace glite::data::transfer::urlcopy;

using boost::shared_ptr;

namespace{

boost::shared_ptr<Mmap> create_stat(const std::string& name,int num_files,SrmCopyStat ** stat){
    try{
        // Delete Previous file (if exists)
        std::string filename = TransferConfig::instance().repository() + name + ".mem";
        unlink(filename.c_str());
    } catch(...){
        // Ignore error
    }

    boost::shared_ptr<Mmap> mmap_ptr = create_stat_SrmCopy(name,num_files);

    // Check Return Value
    CPPUNIT_ASSERT_MESSAGE("Invalid Result" , mmap_ptr.get()     != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Buffer" , mmap_ptr->buffer() != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Size"   , mmap_ptr->size()   != 0);

    // Get the stat object
    (*stat) = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid ID", (*stat)->mm_common.mm_id == SRMCOPY_STAT_ID);

    return mmap_ptr;
}

void stat_setup(
    SrmCopyStat *       stat,
    int                 num_files,
    const std::string&  name,
    const std::string&  src_name_1,
    const std::string&  src_name_2,
    const std::string&  dst_name_1,
    const std::string&  dst_name_2,
    const std::string&  endpoint,
    const std::string&  version){

    // Write something into the buffer
    time(&(stat->mm_common.mm_start_time));
    time(&(stat->mm_common.mm_last_refresh_time));
#ifdef _USE_GLOBAL_TIMEOUT
    stat->mm_common.mm_global_timeout  = 1000;
#endif
    stat->mm_common.mm_overwrite       = 1;
    // User Description
    copy_message(stat->mm_common.mm_user_tag,"SrmCopyCore UnitTest");


    // flag to enable/disable checks for existance of target directory
    stat->mm_common.mm_enable_dir_check = 1;
    // flag to enable/disable ping
    stat->mm_common.mm_enable_ping_source = 1;
    stat->mm_common.mm_enable_ping_dest = 1;

    // Common properties
    copy_url(stat->mm_srmcopy.mm_src_srm_endpoint,endpoint.c_str());
    copy_version(stat->mm_srmcopy.mm_src_srm_version,version.c_str());
    copy_url(stat->mm_srmcopy.mm_dest_srm_endpoint,endpoint.c_str());
    copy_version(stat->mm_srmcopy.mm_dest_srm_version,version.c_str());
    // File 0
    std::string src_surl = endpoint + "?SFN=/test/" + src_name_1 + "." + version;
    src_surl = (std::string)"srm" + src_surl.substr(5);
    copy_url(stat->mm_files[0].mm_src_name,src_surl.c_str());
    std::string dest_surl = endpoint + "?SFN=/test/" + dst_name_1 + "." + version;
    dest_surl = (std::string)"srm" + dest_surl.substr(5);
    copy_url(stat->mm_files[0].mm_dest_name,dest_surl.c_str());

    if (num_files >= 2){
        // File 1
        src_surl = endpoint + "?SFN=/test/" + src_name_2 + "." + version;
        src_surl = (std::string)"srm" + src_surl.substr(5);
        copy_url(stat->mm_files[1].mm_src_name,src_surl.c_str());
        dest_surl = endpoint + "?SFN=/test/" + dst_name_2 + "." + version;
        dest_surl = (std::string)"srm" + dest_surl.substr(5);
        copy_url(stat->mm_files[1].mm_dest_name,dest_surl.c_str());
    }

    // Timeouts
    stat->mm_srmcopy.mm_http_timeout    = 40;
    stat->mm_srmcopy.mm_copy_timeout    = 600;
    stat->mm_srmcopy.mm_refresh_timeout = 180;
    // Transfer
    stat->mm_srmcopy.mm_push_mode        = 0;

}

} // namespace

/*
 * Test SetUp
 */
void SrmCopyCoreTestCase_Common::statSetUp(
    const std::string&  name,
    const std::string&  src_name_1,
    const std::string&  src_name_2,
    const std::string&  dst_name_1,
    const std::string&  dst_name_2,
    const std::string&  endpoint,
    const std::string&  version){

    // Create a stat File and return the Mmap object for it
    int num_files = 1;
    if (src_name_2.length() || dst_name_2.length())
        num_files = 2;

    SrmCopyStat *stat = 0;
    boost::shared_ptr<Mmap> mmap_ptr = create_stat(name,num_files,&stat);

    stat_setup(stat,num_files,name,src_name_1,src_name_2,dst_name_1,dst_name_2,endpoint,version);

}

void SrmCopyCoreTestCase_Common::statSetUpWithTokens(
    const std::string&  name,
    const std::string&  src_name_1,
    const std::string&  src_name_2,
    const std::string&  dst_name_1,
    const std::string&  dst_name_2,
    const std::string&  endpoint,
    const std::string&  version,
    const std::string&  source_space_token,
    const std::string&  dest_space_token){

    // Create a stat File and return the Mmap object for it
    int num_files = 1;
    if (src_name_2.length() || dst_name_2.length())
        num_files = 2;

    SrmCopyStat *stat = 0;
    boost::shared_ptr<Mmap> mmap_ptr = create_stat(name,num_files,&stat);

    stat_setup(stat,num_files,name,src_name_1,src_name_2,dst_name_1,dst_name_2,endpoint,version);

    // Tokens
    copy_message(stat->mm_common.mm_source_token_descr,source_space_token.c_str());
    copy_message(stat->mm_common.mm_dest_token_descr,dest_space_token.c_str());

}

void SrmCopyCoreTestCase_Common::statSetUpWlcg(
    const std::string&  name,
    const std::string&  src_name_1,
    const std::string&  src_name_2,
    const std::string&  dst_name_1,
    const std::string&  dst_name_2,
    const std::string&  endpoint,
    const std::string&  version,
    int                 copy_pin_lifetime,
    int                 lan_connection,
    int                 fail_nearline_files){

    // Create a stat File and return the Mmap object for it
    int num_files = 1;
    if (src_name_2.length() || dst_name_2.length())
        num_files = 2;

    SrmCopyStat *stat = 0;
    boost::shared_ptr<Mmap> mmap_ptr = create_stat(name,num_files,&stat);

    stat_setup(stat,num_files,name,src_name_1,src_name_2,dst_name_1,dst_name_2,endpoint,version);
    // added with stat version 3.3.0
    stat->mm_common.mm_copy_pin_lifetime      = copy_pin_lifetime;
    stat->mm_common.mm_lan_connection         = lan_connection;
    stat->mm_common.mm_fail_nearline_files    = fail_nearline_files;

}

void SrmCopyCoreTestCase_Common::statSetUpCksm(
    const std::string&  name,
    const std::string&  src_name_1,
    const std::string&  src_name_2,
    const std::string&  dst_name_1,
    const std::string&  dst_name_2,
    const std::string&  endpoint,
    const std::string&  version,
    const std::string&  algorithm_1,
    const std::string&  value_1,
    const std::string&  algorithm_2,
    const std::string&  value_2,
    const int compare)
{
    // Create a stat File and return the Mmap object for it
    int num_files = 1;
    if (src_name_2.length() || dst_name_2.length())
        num_files = 2;
    SrmCopyStat *stat = 0;
    boost::shared_ptr<Mmap> mmap_ptr = create_stat(name,num_files,&stat);
    stat_setup(stat,num_files,name,src_name_1,src_name_2,dst_name_1,dst_name_2,endpoint,version);

    // checksum
    stat->mm_files[0].mm_compare_checksum = compare;
    copy_checksum_algorithm(stat->mm_files[0].mm_checksum_user_algorithm,algorithm_1.c_str());
    copy_checksum_value(stat->mm_files[0].mm_checksum_user_value,value_1.c_str());
    if (num_files > 1){
        stat->mm_files[1].mm_compare_checksum = compare;
        copy_checksum_algorithm(stat->mm_files[1].mm_checksum_user_algorithm,algorithm_2.c_str());
        copy_checksum_value(stat->mm_files[1].mm_checksum_user_value,value_2.c_str());
    }
}

/*
 * Test Prepare Method
 */
void SrmCopyCoreTestCase_Common::prepare(const std::string& name){

    time_t current;
    time(&current);

    {
    // Create Tx Core Object
        SrmCopyCore tx(name);

    // Prepare For transfer
        tx.prepare();
    }
    // Check Results
    // --------------

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(name,true);
    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == SRMCOPY_STAT_ID);

    // Check Status
    check_status("Invalid Status (prepare)", stat->mm_srmcopy.mm_status, UrlCopyStatus::READY);
    check_category("Invalid Error (prepare)" , stat->mm_srmcopy.mm_error.mm_category, UrlCopyError::SUCCESS);
    CPPUNIT_ASSERT_MESSAGE("Invalid Token" , strlen(stat->mm_srmcopy.mm_token) == 0);
    // Preparation times
    CPPUNIT_ASSERT_MESSAGE("Invalid Source Prep. Start", stat->mm_srmcopy.mm_src_prep_start_time >= current);
    CPPUNIT_ASSERT_MESSAGE("Invalid Source Prep. Done" , stat->mm_srmcopy.mm_src_prep_done_time >= current);
    CPPUNIT_ASSERT_MESSAGE("Invalid Dest. Prep. Start" , stat->mm_srmcopy.mm_dest_prep_start_time >= current);
    CPPUNIT_ASSERT_MESSAGE("Invalid Dest. Prep. Done"  , stat->mm_srmcopy.mm_dest_prep_done_time >= current);

    // File 0
    check_status("Invalid Status #0 (prepare)"      , stat->mm_files[0].mm_status, UrlCopyStatus::READY);
    check_category("Invalid Error #0 (prepare)"     , stat->mm_files[0].mm_error.mm_category, UrlCopyError::SUCCESS);
    CPPUNIT_ASSERT_MESSAGE("Invalid Dest. Exists #0", stat->mm_files[0].mm_dest_exists == 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Src. size #0"   , stat->mm_files[0].mm_src_size > 0);

    // File 1
    check_status("Invalid Status #1 (prepare)"      , stat->mm_files[1].mm_status, UrlCopyStatus::READY);
    check_category("Invalid Error #1 (prepare)"     , stat->mm_files[1].mm_error.mm_category, UrlCopyError::SUCCESS);
    CPPUNIT_ASSERT_MESSAGE("Invalid Dest. Exists #1", stat->mm_files[1].mm_dest_exists == 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Src. size #1"   , stat->mm_files[1].mm_src_size > 0);


}

/*
 * Test Transfer Method
 */
void SrmCopyCoreTestCase_Common::transfer(const std::string& name){

    time_t current;
    time(&current);

    {
    // Create Tx Core Object
        SrmCopyCore tx(name);

    // Prepare For transfer
        tx.transfer();
    }

    // Check Results
    // --------------

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(name,true);
    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat id", stat->mm_common.mm_id == SRMCOPY_STAT_ID);

    // Check Status
    check_status("Invalid Status (transfer)", stat->mm_srmcopy.mm_status, UrlCopyStatus::DONE);
    check_category("Invalid Error (transfer)" , stat->mm_srmcopy.mm_error.mm_category, UrlCopyError::SUCCESS);
    CPPUNIT_ASSERT_MESSAGE("Invalid Token" , strlen(stat->mm_srmcopy.mm_token) != 0);

    // Transfer times
    CPPUNIT_ASSERT_MESSAGE("Invalid Tx Start", stat->mm_srmcopy.mm_start_time >= current);
    CPPUNIT_ASSERT_MESSAGE("Invalid Tx Done" , stat->mm_srmcopy.mm_done_time >= current);

    // File 0
    check_status("Invalid Status #0 (transfer)"     , stat->mm_files[0].mm_status, UrlCopyStatus::DONE);
    check_category("Invalid Error #0 (transfer)"    , stat->mm_files[0].mm_error.mm_category, UrlCopyError::SUCCESS);
    CPPUNIT_ASSERT_MESSAGE("Invalid Dest. size #0"  , stat->mm_files[0].mm_size > 0);

    // File 1
    check_status("Invalid Status #1 (transfer)"     , stat->mm_files[1].mm_status, UrlCopyStatus::DONE);
    check_category("Invalid Error #1 (transfer)"    , stat->mm_files[1].mm_error.mm_category, UrlCopyError::SUCCESS);
    CPPUNIT_ASSERT_MESSAGE("Invalid Dest. size #1"  , stat->mm_files[1].mm_size > 0);


}

/*
 * Test Complete Method
 */
void SrmCopyCoreTestCase_Common::complete(const std::string& name){

    time_t current;
    time(&current);

    {
    // Create Tx Core Object
        SrmCopyCore tx(name);

    // Complete The request
        tx.complete();
    }
    // Check Results
    // --------------

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(name,true);
    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat id", stat->mm_common.mm_id == SRMCOPY_STAT_ID);

    //printf("Request status is %s\nError: %s - %s\n",
    //  UrlCopyStatus::toString((UrlCopyStatus::Value)stat->mm_srmcopy.mm_status),
    //  UrlCopyError::categoryToString((UrlCopyError::Category)stat->mm_srmcopy.mm_error.mm_category),
    //  stat->mm_srmcopy.mm_error.mm_message);

    // Check Status
    check_status("Invalid Status (complete)", stat->mm_srmcopy.mm_status, UrlCopyStatus::COMPLETED);
    check_category("Invalid Error (complete)" , stat->mm_srmcopy.mm_error.mm_category, UrlCopyError::SUCCESS);
    CPPUNIT_ASSERT_MESSAGE("Invalid Token" , strlen(stat->mm_srmcopy.mm_token) != 0);

    // Final times
    CPPUNIT_ASSERT_MESSAGE("Invalid Final Start", stat->mm_srmcopy.mm_final_start_time >= current);
    CPPUNIT_ASSERT_MESSAGE("Invalid Final Done" , stat->mm_srmcopy.mm_final_done_time >= current);

    // File 0
    check_status("Invalid Status #0 (complete)"     , stat->mm_files[0].mm_status, UrlCopyStatus::COMPLETED);
    check_category("Invalid Error #0 (complete)"       , stat->mm_files[0].mm_error.mm_category, UrlCopyError::SUCCESS);

    // File 1
    check_status("Invalid Status #1 (complete)"     , stat->mm_files[1].mm_status, UrlCopyStatus::COMPLETED);
    check_category("Invalid Error #1 (complete)"       , stat->mm_files[1].mm_error.mm_category, UrlCopyError::SUCCESS);
}

