/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: SrmCopyCoreTokens Test Case
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *           Zsolt Molnar <zsolt.molnar@cern.ch>
 *  Version info: $Id: SrmCopyCoreTokensTestCase.cpp,v 1.4 2009/03/19 13:37:09 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#include "SrmCopyCoreTokensTestCase.h"

/*
 * The SRM endpoints
 */
extern std::string g_srm_11_endpoint;
extern std::string g_srm_22_endpoint;

namespace {
const char SAMPLE_TEXT[] = "Hello World! This is a test\nWrite some data for Transfer Srm Copy tests\n";
const char EMPTY_FILE[] = "";
static const char * source_1 = "srmcopycore_multiple_tokens_source_A.dat";
static const char * source_2 = "srmcopycore_multiple_tokens_source_B.dat";
}

void SrmCopyCoreTokensTestCase::test_space_token_setup(){
    __TIMER;
    CPPUNIT_ASSERT_MESSAGE("Invalid Endpoint", g_srm_22_endpoint.empty() == false);
    CPPUNIT_ASSERT_MESSAGE("Invalid Endpoint", g_srm_11_endpoint.empty() == false);
    putFile(source_1,g_srm_22_endpoint,"2.2",SAMPLE_TEXT);
    putFile(source_2,g_srm_22_endpoint,"2.2",SAMPLE_TEXT);
}

/**
 * test for bug #33652 (FTS unable to handle multiple space tokens) : destination
 * check that in case only one space is returned from the description and the
 * space does not have enough unused space the transfer fails
 */
void SrmCopyCoreTokensTestCase::test_space_token_01(){
    __TIMER;
    static const char * test_name = "srmcopycore_unit_test_space_token_01";
    static const char * dest_1 = "srmcopycore_multiple_tokens_dest_A_01.dat";
    static const char * dest_2 = "srmcopycore_multiple_tokens_dest_B_01.dat";

    static const unsigned long mb = 1024 * 1024;
    static const std::string TOKEN_A = make_token(10*mb,10*mb,0);
    statSetUpWithTokens(
        test_name,
        source_1,source_2,
        dest_1,dest_2,
        g_srm_22_endpoint,"2.2",
        "Description_for:TEST-SOURCE-TOKEN",
        make_description(TOKEN_A));

    {
        // Create Tx Core Object
        SrmCopyCore tx(test_name);
        // Prepare For transfer
        tx.prepare();
    }

    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);

    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat id", stat->mm_common.mm_id == SRMCOPY_STAT_ID);

    // Check Status
    check_status("Invalid Status", stat->mm_srmcopy.mm_status, UrlCopyStatus::PREPARING);
    check_scope("Invalid Scope", stat->mm_srmcopy.mm_error.mm_scope, UrlCopyError::SCOPE_DESTINATION);
    check_phase("Invalid Phase", stat->mm_srmcopy.mm_error.mm_phase, UrlCopyError::PHASE_PREPARATION);
    check_category("Invalid Error", stat->mm_srmcopy.mm_error.mm_category, UrlCopyError::ERROR_NO_SPACE_LEFT);
    CHECK_STRING(stat->mm_common.mm_dest_space_token,"");
}

/**
 * test for bug #33652 (FTS unable to handle multiple space tokens)
 * check that if multiple spaces are returned from the token description, is
 * selected the space that:
 * - has the minimum *unused space*
 * - has an unused space equal to at least file size + 1MB
 */
void SrmCopyCoreTokensTestCase::test_space_token_02(){
    __TIMER;
    static const char * test_name = "srmcopycore_unit_test_space_token_02";
    static const char * dest_1 = "srmcopycore_multiple_tokens_dest_A_02.dat";
    static const char * dest_2 = "srmcopycore_multiple_tokens_dest_B_02.dat";

    static const unsigned long mb = 1024 * 1024;
    static const std::string TOKEN_A = make_token(10*mb,10*mb,0);
    static const std::string TOKEN_B = make_token(10*mb,10*mb,3*mb);
    statSetUpWithTokens(
        test_name,
        source_1,source_2,
        dest_1,dest_2,
        g_srm_22_endpoint,"2.2",
        "Description_for:TEST-SOURCE-TOKEN",
        make_description(TOKEN_A,TOKEN_B));
    prepare(test_name);
    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat", stat != 0);
    CHECK_STRING(stat->mm_common.mm_dest_space_token,TOKEN_B.c_str());
}

/**
 * test for bug #33652 (FTS unable to handle multiple space tokens)
 * check that if multiple spaces are returned from the token description and
 * none of them satisfies the criteria of test case test_space_token_02, then
 * the transfer fails
 */
void SrmCopyCoreTokensTestCase::test_space_token_03(){
    __TIMER;
    static const char * test_name = "srmcopycore_unit_test_space_token_03";
    static const char * dest_1 = "srmcopycore_multiple_tokens_dest_A_03.dat";
    static const char * dest_2 = "srmcopycore_multiple_tokens_dest_B_03.dat";

    static const unsigned long mb = 1024 * 1024;
    static const std::string TOKEN_A = make_token(10*mb,10*mb,0);
    static const std::string TOKEN_B = make_token(10*mb,10*mb,1*mb);
    static const std::string TOKEN_C = make_token(10*mb,10*mb,1024);
    statSetUpWithTokens(
        test_name,
        source_1,source_2,
        dest_1,dest_2,
        g_srm_22_endpoint,"2.2",
        "Description_for:TEST-SOURCE-TOKEN",
        make_description(TOKEN_A,TOKEN_B,TOKEN_C    ));
    {
        // Create Tx Core Object
        SrmCopyCore tx(test_name);
        // Prepare For transfer
        tx.prepare();
    }

    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);

    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat id", stat->mm_common.mm_id == SRMCOPY_STAT_ID);

    // Check Status
    check_status("Invalid Status", stat->mm_srmcopy.mm_status, UrlCopyStatus::PREPARING);
    check_scope("Invalid Scope", stat->mm_srmcopy.mm_error.mm_scope, UrlCopyError::SCOPE_DESTINATION);
    check_phase("Invalid Phase", stat->mm_srmcopy.mm_error.mm_phase, UrlCopyError::PHASE_PREPARATION);
    check_category("Invalid Error", stat->mm_srmcopy.mm_error.mm_category, UrlCopyError::ERROR_NO_SPACE_LEFT);
    CHECK_STRING(stat->mm_common.mm_dest_space_token,"");
}

/**
 * check that for source tokens the first is selected, regardless of file size
 * and available space
 */
void SrmCopyCoreTokensTestCase::test_space_token_04(){
    __TIMER;
    static const char * test_name = "srmcopycore_unit_test_space_token_04";
    static const char * dest_1 = "srmcopycore_multiple_tokens_dest_A_04.dat";
    static const char * dest_2 = "srmcopycore_multiple_tokens_dest_B_04.dat";

    static const unsigned long mb = 1024 * 1024;
    static const std::string TOKEN_A = make_token(10*mb,10*mb,0);
    static const std::string TOKEN_B = make_token(100*mb,10*mb,10*mb);
    static const std::string TOKEN_C = make_token(1000*mb,10*mb,100*mb);
    statSetUpWithTokens(
        test_name,
        source_1,source_2,
        dest_1,dest_2,
        g_srm_22_endpoint,"2.2",
        make_description(TOKEN_A,TOKEN_B,TOKEN_C),
        "");
    prepare(test_name);
    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat", stat != 0);
    CHECK_STRING(stat->mm_common.mm_source_space_token,TOKEN_A.c_str());
}

/**
 * check that if an invalid source token is specified, the transfer fails
 */
void SrmCopyCoreTokensTestCase::test_space_token_05(){
    __TIMER;
    static const char * test_name = "srmcopycore_unit_test_space_token_05";
    static const char * dest_1 = "srmcopycore_multiple_tokens_dest_A_05.dat";
    static const char * dest_2 = "srmcopycore_multiple_tokens_dest_B_05.dat";

    statSetUpWithTokens(
        test_name,
        source_1,source_2,
        dest_1,dest_2,
        g_srm_22_endpoint,"2.2",
        "INVALID-SOURCE-TOKEN",
        "");
    {
        // Create Tx Core Object
        SrmCopyCore tx(test_name);
        // Prepare For transfer
        tx.prepare();
    }

    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);

    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat id", stat->mm_common.mm_id == SRMCOPY_STAT_ID);

    // Check Status
    check_status("Invalid Status", stat->mm_srmcopy.mm_status, UrlCopyStatus::PREPARING);
    check_scope("Invalid Scope", stat->mm_srmcopy.mm_error.mm_scope, UrlCopyError::SCOPE_SOURCE);
    check_phase("Invalid Phase", stat->mm_srmcopy.mm_error.mm_phase, UrlCopyError::PHASE_PREPARATION);
    check_category("Invalid Error", stat->mm_srmcopy.mm_error.mm_category, UrlCopyError::ERROR_USER);
    CHECK_STRING(stat->mm_common.mm_dest_space_token,"");
}

/**
 * check that if an invalid destination token is specified, the transfer fails
 */
void SrmCopyCoreTokensTestCase::test_space_token_06(){
    __TIMER;
    static const char * test_name = "srmcopycore_unit_test_space_token_06";
    static const char * dest_1 = "srmcopycore_multiple_tokens_dest_A_06.dat";
    static const char * dest_2 = "srmcopycore_multiple_tokens_dest_B_06.dat";

    statSetUpWithTokens(
        test_name,
        source_1,source_2,
        dest_1,dest_2,
        g_srm_22_endpoint,"2.2",
        "",
        "INVALID-DEST-TOKEN");
    {
        // Create Tx Core Object
        SrmCopyCore tx(test_name);
        // Prepare For transfer
        tx.prepare();
    }

    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);

    // Get the stat object
    SrmCopyStat * stat = reinterpret_cast<SrmCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat id", stat->mm_common.mm_id == SRMCOPY_STAT_ID);

    // Check Status
    check_status("Invalid Status", stat->mm_srmcopy.mm_status, UrlCopyStatus::PREPARING);
    check_scope("Invalid Scope", stat->mm_srmcopy.mm_error.mm_scope, UrlCopyError::SCOPE_DESTINATION);
    check_phase("Invalid Phase", stat->mm_srmcopy.mm_error.mm_phase, UrlCopyError::PHASE_PREPARATION);
    check_category("Invalid Error", stat->mm_srmcopy.mm_error.mm_category, UrlCopyError::ERROR_USER);
    CHECK_STRING(stat->mm_common.mm_dest_space_token,"");
}
