/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: test main function
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: TestTransferUrlCopy.cpp,v 1.12 2009/12/03 17:21:45 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include <glite/data/config/service/glite-config.h>
#include <glite/data/config/service/glite-config-register.hxx>

#include "logger-utility.h"
#include "test-utility.h"

#include <stdlib.h>
#include <string>
#include <map>
#include <iostream>
#include <unistd.h>
#include <stdarg.h>
#include <cppunit/extensions/TestFactoryRegistry.h>
#include <cppunit/ui/text/TestRunner.h>
#include <cppunit/CompilerOutputter.h>

using std::cout;
using std::cerr;
using std::cin;
using std::endl;

using namespace test_utils;

namespace {
const char * const CONFIG_NAME = "glite-transfer-urlcopy-test";
const char * const PROGNAME    = "glite_transfer_urlcopy_test";
const char * const PROGVERSION = "0.0";
const char * const PROGAUTHOR  = "(C) EGEE";

}

/*
 * The SRM endpoints
 */
std::string g_srm_11_endpoint;
std::string g_srm_22_endpoint;
std::string g_tmp_folder;
std::string g_srm11Pool;
std::string g_srm22Pool;

// ------------------------------------------------------------------------------
// -----------------------------------------------------------------------------
// Configuration for the Unit Test
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

#include <glite/data/config/service/ComponentConfiguration.hxx>
#include <glite/data/config/service/ParamReader.hxx>

using namespace glite::config;

/**
 * TransferUrlCopyTestComponent
 */
class TransferUrlCopyTestComponent: public ComponentConfiguration {
public:
    /**
     * Initialize the Component
     */
    virtual int init(const Params& params){return 0;}

    /**
     * Configure the Component
     */
    virtual int config(const Params& params){
        ParamReader reader(getName(),params);
        reader.getParameter("srm11Pool",g_srm11Pool).required();
        reader.getParameter("srm22Pool",g_srm22Pool).required();
        return 0;
    }

    /**
     * Start the Component
     */
    virtual int start(){
        //log_setup(CONFIG_NAME,glite::data::transfer::urlcopy::LOGLEVEL_DEBUG,
        //    true,"test-transfer-url-copy", false, "", 0, "");
        
        // Console logging on
        log_setup(CONFIG_NAME,glite::data::transfer::urlcopy::LOGLEVEL_DEBUG,
            false,"test-transfer-url-copy", false, "", 0, "");
        return 0;
    }

    /**
     * Stop the Component
     */
    virtual int stop(){
        // TODO Finalize the logger
        return 0;
    }
 
    /**
     * Finalize the Component
     */
    virtual int fini(){return 0;}

    /**
     * Return Singleton Instance
     */
    static TransferUrlCopyTestComponent& instance() {
        static TransferUrlCopyTestComponent s_instance;
        return s_instance;
    }
        
private:
    /**
     * Constructor
     */
    TransferUrlCopyTestComponent():ComponentConfiguration(CONFIG_NAME){
    }
    /**
     * Destructor
     */
    virtual ~TransferUrlCopyTestComponent(){}
    
private:

};

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

/*
 * print_usage_and_die
 *
 * Print Program Usage and ther exit
 */
void print_usage(){
    cout << endl;
    cout << "<" << PROGNAME << "> Version " << PROGVERSION << " by " << PROGAUTHOR;
    cout << endl;
    cout << "usage: " << PROGNAME << "[-l <glite_location_user>] [-s] -1 <srm1.1 endpoint> -2 <srm2.2 endpoint> [test classes to run]" << endl;
    cout << " Where:" << endl;
    cout << " -l : specify the glite location. Configuration files should be located under the etc folder on that path" << endl;
    cout << " -s : Enable the Compiler Outputter" << endl;
    cout << " -1 : SRM 1.1 enpoint" << endl;
    cout << " -2 : SRM 2.2 enpoint" << endl;
    cout << " -c : Configuration name" << endl;
    cout << " -t : folder for temporary files" << endl;
    cout << " Where:" << endl;
    cout << " Optional parameters:" << endl;
    cout << "  -h : print this screen" << endl;
    cout << " Test classes:" << endl;
    
    const TestsMap::Names& n = TestsMap::instance().names();
    for (TestsMap::Names::const_iterator it = n.begin();it != n.end();++it){
        cout << "     " << (*it) << endl;
    }

}

/*
 * main
 *
 * Transfer FTS Main function
 */
int main(int argc, char* argv[]) {
    
    // Get Options
    const char * location = 0;
    const char * config   = CONFIG_NAME;
    bool selfTest = false;
    int flag;
    while ((flag = getopt (argc, argv, "l:1:2:c:st:h")) != -1) {
        switch (flag) {
            case 'l':
                location = optarg;
            break;
            case 'c':
                config = optarg;
            break;
            case 's':
                selfTest = true;
            break;
            case '1':
                if(0 != optarg){
                    g_srm_11_endpoint = optarg;
                }
            break;
            case '2':
                if(0 != optarg){
                    g_srm_22_endpoint = optarg;
                }
            break;
            case 't':
                if(0 != optarg){
                    g_tmp_folder = optarg;
                }
                break;
            case 'h':
            default:
                print_usage();
            break;
        } // End switch
    } // End while
    if (optind < argc){
        // enable only specified test classes
        TestsMap::instance().disableAll();
        for (;optind < argc;++optind){
            std::string test_name = argv[optind];
            if (!TestsMap::instance().enable(test_name)){
                cerr << "Warning: Invalid test class name: " << test_name << endl;
                print_usage();
            }
        }
    }

    if (g_srm_11_endpoint.empty() ||
        g_srm_22_endpoint.empty() ||
        g_tmp_folder.empty()) {
        print_usage();
    }
    if(0 != location){
        // Set "GLite Location User"
        setenv("GLITE_LOCATION",location,1);
    }
    // Register the Static Component
    glite_config_register(&(TransferUrlCopyTestComponent::instance()));
    // Initialize the Application
    int initres = glite_config_initialize(config);
    if(initres < 0){
        cerr << "Warning: Initialization failed " << endl;
    }
    
    CPPUNIT_NS :: TestResult testresult;
    CPPUNIT_NS :: TestResultCollector collectedresults;
    testresult.addListener (&collectedresults);
    CPPUNIT_NS :: BriefTestProgressListener progress;
    testresult.addListener (&progress);
    CPPUNIT_NS :: TestRunner testrunner;
    testrunner.addTest (CPPUNIT_NS :: TestFactoryRegistry :: getRegistry ().makeTest ());
    testrunner.run (testresult);
    CPPUNIT_NS :: CompilerOutputter compileroutputter (&collectedresults, std::cerr);
    compileroutputter.write ();
    bool wasSucessful = collectedresults.wasSuccessful();

    // Finalize
    glite_config_finalize();
    
    return wasSucessful ? 0 : 1;
}
