/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: UrlCopyCoreTokens Test Case
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: UrlCopyCoreCheckSumTestCase.cpp,v 1.13 2009/10/26 16:32:20 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#include "UrlCopyCore.h"
#include "Mmap.h"
#include "stat-utility.h"
#include "TransferConfig.h"
#include "ChecksumCheckerUtils.h"
#include "ChecksumChecker.h"
#include "Checksum.h"
#include <glite/data/agents/AgentExceptions.h>

#include <glite/data/srm/util/Context.h>
#include <glite/data/srm/util/PrepareToPut.h>
#include <glite/data/srm/util/SrmMkDir.h>

#include <memory>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <boost/shared_ptr.hpp>
#include <boost/algorithm/string.hpp>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <sstream>

#include "UrlCopyCoreCheckSumTestCase.h"

using namespace glite::data::agents;
using namespace glite::data::srm::util;
using namespace glite::data::transfer::urlcopy;

using boost::shared_ptr;

/*
 * The SRM endpoints
 */
extern std::string g_srm_11_endpoint;
extern std::string g_srm_22_endpoint;

namespace {
const char SAMPLE_TEXT[] = "Hello World! This is a test\nWrite some data for Transfer Url Copy tests\n";
const char EMPTY_FILE[] = "";
static const char * V = "2.2";

static const char * algorithm   = "ADLER32";
static const char * value       = "A123456789ABCDEF";
std::string source_file         = make_cksm_file("urlcopy-src.dat",algorithm,value);
std::string source_no_checksum  = "urlcopy-src-nochecksum.dat";
}

/**
 * setup the source file for the test.
 */
void UrlCopyCoreCheckSumTestCase::test_setup(){
    putFile(source_file,g_srm_22_endpoint,V,SAMPLE_TEXT);
    putFile(source_no_checksum,g_srm_22_endpoint,V,SAMPLE_TEXT);
}



void UrlCopyCoreCheckSumTestCase::testUseCase0()
{
	__TIMER;
    static const char * test_name   = "testUseCase0";
    std::string dest_file = make_cksm_file("urlcopy-dst01.dat", algorithm, value);

    statSetUpChecksum(
    	test_name, source_file, g_srm_22_endpoint, V, dest_file,
    	g_srm_22_endpoint, V, "", "", false);

    prepare(test_name);
    transfer(test_name);
    complete(test_name);
    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name, true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);
    // Check that checksum data is written in output structures
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm, "");
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value, "");
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm, "");
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value, "");
    CPPUNIT_ASSERT_MESSAGE("Destination file was deleted.", fileExists(dest_file,g_srm_22_endpoint,V));
}

void UrlCopyCoreCheckSumTestCase::testUseCase1_src_dst_checksumMatch()
{
	__TIMER;
    static const char * test_name   = "testUseCase1_src_checksumMatch";
    std::string dest_file = make_cksm_file("urlcopy-dst01.dat", algorithm, value);

    statSetUpChecksum(
    	test_name, source_file, g_srm_22_endpoint, V, dest_file,
    	g_srm_22_endpoint, V, algorithm, value, false);

    prepare(test_name);
    transfer(test_name);
    complete(test_name);
    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name, true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);
    // Check that checksum data is written in output structures
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value, value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value, value);
}

void UrlCopyCoreCheckSumTestCase::testUseCase1_src_checksumNotAvailable()
{
	__TIMER;
    static const char * test_name   = "testUseCase1_src_checksumNotAvailable";
    std::string dest_file = make_cksm_file("urlcopy-dst01.dat", algorithm, value);

    statSetUpChecksum(
    	test_name, source_no_checksum, g_srm_22_endpoint, V, dest_file,
    	g_srm_22_endpoint, V, algorithm, value, false);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP(algorithm, value);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);

    prepare(test_name);
    transfer(test_name);
    complete(test_name);
    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name, true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);
    // Check that checksum data is written in output structures
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm, "");
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value, "");
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value, value);
    CPPUNIT_ASSERT(!getter->fallbackHappened());
}

void UrlCopyCoreCheckSumTestCase::testUseCase1_src_checksumDifferent()
{
	__TIMER;
    static const char * test_name   = "testUseCase1_src_checksumDifferent";
    static const char * u_value     = "dfdfdfdfdfdfdfdf";

    std::string dest_file = make_cksm_file("urlcopy-dst03.dat", algorithm, value);

    statSetUpChecksum(test_name, source_file, g_srm_22_endpoint, V,
    	dest_file, g_srm_22_endpoint, V, algorithm, u_value, false);

    {
        UrlCopyCore tx(test_name);
        tx.prepare();
    }

    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    check_status("Invalid Status",  stat->mm_urlcopy.mm_status,             UrlCopyStatus::PREPARING);
    check_scope("Invalid Scope",    stat->mm_urlcopy.mm_error.mm_scope,     UrlCopyError::SCOPE_SOURCE);
    check_phase("Invalid Phase",    stat->mm_urlcopy.mm_error.mm_phase,     UrlCopyError::PHASE_PREPARATION);
    check_category("Invalid Error", stat->mm_urlcopy.mm_error.mm_category,  UrlCopyError::ERROR_CHECKSUM_MISMATCH);

    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value,    value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm,"");
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value,    "");
}

void UrlCopyCoreCheckSumTestCase::testUseCase1_dst_gridFTPfallback()
{
	__TIMER;
    static const char* test_name   = "testUseCase1_dst_gridFTPfallback";
    std::string dest_file   = "urlcopy-dst07.dat";

    statSetUpChecksum(test_name, source_file, g_srm_22_endpoint, V,
    	dest_file, g_srm_22_endpoint, V, algorithm, value, false);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP(algorithm, value);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);

    prepare(test_name);
    transfer(test_name);
    {
        UrlCopyCore tx(test_name);
        tx.complete();
    }

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT(getter->fallbackHappened());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value, value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value, value);
    CPPUNIT_ASSERT(getter->fallbackHappened());
    CPPUNIT_ASSERT_MESSAGE("Destination file was deleted.", fileExists(dest_file,g_srm_22_endpoint,V));
}

void UrlCopyCoreCheckSumTestCase::testUseCase1_dst_checksumNotAvailable()
{
	__TIMER;
    static const char* test_name   = "testUseCase1_dst_checksumNotAvailable";
    std::string dest_file   = "urlcopy-dst07.dat";

    statSetUpChecksum(test_name, source_file, g_srm_22_endpoint, V,
    	dest_file, g_srm_22_endpoint, V, algorithm, value, false);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP("", "", UrlCopyError::ERROR_CHECKSUM_UNKNOWN);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);

    prepare(test_name);
    transfer(test_name);
    {
        UrlCopyCore tx(test_name);
        tx.complete();
    }

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT(getter->fallbackHappened());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    CPPUNIT_ASSERT(getter->fallbackHappened());
    check_status("Invalid Status",  stat->mm_urlcopy.mm_status,             UrlCopyStatus::COMPLETED);
    check_scope("Invalid Scope",    stat->mm_urlcopy.mm_error.mm_scope,     UrlCopyError::SCOPE_DESTINATION);
    check_phase("Invalid Phase",    stat->mm_urlcopy.mm_error.mm_phase,     UrlCopyError::PHASE_FINALIZATION);
    check_category("Invalid Error", stat->mm_urlcopy.mm_error.mm_category,  UrlCopyError::ERROR_CHECKSUM_UNKNOWN);

    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value,    value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm,"");
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value,    "");

    CPPUNIT_ASSERT_MESSAGE("Destination file was not deleted.",!fileExists(dest_file,g_srm_22_endpoint,V));
}

void UrlCopyCoreCheckSumTestCase::testUseCase1_dst_checksumDifferent()
{
	__TIMER;
    static const char * test_name   = "testUseCase1_dst_checksumDifferent";

    static const char * d_value     = "AAAAAAAAAAA2";
    std::string dest_file   = make_cksm_file("urlcopy-dst06.dat",algorithm,d_value);

    statSetUpChecksum(test_name,source_file,g_srm_22_endpoint,V,dest_file,g_srm_22_endpoint,V,
        algorithm,value, false);

    prepare(test_name);
    transfer(test_name);
    {
        UrlCopyCore tx(test_name);
        tx.complete();
    }

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    check_status("Invalid Status",  stat->mm_urlcopy.mm_status,             UrlCopyStatus::COMPLETED);
    check_scope("Invalid Scope",    stat->mm_urlcopy.mm_error.mm_scope,     UrlCopyError::SCOPE_DESTINATION);
    check_phase("Invalid Phase",    stat->mm_urlcopy.mm_error.mm_phase,     UrlCopyError::PHASE_FINALIZATION);
    check_category("Invalid Error", stat->mm_urlcopy.mm_error.mm_category,  UrlCopyError::ERROR_CHECKSUM_MISMATCH);

    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value,    value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value,    d_value);
    CPPUNIT_ASSERT_MESSAGE("Destination file was not deleted.",!fileExists(dest_file,g_srm_22_endpoint,V));
}

void UrlCopyCoreCheckSumTestCase::testUseCase2_src_dst_checksumMatch()
{
	__TIMER;
    static const char * test_name   = "testUseCase2_src_dst_checksumMatch";
    std::string dest_file = make_cksm_file("urlcopy-dst01.dat", algorithm, value);

    statSetUpChecksum(
    	test_name, source_file, g_srm_22_endpoint, V, dest_file,
    	g_srm_22_endpoint, V, "", "", true);

    prepare(test_name);
    transfer(test_name);
    complete(test_name);
    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name, true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);
    // Check that checksum data is written in output structures
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value, value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value, value);
    CPPUNIT_ASSERT_MESSAGE("Destination file was deleted.", fileExists(dest_file,g_srm_22_endpoint,V));
}

void UrlCopyCoreCheckSumTestCase::testUseCase2_src_checksumNotAvailable_gridFTP()
{
	__TIMER;
    static const char * test_name   = "testUseCase2_src_checksumNotAvailable_gridFTP";
    std::string dest_file = make_cksm_file("urlcopy-dst01.dat", algorithm, value);

    statSetUpChecksum(
    	test_name, source_no_checksum, g_srm_22_endpoint, V, dest_file,
    	g_srm_22_endpoint, V, "", "", true);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP(algorithm, value);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);

    prepare(test_name);
    transfer(test_name);
    complete(test_name);
    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name, true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);
    // Check that checksum data is written in output structures
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value, value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value, value);
    CPPUNIT_ASSERT(getter->fallbackHappened());
}

void UrlCopyCoreCheckSumTestCase::testUseCase2_src_checksumNotAvailable()
{
	__TIMER;
    static const char * test_name   = "testUseCase2_src_checksumNotAvailable";
    std::string dest_file = make_cksm_file("urlcopy-dst01.dat", algorithm, value);

    statSetUpChecksum(
    	test_name, source_no_checksum, g_srm_22_endpoint, V, dest_file,
    	g_srm_22_endpoint, V, "", "", true);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP(algorithm, value, UrlCopyError::ERROR_CHECKSUM_UNKNOWN);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);
    CPPUNIT_ASSERT(!getter->fallbackHappened());

    {
        UrlCopyCore tx(test_name);
        tx.prepare();
    }

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    check_status("Invalid Status",  stat->mm_urlcopy.mm_status,             UrlCopyStatus::PREPARING);
    check_scope("Invalid Scope",    stat->mm_urlcopy.mm_error.mm_scope,     UrlCopyError::SCOPE_SOURCE);
    check_phase("Invalid Phase",    stat->mm_urlcopy.mm_error.mm_phase,     UrlCopyError::PHASE_PREPARATION);
    check_category("Invalid Error", stat->mm_urlcopy.mm_error.mm_category,  UrlCopyError::ERROR_CHECKSUM_UNKNOWN);

    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm,"");
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value,    "");
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm,"");
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value,    "");
    CPPUNIT_ASSERT(getter->fallbackHappened());
}

void UrlCopyCoreCheckSumTestCase::testUseCase2_dst_checksumNotAvailable_gridFTP()
{
	__TIMER;
    static const char * test_name   = "testUseCase2_dst_checksumNotAvailable_gridFTP";
    std::string dest_file = "urlcopy-dst07.dat";

    statSetUpChecksum(
    	test_name, source_file, g_srm_22_endpoint, V, dest_file,
    	g_srm_22_endpoint, V, "", "", true);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP(algorithm, value);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);
    CPPUNIT_ASSERT(!getter->fallbackHappened());

    prepare(test_name);
    transfer(test_name);
    {
        UrlCopyCore tx(test_name);
        tx.complete();
    }

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT(getter->fallbackHappened());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    CPPUNIT_ASSERT(getter->fallbackHappened());
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value,    value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value,    value);
    CPPUNIT_ASSERT_MESSAGE("Destination file was deleted.", fileExists(dest_file,g_srm_22_endpoint,V));
}

void UrlCopyCoreCheckSumTestCase::testUseCase2_dst_checksumNotAvailable()
{
	__TIMER;
    static const char * test_name   = "testUseCase2_dst_checksumNotAvailable";
    std::string dest_file = "urlcopy-dst07.dat";

    statSetUpChecksum(
    	test_name, source_file, g_srm_22_endpoint, V, dest_file,
    	g_srm_22_endpoint, V, "", "", true);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP("", "", UrlCopyError::ERROR_CHECKSUM_UNKNOWN);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);
    CPPUNIT_ASSERT(!getter->fallbackHappened());

    prepare(test_name);
    transfer(test_name);
    {
        UrlCopyCore tx(test_name);
        tx.complete();
    }

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT(getter->fallbackHappened());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    check_status("Invalid Status",  stat->mm_urlcopy.mm_status,             UrlCopyStatus::COMPLETED);
    check_scope("Invalid Scope",    stat->mm_urlcopy.mm_error.mm_scope,     UrlCopyError::SCOPE_DESTINATION);
    check_phase("Invalid Phase",    stat->mm_urlcopy.mm_error.mm_phase,     UrlCopyError::PHASE_FINALIZATION);
    check_category("Invalid Error", stat->mm_urlcopy.mm_error.mm_category,  UrlCopyError::ERROR_CHECKSUM_UNKNOWN);

    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value,    value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm,"");
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value,    "");
    CPPUNIT_ASSERT(getter->fallbackHappened());
    CPPUNIT_ASSERT_MESSAGE("Destination file was not deleted.", !fileExists(dest_file,g_srm_22_endpoint,V));
}

void UrlCopyCoreCheckSumTestCase::testUseCase2_checksumDifferent()
{
	__TIMER;
    static const char * test_name   = "testUseCase2_checksumDifferent";
    static const char * d_value     = "AAAAAAAAAAA2";
    std::string dest_file   = make_cksm_file("urlcopy-dst06.dat", algorithm, d_value);

    statSetUpChecksum(
    	test_name, source_file, g_srm_22_endpoint, V, dest_file,
    	g_srm_22_endpoint, V, "", "", true);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP("", "", UrlCopyError::ERROR_CHECKSUM_UNKNOWN);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);
    CPPUNIT_ASSERT(!getter->fallbackHappened());

    prepare(test_name);
    transfer(test_name);
    {
        UrlCopyCore tx(test_name);
        tx.complete();
    }

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT(!getter->fallbackHappened());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    check_status("Invalid Status",  stat->mm_urlcopy.mm_status,             UrlCopyStatus::COMPLETED);
    check_scope("Invalid Scope",    stat->mm_urlcopy.mm_error.mm_scope,     UrlCopyError::SCOPE_DESTINATION);
    check_phase("Invalid Phase",    stat->mm_urlcopy.mm_error.mm_phase,     UrlCopyError::PHASE_FINALIZATION);
    check_category("Invalid Error", stat->mm_urlcopy.mm_error.mm_category,  UrlCopyError::ERROR_CHECKSUM_MISMATCH);

    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value,    value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value,    d_value);
    CPPUNIT_ASSERT_MESSAGE("Destination file was not deleted.", !fileExists(dest_file,g_srm_22_endpoint,V));
}

void UrlCopyCoreCheckSumTestCase::testUseCase3_src_dst_checksumMatch()
{
	__TIMER;
    static const char * test_name   = "testUseCase3_src_dst_checksumMatch";
    std::string dest_file = make_cksm_file("urlcopy-dst01.dat", algorithm, value);

    statSetUpChecksum(
    	test_name, source_file, g_srm_22_endpoint, V, dest_file,
    	g_srm_22_endpoint, V, algorithm, value, true);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP("", "", UrlCopyError::ERROR_CHECKSUM_UNKNOWN);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);
    CPPUNIT_ASSERT(!getter->fallbackHappened());

    prepare(test_name);
    transfer(test_name);
    complete(test_name);
    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name, true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);
    // Check that checksum data is written in output structures
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value, value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value, value);
    CPPUNIT_ASSERT_MESSAGE("Destination file was deleted.", fileExists(dest_file,g_srm_22_endpoint,V));
    CPPUNIT_ASSERT(!getter->fallbackHappened());
}

void UrlCopyCoreCheckSumTestCase::testUseCase3_src_checksumNotAvailable_gridFTP()
{
	__TIMER;
    static const char * test_name   = "testUseCase3_src_checksumNotAvailable_gridFTP";
    std::string dest_file = make_cksm_file("urlcopy-dst01.dat", algorithm, value);

    statSetUpChecksum(
    	test_name, source_no_checksum, g_srm_22_endpoint, V, dest_file,
    	g_srm_22_endpoint, V, algorithm, value, true);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP(algorithm, value);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);
    CPPUNIT_ASSERT(!getter->fallbackHappened());

    prepare(test_name);
    CPPUNIT_ASSERT(getter->fallbackHappened());
    transfer(test_name);
    complete(test_name);
    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name, true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);
    // Check that checksum data is written in output structures
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value, value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value, value);
    CPPUNIT_ASSERT_MESSAGE("Destination file does not exist!", fileExists(dest_file,g_srm_22_endpoint,V));
}

void UrlCopyCoreCheckSumTestCase::testUseCase3_src_checksumNotAvailable()
{
	__TIMER;
    static const char * test_name   = "testUseCase3_src_checksumNotAvailable";
    std::string dest_file = make_cksm_file("urlcopy-dst01.dat", algorithm, value);

    statSetUpChecksum(
    	test_name, source_no_checksum, g_srm_22_endpoint, V, dest_file,
    	g_srm_22_endpoint, V, algorithm, value, true);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP(algorithm, value, UrlCopyError::ERROR_CHECKSUM_UNKNOWN);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);
    CPPUNIT_ASSERT(!getter->fallbackHappened());

    {
        UrlCopyCore tx(test_name);
        tx.prepare();
    }

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    check_status("Invalid Status",  stat->mm_urlcopy.mm_status,             UrlCopyStatus::PREPARING);
    check_scope("Invalid Scope",    stat->mm_urlcopy.mm_error.mm_scope,     UrlCopyError::SCOPE_SOURCE);
    check_phase("Invalid Phase",    stat->mm_urlcopy.mm_error.mm_phase,     UrlCopyError::PHASE_PREPARATION);
    check_category("Invalid Error", stat->mm_urlcopy.mm_error.mm_category,  UrlCopyError::ERROR_CHECKSUM_UNKNOWN);

    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm,"");
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value,    "");
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm,"");
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value,    "");
    CPPUNIT_ASSERT(getter->fallbackHappened());
    CPPUNIT_ASSERT_MESSAGE("Destination file exists!", fileExists(dest_file,g_srm_22_endpoint,V));
}

void UrlCopyCoreCheckSumTestCase::testUseCase3_src_checksumDifferent()
{
	__TIMER;
    static const char * test_name   = "testUseCase3_src_checksumDifferent";
    static const char * u_value     = "dfdfdfdfdfdfdfdf";

    std::string dest_file = make_cksm_file("urlcopy-dst03.dat", algorithm, value);

    statSetUpChecksum(test_name, source_file, g_srm_22_endpoint, V,
    	dest_file, g_srm_22_endpoint, V, algorithm, u_value, true);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP(algorithm, value, UrlCopyError::ERROR_CHECKSUM_UNKNOWN);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);
    CPPUNIT_ASSERT(!getter->fallbackHappened());

    {
        UrlCopyCore tx(test_name);
        tx.prepare();
    }

    CPPUNIT_ASSERT(!getter->fallbackHappened());
    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    check_status("Invalid Status",  stat->mm_urlcopy.mm_status,             UrlCopyStatus::PREPARING);
    check_scope("Invalid Scope",    stat->mm_urlcopy.mm_error.mm_scope,     UrlCopyError::SCOPE_SOURCE);
    check_phase("Invalid Phase",    stat->mm_urlcopy.mm_error.mm_phase,     UrlCopyError::PHASE_PREPARATION);
    check_category("Invalid Error", stat->mm_urlcopy.mm_error.mm_category,  UrlCopyError::ERROR_CHECKSUM_MISMATCH);

    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value,    value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm,"");
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value,    "");
    CPPUNIT_ASSERT_MESSAGE("Destination file exists!", !fileExists(dest_file,g_srm_22_endpoint,V));
}

void UrlCopyCoreCheckSumTestCase::testUseCase3_dst_checksumNotAvailable_gridFTP()
{
	__TIMER;
    static const char * test_name   = "testUseCase3_dst_checksumNotAvailable_gridFTP";
    std::string dest_file = "urlcopy-dst07.dat";

    statSetUpChecksum(
    	test_name, source_file, g_srm_22_endpoint, V, dest_file,
    	g_srm_22_endpoint, V, algorithm, value, true);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP(algorithm, value);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);
    CPPUNIT_ASSERT(!getter->fallbackHappened());

    prepare(test_name);
    transfer(test_name);
    {
        UrlCopyCore tx(test_name);
        tx.complete();
    }

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT(getter->fallbackHappened());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    CPPUNIT_ASSERT(getter->fallbackHappened());
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value,    value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value,    value);
    CPPUNIT_ASSERT_MESSAGE("Destination file was deleted.", fileExists(dest_file,g_srm_22_endpoint,V));
}

void UrlCopyCoreCheckSumTestCase::testUseCase3_dst_checksumNotAvailable()
{
	__TIMER;
    static const char * test_name   = "testUseCase2_dst_checksumNotAvailable";
    std::string dest_file = "urlcopy-dst07.dat";

    statSetUpChecksum(
    	test_name, source_file, g_srm_22_endpoint, V, dest_file,
    	g_srm_22_endpoint, V, algorithm, value, true);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP("", "", UrlCopyError::ERROR_CHECKSUM_UNKNOWN);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);
    CPPUNIT_ASSERT(!getter->fallbackHappened());

    prepare(test_name);
    transfer(test_name);
    {
        UrlCopyCore tx(test_name);
        tx.complete();
    }

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT(getter->fallbackHappened());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    check_status("Invalid Status",  stat->mm_urlcopy.mm_status,             UrlCopyStatus::COMPLETED);
    check_scope("Invalid Scope",    stat->mm_urlcopy.mm_error.mm_scope,     UrlCopyError::SCOPE_DESTINATION);
    check_phase("Invalid Phase",    stat->mm_urlcopy.mm_error.mm_phase,     UrlCopyError::PHASE_FINALIZATION);
    check_category("Invalid Error", stat->mm_urlcopy.mm_error.mm_category,  UrlCopyError::ERROR_CHECKSUM_UNKNOWN);

    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value,    value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm,"");
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value,    "");
    CPPUNIT_ASSERT(getter->fallbackHappened());
    CPPUNIT_ASSERT_MESSAGE("Destination file was not deleted.", !fileExists(dest_file,g_srm_22_endpoint,V));
}

void UrlCopyCoreCheckSumTestCase::testUseCase3_dst_checksumDifferent()
{
	__TIMER;
    static const char * test_name   = "testUseCase3_dst_checksumDifferent";

    static const char * d_value     = "AAAAAAAAAAA2";
    std::string dest_file   = make_cksm_file("urlcopy-dst06.dat",algorithm, d_value);

    statSetUpChecksum(test_name,source_file,g_srm_22_endpoint,V,dest_file,g_srm_22_endpoint,V,
        algorithm, value, true);

    ChecksumGetter_noGridFTP* getter = new ChecksumGetter_noGridFTP("", "", UrlCopyError::ERROR_CHECKSUM_UNKNOWN);
    ChecksumChecker::Test_ChangeChecksumGetter(getter);
    CPPUNIT_ASSERT(!getter->fallbackHappened());

    prepare(test_name);
    transfer(test_name);
    {
        UrlCopyCore tx(test_name);
        tx.complete();
    }

    CPPUNIT_ASSERT(!getter->fallbackHappened());
    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    check_status("Invalid Status",  stat->mm_urlcopy.mm_status,             UrlCopyStatus::COMPLETED);
    check_scope("Invalid Scope",    stat->mm_urlcopy.mm_error.mm_scope,     UrlCopyError::SCOPE_DESTINATION);
    check_phase("Invalid Phase",    stat->mm_urlcopy.mm_error.mm_phase,     UrlCopyError::PHASE_FINALIZATION);
    check_category("Invalid Error", stat->mm_urlcopy.mm_error.mm_category,  UrlCopyError::ERROR_CHECKSUM_MISMATCH);

    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value,    value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value,    d_value);
    CPPUNIT_ASSERT_MESSAGE("Destination file was not deleted.",!fileExists(dest_file,g_srm_22_endpoint,V));
}

void UrlCopyCoreCheckSumTestCase::testMismatchInAlgorithmName()
{
	__TIMER;
    static const char * test_name   = "testMismatchInAlgorithmName";
    static const char * u_algorithm = "BAAAAAD";

    std::string dest_file = make_cksm_file("urlcopy-dst03.dat", algorithm, value);

    statSetUpChecksum(test_name, source_file, g_srm_22_endpoint, V,
    	dest_file, g_srm_22_endpoint, V, u_algorithm, value, false);

    {
        UrlCopyCore tx(test_name);
        tx.prepare();
    }

    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    check_status("Invalid Status",  stat->mm_urlcopy.mm_status,             UrlCopyStatus::PREPARING);
    check_scope("Invalid Scope",    stat->mm_urlcopy.mm_error.mm_scope,     UrlCopyError::SCOPE_SOURCE);
    check_phase("Invalid Phase",    stat->mm_urlcopy.mm_error.mm_phase,     UrlCopyError::PHASE_PREPARATION);
    check_category("Invalid Error", stat->mm_urlcopy.mm_error.mm_category,  UrlCopyError::ERROR_CHECKSUM_MISMATCH);

    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm,algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value,    value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm,"");
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value,    "");
}

void UrlCopyCoreCheckSumTestCase::testNonStandardAlgorithmName()
{
	__TIMER;
    static const char * test_name   = "testNonStandardAlgorithmName";
    const char* d_algorithm = "AD";

    std::string dest_file = make_cksm_file("urlcopy-dst01.dat", d_algorithm, value);

    statSetUpChecksum(
    	test_name, source_file, g_srm_22_endpoint, V, dest_file,
    	g_srm_22_endpoint, V, "", "", true);

    prepare(test_name);
    transfer(test_name);
    complete(test_name);
    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name, true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);
    // Check that checksum data is written in output structures
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value, value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value, value);
    CPPUNIT_ASSERT_MESSAGE("Destination file was deleted.", fileExists(dest_file,g_srm_22_endpoint,V));
}

void UrlCopyCoreCheckSumTestCase::testChecksumCaseIndependence()
{
	__TIMER;
    Checksum chk_lower("ADLER32", "abcde");
    Checksum chk_upper("ADLER32", "ABCDE");
    CPPUNIT_ASSERT(chk_lower == chk_upper);
}

void UrlCopyCoreCheckSumTestCase::testChecksum()
{
	__TIMER;
    std::string val("ABCDE");
    std::string alg("ADLER32");
    Checksum chk(alg, val);
    CPPUNIT_ASSERT_EQUAL(alg, chk.algorithm());
    CPPUNIT_ASSERT_EQUAL(val, chk.value());
}

void UrlCopyCoreCheckSumTestCase::testChecksumSpecDeviations()
{
    std::string rightAdler("5678");
    Checksum chk_good("ADLER32", rightAdler);
    // Adler padded with spaces. Leading spaces should be removed.
    std::string adlerPaddedWithSpaces("   " + rightAdler);
    Checksum chk_spaces("ADLER32", adlerPaddedWithSpaces);
    CPPUNIT_ASSERT_EQUAL(chk_good, chk_spaces);

    // Adler padded with zeros. Leading zeros should be removed.
    std::string adlerPaddedWithZeros("00" + rightAdler);
    Checksum chk_zeros("ADLER32", adlerPaddedWithZeros);
    CPPUNIT_ASSERT_EQUAL(chk_good, chk_zeros);

    // Adler padded with something else. Should not match
    std::string adlerPaddedWithElse("xx" + rightAdler);
    Checksum chk_else("ADLER32", adlerPaddedWithElse);
    CPPUNIT_ASSERT(chk_good != chk_else);
}

void UrlCopyCoreCheckSumTestCase::testUseCase1_src_dst_deviated_checksumMatch()
{
    __TIMER;
    static const char * test_name   = "testUseCase1_src_checksumMatch";
    static const char * val_space   = "     a123456789abcdef";

    std::string dest_file = make_cksm_file("urlcopy-dst01.dat", algorithm, value);

    statSetUpChecksum(
        test_name, source_file, g_srm_22_endpoint, V, dest_file,
        g_srm_22_endpoint, V, algorithm, val_space, false);

    prepare(test_name);
    transfer(test_name);
    complete(test_name);
    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name, true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);
    // Check that checksum data is written in output structures
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_src_checksum_value, value);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_algorithm, algorithm);
    CHECK_STRING(stat->mm_urlcopy.mm_dst_checksum_value, value);
}

