/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: UrlCopyCore Test Case
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *           Zsolt Molnar <zsolt.molnar@cern.ch>
 *  Version info: $Id: UrlCopyCoreTestCase.cpp,v 1.13 2009/03/19 13:37:09 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include "UrlCopyCore.h"
#include "UrlCopyCoreTestCase.h"
#include "Mmap.h"
#include "stat-utility.h"
#include "TransferConfig.h"
#include <glite/data/agents/AgentExceptions.h>

#include <glite/data/srm/util/Context.h>
#include <glite/data/srm/util/PrepareToPut.h>
#include <glite/data/srm/util/SrmMkDir.h>

#include <memory>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <boost/shared_ptr.hpp>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <sstream>

using namespace glite::data::agents;
using namespace glite::data::srm::util;
using namespace glite::data::transfer::urlcopy;

using boost::shared_ptr;

/*
 * The SRM endpoints
 */
extern std::string g_srm_11_endpoint;
extern std::string g_srm_22_endpoint;

namespace {
const char SAMPLE_TEXT[] = "Hello World! This is a test\nWrite some data for Transfer Url Copy tests\n";
const char EMPTY_FILE[] = "";
}

/*
 * Put Source file for 1.1 tests
 */
void UrlCopyCoreTestCase::test_PutSource_11(){
    __TIMER;
    CPPUNIT_ASSERT_MESSAGE("Invalid Endpoint", g_srm_11_endpoint.empty() == false);
    putFile("urlcopycore_001.dat",g_srm_11_endpoint,"1.1",SAMPLE_TEXT);
}

/*
 * Put Source file for 2.2 tests
 */
void UrlCopyCoreTestCase::test_PutSource_22(){
    __TIMER;
    CPPUNIT_ASSERT_MESSAGE("Invalid Endpoint", g_srm_22_endpoint.empty() == false);
    putFile("urlcopycore_001.dat",g_srm_22_endpoint,"2.2",SAMPLE_TEXT);
}

/*
 * Put Source file for 1.1 & 2.2 tests
 */
void UrlCopyCoreTestCase::test_PutSource_11_22(){
    __TIMER;
    CPPUNIT_ASSERT_MESSAGE("Invalid Endpoint", g_srm_11_endpoint.empty() == false);
    putFile("urlcopycore_003.dat",g_srm_11_endpoint,"1.1",SAMPLE_TEXT);
}

/*
 * Put Source file for 2.2 & 1.1 tests
 */
void UrlCopyCoreTestCase::test_PutSource_22_11(){
    __TIMER;
    CPPUNIT_ASSERT_MESSAGE("Invalid Endpoint", g_srm_22_endpoint.empty() == false);
    putFile("urlcopycore_004.dat",g_srm_22_endpoint,"2.2",SAMPLE_TEXT);
}


/*
 * Test SetUp
 */
void UrlCopyCoreTestCase::test_SetUp_11(){
    __TIMER;
    statSetUp("urlcopycore_unit_test_11","urlcopycore_001.dat",g_srm_11_endpoint,"1.1","urlcopycore_002.dat",g_srm_11_endpoint,"1.1");
}

/*
 * Test SetUp
 */
void UrlCopyCoreTestCase::test_SetUp_22(){
    __TIMER;
    statSetUp("urlcopycore_unit_test_22","urlcopycore_001.dat",g_srm_22_endpoint,"2.2","urlcopycore_002.dat",g_srm_22_endpoint,"2.2");
}

/*
 * Test SetUp
 */
void UrlCopyCoreTestCase::test_SetUp_11_22(){
    __TIMER;
    statSetUp("urlcopycore_unit_test_11_22","urlcopycore_003.dat",g_srm_11_endpoint,"1.1","urlcopycore_003.dat",g_srm_22_endpoint,"2.2");
}

/*
 * Test SetUp
 */
void UrlCopyCoreTestCase::test_SetUp_22_11(){
    __TIMER;
    statSetUp("urlcopycore_unit_test_22_11","urlcopycore_004.dat",g_srm_22_endpoint,"2.2","urlcopycore_004.dat",g_srm_11_endpoint,"1.1");
}


/*
 * Test Prepare Method
 */
void UrlCopyCoreTestCase::testPrepare_11(){
    __TIMER;
    prepare("urlcopycore_unit_test_11");
}

/*
 * Test Prepare Method
 */
void UrlCopyCoreTestCase::testPrepare_22(){
    __TIMER;
    prepare("urlcopycore_unit_test_22");
}

/*
 * Test Prepare Method
 */
void UrlCopyCoreTestCase::testPrepare_11_22(){
    __TIMER;
    prepare("urlcopycore_unit_test_11_22");
}

/*
 * Test Prepare Method
 */
void UrlCopyCoreTestCase::testPrepare_22_11(){
    __TIMER;
    prepare("urlcopycore_unit_test_22_11");
}

/*
 * Test Transfer Method
 */
void UrlCopyCoreTestCase::testTransfer_11(){
    __TIMER;
    transfer("urlcopycore_unit_test_11");
}

/*
 * Test Transfer Method
 */
void UrlCopyCoreTestCase::testTransfer_22(){
    __TIMER;
    transfer("urlcopycore_unit_test_22");
}

/*
 * Test Transfer Method
 */
void UrlCopyCoreTestCase::testTransfer_11_22(){
    __TIMER;
    transfer("urlcopycore_unit_test_11_22");
}

/*
 * Test Transfer Method
 */
void UrlCopyCoreTestCase::testTransfer_22_11(){
    __TIMER;
    transfer("urlcopycore_unit_test_22_11");
}

/*
 * Test Complete Method
 */
void UrlCopyCoreTestCase::testComplete_11(){
    __TIMER;
    complete("urlcopycore_unit_test_11");
}

/*
 * Test Complete Method
 */
void UrlCopyCoreTestCase::testComplete_22(){
    __TIMER;
    complete("urlcopycore_unit_test_22");
}

/*
 * Test Complete Method
 */
void UrlCopyCoreTestCase::testComplete_11_22(){
    __TIMER;
    complete("urlcopycore_unit_test_11_22");
}

/*
 * Test Complete Method
 */
void UrlCopyCoreTestCase::testComplete_22_11(){
    __TIMER;
    complete("urlcopycore_unit_test_22_11");
}

/**
 * test case of source SRM returning 0 length in prepareToGet
 */
static const char * zero_lenPtg_file       = "urlcopycore-this-one-returns-size-zero-on-get.dat";
static const char * zero_lenPtg_file_dest  = "urlcopycore-copy-of-this-one-returns-size-zero-on-get.dat";

void UrlCopyCoreTestCase::test_0LenPtg_11(){
    __TIMER;
    static const char * test_name = "urlcopycore_unit_test_0LenPtg_11";
    CPPUNIT_ASSERT_MESSAGE("Invalid Endpoint", g_srm_11_endpoint.empty() == false);
    putFile(zero_lenPtg_file,g_srm_11_endpoint,"1.1",SAMPLE_TEXT);
    statSetUp(test_name,zero_lenPtg_file,g_srm_11_endpoint,"1.1",zero_lenPtg_file_dest,g_srm_11_endpoint,"1.1");
    prepare(test_name);
    transfer(test_name);
    complete(test_name);
}
void UrlCopyCoreTestCase::test_0LenPtg_22(){
    __TIMER;
    static const char * test_name = "urlcopycore_unit_test_0LenPtg_22";
    CPPUNIT_ASSERT_MESSAGE("Invalid Endpoint", g_srm_22_endpoint.empty() == false);
    putFile(zero_lenPtg_file,g_srm_22_endpoint,"2.2",SAMPLE_TEXT);
    statSetUp(test_name,zero_lenPtg_file,g_srm_22_endpoint,"2.2",zero_lenPtg_file_dest,g_srm_22_endpoint,"2.2");
    prepare(test_name);
    transfer(test_name);
    complete(test_name);
}

/**
 * test case of source file actually having 0 length (file lost).
 * Verify that transfer is not started
 */

static const char * zero_len_file       = "urlcopycore-zero-len.dat";
static const char * zero_len_file_dest  = "urlcopycore-copy-of-zero-len.dat";

void UrlCopyCoreTestCase::prepare0Len(const std::string& test_name){

    time_t current;
    time(&current);

    // Create Tx Core Object
    UrlCopyCore tx(test_name);
    // Prepare For transfer
    tx.prepare();

    // Check Results
    // --------------

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());

    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    // Check Status
    check_status("Invalid Status (prepare0Len)", stat->mm_urlcopy.mm_status, UrlCopyStatus::PREPARING);
    check_category("Invalid UrlCopy Error (prepare0Len)",stat->mm_urlcopy.mm_error.mm_category,UrlCopyError::ERROR_GENERAL_FAILURE);

    // Source File
    CPPUNIT_ASSERT_MESSAGE("Invalid Source Token"      , strlen(stat->mm_urlcopy.mm_source.mm_token) != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Source Turl"       , strlen(stat->mm_urlcopy.mm_source.mm_tx_name) != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Source Prep. Start", stat->mm_urlcopy.mm_source.mm_prep_start_time >= current);
    CPPUNIT_ASSERT_MESSAGE("Invalid Source Prep. Done" , stat->mm_urlcopy.mm_source.mm_prep_done_time >= current);
    check_category("Invalid Source Error (prepare0Len)",stat->mm_urlcopy.mm_source.mm_error.mm_category,UrlCopyError::ERROR_GENERAL_FAILURE);
    check_category("Invalid Destination Error (prepare0Len)",stat->mm_urlcopy.mm_destination.mm_error.mm_category,UrlCopyError::SUCCESS);

    tx.complete();


}

void UrlCopyCoreTestCase::test_0Len_11(){
    __TIMER;
    static const char * test_name = "urlcopycore_unit_test_0Len_11";
    CPPUNIT_ASSERT_MESSAGE("Invalid Endpoint", g_srm_11_endpoint.empty() == false);
    putFile(zero_len_file,g_srm_11_endpoint,"1.1",EMPTY_FILE);
    statSetUp(test_name,zero_len_file,g_srm_11_endpoint,"1.1",zero_len_file_dest,g_srm_11_endpoint,"1.1");
    prepare0Len(test_name);
}
void UrlCopyCoreTestCase::test_0Len_22(){
    __TIMER;
    static const char * test_name = "urlcopycore_unit_test_0Len_22";
    CPPUNIT_ASSERT_MESSAGE("Invalid Endpoint", g_srm_22_endpoint.empty() == false);
    putFile(zero_len_file,g_srm_22_endpoint,"2.2",EMPTY_FILE);
    statSetUp(test_name,zero_len_file,g_srm_22_endpoint,"2.2",zero_len_file_dest,g_srm_22_endpoint,"2.2");
    prepare0Len(test_name);
}

/**
 * Verify that the transfer succeeds if a directory does not exist on stat
 * but exists on mkdir (happening e.g. if another transfer created the
 * directory in the meanwhile)
 */
void UrlCopyCoreTestCase::test_MkDir_Duplication(){
    __TIMER;
    static const char * test_name = "urlcopycore_unit_test_mkdir_duplication_22";
    static const char * source_file = "mkdir_duplication.dat";
    static const char * dest_file = "test-path/this-one-returns-SRM_DUPLICATION_ERROR-on-mkdir/copy-of-mkdir_duplication.dat";
    CPPUNIT_ASSERT_MESSAGE("Invalid Endpoint", g_srm_22_endpoint.empty() == false);
    putFile(source_file,g_srm_22_endpoint,"2.2",SAMPLE_TEXT);
    statSetUp(test_name,source_file,g_srm_22_endpoint,"2.2",dest_file,g_srm_22_endpoint,"2.2");
    prepare(test_name);
    transfer(test_name);
    complete(test_name);
}

/**
 * first set file status request failing
 */
void UrlCopyCoreTestCase::test_SetFileStatusFailure_11(){
    __TIMER;
    static const char * test_name = "urlcopycore_unit_test_setfilestatus_failure_11";
    static const char * source_file = "test-set-status-failure.dat";
    static const char * dest_file = "this-one-fails-on-first-set-status.dat";
    CPPUNIT_ASSERT_MESSAGE("Invalid Endpoint", g_srm_11_endpoint.empty() == false);
    putFile(source_file,g_srm_11_endpoint,"1.1",SAMPLE_TEXT);
    statSetUp(test_name,source_file,g_srm_11_endpoint,"1.1",dest_file,g_srm_11_endpoint,"1.1");
    prepare(test_name);
    transfer(test_name);
    complete(test_name);
}
