/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: UrlCopyCore Test Case
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: UrlCopyCoreTestCase_Common.cpp,v 1.8 2009/07/30 19:32:55 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#include "UrlCopyCoreTestCase_Common.h"
#include "UrlCopyCore.h"
#include "test-utility.h"
#include "Mmap.h"
#include "stat-utility.h"
#include "TransferConfig.h"
#include <glite/data/agents/AgentExceptions.h>

#include <glite/data/srm/util/Context.h>
#include <glite/data/srm/util/PrepareToPut.h>
#include <glite/data/srm/util/SrmMkDir.h>

#include <memory>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <boost/shared_ptr.hpp>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <sstream>

using namespace glite::data::agents;
using namespace glite::data::srm::util;
using namespace glite::data::transfer::urlcopy;

using boost::shared_ptr;

namespace{

boost::shared_ptr<Mmap> create_stat(const std::string& name,UrlCopyStat ** stat){
    try{
        // Delete Previous file (if exists)
        std::string filename = TransferConfig::instance().repository() + name + ".mem";
        unlink(filename.c_str());
    } catch(...){
        // Ignore error
    }

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = create_stat_UrlCopy(name);

    // Check Return Value
    CPPUNIT_ASSERT_MESSAGE("Invalid Result" , mmap_ptr.get()     != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Buffer" , mmap_ptr->buffer() != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Size"   , mmap_ptr->size()   != 0);

    // Get the stat object
    (*stat) = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid ID", (*stat)->mm_common.mm_id == URLCOPY_STAT_ID);
    return mmap_ptr;
}

void stat_setup(
    UrlCopyStat *       stat,
    const std::string&  name,
    const std::string&  src_name,
    const std::string&  src_endpoint,
    const std::string&  src_version,
    const std::string&  dst_name,
    const std::string&  dst_endpoint,
    const std::string&  dst_version,
    const bool          isSrcADir = false){

    // Write something into the buffer
    time(&(stat->mm_common.mm_start_time));
    time(&(stat->mm_common.mm_last_refresh_time));
#ifdef _USE_GLOBAL_TIMEOUT
    stat->mm_common.mm_global_timeout  = 1000;
#endif
    stat->mm_common.mm_overwrite       = 1;
    // User Description
    copy_message(stat->mm_common.mm_user_tag,"UrlCopyCore UnitTest");
    // Source File
    copy_url(stat->mm_urlcopy.mm_source.mm_srm_endpoint,src_endpoint.c_str());


    // flag to enable/disable checks for existance of target directory
    stat->mm_common.mm_enable_dir_check = 1;
    // flag to enable/disable ping
    stat->mm_common.mm_enable_ping_source = 1;
    stat->mm_common.mm_enable_ping_dest = 1;


    copy_version(stat->mm_urlcopy.mm_source.mm_srm_version,src_version.c_str());
    std::string src_surl = src_endpoint + "?SFN=/test/";

    if (!isSrcADir)
    	src_surl += src_name + "." + src_version;

    src_surl = (std::string)"srm" + src_surl.substr(5);
    copy_url(stat->mm_urlcopy.mm_source.mm_name,src_surl.c_str());
    stat->mm_urlcopy.mm_source.mm_http_timeout  = 40;
    stat->mm_urlcopy.mm_source.mm_prep_timeout  = 300;
    stat->mm_urlcopy.mm_source.mm_final_timeout = 300;
    // Destination File
    copy_url(stat->mm_urlcopy.mm_destination.mm_srm_endpoint,dst_endpoint.c_str());
    copy_version(stat->mm_urlcopy.mm_destination.mm_srm_version,dst_version.c_str());
    std::string dst_surl = dst_endpoint + "?SFN=/test/" + dst_name + "." + dst_version;
    dst_surl = (std::string)"srm" + dst_surl.substr(5);
    copy_url(stat->mm_urlcopy.mm_destination.mm_name,dst_surl.c_str());
    stat->mm_urlcopy.mm_destination.mm_http_timeout  = 40;
    stat->mm_urlcopy.mm_destination.mm_prep_timeout  = 300;
    stat->mm_urlcopy.mm_destination.mm_final_timeout = 300;
    // Transfer
    stat->mm_urlcopy.mm_tx.mm_streams        = 1;
    stat->mm_urlcopy.mm_tx.mm_tcp_bs         = 64 * 1024;
    stat->mm_urlcopy.mm_tx.mm_block_size     = 2 * 1024 * 1024;
    stat->mm_urlcopy.mm_tx.mm_timeout        = 1800;
    stat->mm_urlcopy.mm_tx.mm_marker_timeout = 300;

    // added in 2.1.0
    stat->mm_urlcopy.mm_tx.mm_first_marker_timeout = 0;

}

} // namespace

/**
 * Test SetUp
 */
void UrlCopyCoreTestCase_Common::statSetUp(
    const std::string&  name,
    const std::string&  src_name,
    const std::string&  src_endpoint,
    const std::string&  src_version,
    const std::string&  dst_name,
    const std::string&  dst_endpoint,
    const std::string&  dst_version,
    const bool isSrcADir){

    UrlCopyStat *stat = 0;
    boost::shared_ptr<Mmap> mmap_ptr = create_stat(name,&stat);
    stat_setup(stat,name,src_name,src_endpoint,src_version,dst_name,dst_endpoint,dst_version,isSrcADir);

}

void UrlCopyCoreTestCase_Common::statSetUpWithTokens(
    const std::string&  name,
    const std::string&  src_name,
    const std::string&  src_endpoint,
    const std::string&  src_version,
    const std::string&  dst_name,
    const std::string&  dst_endpoint,
    const std::string&  dst_version,
    const std::string&  src_token_description,
    const std::string&  dst_token_description){

    UrlCopyStat *stat = 0;
    boost::shared_ptr<Mmap> mmap_ptr = create_stat(name,&stat);
    stat_setup(stat,name,src_name,src_endpoint,src_version,dst_name,dst_endpoint,dst_version);

    // Tokens
    copy_message(stat->mm_common.mm_source_token_descr,src_token_description.c_str());
    copy_message(stat->mm_common.mm_dest_token_descr,dst_token_description.c_str());
}

void UrlCopyCoreTestCase_Common::statSetUpWlcg(
    const std::string&  name,
    const std::string&  src_name,
    const std::string&  src_endpoint,
    const std::string&  src_version,
    const std::string&  dst_name,
    const std::string&  dst_endpoint,
    const std::string&  dst_version,
    int                 copy_pin_lifetime,
    int                 lan_connection,
    int                 fail_nearline_files){

    UrlCopyStat *stat = 0;
    boost::shared_ptr<Mmap> mmap_ptr = create_stat(name,&stat);
    stat_setup(stat,name,src_name,src_endpoint,src_version,dst_name,dst_endpoint,dst_version);

    // added with stat version 3.3.0
    stat->mm_common.mm_copy_pin_lifetime    = copy_pin_lifetime;
    stat->mm_common.mm_lan_connection       = lan_connection;
    stat->mm_common.mm_fail_nearline_files  = fail_nearline_files;

}

void UrlCopyCoreTestCase_Common::statSetUpChecksum(
    const std::string&  name,
    const std::string&  src_name,
    const std::string&  src_endpoint,
    const std::string&  src_version,
    const std::string&  dst_name,
    const std::string&  dst_endpoint,
    const std::string&  dst_version,
    const std::string&  user_algorithm,
    const std::string&  user_value,
    const bool compare)
{
    UrlCopyStat *stat = 0;
    boost::shared_ptr<Mmap> mmap_ptr = create_stat(name,&stat);
    stat_setup(stat,name,src_name,src_endpoint,src_version,dst_name,dst_endpoint,dst_version);

    // checksum data
    stat->mm_urlcopy.mm_compare_checksum = static_cast<int>(compare);
    copy_checksum_algorithm(stat->mm_urlcopy.mm_checksum_user_algorithm,user_algorithm.c_str());
    copy_checksum_value(stat->mm_urlcopy.mm_checksum_user_value,user_value.c_str());

}

/**
 * Test Prepare Method
 */
void UrlCopyCoreTestCase_Common::prepare(const std::string& name){
    time_t current;
    time(&current);

    {
    // Create Tx Core Object
    UrlCopyCore tx(name);

    // Prepare For transfer
    tx.prepare();
    }
    // Check Results
    // --------------

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    // Check Status
    check_status("Invalid Status (prepare)", stat->mm_urlcopy.mm_status, UrlCopyStatus::READY);
    check_category("Invalid Error (prepare)" , stat->mm_urlcopy.mm_error.mm_category, UrlCopyError::SUCCESS);

    // Source File
    CPPUNIT_ASSERT_MESSAGE("Invalid Source Token"      , strlen(stat->mm_urlcopy.mm_source.mm_token) != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Source Turl"       , strlen(stat->mm_urlcopy.mm_source.mm_tx_name) != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Source Prep. Start", stat->mm_urlcopy.mm_source.mm_prep_start_time >= current);
    CPPUNIT_ASSERT_MESSAGE("Invalid Source Prep. Done" , stat->mm_urlcopy.mm_source.mm_prep_done_time >= current);
    check_category("Invalid Source Error (prepare)"    , stat->mm_urlcopy.mm_source.mm_error.mm_category, UrlCopyError::SUCCESS);

    // Destination File
    CPPUNIT_ASSERT_MESSAGE("Invalid Dest. Token"      , strlen(stat->mm_urlcopy.mm_destination.mm_token) != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Dest. Turl"       , strlen(stat->mm_urlcopy.mm_destination.mm_tx_name) != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Dest. Prep. Start", stat->mm_urlcopy.mm_destination.mm_prep_start_time >= current);
    CPPUNIT_ASSERT_MESSAGE("Invalid Dest. Prep. Done" , stat->mm_urlcopy.mm_destination.mm_prep_done_time >= current);
    check_category("Invalid Dest. Error (prepare)"    , stat->mm_urlcopy.mm_destination.mm_error.mm_category, UrlCopyError::SUCCESS);
}

/**
 * Test Transfer Method
 */
void UrlCopyCoreTestCase_Common::transfer(const std::string& name){

    time_t current;
    time(&current);

    {
    // Create Tx Core Object
    UrlCopyCore tx(name);

    // Prepare For transfer
    tx.transfer();
    }

    // Check Results
    // --------------

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat id", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    // Check Status
    check_status("Invalid Status (transfer)", stat->mm_urlcopy.mm_status, UrlCopyStatus::DONE);
    check_category("Invalid Error (transfer)" , stat->mm_urlcopy.mm_error.mm_category, UrlCopyError::SUCCESS);

    // Transfer File
    CPPUNIT_ASSERT_MESSAGE("Invalid Tx Size"    , stat->mm_urlcopy.mm_tx.mm_size > 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Tx Start"   , stat->mm_urlcopy.mm_tx.mm_start_time >= current);
    CPPUNIT_ASSERT_MESSAGE("Invalid Tx Done"    , stat->mm_urlcopy.mm_tx.mm_done_time >= current);
    check_category("Invalid Tx Error (transfer)", stat->mm_urlcopy.mm_tx.mm_error.mm_category, UrlCopyError::SUCCESS);

}

/**
 * Test Complete Method
 */
void UrlCopyCoreTestCase_Common::complete(const std::string& name){

    time_t current;
    time(&current);

    {
    // Create Tx Core Object
    UrlCopyCore tx(name);

    // Complete The request
    tx.complete();
    }
    // Check Results
    // --------------

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat id", stat->mm_common.mm_id == URLCOPY_STAT_ID);

    // Check Status
    check_status("Invalid Status (complete)", stat->mm_urlcopy.mm_status, UrlCopyStatus::COMPLETED);
    check_category("Invalid Category (complete)",stat->mm_urlcopy.mm_error.mm_category,UrlCopyError::SUCCESS);

    // Source File
    CPPUNIT_ASSERT_MESSAGE("Invalid Source Token"      , strlen(stat->mm_urlcopy.mm_source.mm_token) != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Source Turl"       , strlen(stat->mm_urlcopy.mm_source.mm_tx_name) != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Source Final Start", stat->mm_urlcopy.mm_source.mm_final_start_time >= current);
    CPPUNIT_ASSERT_MESSAGE("Invalid Source Final Done" , stat->mm_urlcopy.mm_source.mm_final_done_time >= current);
    check_category("Invalid Source Error (complete)"   , stat->mm_urlcopy.mm_source.mm_error.mm_category, UrlCopyError::SUCCESS);

    // Destination File
    CPPUNIT_ASSERT_MESSAGE("Invalid Dest. Token"      , strlen(stat->mm_urlcopy.mm_destination.mm_token) != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Dest. Turl"       , strlen(stat->mm_urlcopy.mm_destination.mm_tx_name) != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Dest. Final Start", stat->mm_urlcopy.mm_destination.mm_final_start_time >= current);
    CPPUNIT_ASSERT_MESSAGE("Invalid Dest. Final Done" , stat->mm_urlcopy.mm_destination.mm_final_done_time >= current);
    check_category("Invalid Dest. Error (complete)"   , stat->mm_urlcopy.mm_destination.mm_error.mm_category, UrlCopyError::SUCCESS);
}

