/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: UrlCopyCoreTokens Test Case
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *           Zsolt Molnar <zsolt.molnar@cern.ch>
 *  Version info: $Id: UrlCopyCoreTokensTestCase.cpp,v 1.4 2009/03/19 13:37:09 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#include "UrlCopyCoreTokensTestCase.h"
#include "UrlCopyCore.h"
#include "Mmap.h"
#include "stat-utility.h"
#include "TransferConfig.h"
#include <glite/data/agents/AgentExceptions.h>

#include <glite/data/srm/util/Context.h>
#include <glite/data/srm/util/PrepareToPut.h>
#include <glite/data/srm/util/SrmMkDir.h>

#include <memory>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <boost/shared_ptr.hpp>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <sstream>

using namespace glite::data::agents;
using namespace glite::data::srm::util;
using namespace glite::data::transfer::urlcopy;

using boost::shared_ptr;

/*
 * The SRM endpoints
 */
extern std::string g_srm_11_endpoint;
extern std::string g_srm_22_endpoint;

namespace {
const char SAMPLE_TEXT[] = "Hello World! This is a test\nWrite some data for Transfer Url Copy tests\n";
const char EMPTY_FILE[] = "";
}

void UrlCopyCoreTokensTestCase::test_space_token_setup(){
    __TIMER;
    static const char * source_file = "urlcopycore_multiple_tokens_source.dat";
    putFile(source_file,g_srm_22_endpoint,"2.2",SAMPLE_TEXT);
}

/**
 * test for bug #33652 (FTS unable to handle multiple space tokens) : destination
 * check that in case only one space is returned from the description and the
 * space does not have enough unused space the transfer fails
 */
void UrlCopyCoreTokensTestCase::test_space_token_01(){
    __TIMER;
    static const char * test_name = "urlcopycore_unit_test_space_token_01";
    static const char * source_file = "urlcopycore_multiple_tokens_source.dat";
    static const char * dest_file = "test_space_token_01_dest.dat";
    static const unsigned long mb = 1024 * 1024;
    static const std::string TOKEN_A = make_token(10*mb,10*mb,30);
    statSetUpWithTokens(
        test_name,
        source_file,g_srm_22_endpoint,"2.2",
        dest_file,g_srm_22_endpoint,"2.2",
        "Description_for:TEST-SOURCE-TOKEN",
        make_description(TOKEN_A));

    {
        // Create Tx Core Object
        UrlCopyCore tx(test_name);
        // Prepare For transfer
        tx.prepare();
    }
    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);
    // Check Status
    check_status("Invalid Status", stat->mm_urlcopy.mm_status, UrlCopyStatus::PREPARING);
    check_scope("Invalid Scope", stat->mm_urlcopy.mm_error.mm_scope, UrlCopyError::SCOPE_DESTINATION);
    check_phase("Invalid Phase", stat->mm_urlcopy.mm_error.mm_phase, UrlCopyError::PHASE_PREPARATION);
    check_category("Invalid Error", stat->mm_urlcopy.mm_error.mm_category, UrlCopyError::ERROR_NO_SPACE_LEFT);
    CHECK_STRING(stat->mm_common.mm_dest_space_token,"");
}

/**
 * test for bug #33652 (FTS unable to handle multiple space tokens)
 * check that if multiple spaces are returned from the token description, is
 * selected the space that:
 * - has the minimum *unused space*
 * - has an unused space equal to at least file size + 1MB
 */
void UrlCopyCoreTokensTestCase::test_space_token_02(){
    __TIMER;
    static const char * test_name = "urlcopycore_unit_test_space_token_02";
    static const char * source_file = "urlcopycore_multiple_tokens_source.dat";
    static const char * dest_file = "test_space_token_02_dest.dat";
    static const unsigned long mb = 1024 * 1024;
    static const std::string TOKEN_A = make_token(10*mb,10*mb,0);
    static const std::string TOKEN_B = make_token(10*mb,10*mb,2*mb);
    statSetUpWithTokens(
        test_name,
        source_file,g_srm_22_endpoint,"2.2",
        dest_file,g_srm_22_endpoint,"2.2",
        "Description_for:TEST-SOURCE-TOKEN",
        make_description(TOKEN_A,TOKEN_B));
    prepare(test_name);
    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat", stat != 0);
    CHECK_STRING(stat->mm_common.mm_dest_space_token,TOKEN_B.c_str());
}

/**
 * test for bug #33652 (FTS unable to handle multiple space tokens)
 * check that if multiple spaces are returned from the token description and
 * none of them satisfies the criteria of test case test_space_token_02, then
 * the transfer fails
 */
void UrlCopyCoreTokensTestCase::test_space_token_03(){
    __TIMER;
    static const char * test_name = "urlcopycore_unit_test_space_token_03";
    static const char * source_file = "urlcopycore_multiple_tokens_source.dat";
    static const char * dest_file = "test_space_token_03_dest.dat";
    static const unsigned long mb = 1024 * 1024;
    static const std::string TOKEN_A = make_token(10*mb,10*mb,0);
    static const std::string TOKEN_B = make_token(10*mb,10*mb,10);
    static const std::string TOKEN_C = make_token(10*mb,10*mb,1*mb);
    statSetUpWithTokens(
        test_name,
        source_file,g_srm_22_endpoint,"2.2",
        dest_file,g_srm_22_endpoint,"2.2",
        "Description_for:TEST-SOURCE-TOKEN",
        make_description(TOKEN_A,TOKEN_B,TOKEN_C));

    {
        // Create Tx Core Object
        UrlCopyCore tx(test_name);
        // Prepare For transfer
        tx.prepare();
    }
    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);
    // Check Status
    check_status("Invalid Status", stat->mm_urlcopy.mm_status, UrlCopyStatus::PREPARING);
    check_scope("Invalid Scope", stat->mm_urlcopy.mm_error.mm_scope, UrlCopyError::SCOPE_DESTINATION);
    check_phase("Invalid Phase", stat->mm_urlcopy.mm_error.mm_phase, UrlCopyError::PHASE_PREPARATION);
    check_category("Invalid Error", stat->mm_urlcopy.mm_error.mm_category, UrlCopyError::ERROR_NO_SPACE_LEFT);
    CHECK_STRING(stat->mm_common.mm_dest_space_token,"");
}

/**
 * check that for source tokens the first is selected, regardless of file size
 * and available space
 */
void UrlCopyCoreTokensTestCase::test_space_token_04(){
    __TIMER;
    static const char * test_name = "urlcopycore_unit_test_space_token_04";
    static const char * source_file = "urlcopycore_multiple_tokens_source.dat";
    static const char * dest_file = "test_space_token_04_dest.dat";
    static const unsigned long mb = 1024 * 1024;
    static const std::string TOKEN_A = make_token(10*mb,10*mb,0);
    static const std::string TOKEN_B = make_token(100*mb,10*mb,1);
    static const std::string TOKEN_C = make_token(1000*mb,10*mb,2);
    statSetUpWithTokens(
        test_name,
        source_file,g_srm_22_endpoint,"2.2",
        dest_file,g_srm_22_endpoint,"2.2",
        make_description(TOKEN_A,TOKEN_B,TOKEN_C),
        "");

    prepare(test_name);
    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat", stat != 0);
    CHECK_STRING(stat->mm_common.mm_source_space_token,TOKEN_A.c_str());
}

/**
 * check that if an invalid source token is specified, the transfer fails
 */
void UrlCopyCoreTokensTestCase::test_space_token_05(){
    __TIMER;
    static const char * test_name = "urlcopycore_unit_test_space_token_05";
    static const char * source_file = "urlcopycore_multiple_tokens_source.dat";
    static const char * dest_file = "test_space_token_05_dest.dat";
    statSetUpWithTokens(
        test_name,
        source_file,g_srm_22_endpoint,"2.2",
        dest_file,g_srm_22_endpoint,"2.2",
        "INVALID-SOURCE-TOKEN",
        "");
    {
        // Create Tx Core Object
        UrlCopyCore tx(test_name);
        // Prepare For transfer
        tx.prepare();
    }
    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);
    // Check Status
    check_status("Invalid Status", stat->mm_urlcopy.mm_status, UrlCopyStatus::PREPARING);
    check_scope("Invalid Scope", stat->mm_urlcopy.mm_error.mm_scope, UrlCopyError::SCOPE_SOURCE);
    check_phase("Invalid Phase", stat->mm_urlcopy.mm_error.mm_phase, UrlCopyError::PHASE_PREPARATION);
    check_category("Invalid Error", stat->mm_urlcopy.mm_error.mm_category, UrlCopyError::ERROR_USER);
    CHECK_STRING(stat->mm_common.mm_source_space_token,"");
}

/**
 * check that if an invalid destination token is specified, the transfer fails
 */
void UrlCopyCoreTokensTestCase::test_space_token_06(){
    __TIMER;
    static const char * test_name = "urlcopycore_unit_test_space_token_06";
    static const char * source_file = "urlcopycore_multiple_tokens_source.dat";
    static const char * dest_file = "test_space_token_06_dest.dat";
    statSetUpWithTokens(
        test_name,
        source_file,g_srm_22_endpoint,"2.2",
        dest_file,g_srm_22_endpoint,"2.2",
        "",
        "INVALID-DEST-TOKEN");
    {
        // Create Tx Core Object
        UrlCopyCore tx(test_name);
        // Prepare For transfer
        tx.prepare();
    }
    // Check Results

    // Create a stat File and return the Mmap object for it
    boost::shared_ptr<Mmap> mmap_ptr = open_stat(test_name,true);
    // Get the stat object
    UrlCopyStat * stat = reinterpret_cast<UrlCopyStat *>(mmap_ptr->buffer());
    CPPUNIT_ASSERT_MESSAGE("Invalid stat"   , stat != 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid stat ID", stat->mm_common.mm_id == URLCOPY_STAT_ID);
    // Check Status
    check_status("Invalid Status", stat->mm_urlcopy.mm_status, UrlCopyStatus::PREPARING);
    check_scope("Invalid Scope", stat->mm_urlcopy.mm_error.mm_scope, UrlCopyError::SCOPE_DESTINATION);
    check_phase("Invalid Phase", stat->mm_urlcopy.mm_error.mm_phase, UrlCopyError::PHASE_PREPARATION);
    check_category("Invalid Error", stat->mm_urlcopy.mm_error.mm_category, UrlCopyError::ERROR_USER);
    CHECK_STRING(stat->mm_common.mm_dest_space_token,"");
}
