/*
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Mmap Test Case
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *           Zsolt Molnar <zsolt.molnar@cern.ch>
 *  Version info: $Id: UtilitiesTestCase.cpp,v 1.6 2009/03/19 13:37:09 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#include "UtilitiesTestCase.h"
#include "transfer-utility.h"
#include <stdio.h>
#include <fstream>
#include <iostream>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/algorithm/string/find.hpp>

using namespace glite::data::transfer::urlcopy;

extern std::string g_tmp_folder;

const std::string UtilitiesTestCase::ExecName = "glite-url-copy-exec";
const std::string UtilitiesTestCase::TxId = "CERN-TRIUMF__2007-02-27-1004_m8PnmI";

void UtilitiesTestCase::write_cmd_line_param(std::ofstream& stream,const std::string& line){
    stream << line.c_str();
    stream.put('\0');
}

std::string temp_file_path(const std::string& file_name){
    return g_tmp_folder + "/" + file_name;
}

/**
 * test case setup
 */
void UtilitiesTestCase::setUp(){
    test_file = temp_file_path("existing-cmd-line");
    test_file_nonexisting = temp_file_path("non-existing-cmd-line");
    std::ofstream test_file_stream;
    try{
        test_file_stream.open(test_file.c_str());
        write_cmd_line_param(test_file_stream,ExecName);
        write_cmd_line_param(test_file_stream,TxId);
    }
    catch (...){
    }
    test_file_stream.close();
}

/**
 * test case teardown
 */
void UtilitiesTestCase::tearDown(){
    remove(test_file.c_str());
}

/**
 * matching exec name and tx id
 */
void UtilitiesTestCase::test_check_cmdline_file_match(){
    __TIMER;
    std::string reason;
    bool res = check_cmdline_file(test_file,ExecName,TxId,reason);
    CPPUNIT_ASSERT_MESSAGE("test_check_cmdline_file_match failed: " + reason,res);
}

/**
 * missing file
 */
void UtilitiesTestCase::test_check_cmdline_file_missing_file(){
    __TIMER;
    std::string reason;
    bool res = check_cmdline_file(test_file_nonexisting,ExecName,TxId,reason);
    CPPUNIT_ASSERT_MESSAGE("test_check_cmdline_file_missing_file failed",!res);
    CPPUNIT_ASSERT_MESSAGE("invalid reason: " + reason,
        boost::starts_with(reason,"could not open file"));
}

/**
 * different exec name
 */
void UtilitiesTestCase::test_check_cmdline_file_different_exec_name(){
    __TIMER;
    std::string reason;
    bool res = check_cmdline_file(test_file,"invalid-exec",TxId,reason);
    CPPUNIT_ASSERT_MESSAGE("test_check_cmdline_file_different_exec_name failed",!res);
    CPPUNIT_ASSERT_MESSAGE("invalid reason: " + reason,
        boost::starts_with(reason,"different exec names"));
}

/**
 * different tx id
 */
void UtilitiesTestCase::test_check_cmdline_file_different_tx_id(){
    __TIMER;
    std::string reason;
    bool res = check_cmdline_file(test_file,ExecName,"another-tx-id",reason);
    CPPUNIT_ASSERT_MESSAGE("test_check_cmdline_file_different_tx_id failed",!res);
    CPPUNIT_ASSERT_MESSAGE("invalid reason: " + reason,
        boost::starts_with(reason,"different tx ids"));
}
