/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Utility for the Test Cases
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: test-utility.cpp,v 1.13 2009/10/26 16:32:20 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include "test-utility.h"
#include "Mmap.h"
#include "stat-utility.h"
#include "TransferConfig.h"
#include "Checksum.h"
#include <glite/data/agents/AgentExceptions.h>

#include <glite/data/srm/util/Context.h>
#include <glite/data/srm/util/PrepareToPut.h>
#include <glite/data/srm/util/SrmMkDir.h>

#include <memory>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <boost/shared_ptr.hpp>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <stdlib.h>
#include <dirent.h>
#include <libgen.h>
#include <errno.h>


using namespace glite::data::agents;
using namespace glite::data::srm::util;
using namespace glite::data::transfer::urlcopy;
using boost::shared_ptr;
using namespace test_utils;

extern std::string g_srm11Pool;
extern std::string g_srm22Pool;
extern std::string g_srm_11_endpoint;
extern std::string g_srm_22_endpoint;

namespace{
    /**
     * SRM context specialization (for logging)
     */
    class TestContext : public glite::data::srm::util::Context{
    };

/**
 * Put Source file for tests
 * Contacts the stub SRM endpoint to make the appropriate srmMkDir requests;
 * then performs an srmPrepareToPut to get a TURL (which is a local address) and
 * writes a file at the given physical location.
 */
void srmPutFile(
    const std::string& file_name,
    const std::string& endpoint,
    const std::string& version,
    const char * const text){

    // Source path
    std::string dir_surl = endpoint + "?SFN=/test/";
    dir_surl = (std::string)"srm" + dir_surl.substr(5);
    // Source file
    std::string surl = dir_surl + file_name + "." + version;

    // Create Context
    Context::Ptr ctx(new TestContext());

    // Create an Srm Context
    ctx->endpoint = endpoint;
    ctx->version  = version;

    try{
        // Create the Directory
        // -------------------------------------

        // Create MkDir Request
        boost::shared_ptr<SrmMkDir> mkdir_request(SrmMkDir::create(ctx));
        // Set Request Parameters
        mkdir_request->directoryPath = dir_surl;
        // Send Request
        SrmStatus mkdir_result = mkdir_request->execute();
        // Check Return Value
        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code",
            (mkdir_result.code == SrmStatus::SRM_SUCCESS) ||
            (mkdir_result.code == SrmStatus::SRM_DUPLICATION_ERROR));

        // Open the file and put something there
        // -------------------------------------

        // Create Put Request
        boost::shared_ptr<PrepareToPut> request(PrepareToPut::create(ctx));
        // Set Request Parameters
        request->files.resize(1);
        request->files[0].surl = surl;
        request->transferParameters.protocol = "file";
        // Send Request
        request->execute();
        // Check Return Value
        bool is_request_failed = true;
        if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
           (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
           (request->status.code == SrmStatus::SRM_SUCCESS)){
               is_request_failed = false;
        }
        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_failed == false);
        while((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
              (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
            // Sleep for a while
            // sleep(1);
            // Keep polling for the status
            request->update();
            // Check Return Value
            if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
               (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
               (request->status.code == SrmStatus::SRM_SUCCESS)){
                   is_request_failed = false;
            }
            CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_failed == false);
        }

        if (request->status.code != SrmStatus::SRM_SUCCESS){
            std::stringstream ss;
            ss << "Result Code is " << request->status.code << " (should now be SUCCESS)";
            CPPUNIT_FAIL(ss.str().c_str());
        }

        if(request->status.code == SrmStatus::SRM_SUCCESS){
            // Check that the file properties are set correctly
            CPPUNIT_ASSERT_MESSAGE("Invalid File Status", request->files[0].status.code == SrmStatus::SRM_SPACE_AVAILABLE);
            CPPUNIT_ASSERT_MESSAGE("No turl retrieved", request->files[0].turl.empty() == false);
        }
        // Open the file
        if(0 == strncmp(request->files[0].turl.c_str(),"file://",7)){
            std::string path = request->files[0].turl.substr(7);
            int fd = open(path.c_str(),O_RDWR);
            CPPUNIT_ASSERT_MESSAGE("Cannot Open file", fd != -1);
            write(fd,text,strlen(text));
            close(fd);
        } else {
           CPPUNIT_FAIL("Invalid protocol for TURL: Not [file]");
        }
        // Complete the request
        request->done();
        // Check the Request status
        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [DONE]", request->status.code == SrmStatus::SRM_SUCCESS);
        // Check the File status
        CPPUNIT_ASSERT_MESSAGE("Invalid File Status", request->files[0].status.code == SrmStatus::SRM_SUCCESS);
    } catch(const std::exception& exc){
        CPPUNIT_FAIL(exc.what());
    } catch(...){
        CPPUNIT_FAIL("Got Unknown Error");
    }
}

/*
 * Get the parent Directory name
 */
std::string get_dirname(const std::string& dir_path){
    char * parent_name_str = strdup(dir_path.c_str());
    dirname(parent_name_str);
    std::string parent_name = parent_name_str;
    free(parent_name_str);
    return parent_name;
}

/*
 * Create a Directory
 */
int create_dir_path(const std::string& dir_name, bool recursive){
    int r = -1;
    if(false == recursive){
        // Ok, parent exists or has been created: Create teh directory
        r = mkdir(dir_name.c_str(),0700);
    } else {
        // Get the parent for that dir
        std::string parent_name = get_dirname(dir_name);
        // Check if parent_name_str exist
        r = access(parent_name.c_str(), R_OK);
        if(0 == r){
            // Ok, parent exists: Create the directory
            r = mkdir(dir_name.c_str(),0700);
        } else {
            // Directory doesn't exist: try to create the parent dir
            r = create_dir_path(parent_name,true);
            if(0 == r){
                // Ok, parent exists or has been created: Create the directory
                r= mkdir(dir_name.c_str(),0700);
            }
        }
    }
    if(r != 0){
        return errno;
    }
    return 0;
}

std::string combine_path(const std::string& path1,const std::string& path2){
    static const char separator = '/';
    std::stringstream ret;
    ret << path1;
    if (path1.length() && path1[path1.length()] != separator &&
        path2.length() && path2[0] != separator){
        ret << separator;
    }
    ret << path2;
    return ret.str();
}

std::string get_file_path(const std::string& dir,const std::string& file,const std::string& version,bool create_path){
    std::string path = combine_path(dir,"test"); // note the 'SFN=test' in the srmPutFile function
    if (create_path){
        create_dir_path(path,true);
    }
    path = combine_path(path,file) + "." + version;
    return path;
}

void writeFile(const std::string& dir,const std::string& file,const std::string& version,const char* const text){
    std::string path = get_file_path(dir,file,version,true);
    std::ofstream out(path.c_str(),std::ios_base::trunc);
    out << text;
    out.close();
}

bool statFile(const std::string& dir,const std::string& file,const std::string& version){
    std::string path = get_file_path(dir,file,version,false);
    struct stat buf;
    int res = stat(path.c_str(),&buf);
    return (0 == res);
}

} // namespace

timer::timer(const char * fname) : m_fname(fname){
    gettimeofday(&m_start,0);
}
timer::~timer(){
    if (getenv("PRINT_TEST_TIMES")){
        gettimeofday(&m_end,0);
        int sec = m_end.tv_sec - m_start.tv_sec;
        int msec = (int)((m_end.tv_usec - m_start.tv_usec) / 1000);
        if (msec < 0){
            msec += 1000;
            sec -= 1;
        }
        std::cerr << "# " << m_fname << " : " << sec << "." << std::setw(3) << std::setfill('0') << msec << " sec." << std::endl;
    }
}

//------------------------------------------------------------------------------
//                                  TestsMap
//------------------------------------------------------------------------------

TestsMap& TestsMap::instance(){
    static TestsMap s_instance;
    return s_instance;
}

TestsMap::TestsMap(){
}

void TestsMap::addSuite(const std::string& name){
    m_suites[name] = true;
    m_names.push_back(name);
}

const TestsMap::Names& TestsMap::names(){
    return m_names;
}

void TestsMap::disableAll(){
    for (Suites::iterator it = m_suites.begin();it != m_suites.end();++it){
        const std::string& name = it->first;
        m_suites[name] = false;
    }
}

bool TestsMap::enable(const std::string& name){
    if (m_suites.end() == m_suites.find(name)){
        return false;
    }
    m_suites[name] = true;
    return true;
}


/**
 * Put Source file for tests
 */
void putFile(
    const std::string& file_name,
    const std::string& endpoint,
    const std::string& version,
    const char * const text){
    // if the endpoint is one of the two stub endpoints, write the file on the
    // file system (to speed up tests)
    if (endpoint == g_srm_11_endpoint){
        writeFile(g_srm11Pool,file_name,version,text);
    } else if (endpoint == g_srm_22_endpoint){
        writeFile(g_srm22Pool,file_name,version,text);
    } else {
        // this should never happen
        srmPutFile(file_name,endpoint,version,text);
    }
}

/**
 * Check if file exists
 * @return true if file exists, false otherwise
 */
bool fileExists(
    const std::string& file_name,
    const std::string& endpoint,
    const std::string& version){
    // if the endpoint is one of the two stub endpoints, check on the file
    // system (to speed up tests)
    if (endpoint == g_srm_11_endpoint){
        return statFile(g_srm11Pool,file_name,version);
    } else if (endpoint == g_srm_22_endpoint){
        return statFile(g_srm11Pool,file_name,version);
    } else {
        // this should never happen
        CPPUNIT_FAIL("Unknown endpoint; this condition should never be met.");
    }

    return false;
}

/**
 * check_status
 */
void check_status(
    const char * msg,
    const int value,
    const UrlCopyStatus::Value expected) {
    if (expected != value){
        std::stringstream message;
        message << msg << " - got : " << (UrlCopyStatus::Value)value << " expected : " << expected;
        CPPUNIT_FAIL(message.str().c_str());
    }
}

/**
 * check_category
 */
void check_category(
    const char * msg,
    int value,
    UrlCopyError::Category expected) {
    if (expected != value){
        std::stringstream message;
        message << msg << " - got : " << (UrlCopyError::Category)value << " expected : " << expected;
        CPPUNIT_FAIL(message.str().c_str());
    }
}

/**
 * check_phase
 */
void check_phase(
    const char * msg,
    int value,
    UrlCopyError::Phase expected){
    if (expected != value){
        std::stringstream message;
        message << msg << " - got : " << (UrlCopyError::Phase)value << " expected : " << expected;
        CPPUNIT_FAIL(message.str().c_str());
    }
}

/**
 * check_scope
 */
void check_scope(
    const char * msg,
    int value,
    UrlCopyError::Scope expected){
    if (expected != value){
        std::stringstream message;
        message << msg << " - got : " << (UrlCopyError::Scope)value << " expected : " << expected;
        CPPUNIT_FAIL(message.str().c_str());
    }
}


/**
 * create a token description string in the form used by the SRM stub services
 */
std::string make_token(
    unsigned int total_space,
    unsigned int guaranteed_space,
    unsigned int unused_space){
    std::stringstream token;
    token << "rp=CUSTODIAL:al=ONLINE:o=dteam:ts=" << total_space << ":gs=" <<
        guaranteed_space << ":us=" << unused_space << ":la=1000:ll=999";
    std::string ret = token.str();
    return ret;
}

/**
 * Create a token description string in the form used by the SRM stub services
 * token description corresponding to one space
 */
std::string make_description(const std::string& token1){
    std::stringstream td;
    td << "Description_for:" << token1;
    std::string ret = td.str();
    return ret;
}
/**
 * Create a token description string in the form used by the SRM stub services
 * token description corresponding to two spaces
 */
std::string make_description(const std::string& token1,const std::string& token2){
    std::stringstream td;
    td << "Description_for:" << token1 << ";" << token2;
    std::string ret = td.str();
    return ret;
}
/**
 * Create a token description string in the form used by the SRM stub services
 * token description corresponding to three spaces
 */
std::string make_description(const std::string& token1,const std::string& token2,const std::string& token3){
    std::stringstream td;
    td << "Description_for:" << token1 << ";" << token2 << ";" << token3;
    std::string ret = td.str();
    return ret;
}

/**
 * create a file name so that the stub SRM service returns the given values of checksum
 * algorithm and value on an srmLs
 */
std::string make_cksm_file(
    const std::string& name,
    const std::string& algorithm,
    const std::string& value){
    std::stringstream st;
    st << "cksm-" << algorithm << "-" << value << "-" << name;
    std::string ret = st.str();
    return ret;
}

ChecksumGetter_noGridFTP::ChecksumGetter_noGridFTP(
	const std::string& alg,
	const std::string& val,
	const UrlCopyError::Category& c
)
	: _alg(alg),
	  _val(val),
	  _c(c),
	  _isFallbackHappened(false)
{
	// EMPTY
}

void ChecksumGetter_noGridFTP::getChecksumFromGridFTP(
		ChecksumCheckerUtils::FilePair& filePair,
		ChecksumCheckerUtils::File& file,
		const unsigned int timeout,
		RefreshTimeUpdater::IUpdateRefreshTime*,
		UrlCopyError::Category& c,
		std::string&)
{
	c = _c;

	if (c == UrlCopyError::SUCCESS) {
	    Checksum ch(_alg, _val);
		ch.copyChecksumAlgorithm(file.algorithm);
		ch.copyChecksumValue(file.value);
	}

	_isFallbackHappened = true;
}

bool ChecksumGetter_noGridFTP::fallbackHappened() const
{
	return _isFallbackHappened;
}

