/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Transfer Url Copy: Utility for the Test Cases
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: test-utility.h,v 1.10 2009/07/30 19:32:55 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#ifndef GLITE_DATA_TRANSFER_URL_COPY_TEST_UTILITY_H_
#define GLITE_DATA_TRANSFER_URL_COPY_TEST_UTILITY_H_

#include <string>
#include <vector>
#include <map>
#include <iostream>
#include <glite/data/transfer/urlcopy/UrlCopyError.h>
#include <glite/data/transfer/urlcopy/UrlCopyStatus.h>
#include <sys/time.h>

#include <log4cpp/Category.hh>
#include "cppunit_wrapper.h"

#define DEBUG(a) log4cpp::Category::getInstance("test-transfer-url-copy").debugStream() << a;

namespace test_utils{

class timer{
public:
    timer(const char * fname);
    ~timer();
protected:
    std::string m_fname;
    struct timeval m_start;
    struct timeval m_end;
};

class TestsMap{
public:
    class RegisterTestClass{
    public:
        RegisterTestClass(const std::string& name){
            TestsMap::instance().addSuite(name);
        }
    };
    typedef std::vector<std::string> Names;
    static TestsMap& instance();
    void addSuite(const std::string& name);
    const Names& names();
    void disableAll();
    bool enable(const std::string& name);

    CppUnit::Test * suite() {
        CppUnit::TestFactoryRegistry& registry = CppUnit::TestFactoryRegistry::getRegistry();
        for (Suites::const_iterator it = m_suites.begin();it != m_suites.end();++it){
            const std::string& name = it->first;
            bool enabled = it->second;
            if (enabled){
                CppUnit::TestFactoryRegistry& testCase = CppUnit::TestFactoryRegistry::getRegistry(name);
                registry.registerFactory(&testCase);
            }
        }
        return registry.makeTest();
    }

private:
    TestsMap();
    typedef std::map<std::string,bool> Suites;
    Suites m_suites;
    Names m_names;
};

}

#define TEST_CLASS(class_name) \
    CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( class_name, #class_name ); \
    test_utils::TestsMap::RegisterTestClass g_##class_name(#class_name);

#define __TIMER test_utils::timer __timer(__PRETTY_FUNCTION__)


/**
 * Put Source file for tests
 * Contacts the stub SRM endpoint to make the appropriate srmMkDir requests;
 * then performs an srmPrepareToPut to get a TURL (which is a local address) and
 * writes a file at the given physical location.
 */
void putFile(const std::string& file_name,const std::string& endpoint,const std::string& version, const char * const text);

/**
 * Check if file exists
 * @return true if file exists, false otherwise
 */
bool fileExists(const std::string& file_name,const std::string& endpoint,const std::string& version);


/**
 * check_status
 */
void check_status(
    const char * msg,
    int value,
    glite::data::transfer::urlcopy::UrlCopyStatus::Value expected);

/**
 * check_category
 */
void check_category(
    const char * msg,
    int value,
    glite::data::transfer::urlcopy::UrlCopyError::Category expected);

/**
 * check_phase
 */
void check_phase(
    const char * msg,
    int value,
    glite::data::transfer::urlcopy::UrlCopyError::Phase expected);

/**
 * check_scope
 */
void check_scope(
    const char * msg,
    int value,
    glite::data::transfer::urlcopy::UrlCopyError::Scope expected);


/**
 * create a token description string in the form used by the SRM stub services
 * "rp=(rp):al=(al):o=(o):ts=(ts):gs=(gs):us=(us):la=(la):ll=(ll)", where:
 * rp = retention policy
 * al = access latency
 * o = owner
 * ts/gs/us = total/guaranteed/unused space
 * la/ll = lifetime assigned/left
 */
std::string make_token(
    unsigned int total_space,
    unsigned int guaranteed_space,
    unsigned int unused_space);

/**
 * Create a token description string in the form used by the SRM stub services
 * token description corresponding to one space
 */
std::string make_description(const std::string& token1);
/**
 * Create a token description string in the form used by the SRM stub services
 * token description corresponding to two spaces
 */
std::string make_description(const std::string& token1,const std::string& token2);
/**
 * Create a token description string in the form used by the SRM stub services
 * token description corresponding to three spaces
 */
std::string make_description(const std::string& token1,const std::string& token2,const std::string& token3);

/**
 * create a file name so that the stub SRM service returns the given values of checksum
 * algorithm and value on an srmLs
 */
std::string make_cksm_file(
    const std::string& name,
    const std::string& algorithm,
    const std::string& value);

#define CHECK_VALUE(value,expected) \
    if (value != expected){ \
        std::stringstream err; \
        err << "Invalid value for " << #value << ": got '" << value << "' expected '" << expected << "'"; \
        CPPUNIT_FAIL(err.str().c_str()); \
    }
#define CHECK_STRING(value,expected) \
    if (0 != strcmp(value,expected)){ \
        std::stringstream err; \
        err << "Invalid value for " << #value << ": got '" << value << "' expected '" << expected << "'"; \
        CPPUNIT_FAIL(err.str().c_str()); \
    }


#include "IChecksumGetter.h"

using namespace glite::data::transfer::urlcopy;

class ChecksumGetter_noGridFTP : public IChecksumGetter
{
public:
	ChecksumGetter_noGridFTP(
		const std::string& alg,
		const std::string& val,
		const UrlCopyError::Category& c = UrlCopyError::SUCCESS);

	virtual ~ChecksumGetter_noGridFTP() { /* EMPTY */ }

	virtual void getChecksumFromGridFTP(
		ChecksumCheckerUtils::FilePair& filePair,
		ChecksumCheckerUtils::File& file,
		const unsigned int timeout,
		RefreshTimeUpdater::IUpdateRefreshTime* updater,
		UrlCopyError::Category& c,
		std::string& m);

	bool fallbackHappened() const;

private:
	const std::string _alg;
	const std::string _val;
	const UrlCopyError::Category _c;
	bool _isFallbackHappened;
};


#endif //GLITE_DATA_TRANSFER_URL_COPY_TEST_UTILITY_H_
